module bound_mod
  use genrev3d_types
  use bound_types
  use wave_types

  implicit none
  logical, private,save :: verb,debug1,debug2
  logical, private,save :: bound_z,bound_mx,bound_my,bound_hx,bound_hy
  integer, private,save :: n_z_bound,n_mx_bound,n_my_bound
  integer, private,save :: n_hx_bound,n_hy_bound

  real, allocatable, dimension (:), private,save :: z_bound,mx_bound,my_bound
  real, allocatable, dimension (:), private,save :: hx_bound,hy_bound

!  integer, external :: omp_get_thread_num

contains
  subroutine init_bound(gen_par,bound_par)
    implicit none
    type (genrev3d_param_type) gen_par
    type (bound_param_type) bound_par

    integer ierr,i_z,i_mx,i_my,i_hx,i_hy
    real dev

    verb=gen_par%verb
    debug1=gen_par%debug1
    debug2=gen_par%debug2

     if(gen_par%debug2) write(0,*)'In init_bound'
  if(gen_par%debug2) write(0,*)'init_bound bound_z',bound_par%bound_z
  if(gen_par%debug2) write(0,*)'init_bound bound_mx',bound_par%bound_mx
  if(gen_par%debug2) write(0,*)'init_bound bound_my',bound_par%bound_my
  if(gen_par%debug2) write(0,*)'init_bound bound_hx',bound_par%bound_hx
  if(gen_par%debug2) write(0,*)'init_bound bound_hy',bound_par%bound_hy

    bound_z=bound_par%bound_z
    bound_mx=bound_par%bound_mx
    bound_my=bound_par%bound_my
    bound_hx=bound_par%bound_hx
    bound_hy=bound_par%bound_my

    n_z_bound =bound_par%n_z_bound
    n_mx_bound =bound_par%n_mx_bound
    n_my_bound =bound_par%n_my_bound
    n_hx_bound =bound_par%n_hx_bound

     if(bound_par%bound_z) then
       allocate (z_bound(n_z_bound),stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of z_bound")
       if(gen_par%debug2) write(0,*)'Allocate z_bound with shape=',shape(z_bound)
  
       do i_z=1,n_z_bound
         dev=(i_z*bound_par%z_bound_stdev)/(n_z_bound)
         z_bound(i_z)=bound_par%z_bound_min+ &
                        (1.-bound_par%z_bound_min)*exp(-dev**2)
       end do
       if(debug2) then
         call snap('z_bound.H',n_z_bound,1,z_bound)
       end if
     end if

    if(bound_par%bound_mx) then
     allocate (mx_bound(n_mx_bound),stat=ierr)  
     if(ierr .ne. 0) call seperr("Problems with allocation of mx_bound")
     if(gen_par%debug2) write(0,*)'Allocate mx_bound with shape=',shape(mx_bound)

     do i_mx=1,n_mx_bound
       dev=(i_mx*bound_par%mx_bound_stdev)/(n_mx_bound)
       mx_bound(i_mx)=bound_par%mx_bound_min+ &
                      (1.-bound_par%mx_bound_min)*exp(-dev**2)
     end do
     if(debug2) then
       call snap('mx_bound.H',n_mx_bound,1,mx_bound)
     end if
    end if

    if(bound_par%bound_my) then
     allocate (my_bound(n_my_bound),stat=ierr)  
     if(ierr .ne. 0) call seperr("Problems with allocation of my_bound")
     if(gen_par%debug2) write(0,*)'Allocate my_bound with shape=',shape(my_bound)

     do i_my=1,n_my_bound
       dev=(i_my*bound_par%my_bound_stdev)/(n_my_bound)
       my_bound(i_my)=bound_par%my_bound_min+ &
                      (1.-bound_par%my_bound_min)*exp(-dev**2)
     end do
     if(debug2) then
       call snap('my_bound.H',n_my_bound,1,my_bound)
     end if
    end if

    if(bound_par%bound_hx) then
     allocate (hx_bound(n_hx_bound),stat=ierr)  
     if(ierr .ne. 0) call seperr("Problems with allocation of hx_bound")
     if(gen_par%debug2) write(0,*)'Allocate hx_bound with shape=',shape(hx_bound)

     do i_hx=1,n_hx_bound
       dev=(i_hx*bound_par%hx_bound_stdev)/(n_hx_bound)
       hx_bound(i_hx)=bound_par%hx_bound_min+ &
                      (1.-bound_par%hx_bound_min)*exp(-dev**2)
     end do
     if(debug2) then
       call snap('hx_bound.H',n_hx_bound,1,hx_bound)
     end if
    end if

    if(bound_par%bound_hy) then
     allocate (hy_bound(n_hy_bound),stat=ierr)  
     if(ierr .ne. 0) call seperr("Problems with allocation of hy_bound")
     if(gen_par%debug2) write(0,*)'Allocate hy_bound with shape=',shape(hy_bound)

     do i_hy=1,n_hy_bound
       dev=(i_hy*bound_par%hy_bound_stdev)/(n_hy_bound)
       hy_bound(i_hy)=bound_par%hy_bound_min+ &
                      (1.-bound_par%hy_bound_min)*exp(-dev**2)
     end do
     if(debug2) then
       call snap('hy_bound.H',n_hy_bound,1,hy_bound)
     end if
    end if

!write(0,*)'thread =',omp_get_thread_num()
!
!!$omp flush(bound_mx,bound_my,bound_hx,bound_hy)
!!$omp flush(n_mx_bound,n_my_bound,n_hx_bound,n_hy_bound)
!!$omp flush(mx_bound,my_bound,hx_bound,hy_bound)


  end subroutine init_bound

  subroutine apply_bound(wave_par,wave_block_slice)
    implicit none
    type (wave_param_type), intent(in) :: wave_par
    real, dimension (:,:,:,:,:), intent(inout) :: wave_block_slice

    integer, automatic :: ierr,i_z,i_mx,i_my,i_hx,i_hy
!    integer :: ierr,i_mx,i_my,i_hx,i_hy

!write(0,*)'hy_bound with shape=',shape(hy_bound)
!write(0,*)'wave_block_slice with shape=',shape(wave_block_slice)

!write(0,*)'n_z =',wave_par%n_z,omp_get_thread_num()
!write(0,*)'n_z_bound =',n_z_bound,omp_get_thread_num()
!write(0,*)'n_mx =',wave_par%n_mx,omp_get_thread_num()
!write(0,*)'n_mx_bound =',n_mx_bound,omp_get_thread_num()
!write(0,*)'n_my =',wave_par%n_my,omp_get_thread_num()
!write(0,*)'n_my_bound =',n_my_bound,omp_get_thread_num()
!write(0,*)'n_hx =',wave_par%n_hx,omp_get_thread_num()
!write(0,*)'n_hx_bound =',n_hx_bound,omp_get_thread_num()
!write(0,*)'n_hy =',wave_par%n_hy,omp_get_thread_num()
!write(0,*)'n_hy_bound =',n_hy_bound,omp_get_thread_num()


    if(bound_z) then
      if(debug2) write(0,*)'Apply bound on z axis'
      do i_z=1,n_z_bound
        wave_block_slice(i_z,:,:,:,:)= &
          wave_block_slice(i_z,:,:,:,:)* &
            z_bound(n_z_bound-i_z+1)
        wave_block_slice(wave_par%n_z-i_z+1,:,:,:,:)= &
          wave_block_slice(wave_par%n_z-i_z+1,:,:,:,:)* &
            z_bound(n_z_bound-i_z+1)
      end do
    end if

!write(0,*)'##100'
    if(bound_mx) then
      if(debug2) write(0,*)'Apply bound on mx axis'
!write(0,*)'n_mx_bound',n_mx_bound
!write(0,*)'wave_par-n_mx',wave_par%n_mx
!write(0,*)'shape(wave_block_slice)=',shape(wave_block_slice)
!write(0,*)'shape(mx_bound)=',shape(mx_bound)
      do i_mx=1,n_mx_bound
        wave_block_slice(:,i_mx,:,:,:)= &
          wave_block_slice(:,i_mx,:,:,:)* &
            mx_bound(n_mx_bound-i_mx+1)
        wave_block_slice(:,wave_par%n_mx-i_mx+1,:,:,:)= &
          wave_block_slice(:,wave_par%n_mx-i_mx+1,:,:,:)* &
            mx_bound(n_mx_bound-i_mx+1)
      end do
    end if
!write(0,*)'##200'

    if(bound_my) then
      if(debug2) write(0,*)'Apply bound on my axis'
      do i_my=1,n_my_bound
        wave_block_slice(:,:,i_my,:,:)= &
          wave_block_slice(:,:,i_my,:,:)* &
            my_bound(n_my_bound-i_my+1)
        wave_block_slice(:,:,wave_par%n_my-i_my+1,:,:)= &
          wave_block_slice(:,:,wave_par%n_my-i_my+1,:,:)* &
            my_bound(n_my_bound-i_my+1)
      end do
    end if

!write(0,*)'##300'
    if(bound_hx) then
      if(debug2) write(0,*)'Apply bound on hx axis'
      do i_hx=1,n_hx_bound
        wave_block_slice(:,:,:,i_hx,:)= &
          wave_block_slice(:,:,:,i_hx,:)* &
            hx_bound(n_hx_bound-i_hx+1)
        wave_block_slice(:,:,:,wave_par%n_hx-i_hx+1,:)= &
          wave_block_slice(:,:,:,wave_par%n_hx-i_hx+1,:)* &
            hx_bound(n_hx_bound-i_hx+1)
      end do
    end if
!write(0,*)'##400'


    if(bound_hy) then
      if(debug2) write(0,*)'Apply bound on hy axis'
!      do i_hy=1,n_hy_bound
!            write(0,*)'hy_bound=',hy_bound(i_hy),omp_get_thread_num()
!            write(0,*)'hy_bound(2)=',hy_bound(n_hy_bound-i_hy+1),omp_get_thread_num()
!      end do
      do i_hy=1,n_hy_bound
        wave_block_slice(:,:,:,:,i_hy)= &
          wave_block_slice(:,:,:,:,i_hy)* &
            hy_bound(n_hy_bound-i_hy+1)
        wave_block_slice(:,:,:,:,wave_par%n_hy-i_hy+1)= &
          wave_block_slice(:,:,:,:,wave_par%n_hy-i_hy+1)* &
            hy_bound(n_hy_bound-i_hy+1)
      end do
    end if

!write(0,*)'##500'

  end subroutine apply_bound

  subroutine clean_bound()
    implicit none
    integer ierr

    if(allocated(z_bound)) then
       deallocate (z_bound,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating z_bound')
    end if

    if(allocated(mx_bound)) then
       deallocate (mx_bound,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating mx_bound')
    end if

    if(allocated(my_bound)) then
       deallocate (my_bound,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating my_bound')
    end if

    if(allocated(hx_bound)) then
       deallocate (hx_bound,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating hx_bound')
    end if

    if(allocated(hy_bound)) then
       deallocate (hy_bound,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating hy_bound')
    end if

  end subroutine clean_bound

end module bound_mod
