module data_mod

  use genrev3d_types
  use data_types
  use wave_types
  use rev_extrap_types
  use slow_types


  implicit none

  integer, external ::  sreed,srite,sseek_block
  logical, private :: adj,add
  integer, private :: nx_data,ny_data
  integer, private :: n_bytes_io,i_x_0_data,i_y_0_data

  real, allocatable, dimension (:,:,:,:), private :: data_io
  real, allocatable, dimension (:), private,save :: off_x_taper

contains

  !-----------------------------------------------------------------------------
  subroutine init_io_data(gen_par,data_par,wave_par)

    type (genrev3d_param_type) gen_par
    type (data_param_type) data_par
    type (wave_param_type) wave_par

    integer ierr
    integer i_off_x
    real dev

    if (gen_par%source_type(1:6) == "point_") then

      nx_data=data_par%n_off_x
      ny_data=data_par%n_off_y
      i_x_0_data=(data_par%off_x_0-wave_par%off_x_0)/wave_par%d_mx + 1
      i_y_0_data=(data_par%off_y_0-wave_par%off_y_0)/wave_par%d_my + 1

    else
      nx_data=data_par%n_mx
      ny_data=data_par%n_my
      i_x_0_data=(data_par%mx_0-wave_par%mx_0)/data_par%d_mx + 1
      i_y_0_data=(data_par%my_0-wave_par%my_0)/data_par%d_my + 1

    end if
  
    write(0,*)'nx_data=',nx_data
    write(0,*)'ny_data=',ny_data
    write(0,*)'i_x_0_data=',i_x_0_data
    write(0,*)'i_y_0_data=',i_y_0_data

    n_bytes_io=nx_data*ny_data*data_par%esize
    write(0,*)'data n_bytes_io=',n_bytes_io

    allocate (data_io(nx_data,ny_data, 1,1), stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of data_io")
    if(gen_par%debug2) write(0,*)'Allocate data_io with shape=',shape(data_io)

#ifdef __CONVEY
!$cny migrate_coproc(data_io,n_bytes_io)
#endif

    if (data_par%n_off_x_taper >0) then
      allocate (off_x_taper(data_par%n_off_x_taper), stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of off_x_taper")
      if(gen_par%debug2) &
         write(0,*)'Allocate off_x_taper with shape=',shape(off_x_taper)
       do i_off_x=1,data_par%n_off_x_taper 
         dev=(i_off_x*data_par%off_x_taper_stdev)/data_par%n_off_x_taper 
         off_x_taper(i_off_x)=data_par%off_x_taper_min+ &
                        (1.-data_par%off_x_taper_min)*exp(-dev**2)
       end do
       call snap('Off_x_taper.H',data_par%n_off_x_taper,1,off_x_taper)
#ifdef __CONVEY
!$cny migrate_coproc(off_x_taper,data_par%n_off_x_taper*data_par%esize)
#endif
    end if


  end subroutine init_io_data


  !-----------------------------------------------------------------------------
  subroutine output_data(tag_data,gen_par,data_par,wave_par,rev_extrap_par, &
                         wave_block)

   character(len=128) :: tag_data
   type (genrev3d_param_type) gen_par
   type (data_param_type) data_par
   type (wave_param_type) wave_par
   type (rev_extrap_param_type) rev_extrap_par
   real,  dimension (:,:,:,:,:,:)   :: wave_block

   integer :: len_tag,i_t0,i_t_data,ret_sseek,nsseek_blocks,i_z_data

   len_tag=len_trim(tag_data)

   if(trim(tag_data) == trim(data_par%tag)) then
      write(0,*)'Writing receiver data'
      i_z_data=wave_par%i_z_rec
      i_t_data=nint((rev_extrap_par%t-data_par%t_0)/data_par%d_t)+1
   else
      write(0,*)'Writing source data'
      i_z_data=wave_par%i_z_sou
!      i_t_data=(data_par%n_t - nint((rev_extrap_par%t-data_par%t_sou_0)/(-data_par%d_t)))
      i_t_data=nint((rev_extrap_par%t-data_par%t_sou_0)/data_par%d_t)+1
   end if

   i_t0=wave_par%i_t_step(1)

   if(i_t_data < 1) then
      write(0,*)'i_t_data=',i_t_data
     call seperr("Writing for data before origin")
   end if
   if(i_t_data > data_par%n_t ) then
     write(0,*)'i_t_data=',i_t_data
     call seperr("Writing data after end")
   end if

   nsseek_blocks=(i_t_data-1) +  &
                 ((data_par%i_s1_par-1)+(data_par%i_s2_par-1)*data_par%n_s1_par)*data_par%n_t

!   write(0,*)'i_t_data=',i_t_data
!   write(0,*)'nsseek_blocks=',nsseek_blocks

   ret_sseek=sseek_block(trim(tag_data),nsseek_blocks,n_bytes_io,0) 
   if (ret_sseek .ne. nsseek_blocks) then
   !{
     write(0,*)'ret_sseek=',ret_sseek
     write(0,*)'tag_data=',trim(tag_data)
     call seperr("Error block seeking data")
   !}
   end if


!$cny begin_coproc
   if(ny_data > 1) then
     data_io(1:nx_data,1:ny_data,1,1)= &
         wave_block(i_z_data, &
           i_x_0_data:i_x_0_data+nx_data, &
           i_y_0_data:i_y_0_data+ny_data,1,1,i_t0)
   else
     data_io(1:nx_data,1,1,1)= &
         wave_block(i_z_data, &
           i_x_0_data:i_x_0_data+nx_data, &
           1,1,1,i_t0)
   end if
!$cny end_coproc

#ifdef __CONVEY
!$cny migrate_host(data_io, n_bytes_io)
#endif
   if(srite(trim(tag_data),data_io,n_bytes_io) .ne. n_bytes_io ) then
     write(0,*)'tag_data=',trim(tag_data)
     write(0,*)'i_t_data=',i_t_data
     call seperr('Problems with writing tag_data file')
   end if
#ifdef __CONVEY
!$cny migrate_coproc(data_io, n_bytes_io)
#endif


  return
  end subroutine output_data

  !-----------------------------------------------------------------------------
  subroutine input_data(tag_data,gen_par,data_par,wave_par,rev_extrap_par, &
                         force_block,wave_block)

   character(len=128) :: tag_data
   type (genrev3d_param_type) gen_par
   type (data_param_type) data_par
   type (wave_param_type) wave_par
   type (rev_extrap_param_type) rev_extrap_par
   real,  dimension (:,:,:,:)   :: force_block
   real, dimension (:,:,:,:,:,:)   :: wave_block

   integer :: len_tag,i_t0,i_t_data,ret_sseek,nsseek_blocks
   integer :: i_off_x
!   real   :: wave(wave_par%n_z_pad,wave_par%n_mx_pad,wave_par%n_my_pad)

   len_tag=len_trim(tag_data)

   if(trim(tag_data) == trim(data_par%tag)) then
!      write(0,*)'Reading receiver data'
!      i_z_data=i_z_rec
      i_t_data=nint((rev_extrap_par%t-data_par%t_0)/data_par%d_t)+1
   else
!      write(0,*)'Reading source data'
!      i_z_data=i_z_sou
      i_t_data=nint((rev_extrap_par%t-data_par%t_sou_0)/data_par%d_t)+1
   end if

   i_t0=wave_par%i_t_step(1)

!   i_t_data=nint((rev_extrap_par%t+rev_extrap_par%d_t-data_par%t_0)/data_par%d_t)+1
!   write(0,*)'i_t_data=',i_t_data

   if(i_t_data < 1) call seperr("Asking for data before origin")
   if(i_t_data > data_par%n_t ) call seperr("Asking for data after end")

   nsseek_blocks=(i_t_data-1) +  &
    ((data_par%i_s1_par-1)+(data_par%i_s2_par-1)*data_par%n_s1_par)*data_par%n_t

   ret_sseek=sseek_block(trim(tag_data),nsseek_blocks,n_bytes_io,0) 
   if (ret_sseek .ne. nsseek_blocks) then
   !{
     write(0,*)'ret_sseek=',ret_sseek
     write(0,*)'tag_data=',trim(tag_data)
     call seperr("Error block seeking data")
   !}
   end if

#ifdef __CONVEY
!$cny migrate_host(data_io, n_bytes_io)
#endif
   if(sreed(trim(tag_data),data_io,n_bytes_io) .ne. n_bytes_io ) then
     write(0,*)'tag_data=',trim(tag_data)
     write(0,*)'i_t_data=',i_t_data
     call seperr('Problems with reading tag_data file')
   end if
#ifdef __CONVEY
!$cny migrate_coproc(data_io, n_bytes_io)
#endif

#ifdef __CONVEY
!$cny begin_coproc
#endif

   if(ny_data > 1) then
   !{
     force_block(1,i_x_0_data:i_x_0_data+nx_data, &
                 i_y_0_data:i_y_0_data+ny_data,1)=&
          data_io(1:nx_data,1:ny_data,1,1)
   !}
   else
   !{
     if (data_par%n_off_x_taper >0) then
     !{
      do i_off_x=1,data_par%n_off_x_taper 
        data_io(i_off_x,1,1,1)= &
          data_io(i_off_x,1,1,1)* &
            off_x_taper(data_par%n_off_x_taper-i_off_x+1)
        data_io(nx_data-i_off_x+1,1,1,1)= &
          data_io(nx_data-i_off_x+1,1,1,1)* &
            off_x_taper(data_par%n_off_x_taper-i_off_x+1)
      end do
     !}
     end if
!       wave_block(i_t0,i_z_data, &
!          i_x_0_data:i_x_0_data+data_par%n_off_x, &
!          1,1,1)=&
!            data_io(1:data_par%n_off_x,1,1,1)
!     wave_block(i_t0,i_z_data, &
!        i_x_0_data:i_x_0_data+nx_data, &
!        1,1,1)=&
!          data_io(1:nx_data,1,1,1) 
     force_block(1,i_x_0_data:i_x_0_data+nx_data,1,1)= data_io(1:nx_data,1,1,1)
!       wave_block(i_t0,i_z_data, &
!          i_x_0_data:i_x_0_data+data_par%n_off_x, &
!          1,1,1)

   !}
   end if

#ifdef __CONVEY
!$cny end_coproc
#endif

   return
   end subroutine input_data

  !-----------------------------------------------------------------------------


  !-----------------------------------------------------------------------------
  subroutine clean_io_data()
    integer ierr

    if(allocated(data_io)) then
       deallocate (data_io,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating data_io')
    end if

    return
  end subroutine clean_io_data


end module data_mod
