!*$*OPTIMIZE(5)
!*$*ROUNDOFF(3)
!*$*SCALAROPTIMIZE(3)

module genrev3d_mod
  use genrev3d_types
  use data_types
  use image_types
  use slow_types
  use wave_types
  use wavelet_types
  use rev_extrap_types
  use bound_types

  use timers_mod
#ifdef __CONVEY
  use cny_timers_mod
#endif

  use rev_extrap_gen_mod
  use rev_extrap_gen_lin_imag_mod
  use image_mod
  use wave_mod
  use wavelet_mod
  use data_mod
  use bound_mod
!  use slow_mod
!  use filter_data_mod

  implicit none

  type(genrev3d_param_type), private :: gen_par
  type (genrev3d_info_type), private::  info_rev
  type (data_param_type), private:: data_par
  type (image_param_type), private:: image_par
  type (wave_param_type), private:: wave_par
  type (wavelet_param_type), private:: wavelet_par
  type (slow_param_type), private:: slow_par
  type (rev_extrap_param_type), private:: rev_extrap_par
  type (bound_param_type), private:: bound_par

contains
  function genrev3d(adj,add,tag_mod,tag_data, &
  rev_extrap_oper,rev_der2time_oper, rev_scatter_oper, &
  oper4,oper5) result(stat)

    integer stat
    logical adj,add
    integer, save :: previous_adj=0
    character(len=128) :: tag_data,tag_mod
    integer rev_extrap_oper,rev_der2time_oper,rev_scatter_oper
    integer oper4,oper5
    optional rev_extrap_oper,rev_der2time_oper,rev_scatter_oper
    optional oper4,oper5
    external rev_extrap_oper,rev_der2time_oper,rev_scatter_oper
    external oper4,oper5

    integer ierr,i_t,total_bytes_allocated
    integer n_block_w,n_block_z,i_block_w,i_block_z,i_w_block_par
    real z_rev_0,nint_w_min_0

    real,    allocatable, dimension (:,:,:,:,:,:)   :: slow_block,dslow_block
    real, allocatable, dimension (:,:,:,:)   :: force_block
    real, allocatable, dimension (:,:,:,:,:,:)   :: wave_block
    real, allocatable, dimension (:,:,:,:,:,:)   :: wave_back_block
    real, allocatable, dimension (:,:,:,:) :: image_mem,image_mem_hz 
    real, allocatable, dimension (:,:,:,:)   :: wavelet

    total_bytes_allocated=0

    call SEP_timer_start(1)

    rev_extrap_par%adj=adj
    rev_extrap_par%add=add

    if(gen_par%debug1) write(0,*)'rev_extrap_par%adj=',rev_extrap_par%adj
    if(gen_par%debug1) write(0,*)'rev_extrap_par%add=',rev_extrap_par%add

    ! deals with changes in tags
    select case (gen_par%domain_range(1:8))
    case ('image_da','lin_imag')
       if(trim(tag_mod) /= trim(data_par%tag) ) then
          write(0,*)'Change of tag_data from ',trim(data_par%tag), &
          ' to ',trim(tag_mod)
          data_par%tag=trim(tag_mod)
       end if
       if(trim(tag_data) /= trim(image_par%tag) ) then
          write(0,*)'Change of tag_image from ',trim(image_par%tag), &
          ' to ',trim(tag_data)
          image_par%tag=trim(tag_data)
       end if
     case ('image_gl')
       if(trim(tag_mod) /= trim(data_par%tag) ) then
          write(0,*)'Change of tag_data from ',trim(data_par%tag), &
          ' to ',trim(tag_mod)
          data_par%tag=trim(tag_mod)
       end if
       if(trim(tag_data) /= trim(image_par%tag) ) then
          if(image_par%two_images) then
            write(0,*)'Change of tag_image_sou from ',trim(image_par%tag_sou), &
            ' to Image_sou'
            write(0,*)'Change of tag_image_rec from ',trim(image_par%tag_rec), &
            ' to Image_rec'
          else
            write(0,*)'Change of tag_image from ',trim(image_par%tag), &
            ' to ',trim(tag_data)
            image_par%tag=trim(tag_data)
          end if
       end if
    case ('vel_data')
    case default
       write(0,*)'gen_par-domain_range=',trim(gen_par%domain_range)
       call seperr('genrev3d: Not supported domain_range')
    end select

    if(((previous_adj ==  1) .and. &
    (.not. rev_extrap_par%adj)) .or. &
    ((previous_adj == -1) .and. (rev_extrap_par%adj))) then
       !{
       select case (gen_par%domain_range(1:8))
       case ('image_da','lin_imag')
          if(.not. rev_extrap_par%adj) then
             !{
             data_par%file = DATA_OUTPUT
             image_par%file = IMAGE_INPUT
             !}
          else
             !{
             data_par%file = DATA_INPUT
             image_par%file = IMAGE_OUTPUT
             !}
          end if
       case ('image_gl')
          if(.not. rev_extrap_par%adj) then
             !{
             data_par%file = DATA_OUTPUT
             image_par%file = IMAGE_INPUT
             !}
          else
             !{
             data_par%file = DATA_INPUT
             image_par%file = IMAGE_OUTPUT
             !}
          end if
       case ('vel_data')
       case default
          write(0,*)'gen_par-domain_range=',trim(gen_par%domain_range)
          call seperr('genrev3d: Not supported domain_range')
       end select
       !}
    end if
    if(rev_extrap_par%adj) then
       previous_adj=1
    else
       previous_adj=-1
    end if

    if(gen_par%debug2) write(0,*)'data_file=',data_par%file
    if(gen_par%debug2) write(0,*)'image_file=',image_par%file
    if(gen_par%debug2) write(0,*)'slow_file=',slow_par%file
    if(gen_par%debug2) write(0,*)'file_wavefield',wave_par%file
    if(gen_par%debug2) write(0,*)'file_wavelet',wavelet_par%file


    if(data_par%file /= DATA_NO_FILE) then
       call init_io_data(gen_par,data_par,wave_par)
    end if
    if(image_par%file /= IMAGE_NO_FILE) then
       call init_io_image(gen_par,rev_extrap_par,wave_par,image_par)
    end if
    if(slow_par%file /= SLOW_NO_FILE) then
       call init_io_slow(gen_par,wave_par,slow_par)
    end if
    if(wave_par%file /= WAVE_NO_FILE) then
       call init_io_wave(gen_par,wave_par)
    end if
    if(wavelet_par%file /= WAVELET_NO_FILE) then
       call init_io_wavelet(gen_par,wavelet_par)
       allocate (wavelet(wavelet_par%n_t,wavelet_par%n_z, &
         wavelet_par%n_off_x,wavelet_par%n_off_y), &
         stat=ierr)
         !##############
         if(gen_par%debug1) write(0,*) 'wavelet shape=',shape(wavelet),ierr
         !##############
         if(ierr .ne. 0) call seperr("Problems with allocation of wavelet")
         if(gen_par%debug2) write(0,*)'Allocate wavelet with shape=',shape(wavelet)
         total_bytes_allocated=total_bytes_allocated+ &
           product(4*shape(wavelet))
write(0,*)'total_bytes_allocated=',total_bytes_allocated
    else
        allocate (wavelet(1,1,1,1),stat=ierr)
    end if

    if(slow_par%file /= SLOW_NO_FILE) then
       allocate (slow_block(wave_par%n_z, &
       wave_par%n_mx,wave_par%n_my, &
       wave_par%n_hx,wave_par%n_hy, &
       slow_par%n_comp), &
       stat=ierr)  
       if(gen_par%debug2) & 
          write(0,*)'Allocate slow_block with shape=',shape(slow_block)
       if(ierr .ne. 0) call seperr("Problems with allocation of slow_block")
       total_bytes_allocated=total_bytes_allocated+ &
         product(4*shape(slow_block))
    else
       allocate (slow_block(1,1,1,1,1,1),stat=ierr)  
       if(gen_par%debug2) & 
          write(0,*)'Allocate slow_block with shape=',shape(slow_block)
       if(ierr .ne. 0) call seperr("Problems with allocation of slow_block")
    end if

    allocate (force_block(wavelet_par%n_z,wave_par%n_mx,wave_par%n_my,1), stat=ierr)  
    if(gen_par%debug1) write(0,*)'Allocate force_block with shape=',shape(force_block)
    if(ierr .ne. 0) call seperr("Problems with allocation of force_block")
write(0,*)'Before !$cny migrate_coproc(force_block)'
#ifdef __CONVEY
!$cny migrate_coproc(force_block, wavelet_par%n_z*wave_par%n_mx*wave_par%n_my*4)
write(0,*)'after !$cny migrate_coproc(force_block)'
#endif

    allocate (wave_block(wave_par%n_z,wave_par%n_mx,wave_par%n_my, &
    wave_par%n_hx,wave_par%n_hy,wave_par%n_t_step),stat=ierr)  

    if(gen_par%debug1) write(0,*)'Allocate wave_block with shape=',shape(wave_block)
    if(ierr .ne. 0) call seperr("Problems with allocation of wave_block")
    total_bytes_allocated=total_bytes_allocated+ &
      product(4*shape(wave_block))

    wave_block=0.
#ifdef __CONVEY
write(0,*)'Before !$cny migrate_coproc(wave_block)'
!$cny migrate_coproc(wave_block, wave_par%n_t_step*wave_par%n_z*wave_par%n_mx*wave_par%n_my*wave_par%n_hx*wave_par%n_hy*4)
write(0,*)'after !$cny migrate_coproc(wave_block)'
#endif

    allocate (wave_par%i_t_step(wave_par%n_t_step),stat=ierr)
    if(gen_par%debug1) write(0,*)'Allocate i_t_step with shape=', &
                        shape(wave_par%i_t_step)
    if(ierr .ne. 0) call seperr("Problems with allocation of i_t_step")
    do i_t=1,wave_par%n_t_step
      wave_par%i_t_step(i_t)=i_t
    end do
!    write(0,*)'i_t_step=',wave_par%i_t_step

    if (gen_par%domain_range(1:8) == 'lin_imag') then

      allocate (dslow_block(wave_par%n_z, &
      wave_par%n_mx,wave_par%n_my, &
      slow_par%n_hx,slow_par%n_hy, &
      slow_par%n_comp), &
      stat=ierr)  
      if(gen_par%debug2) &
          write(0,*)'Allocate dslow_block with shape=',shape(dslow_block)
      if(ierr .ne. 0) call seperr("Problems with allocation of dslow_block")
      total_bytes_allocated=total_bytes_allocated+ &
        product(4*shape(dslow_block))

      dslow_block=0.

      allocate (wave_back_block(wave_par%n_z,wave_par%n_mx,wave_par%n_my, &
      wave_par%n_hx,wave_par%n_hy,1),stat=ierr)  
      if(gen_par%debug2)  &
          write(0,*)'Allocate wave_back_block with shape=',shape(wave_block)
      if(ierr .ne. 0) call seperr("Problems with allocation of wave_back_block")
      total_bytes_allocated=total_bytes_allocated+ &
        product(4*shape(wave_back_block))
      wave_back_block=0.
    else
      allocate (wave_back_block(1,1,1,1,1,1),stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of wave_back_block")
      wave_back_block=0.

      allocate (dslow_block(1,1,1,1,1,1),stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of dslow_block")
      dslow_block=0.
    end if

    allocate (wave_par%i_t_step_back(1),stat=ierr)
    if(gen_par%debug1) write(0,*)'Allocate i_t_step_back with shape=', &
                        shape(wave_par%i_t_step_back)
    if(ierr .ne. 0) call seperr("Problems with allocation of i_t_step_back")
    wave_par%i_t_step_back(1)=1
!    write(0,*)'i_t_step_back=',wave_par%i_t_step_back

    if(image_par%image_memory) then
    !{
      allocate (image_mem(image_par%n_z,image_par%n_mx,image_par%n_hx, &
          image_par%n_t_lag), stat=ierr)
      if(gen_par%debug2) write(0,*)'Allocate image_mem with shape=', &
          shape(image_mem)
      if(ierr .ne. 0) call seperr("Problems with allocation of image_mem")
      total_bytes_allocated=total_bytes_allocated+ &
        product(4*shape(image_mem))
      image_mem=0.
      if(.not. gen_par%image_zero_offset) then
      !{
        if (image_par%write_image_hz) then
          allocate (image_mem_hz(image_par%n_z,image_par%n_mx,image_par%n_hz, &
             image_par%n_t_lag), stat=ierr)
          if(gen_par%debug2) write(0,*)'Allocate image_mem_hz with shape=', &
            shape(image_mem_hz)
          if(ierr .ne. 0)call seperr("Problems with allocation of image_mem_hz")
          total_bytes_allocated=total_bytes_allocated+ &
            product(4*shape(image_mem_hz))
          image_mem_hz=0.
        end if
      !}
      end if
    !}
    end if

    if(gen_par%debug1) write(0,*)'total_bytes_allocated=',total_bytes_allocated

    if (gen_par%domain_range(1:8) == 'lin_imag') then
      stat= rev_extrap_gen_lin_imag( &
      gen_par,data_par,image_par,rev_extrap_par,slow_par,wave_par,wavelet_par, &
      rev_extrap_oper,rev_der2time_oper,rev_scatter_oper, &
      slow_block,&
      dslow_block,&
      force_block,&
      wave_block,&
      wave_back_block,&
      wavelet,&
      image_mem, &
      image_mem_hz) 
    else
      stat= rev_extrap_gen( &
      gen_par,data_par,image_par,rev_extrap_par,slow_par,wave_par,wavelet_par, &
      rev_extrap_oper,rev_der2time_oper,rev_scatter_oper, &
      slow_block,&
      force_block,&
      wave_block,&
      wave_back_block,&
      wavelet,&
      image_mem, &
      image_mem_hz) 
    end if

    if(ierr .ne. 0)  &
       call seperr("genrev3d_mod: Problems with rev_extrap_gen_par")

    if(allocated(wavelet)) deallocate (wavelet,stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with deallocation of wavelet")
    if(allocated(wave_block)) deallocate (wave_block,stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with deallocation of wave_block")
    if(allocated(wave_back_block)) deallocate (wave_back_block,stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with deallocation of wave_back_block")
    if(allocated(force_block)) deallocate (force_block,stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with deallocation of force_block")
    if(allocated(slow_block)) deallocate (slow_block,stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with deallocation of slow_block")
    if(allocated(dslow_block)) deallocate (dslow_block,stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with deallocation of dslow_block")
    if(allocated(image_mem)) deallocate (image_mem,stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with deallocation of image_mem")
    if(allocated(image_mem_hz)) deallocate (image_mem_hz,stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with deallocation of image_mem_hz")

    call SEP_timer_stop(1)

    call SEP_print_timers()

#ifdef __CONVEY
    call CNY_print_timers()
#endif

    return
  end function genrev3d

  subroutine genrev3d_init(verb,init_rev_extrap_oper)
    interface 
       subroutine param_gen(gen_par)
         use genrev3d_types, only: genrev3d_param_type
         type (genrev3d_param_type) gen_par
       end subroutine param_gen
    end interface
    interface 
       subroutine param_wavelet(gen_par,wavelet_par)
         use genrev3d_types, only: genrev3d_param_type
         use wavelet_types, only: wavelet_param_type
         type (genrev3d_param_type) gen_par
         type (wavelet_param_type) wavelet_par
       end subroutine param_wavelet
    end interface
    interface 
       subroutine param_data(gen_par,data_par)
         use genrev3d_types, only: genrev3d_param_type
         use data_types, only: data_param_type
         type (genrev3d_param_type) gen_par
         type (data_param_type) data_par
       end subroutine param_data
    end interface
    interface 
       subroutine param_extrap(gen_par,data_par,image_par,wavelet_par, &
                              rev_extrap_par)
         use genrev3d_types, only: genrev3d_param_type
         use data_types, only: data_param_type
         use image_types, only: image_param_type
         use wavelet_types, only: wavelet_param_type
         use rev_extrap_types, only: rev_extrap_param_type
         type (genrev3d_param_type) gen_par
         type (data_param_type) data_par
         type (image_param_type) image_par
         type (wavelet_param_type) wavelet_par
         type (rev_extrap_param_type) rev_extrap_par
       end subroutine param_extrap
    end interface
    interface 
       subroutine param_image(gen_par,data_par,image_par,wavelet_par)
         use genrev3d_types, only: genrev3d_param_type
         use data_types, only: data_param_type
         use image_types, only: image_param_type
         use wavelet_types, only: wavelet_param_type
         type (genrev3d_param_type) gen_par
         type (data_param_type) data_par
         type (image_param_type) image_par
         type (wavelet_param_type) wavelet_par
       end subroutine param_image
    end interface
    interface 
       subroutine param_wave(gen_par,data_par,rev_extrap_par,image_par, & 
                             wavelet_par,wave_par)
         use genrev3d_types, only: genrev3d_param_type
         use data_types, only: data_param_type
         use rev_extrap_types, only: rev_extrap_param_type
         use image_types, only: image_param_type
         use wavelet_types, only: wavelet_param_type
         use wave_types, only: wave_param_type
         type (genrev3d_param_type) gen_par
         type (data_param_type) data_par
         type (rev_extrap_param_type) rev_extrap_par
         type (image_param_type) image_par
         type (wavelet_param_type) wavelet_par
         type (wave_param_type) wave_par
       end subroutine param_wave
    end interface
    interface 
       subroutine param_slow(gen_par,wave_par,image_par,slow_par)
         use genrev3d_types, only: genrev3d_param_type
         use wave_types, only: wave_param_type
         use image_types, only: image_param_type
         use slow_types, only: slow_param_type
         type (genrev3d_param_type) gen_par
         type (wave_param_type) wave_par
         type (image_param_type) image_par
         type (slow_param_type) slow_par
       end subroutine param_slow
    end interface
    interface 
       subroutine param_bound(gen_par,bound_par)
         use genrev3d_types, only: genrev3d_param_type
         use bound_types, only: bound_param_type
         type (genrev3d_param_type) gen_par
         type (bound_param_type) bound_par
       end subroutine param_bound
    end interface
    interface 
       subroutine param_check(gen_par,data_par,wave_par,wavelet_par,slow_par, &
                              rev_extrap_par,image_par)
         use genrev3d_types, only: genrev3d_param_type
         use data_types, only: data_param_type
         use wave_types, only: wave_param_type
         use wavelet_types, only: wavelet_param_type
         use slow_types, only: slow_param_type
         use rev_extrap_types, only: rev_extrap_param_type
         use image_types, only: image_param_type
         type (genrev3d_param_type) gen_par
         type (data_param_type) data_par
         type (wave_param_type) wave_par
         type (wavelet_param_type) wavelet_par
         type (slow_param_type) slow_par
         type (rev_extrap_param_type) rev_extrap_par
         type (image_param_type) image_par
       end subroutine param_check
    end interface

    interface
       function init_rev_extrap_oper(gen_par,data_par,slow_par,rev_extrap_par,&
                        wave_par,wavelet_par,bound_par) &
                result(info)
         use genrev3d_types, only: genrev3d_param_type,genrev3d_info_type
         use data_types, only: data_param_type
         use slow_types, only: slow_param_type
         use rev_extrap_types, only: rev_extrap_param_type
         use wave_types, only: wave_param_type
         use wavelet_types, only: wavelet_param_type
         use bound_types, only: bound_param_type
         type (genrev3d_info_type) info
         type (genrev3d_param_type) gen_par
         type (slow_param_type) slow_par
         type (data_param_type) data_par
         type (wave_param_type) wave_par
         type (rev_extrap_param_type) rev_extrap_par
         type (wavelet_param_type) wavelet_par
         type (bound_param_type) bound_par
       end  function init_rev_extrap_oper
    end interface

    logical :: verb
    character(len=128) :: desc_timer
! Now done at declaration
    ! initialize the default parameters
!    gen_par = &
!    genrev3d_param_type ( &
!    verb_init, &
!    debug1_init, &
!    debug2_init, &
!    timers_on_init, &
!    image_zero_offset_init, &
!    image_zero_time_init, &
!    surf_offset_cig_init, &
!    first_adj_init, &
!    first_add_init, &
!    check_stability_init, &
!    layered_mig_vel_init, &
!    skip_source_prop_init, &
!    trim(source_type_init), &
!    trim(oper_type_init), &
!    trim(domain_range_init), &
!    n_threads_init, &
!    max_memory_init, &
!    non_data_memory_init, &
!    n_data_dim_init, &
!    slow_const_init, &
!    sou_depth_init, &
!    rec_depth_init &
!     )


    if(verb) then
       write(0,*)'*******************************'
       write(0,*)'*******************************'
       write(0,*)'Default parameters for Genrev3d'
       write(0,*)'verb=',gen_par%verb
       write(0,*)'debug1=',gen_par%debug1
       write(0,*)'debug2=',gen_par%debug2
       write(0,*)'timers_on=',gen_par%timers_on
       write(0,*)'image_zero_offset=',gen_par%image_zero_offset
       write(0,*)'image_zero_time=',gen_par%image_zero_time
       write(0,*)'first_adj=',gen_par%first_adj
       write(0,*)'first_add=',gen_par%first_add
       write(0,*)'check_stability=',gen_par%check_stability
       write(0,*)'layered_mig_vel=',gen_par%layered_mig_vel
       write(0,*)'skip_source_prop=',gen_par%skip_source_prop
       write(0,*)'source_type=',trim(gen_par%source_type)
       write(0,*)'oper_type=',trim(gen_par%oper_type)
       write(0,*)'domain_range=',trim(gen_par%domain_range)
       write(0,*)'n_threads=',gen_par%n_threads
       write(0,*)'max_memory=',gen_par%max_memory
       write(0,*)'non_data_memory=',gen_par%non_data_memory
       write(0,*)'n_data_dim=',gen_par%n_data_dim
       write(0,*)'slow_const=',gen_par%slow_const
       write(0,*)'sou_depth=',gen_par%sou_depth
       write(0,*)'rec_depth=',gen_par%rec_depth
       write(0,*)'refl_dip=',gen_par%refl_dip
       write(0,*)'deep_refl_dip=',gen_par%deep_refl_dip
       write(0,*)'deep_refl_depth=',gen_par%deep_refl_depth
       write(0,*)'*******************************'
    end if

    call param_gen(gen_par)
    call param_wavelet(gen_par,wavelet_par)
    call param_data(gen_par,data_par)
    call param_image(gen_par,data_par,image_par,wavelet_par)
    call param_extrap(gen_par,data_par,image_par,wavelet_par,rev_extrap_par)
    call param_wave(gen_par,data_par,rev_extrap_par,image_par,wavelet_par, &
                    wave_par)
    call param_slow(gen_par,wave_par,image_par,slow_par)

    if(gen_par%apply_bound) then
      call param_bound(gen_par,bound_par)
      call init_bound(gen_par,bound_par)
    else
      bound_par%n_z_bound=0
      bound_par%n_mx_bound=0
      bound_par%n_my_bound=0
      bound_par%n_hx_bound=0
      bound_par%n_hy_bound=0
    end if

    call param_check(gen_par,data_par,wave_par,wavelet_par,slow_par, &
                     rev_extrap_par,image_par)


!    call SEP_switch_timers(gen_par%timers_on,gen_par%n_threads)
    call SEP_switch_timers(gen_par%timers_on,1)

    desc_timer='All computation within genrev_3d'
    call SEP_timer_init(1,desc_timer,0.,0)
    desc_timer='Init Operator'
    call SEP_timer_init(2,desc_timer,0.,0)

    desc_timer='Read slowness function'
    call SEP_timer_init(11,desc_timer,0.,1)
    desc_timer='Read/write wavefield'
    call SEP_timer_init(12,desc_timer,0.,0)
    desc_timer='Read/write images'
    call SEP_timer_init(13,desc_timer,0.,1)
    desc_timer='Read/write data'
    call SEP_timer_init(14,desc_timer,0.,0)
    desc_timer='Total computation in rev_extrap_oper (thread 0)'
    call SEP_timer_init(15,desc_timer,0.,0)

!    desc_timer='Total computation in parallel'
!    call SEP_timer_init(15,desc_timer,0.,0)
!    desc_timer='Total computation in parallel (thread 0)'
!    call SEP_timer_init(16,desc_timer,0.,0)


#ifdef __CONVEY
    call CNY_switch_timers(gen_par%timers_on)

    desc_timer='Computation inside rev_extrap_oper on CONVEY coprocessor '
    call CNY_timer_init(1,desc_timer,0.,0)
    desc_timer='Computation outside rev_extrap_oper on CONVEY coprocessor '
    call CNY_timer_init(2,desc_timer,0.,0)
    desc_timer='Computation inside rev_extrap_oper loop #1 '
    call CNY_timer_init(3,desc_timer,0.,0)
    desc_timer='Computation inside rev_extrap_oper loop #2 '
    call CNY_timer_init(4,desc_timer,0.,0)
    desc_timer='Computation inside rev_extrap_oper loop #3 '
    call CNY_timer_init(5,desc_timer,0.,0)
    desc_timer='Computation inside rev_extrap_oper loop #4 '
    call CNY_timer_init(6,desc_timer,0.,0)
    desc_timer='Computation inside rev_extrap_oper loop #5 '
    call CNY_timer_init(7,desc_timer,0.,0)
#endif


    call SEP_timer_start(2)
    info_rev=init_rev_extrap_oper(gen_par,data_par,slow_par,rev_extrap_par,&
                 wave_par,wavelet_par,bound_par)
    call SEP_timer_stop(2)
     
    return
  end subroutine genrev3d_init

  subroutine genrev3d_clean(verb,clean_rev_extrap_oper)
    interface
       function clean_rev_extrap_oper() result(stat)
         integer stat
       end  function clean_rev_extrap_oper
    end interface

    logical :: verb
    integer stat

    stat=clean_rev_extrap_oper()
    if(verb .and. (stat /= 0)) write(0,*)'Problems in clean_rev_extrap_oper'

    call clean_bound()

    return
  end subroutine genrev3d_clean
end module genrev3d_mod
