module genrev3d_types

  integer, parameter ::  status_init=0
  integer, parameter ::  mem_fixed_init=0
  integer, parameter ::  mem_n_w_inp_init=0
  integer, parameter ::  mem_n_mx_inp_init=0
  integer, parameter ::  mem_n_my_inp_init=0
  integer, parameter ::  mem_n_hx_inp_init=0
  integer, parameter ::  mem_n_hy_inp_init=0

  type genrev3d_info_type
     integer :: status
     integer :: mem_fixed
     integer :: mem_n_w_inp
     integer :: mem_n_mx_inp,mem_my_inp
     integer :: mem_n_hx_inp,mem_hy_inp
  end type  genrev3d_info_type

  type(genrev3d_info_type), parameter :: info_default = genrev3d_info_type &
  (status_init, &
  mem_fixed_init, &
  mem_n_w_inp_init, &
  mem_n_mx_inp_init, &
  mem_n_my_inp_init, &
  mem_n_hx_inp_init, &
  mem_n_hy_inp_init )

  logical, parameter :: verb_init=.true.
  logical, parameter :: debug1_init=.false.
  logical, parameter :: debug2_init=.false.
  logical, parameter :: timers_on_init=.false.
  logical, parameter :: image_zero_offset_init=.true.
  logical, parameter :: image_zero_time_init=.false.
  logical, parameter :: surf_offset_cig_init=.false.
  logical, parameter :: write_image_progress_with_shots_init=.false.
  logical, parameter :: first_adj_init=.false.
  logical, parameter :: first_add_init=.false.
  logical, parameter :: check_stability_init=.false.
  logical, parameter :: layered_mig_vel_init=.false.
  logical, parameter :: skip_source_prop_init=.false.
  logical, parameter :: apply_bound_init=.false.
  logical, parameter :: inject_data_init=.true.
  logical, parameter :: transf_geol_offset_init=.false.

  character(len=128), parameter :: source_type_init='point_source'
  character(len=128), parameter :: oper_type_init='const_slow'
  character(len=128), parameter :: domain_range_init='image_data'

  integer, parameter :: n_threads_init=-1
  integer, parameter :: max_memory_init=134217728
  integer, parameter :: non_data_memory_init=0

  integer, parameter :: n_data_dim_init=4

  real, parameter :: slow_const_init=-1.

  real, parameter :: sou_depth_init=0.
  real, parameter :: rec_depth_init=0.

  real, parameter :: refl_dip_init=0.
  real, parameter :: deep_refl_dip_init=0.
  real, parameter :: deep_refl_depth_init=-huge(1.)
  real, parameter :: refl_z0_init=-huge(1.)
  real, parameter :: deep_refl_z0_init=-huge(1.)
  real, parameter :: max_courant_numb_init=1.

  integer, parameter :: integer_init=huge(1)/10.
  real, parameter :: real_init=huge(1.)/10.

  type genrev3d_param_type
     logical :: verb=verb_init
     logical :: debug1=debug1_init
     logical :: debug2=debug2_init
     logical :: timers_on=timers_on_init
     logical :: image_zero_offset=image_zero_offset_init
     logical :: image_zero_time=image_zero_time_init
     logical :: surf_offset_cig=surf_offset_cig_init
     logical :: write_image_progress_with_shots=write_image_progress_with_shots_init
     logical :: first_adj=first_adj_init
     logical :: first_add=first_add_init
     logical :: check_stability=check_stability_init
     logical :: layered_mig_vel=layered_mig_vel_init
     logical :: skip_source_prop=skip_source_prop_init
     logical :: apply_bound=apply_bound_init
     logical :: inject_data=inject_data_init
     logical :: transf_geol_offset=transf_geol_offset_init

! source_type - type of source (point_source, global_source)
! oper_type  - type of propgation operator(const_slow, full_slow)
! domain_range  - domain and range of linear operator (data_image)
     character(len=128) :: source_type=source_type_init
     character(len=128) :: oper_type=oper_type_init
     character(len=128) :: domain_range=domain_range_init

! n_threads  - number of threads
! max_memory  - maximum amount of memory that can be used in total
! non_data_memory  - maximum amount of memory that can be used not for the data
     integer :: n_threads=n_threads_init
     integer :: max_memory=max_memory_init
     integer :: non_data_memory=non_data_memory_init

     integer :: n_data_dim=n_data_dim_init

! slow_const  - constant slowness (if oper_type=const_slow)
! sou_depth  - depth of sources
! rec_depth  - depth of receivers
     real :: slow_const=slow_const_init
     real :: sou_depth=sou_depth_init
     real :: rec_depth=rec_depth_init
     real :: refl_dip=refl_dip_init
     real :: deep_refl_dip=deep_refl_dip_init
     real :: deep_refl_depth=deep_refl_depth_init
     real :: refl_z0=refl_z0_init
     real :: deep_refl_z0=deep_refl_z0_init
     real :: max_courant_numb=max_courant_numb_init

  end type  genrev3d_param_type
end module genrev3d_types
