module image_mod

  use genrev3d_types
  use image_types
  use slow_types
  use wave_types
  use rev_extrap_types
  use wavelet_types


  implicit none

  integer, external ::  sreed,srite,sseek_block

  logical, private :: adj,add,image_zero_offset
  logical, private :: image_in_memory ! Guojian
  logical, private :: write_image_hz  ! Guojian

  integer, private :: n_bytes_image_io,n_bytes_wave_io,i_hx_0,i_hy_0
  integer, private :: n_z_image,n_z_wave
  integer, private :: i_hx_0_image,i_z_0_image
  integer, private :: i_mx_0_image,i_my_0_image
  integer, private :: i_mx_wave_beg,i_mx_wave_end
  integer, private :: i_z_deep_refl

  real, private :: sin_dip,cos_dip,sin_deep_dip,cos_deep_dip
  real, private :: sin_dip_max,cos_dip_max



  real, allocatable, dimension (:,:), private :: image_io,image_io_hz         !Guojian
  real, allocatable, dimension (:,:), private :: wave_io
  
contains

  !-----------------------------------------------------------------------------
  subroutine init_io_image(gen_par,rev_extrap_par,wave_par,image_par)

  implicit none

    type (genrev3d_param_type) gen_par
    type (rev_extrap_param_type) rev_extrap_par
    type (wave_param_type) wave_par
    type (image_param_type) image_par

    integer ierr
    real pi,dip,deep_dip

    adj=rev_extrap_par%adj
    add=rev_extrap_par%add
    write_image_hz=image_par%write_image_hz
    write(0,*)'write_image_hz=',write_image_hz
    image_in_memory=image_par%image_memory
    write(0,*)'image_in_memory=',image_in_memory

    n_z_image=image_par%n_z
    n_z_wave=wave_par%n_z
    i_z_0_image=nint((image_par%z_0-wave_par%z_0)/wave_par%d_z) + 1
write(0,*)'i_z_0_image=',i_z_0_image

    pi=acos(-1.)
    dip=gen_par%refl_dip*pi/180.
    sin_dip=sin(dip)
    cos_dip=cos(dip)

    deep_dip=gen_par%deep_refl_dip*pi/180.
    sin_deep_dip=sin(deep_dip)
    cos_deep_dip=cos(deep_dip)
 
    sin_dip_max=max(sin_dip,sin_deep_dip)
    cos_dip_max=min(cos_dip,cos_deep_dip)

    i_z_deep_refl=nint((gen_par%deep_refl_depth-image_par%z_0)/image_par%d_z) + 1

write(0,*)'sin_dip=',sin_dip
write(0,*)'cos_dip=',cos_dip
write(0,*)'sin_deep_dip=',sin_deep_dip
write(0,*)'cos_deep_dip=',cos_deep_dip
!write(0,*)'i_z_deep_refl=',i_z_deep_refl



    if(gen_par%image_zero_offset) then
    !{
      image_zero_offset=.true.
      n_bytes_image_io=image_par%n_z*image_par%n_mx*image_par%esize
      n_bytes_wave_io=wave_par%n_z*wave_par%n_mx*wave_par%esize
      i_hx_0_image=0

      if(gen_par%debug2) write(0,*)'n_bytes_image_io=',n_bytes_image_io
      if(gen_par%debug2) write(0,*)'n_bytes_wave_io=',n_bytes_wave_io

      allocate (image_io(n_z_image,image_par%n_mx), stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of image_io")
      if(gen_par%debug2) write(0,*)'Allocate image_io with shape=',shape(image_io)
      image_io=0.

      allocate (wave_io(n_z_wave,wave_par%n_mx), stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of wave_io")
      if(gen_par%debug2) write(0,*)'Allocate wave_io with shape=',shape(wave_io)
      wave_io=0.
    !}
    else
    !{
      image_zero_offset=.false.
      n_bytes_image_io=image_par%n_z*image_par%n_mx*image_par%esize
      n_bytes_wave_io=wave_par%n_z*wave_par%n_mx*wave_par%esize

      if(gen_par%debug2) write(0,*)'n_bytes_image_io=',n_bytes_image_io
      if(gen_par%debug2) write(0,*)'n_bytes_wave_io=',n_bytes_wave_io

!         i_hx_0_image=image_par%n_hx/2
      i_hx_0_image=nint(-image_par%hx_0/image_par%d_hx)+1
      if(gen_par%debug2) write(0,*)'(Image) i_hx_0_image=',i_hx_0_image

      allocate (image_io(n_z_image,image_par%n_mx), stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of image_io")
      if(gen_par%debug2) write(0,*)'Allocate image_io with shape=',shape(image_io)
      image_io=0.

      allocate (wave_io(n_z_wave,wave_par%n_mx), stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of wave_io")
      if(gen_par%debug2) write(0,*)'Allocate wave_io with shape=',shape(wave_io)
      wave_io=0.
    !}
    end if
  
  end subroutine init_io_image


  !-----------------------------------------------------------------------------
  subroutine output_image(tag_image,tag_image_hz,gen_par,image_par, &
                          wave_par,rev_extrap_par, &
                          max_it_snap_image, &
                          wave_block,image_mem,image_mem_hz)
! record to memory.
   implicit none

   character(len=128) :: tag_image,tag_image_hz ! 'tag_image_hz' added by Guojian
   type (genrev3d_param_type) gen_par
   type (image_param_type) image_par
   type (wave_param_type) wave_par
   type (rev_extrap_param_type) rev_extrap_par
   integer :: max_it_snap_image !Guojian
   real,  dimension (:,:,:,:,:,:)   :: wave_block
   real, dimension (:,:,:,:) :: image_mem,image_mem_hz 

   character(len=128) :: tag_wave
   logical sum_image_file,write_image
   integer :: len_tag_image,len_tag_image_hz,len_tag_wave !len_tag_image_hz added by Guojian
   integer :: i_z,i_hx,i_hy,i_my,i_mx
   integer :: i_t_image,i_t_wave
   integer :: min_it_snap_image
   integer :: i_t_lag_beg,i_t_lag_end,i_t_lag
   integer :: i_mx_image,i_mx_wave
   integer :: i_hx_shift_sou,i_hx_shift_rec
   integer  nsseek_blocks,ret_sseek,block_wave,ret_sseek_ok
   integer back_i_t_lag,wavepar_itstep1  ! for parallelization
   integer n_threads,i_thread  ! for parallelization

   n_threads=rev_extrap_par%n_threads
   len_tag_image=len_trim(tag_image)
   len_tag_image_hz=len_trim(tag_image_hz)  ! Guojian
   tag_wave=trim(wave_par%tag_back)
   len_tag_wave=len_trim(tag_wave)

!     i_t_lag_beg=max(-i_t_wave+1,nint(image_par%t_0_lag/image_par%d_t_lag))
!     i_t_lag_end=min(image_par%n_t-i_t_wave,i_t_lag_beg+image_par%n_t_lag-1)
   if(image_par%n_t_lag /= 1) then
     i_t_lag_beg=nint(image_par%t_0_lag/image_par%d_t_lag)
     i_t_lag_end=i_t_lag_beg+image_par%n_t_lag-1
   else
     i_t_lag_beg=1
     i_t_lag_end=1
   end if

   i_mx_0_image=nint((wave_par%mx_0-image_par%mx_0)/image_par%d_mx)+1
   i_my_0_image=nint((wave_par%my_0-image_par%my_0)/image_par%d_my)+1
!write(0,*)'i_mx_0_image=',i_mx_0_image
!write(0,*)'i_my_0_image=',i_my_0_image

   i_mx_wave_beg=max(i_hx_0_image+1,-i_mx_0_image+2)
   i_mx_wave_end=min(wave_par%n_mx-(image_par%n_hx-i_hx_0_image), &
                     image_par%n_mx-i_mx_0_image+1)
!write(0,*)'i_mx_wave_beg=',i_mx_wave_beg
!write(0,*)'i_mx_wave_end=',i_mx_wave_end


   if (image_in_memory) then
!    Guojian
      write_image = (rev_extrap_par%i_t_extrap == max_it_snap_image)
!   write(0,*)'write_image=',write_image
      sum_image_file = write_image .and. &
        (((rev_extrap_par%i_s1_par /= 1) .or. (rev_extrap_par%i_s2_par /= 1)) .and. &
         (.not. gen_par%surf_offset_cig) &
        )
!    end Guojian
   else
        min_it_snap_image=i_t_lag_end
        sum_image_file = &
        (((rev_extrap_par%i_t_extrap /= min_it_snap_image) .or.  &
          ((rev_extrap_par%i_s1_par /= 1) .or. (rev_extrap_par%i_s2_par /= 1))) &
        )
   end if
!   write(0,*)'rev_extrap_par%i_t_extrap =',rev_extrap_par%i_t_extrap 
!   write(0,*)'sum_image_file=',sum_image_file


!   i_t_wave=nint((rev_extrap_par%t - image_par%t_0)/image_par%d_t)+1
   i_t_wave=nint((rev_extrap_par%t - wave_par%t_back_0)/wave_par%d_t_back)+1
!     i_t_wave=nint((rev_extrap_par%t - image_par%t_0)/image_par%d_t)+3

     if(gen_par%debug2) write(0,*)'i_t_wave=',i_t_wave

! old 5/2003     i_t_image= image_par%n_t - &
!                i_t_lag_end - &
!                  (nint((rev_extrap_par%t - image_par%t_0)/image_par%d_t))-2

               i_t_image= image_par%n_t - &
              i_t_lag_end - &
                (nint((rev_extrap_par%t - image_par%t_0)/image_par%d_t))

!     i_t_image= image_par%n_t + &
!                i_t_lag_beg - &
!                  (nint((rev_extrap_par%t - image_par%t_0)/image_par%d_t))-2

!                  (nint((rev_extrap_par%t - image_par%t_0)/image_par%d_t)+1)

!     if(gen_par%debug2) write(0,*)'i_t_image=',i_t_image
!     if(gen_par%debug2) write(0,*)'i_mx_0_image=',i_mx_0_image
!     if(gen_par%debug2) write(0,*)'i_mx_wave_beg=',i_mx_wave_beg
!     if(gen_par%debug2) write(0,*)'i_mx_wave_end=',i_mx_wave_end
!     if(gen_par%debug2) write(0,*)'i_my_0_image=',i_my_0_image
!     if(gen_par%debug2) write(0,*)'i_t_lag_beg=',i_t_lag_beg
!     if(gen_par%debug2) write(0,*)'i_t_lag_end=',i_t_lag_end

   if(sum_image_file)  then
   !{
     if(image_par%write_image_progress) then
       nsseek_blocks= -image_par%n_t_lag*image_par%n_my*image_par%n_hx* &
                       image_par%n_hy
!         ret_sseek_ok= (i_t_image-2)* &
!         ret_sseek_ok= (i_t_lag - i_t_lag_beg+ &
       ret_sseek_ok= ( &
                               (i_t_image-2)*image_par%n_t_lag)* &
                               image_par%n_my*image_par%n_hx *image_par%n_hy
     else
       nsseek_blocks= -image_par%n_t_lag*image_par%n_my*image_par%n_hx* &
                       image_par%n_hy
       ret_sseek_ok=0
!         ret_sseek_ok=i_t_lag - i_t_lag_beg
     end if

!       write(0,*)'First seek on Image: nsseek_blocks=',nsseek_blocks

     ret_sseek=sseek_block(trim(tag_image),nsseek_blocks,n_bytes_image_io,1) 
     if (ret_sseek .ne. ret_sseek_ok) then
     !{
       write(0,*)'nsseek_blocks=',nsseek_blocks
       write(0,*)'ret_sseek=',ret_sseek
       write(0,*)'ret_sseek_ok=',ret_sseek_ok
       write(0,*)'tag_image=',trim(tag_image)
       call seperr("Error block seeking image")
     !}
     end if
     if (write_image_hz) then
       ret_sseek=sseek_block(trim(tag_image_hz),nsseek_blocks,n_bytes_image_io,1) 
       if (ret_sseek .ne. ret_sseek_ok) then
       !{
         write(0,*)'nsseek_blocks=',nsseek_blocks
         write(0,*)'ret_sseek=',ret_sseek
         write(0,*)'ret_sseek_ok=',ret_sseek_ok
         write(0,*)'tag_image_hz=',trim(tag_image_hz)
         call seperr("Error block seeking image")
       !}
       end if
     end if

   !}
   end if

   do i_t_lag=i_t_lag_beg,i_t_lag_end
   !{

   if(gen_par%debug2) write(0,*)'i_t_lag=',i_t_lag

   do i_hy=1,image_par%n_hy
   !{
     do i_my=1,image_par%n_my
     !{

!       block_wave= ((i_t_wave-1)+(i_t_lag-1))*wave_par%n_my ! Biondo 2/1/2006
       block_wave= ((i_t_wave-1)+(i_t_lag-1) -1 )*wave_par%n_my ! Biondo 2/1/2006
       nsseek_blocks= block_wave + (i_my-1)*wave_par%n_hx
!     write(0,*)'block_wave=',block_wave
!         write(0,*)'nsseek_blocks (wave)=',nsseek_blocks
       if(block_wave > wave_par%n_t_back-1) then
         write(0,*)'block_wave=',block_wave
         write(0,*)'n_t_back=',wave_par%n_t_back
         call seperr("Tryng to read beyond saved time step in Wavefield_back")
       end if

       ret_sseek=sseek_block(trim(tag_wave),nsseek_blocks,n_bytes_wave_io,0) 
       if (ret_sseek .ne. nsseek_blocks) then
       !{
         write(0,*)'nsseek_blocks=',nsseek_blocks
         write(0,*)'ret_sseek=',ret_sseek
         write(0,*)'tag_wave=',trim(tag_wave)
         call seperr("Error block seeking wave")
       !}
       end if
       if(sreed(trim(tag_wave),wave_io,n_bytes_wave_io)/=n_bytes_wave_io) then
         write(0,*)'tag_wave=',trim(tag_wave)
         call seperr('Problems with reading tag_wave file')
       end if


       do i_hx=1,image_par%n_hx
       !{
         if(image_zero_offset) then
           i_hx_shift_sou=0
           i_hx_shift_rec=0
         else
           i_hx_shift_sou=i_hx_0_image-i_hx
           i_hx_shift_rec=-(i_hx_0_image-i_hx)
         end if
         
         if(sum_image_file) then
         !{

           if(sreed(trim(tag_image),image_io,n_bytes_image_io) /= &
                n_bytes_image_io ) then
             write(0,*)'tag_image=',trim(tag_image)
             call seperr('Problems with reading tag_image file')
           end if
           if (write_image_hz) then
             if(sreed(trim(tag_image_hz),image_io_hz,n_bytes_image_io) /= &
                n_bytes_image_io ) then
               write(0,*)'tag_image=',trim(tag_image_hz)
               call seperr('Problems with reading tag_image_hz file')
             end if
           end if
           if (image_in_memory) then
             image_mem(:,:,i_hx,i_t_lag-i_t_lag_beg+1)= &
                image_mem(:,:,i_hx,i_t_lag-i_t_lag_beg+1)+image_io(:,:)
             if (write_image_hz) then
             image_mem_hz(:,:,i_hx,i_t_lag-i_t_lag_beg+1)= &
                image_mem_hz(:,:,i_hx,i_t_lag-i_t_lag_beg+1)+image_io_hz(:,:)
             end if
           end if

           if(image_par%write_image_progress) then
             nsseek_blocks= image_par%n_t_lag*image_par%n_my*image_par%n_hx* &
                       image_par%n_hy -1
!               ret_sseek_ok= (i_t_image-1)* &
               ret_sseek_ok= (i_t_lag - i_t_lag_beg+ &
                            (i_t_image-1)*image_par%n_t_lag)* &
                            image_par%n_my*image_par%n_hx *image_par%n_hy + &
                            (i_hx-1)* &
                            image_par%n_my*image_par%n_hy 
           else
             nsseek_blocks= -1
             ret_sseek_ok= (i_t_lag -i_t_lag_beg)* &
                            image_par%n_my*image_par%n_hx*image_par%n_hy + &
                           (i_hx-1)* &
                            image_par%n_my*image_par%n_hy 
           end if

!       write(0,*)'in hx loop seek on Image after read: nsseek_blocks=',nsseek_blocks

           ret_sseek=sseek_block(trim(tag_image),nsseek_blocks,n_bytes_image_io,1) 
           if (ret_sseek .ne. ret_sseek_ok) then
           !{
             write(0,*)'nsseek_blocks=',nsseek_blocks
             write(0,*)'ret_sseek=',ret_sseek
             write(0,*)'ret_sseek_ok=',ret_sseek_ok
             write(0,*)'tag_image=',trim(tag_image)
             call seperr("Error block seeking image")
           !}
           end if
!               write(0,*)'ret_sseek=',ret_sseek
           if (write_image_hz) then
             ret_sseek=sseek_block(trim(tag_image_hz),nsseek_blocks,n_bytes_image_io,1) 
             if (ret_sseek .ne. ret_sseek_ok) then
             !{
               write(0,*)'nsseek_blocks=',nsseek_blocks
               write(0,*)'ret_sseek=',ret_sseek
               write(0,*)'ret_sseek_ok=',ret_sseek_ok
               write(0,*)'tag_image_hz=',trim(tag_image_hz)
               call seperr("Error block seeking image")
             !}
             end if
           end if


         !}
         else
         !{
           if (.not. image_in_memory) then
           image_io=0.
           if (image_par%write_image_hz) then
             image_io_hz=0.
           end if
           end if
         !}
         end if
        
         if (image_in_memory) then 
         !{
           back_i_t_lag=i_t_lag-i_t_lag_beg+1
           wavepar_itstep1=wave_par%i_t_step(1)

!$omp parallel default (private), shared (image_mem,image_mem_hz,wave_io,wave_block), &
!$omp& shared(n_threads,i_mx_wave_beg,i_mx_wave_end,i_mx_0_image,n_z_image), &
!$omp& shared(i_hx,back_i_t_lag,i_hx_shift_sou,wavepar_itstep1,i_hx_shift_rec,i_my)
!$omp do

           do i_thread=1,n_threads
             do i_mx_wave=i_mx_wave_beg-1+i_thread,i_mx_wave_end,n_threads
               i_mx_image=i_mx_wave+i_mx_0_image-1 
               do i_z=1,n_z_image
                  image_mem(i_z,i_mx_image,i_hx,back_i_t_lag)=&
                  image_mem(i_z,i_mx_image,i_hx,back_i_t_lag)+&
                  wave_io(i_z+i_z_0_image-1,i_mx_wave+i_hx_shift_sou)* &
                  wave_block(i_z+i_z_0_image-1,i_mx_wave+i_hx_shift_rec,i_my,& 
                            1,1, wavepar_itstep1)
               end do
             end do
           end do

           if (write_image_hz) then
             do i_thread=1,n_threads
               do i_mx_wave=i_mx_wave_beg-1+i_thread,i_mx_wave_end,n_threads
                 i_mx_image=i_mx_wave+i_mx_0_image-1 
                 do i_z=1,n_z_image
                    if (i_z+i_z_0_image+i_hx_shift_sou >0 .and.i_z+i_z_0_image+i_hx_shift_sou <=n_z_wave) then
                      if ( i_z+i_z_0_image+i_hx_shift_rec >0 .and.i_z+i_z_0_image+i_hx_shift_rec <=n_z_wave ) then 
                         image_mem_hz(i_z,i_mx_image,i_hx,back_i_t_lag)=&
                         image_mem_hz(i_z,i_mx_image,i_hx,back_i_t_lag)+&
                         wave_io(i_z+i_z_0_image+i_hx_shift_sou,i_mx_wave)* &
                         wave_block(i_z+i_z_0_image+i_hx_shift_rec, &
                                    i_mx_wave,i_my,1,1, wavepar_itstep1)
                      end if
                    end if
                 end do
               end do
             end do
            end if


!             do i_thread=1,n_threads
!               do i_mx_wave=i_mx_wave_beg,i_mx_wave_end
!                 i_mx_image=i_mx_wave+i_mx_0_image-1 
!                 do i_z=1,n_z_image
!                    image_mem(i_z,i_mx_image,i_hx,i_t_lag-i_t_lag_beg+1)=&
!                    image_mem(i_z,i_mx_image,i_hx,i_t_lag-i_t_lag_beg+1)+&
!                    wave_io(i_z,i_mx_wave+i_hx_shift_sou)* &
!                    wave_block(wave_par%i_t_step(1), &
!                       i_z,i_mx_wave+i_hx_shift_rec,i_my,1,1)      
!                 end do
!               end do
!             end do

!$omp end parallel

         !}
         else
         !{
           do i_mx_wave=i_mx_wave_beg,i_mx_wave_end
             i_mx_image=i_mx_wave+i_mx_0_image-1
             do i_z=1,n_z_image
                image_io(i_z,i_mx_image)=image_io(i_z,i_mx_image) + &
                wave_io(i_z+i_z_0_image-1,i_mx_wave+i_hx_shift_sou)* &
                wave_block(i_z+i_z_0_image-1,i_mx_wave+i_hx_shift_rec,i_my, &
                    1,1,wave_par%i_t_step(1))


             end do
           end do
         !}
         end if

         if (image_in_memory ) then
         !{
           if(write_image) then
              image_io(:,:) =image_mem(:,:,i_hx,i_t_lag-i_t_lag_beg+1)
              image_mem(:,:,i_hx,i_t_lag-i_t_lag_beg+1)=0.
              if (write_image_hz) then
                image_io_hz(:,:) =image_mem_hz(:,:,i_hx,i_t_lag-i_t_lag_beg+1)
                image_mem_hz(:,:,i_hx,i_t_lag-i_t_lag_beg+1)=0.
              end if
              if(srite(trim(tag_image),image_io,n_bytes_image_io) /= &
                n_bytes_image_io ) then
                write(0,*)'tag_image=',trim(tag_image)
                call seperr('Problems with writing tag_image file')
              end if
              if (write_image_hz) then
                if(srite(trim(tag_image_hz),image_io_hz,n_bytes_image_io) /= &
                  n_bytes_image_io ) then
                  write(0,*)'tag_image_hz=',trim(tag_image_hz)
                  call seperr('Problems with writing tag_image_hz file')
                end if
              end if
           end if 
         !}
         else
         !{
           if(srite(trim(tag_image),image_io,n_bytes_image_io) /= &
                n_bytes_image_io ) then
             write(0,*)'tag_image=',trim(tag_image)
             call seperr('Problems with writing tag_image file')
           end if
         !}
         end if

         if(sum_image_file .and. (i_hx /= image_par%n_hx)) then
         !{
           if(image_par%write_image_progress) then
           !{
             nsseek_blocks=-image_par%n_t_lag*image_par%n_my*image_par%n_hx* &
                       image_par%n_hy
!               ret_sseek_ok= (i_t_image-2)* &
             ret_sseek_ok= (i_t_lag - i_t_lag_beg+ &
                            (i_t_image-2)*image_par%n_t_lag)* &
                            image_par%n_my*image_par%n_hx *image_par%n_hy + &
                            (i_hx)* &
                            image_par%n_my*image_par%n_hy 
           !}
           else
           !{
             nsseek_blocks= 0
             ret_sseek_ok= (i_t_lag -i_t_lag_beg)* &
                            image_par%n_my*image_par%n_hx*image_par%n_hy + &
                           (i_hx)* &
                            image_par%n_my*image_par%n_hy 
           !}
           end if

           ret_sseek=sseek_block(trim(tag_image),nsseek_blocks,n_bytes_image_io,1) 
           if (ret_sseek .ne. ret_sseek_ok) then
           !{
             write(0,*)'nsseek_blocks=',nsseek_blocks
             write(0,*)'ret_sseek=',ret_sseek
             write(0,*)'ret_sseek_ok=',ret_sseek_ok
             write(0,*)'tag_image=',trim(tag_image)
             call seperr("Error block seeking image")
           !}
           end if
           if (write_image_hz) then
           !{
             ret_sseek=sseek_block(trim(tag_image_hz),nsseek_blocks,n_bytes_image_io,1) 
             if (ret_sseek .ne. ret_sseek_ok) then
             !{
               write(0,*)'nsseek_blocks=',nsseek_blocks
               write(0,*)'ret_sseek=',ret_sseek
               write(0,*)'ret_sseek_ok=',ret_sseek_ok
               write(0,*)'tag_image_hz=',trim(tag_image_hz)
               call seperr("Error block seeking image")
             !}
             end if
           !}
           end if
         !}
         end if


       !}
       end do
     !}
     end do
   !}
   end do

   !}
   end do

   return	
  end subroutine output_image


  !-----------------------------------------------------------------------------
  subroutine input_image(gen_par,image_par,rev_extrap_par, &
        image_mem,image_mem_hz)

   implicit none

   type (genrev3d_param_type) gen_par
   type (image_param_type) image_par
   type (rev_extrap_param_type) :: rev_extrap_par
   real, dimension (:,:,:,:) :: image_mem,image_mem_hz 

   character(len=128) :: tag_image
   integer :: len_tag_image,i_hx,i_mx_image,i_z


   tag_image=trim(image_par%tag)
   len_tag_image=len_trim(tag_image)


   if(.not. image_in_memory) then
     call seperr('input_image must be with image_in_memory=y')
   end if

   if(gen_par%debug2) write(0,*)'Reading from ',tag_image(1:len_tag_image)
   do i_hx=1,image_par%n_hx
       if(sreed(trim(tag_image),image_io,n_bytes_image_io) /= &
            n_bytes_image_io ) then
         write(0,*)'tag_image=',trim(tag_image)
         write(0,*)'i_hx=',i_hx
         call seperr('Problems with reading tag_image file')
       end if
       do i_mx_image=1,image_par%n_mx
         do i_z=1,n_z_image
           image_mem(i_z,i_mx_image,i_hx,1)=image_io(i_z,i_mx_image)
         end do
       end do
   end do

    return	
  end subroutine input_image

  !-----------------------------------------------------------------------------

  subroutine input_image_lin(gen_par,image_par,wave_par, dslow_block)

   implicit none

   type (genrev3d_param_type) gen_par
   type (image_param_type) image_par
!   type (rev_extrap_param_type) :: rev_extrap_par
   type (wave_param_type) :: wave_par
   real, dimension (:,:,:,:,:,:) :: dslow_block

   character(len=128) :: tag_image
   integer :: len_tag_image,i_hx,i_mx_image,i_mx_wave,i_z
!   integer :: i_thread,n_threads
   integer :: i_mx_0_image,i_mx_wave_beg,i_mx_wave_end
   integer :: i_hx_shift_sou,i_hx_shift_rec

!   n_threads=rev_extrap_par%n_threads
   tag_image=trim(image_par%tag)
   len_tag_image=len_trim(tag_image)

   i_mx_0_image=nint((wave_par%mx_0-image_par%mx_0)/image_par%d_mx)+1

   i_mx_wave_beg=max(i_hx_0_image+1,-i_mx_0_image+2)
   i_mx_wave_end=min(wave_par%n_mx-(image_par%n_hx-i_hx_0_image), &
                     image_par%n_mx-i_mx_0_image+1)

   if(.not. image_in_memory) then
     call seperr('input_image_lin must be with image_in_memory=y')
   end if

   if(gen_par%debug2) write(0,*)'Reading from ',tag_image(1:len_tag_image)
   do i_hx=1,image_par%n_hx
   !{

     if(image_zero_offset) then
       i_hx_shift_sou=0
       i_hx_shift_rec=0
     else
       i_hx_shift_sou=i_hx_0_image-i_hx
       i_hx_shift_rec=-(i_hx_0_image-i_hx)
     end if
       
     if(sreed(trim(tag_image),image_io,n_bytes_image_io) /= &
          n_bytes_image_io ) then
       write(0,*)'tag_image=',trim(tag_image)
       write(0,*)'i_hx=',i_hx
       call seperr('Problems with reading tag_image file')
     end if

!!$omp parallel default (private), shared (dslow_block,image_io), &
!!$omp& shared(n_threads,i_mx_wave_beg,i_mx_wave_end,i_mx_0_image,n_z_image), &
!!$omp& shared(i_hx,i_hx_shift_rec)
!!$omp do

!     do i_thread=1,n_threads
!       do i_mx_wave=i_mx_wave_beg-1+i_thread,i_mx_wave_end,n_threads
!             dslow_block(i_z+i_z_0_image-1,i_mx_wave+i_hx_shift_rec,1,1,1,1)= &
       do i_mx_wave=i_mx_wave_beg,i_mx_wave_end
         i_mx_image=i_mx_wave+i_mx_0_image-1 
         do i_z=1,n_z_image
             dslow_block(i_z+i_z_0_image-1,i_mx_wave,1,1,1,1)= &
                image_io(i_z,i_mx_image)
         end do
       end do
!     end do

!!$omp end parallel

   !}
   end do

   return	
  end subroutine input_image_lin

  !-----------------------------------------------------------------------------

  subroutine add_output_image_lin(gen_par,image_par,rev_extrap_par,wave_par, &
                dslow_block,image_mem,image_mem_hz)

   implicit none

   type (genrev3d_param_type) gen_par
   type (image_param_type) image_par
   type (rev_extrap_param_type) :: rev_extrap_par
   type (wave_param_type) :: wave_par
   real, dimension (:,:,:,:,:,:) :: dslow_block
   real, dimension (:,:,:,:) :: image_mem,image_mem_hz 

   character(len=128) :: tag_image
   logical :: write_image,sum_image
   integer :: len_tag_image,i_hx,i_mx_image,i_mx_wave,i_z
!   integer :: i_thread,n_threads
   integer :: i_mx_0_image,i_mx_wave_beg,i_mx_wave_end

!   n_threads=rev_extrap_par%n_threads
   tag_image=trim(image_par%tag)
   len_tag_image=len_trim(tag_image)

   i_mx_0_image=nint((wave_par%mx_0-image_par%mx_0)/image_par%d_mx)+1

   i_mx_wave_beg=max(i_hx_0_image+1,-i_mx_0_image+2)
   i_mx_wave_end=min(wave_par%n_mx-(image_par%n_hx-i_hx_0_image), &
                     image_par%n_mx-i_mx_0_image+1)

   if(.not. image_in_memory) then
     call seperr('add_output_image_lin must be with image_in_memory=y')
   end if

!   if(gen_par%debug2) write(0,*)'Writing to ',tag_image(1:len_tag_image)
!   if(gen_par%debug2) write(0,*)'maxval(dslow_block)=',maxval(dslow_block)
!   if(gen_par%debug2) write(0,*)'i_mx_wave_beg=',i_mx_wave_beg
!   if(gen_par%debug2) write(0,*)'i_mx_wave_end=',i_mx_wave_end
!   if(gen_par%debug2) write(0,*)'i_mx_0_image=',i_mx_0_image
!   write(0,*)'i_mx_wave_beg=',i_mx_wave_beg
!   write(0,*)'i_mx_wave_end=',i_mx_wave_end
!   write(0,*)'i_mx_0_image=',i_mx_0_image

   write_image = (gen_par%surf_offset_cig .or.  &
                  gen_par%write_image_progress_with_shots .or. &
                  rev_extrap_par%last_shot)

   sum_image = .not. gen_par%surf_offset_cig

   if(gen_par%debug2) write(0,*)'sum_image=',sum_image
   if(gen_par%debug2) write(0,*)'write_image=',write_image

   if(.not. sum_image) image_mem=0.

   do i_hx=1,image_par%n_hx
   !{
     do i_mx_wave=i_mx_wave_beg,i_mx_wave_end
       i_mx_image=i_mx_wave+i_mx_0_image-1 
       do i_z=1,n_z_image
         image_mem(i_z,i_mx_image,i_hx,1)= image_mem(i_z,i_mx_image,i_hx,1) + &
           dslow_block(i_z+i_z_0_image-1,i_mx_wave,i_hx,1,1,1)
       end do
     end do

     if(write_image) then
     !{
       image_io(:,:)=image_mem(:,:,i_hx,1)
       if(srite(trim(tag_image),image_io,n_bytes_image_io) /= &
            n_bytes_image_io ) then
         write(0,*)'tag_image=',trim(tag_image)
         write(0,*)'i_hx=',i_hx
         call seperr('Problems with writing tag_image file')
       end if
     !}
     end if
   !}
   end do

   return	
  end subroutine add_output_image_lin


  !-----------------------------------------------------------------------------
  subroutine add_image(gen_par,image_par,wave_par,rev_extrap_par,wavelet_par, &
                         wave_block,wavelet,image_mem,image_mem_hz)

! record to memory.

   implicit none

   character(len=128) :: tag_wave
   type (genrev3d_param_type) gen_par
   type (image_param_type) image_par
   type (wave_param_type) :: wave_par
   type (rev_extrap_param_type) :: rev_extrap_par
   type (wavelet_param_type) wavelet_par

   real, dimension (:,:,:,:,:,:)   :: wave_block
   real,  dimension (:,:,:,:)   :: wavelet
   real, dimension (:,:,:,:) :: image_mem,image_mem_hz 

   integer :: len_tag_wave,i_hx,i_hy,i_my,i_mx_wave,i_mx_image,i_z
   integer :: i_hx_shift_sou,i_hx_shift_rec,i_hx_shift
   integer :: i_t_image,i_t_wave,i_t_wavelet
   integer :: i_t_lag_beg,i_t_lag_end,i_t_lag
   integer :: nsseek_blocks,ret_sseek,block_wave,ret_sseek_ok,ret_io_bytes
   integer :: i_mx_ref,i_z_ref,min_i_z_image,max_i_z_image

   real :: z_ref,mx_ref,z_ref_0,mx_ref_0,w_mx,w_z
   real :: sin_dip_refl,cos_dip_refl,hx_shift
   real :: z0_dip_refl

   if(.not. image_in_memory) then
     call seperr('add_image must be with image_in_memory=y')
   end if

   tag_wave=trim(wave_par%tag_back)
   len_tag_wave=len_trim(tag_wave)

   i_mx_0_image=nint((wave_par%mx_0-image_par%mx_0)/image_par%d_mx)+1
   i_my_0_image=nint((wave_par%my_0-image_par%my_0)/image_par%d_my)+1
!write(0,*)'i_mx_0_image=',i_mx_0_image
!write(0,*)'i_my_0_image=',i_my_0_image

   i_mx_wave_beg=max(i_hx_0_image+1,-i_mx_0_image+2)
   i_mx_wave_end=min(wave_par%n_mx-(image_par%n_hx-i_hx_0_image), &
                     image_par%n_mx-i_mx_0_image+1)

! changed to be safer with interpolation, could change again
!   i_mx_wave_beg=max(i_hx_0_image+2,-i_mx_0_image+3)
!   i_mx_wave_end=min(wave_par%n_mx-(image_par%n_hx-i_hx_0_image)-1, &
!                     image_par%n_mx-i_mx_0_image)

!write(0,*)'i_mx_wave_beg=',i_mx_wave_beg
!write(0,*)'i_mx_wave_end=',i_mx_wave_end


   if (gen_par%domain_range(1:8) == 'image_da') then
    !{
! read source wavefield
     i_t_wave=nint((rev_extrap_par%t - image_par%t_0)/image_par%d_t)+1
  
     i_t_lag_beg=1
     i_t_lag_end=1
     i_t_lag=1
  
     i_t_image= image_par%n_t - &
                i_t_lag_end - &
               (nint((rev_extrap_par%t - image_par%t_0)/image_par%d_t))
  
     do i_hy=1,image_par%n_hy
     !{
       do i_my=1,image_par%n_my
       !{
  
!         block_wave= (i_t_wave-1+i_t_lag)*wave_par%n_my BB 2/1/2006
         block_wave= ((i_t_wave-1)+(i_t_lag-1))*wave_par%n_my
         nsseek_blocks= block_wave + (i_my-1)*wave_par%n_hx
  !     write(0,*)'i_t_wave=',i_t_wave
  !     write(0,*)'i_t_lag=',i_t_lag
  !     write(0,*)'block_wave=',block_wave
  !     write(0,*)'nsseek_blocks (wave)=',nsseek_blocks
  
         ret_sseek=sseek_block(trim(tag_wave),nsseek_blocks,n_bytes_wave_io,0) 
         if (ret_sseek .ne. nsseek_blocks) then
         !{
           write(0,*)'nsseek_blocks=',nsseek_blocks
           write(0,*)'ret_sseek=',ret_sseek
           write(0,*)'tag_wave=',trim(tag_wave)
           call seperr("Error block seeking wave")
         !}
         end if
         ret_io_bytes=sreed(trim(tag_wave),wave_io,n_bytes_wave_io)
         if(ret_io_bytes /=n_bytes_wave_io) then
           write(0,*)'ret_io_bytes=',ret_io_bytes
           write(0,*)'n_bytes_wave_io=',n_bytes_wave_io
           write(0,*)'tag_wave=',trim(tag_wave)
           call seperr('Problems with reading tag_wave file')
         end if
  
       !}
       end do
     !}
     end do

! Now apply adjoint of imaging

!write(0,*)'i_t_step=',wave_par%i_t_step
     do i_mx_wave=i_mx_wave_beg,i_mx_wave_end
       i_mx_image=i_mx_wave+i_mx_0_image-1
!write(0,*)'i_mx_image=',i_mx_image
!write(0,*)'i_mx_wave=',i_mx_wave
       do i_hx=1,image_par%n_hx
         if(image_zero_offset) then
           i_hx_shift_sou=0
           i_hx_shift_rec=0
         else
           i_hx_shift_sou=i_hx_0_image-i_hx
           i_hx_shift_rec=-(i_hx_0_image-i_hx)
         end if

!write(0,*)'i_hx=',i_hx
!write(0,*)'i_hx_shift_sou=',i_hx_shift_sou
!write(0,*)'i_hx_shift_rec=',i_hx_shift_rec
       do i_z=1,n_z_image
!write(0,*)'i_z=',i_z


          wave_block(i_z+i_z_0_image-1,i_mx_wave+i_hx_shift_rec, &
                             1,1,1,wave_par%i_t_step(1))= &
          wave_block(i_z+i_z_0_image-1,i_mx_wave+i_hx_shift_rec, &
                             1,1,1,wave_par%i_t_step(1)) + &
          image_mem(i_z,i_mx_image,i_hx,1)*wave_io(i_z+i_z_0_image-1,i_mx_wave+i_hx_shift_sou)

        end do
      end do
    end do
   !}
   else
   !{

! Now apply adjoint of imaging
   select case (rev_extrap_par%sou_rec_wave)
     case (SOU_WAVE)
       i_t_wavelet=nint((rev_extrap_par%t + wavelet_par%t_0)/(-wavelet_par%d_t))+1
       write(0,*)'i_t_wavelet=',i_t_wavelet
     case (REC_WAVE)
       i_t_wavelet=nint((rev_extrap_par%t - wavelet_par%t_0)/wavelet_par%d_t)+1
       write(0,*)'i_t_wavelet=',i_t_wavelet
     case default
       write(0,*)'sou_rec_wave=',rev_extrap_par%sou_rec_wave
       call seperr('Not supported sou_rec_wave')
   end select

   do i_mx_wave=i_mx_wave_beg,i_mx_wave_end
     i_mx_image=i_mx_wave+i_mx_0_image-1
!write(0,*)'i_mx_image=',i_mx_image
!write(0,*)'i_mx_wave=',i_mx_wave
     do i_hx=1,image_par%n_hx
       if(image_zero_offset) then
         i_hx_shift=0
       else
         select case (rev_extrap_par%sou_rec_wave)
         case (SOU_WAVE)
           i_hx_shift=i_hx_0_image-i_hx
         case (REC_WAVE)
           i_hx_shift=-(i_hx_0_image-i_hx)
         case default
           write(0,*)'sou_rec_wave=',rev_extrap_par%sou_rec_wave
           call seperr('Not supported sou_rec_wave')
         end select
       end if
       if(gen_par%transf_geol_offset) then
         hx_shift=cos_dip_max*(i_hx_shift*image_par%d_hx)
       else
         hx_shift=(i_hx_shift*image_par%d_hx)
       end if
!write(0,*)'i_mx_wave+i_hx_shift=',i_mx_wave+i_hx_shift
!write(0,*)'i_mx_image=',i_mx_image
!write(0,*)'i_mx_ref=',i_mx_ref
!write(0,*)'w_mx=',w_mx
!write(0,*)'i_hx=',i_hx
!write(0,*)'i_hx_shift=',i_hx_shift
!write(0,*)'hx_shift=',hx_shift
!  !$omp parallel default (shared), &
!  !$omp& private(i_z)
!  !$omp do

!       min_i_z_image=max(1,nint(sin_dip_max*hx_shift/image_par%d_z)+1)
!       max_i_z_image=min(n_z_image,n_z_image-nint(sin_dip_max*hx_shift/image_par%d_z)-1)

       min_i_z_image=max(1,nint((gen_par%refl_z0 + & 
          sin_dip_max*hx_shift + &
          cos_dip_max*(image_par%z_0 - gen_par%refl_z0)) &
         /image_par%d_z+1))

       max_i_z_image=min(n_z_image,nint((gen_par%deep_refl_z0 + & 
          sin_dip_max*hx_shift + &
          cos_dip_max*(n_z_image*image_par%d_z - gen_par%deep_refl_z0)) &
         /image_par%d_z-1))

       min_i_z_image=1
       max_i_z_image=n_z_image

!write(0,*)'min_i_z_image=',min_i_z_image
!write(0,*)'max_i_z_image=',max_i_z_image
       do i_z=min_i_z_image,max_i_z_image
       !{
!write(0,*)'i_z=',i_z

         if(i_z < i_z_deep_refl) then
           sin_dip_refl=sin_dip
           cos_dip_refl=cos_dip
           z0_dip_refl=gen_par%refl_z0
         else
           sin_dip_refl=sin_deep_dip
           cos_dip_refl=cos_deep_dip
           z0_dip_refl=gen_par%deep_refl_z0
         end if
!write(0,*)'z0_dip_refl=',z0_dip_refl

       if(gen_par%transf_geol_offset) then
         hx_shift=cos_dip_refl*i_hx_shift*image_par%d_hx
       else
         hx_shift=i_hx_shift*image_par%d_hx
       end if

         z_ref_0=((i_z+i_z_0_image-2)*wave_par%d_z + wave_par%z_0) 
         mx_ref_0=((i_mx_wave-1)*wave_par%d_mx+wave_par%mx_0)

         if(sin_dip_refl /= 0.) then
           if((abs(hx_shift) <= 10.1) .and.  &
              (mx_ref_0 == 1940) .and. &
              (abs(z0_dip_refl - z_ref_0) <=10.1) .and. &
               (i_t_wavelet == 1) &
              )  then 
write(0,*)'hx_shift=',hx_shift
write(0,*)'Before rotation: z_ref_0=',z_ref_0
write(0,*)'Before rotation: mx_ref_0=',mx_ref_0
           end if

         mx_ref=mx_ref_0 + cos_dip_refl*hx_shift - &
                           sin_dip_refl*(z_ref_0-z0_dip_refl)
!         z_ref =z_ref_0  + sin_dip_refl*hx_shift + &
         z_ref =z0_dip_refl  + sin_dip_refl*hx_shift + &
                           cos_dip_refl*(z_ref_0-z0_dip_refl)

           if((abs(hx_shift) <= 10.1) .and.  &
              (abs(z0_dip_refl - z_ref_0) <=10.1) .and. &
              (mx_ref_0 == 1940) .and. &
               (i_t_wavelet == 1) &
              )  then 
!write(0,*)'hx_shift=',hx_shift
write(0,*)'After rotation: delta_z_ref=',(z_ref-z_ref_0)
write(0,*)'After rotation: delta_mx_ref=',(mx_ref-mx_ref_0)
           end if
         else
           mx_ref=mx_ref_0 + hx_shift 
           z_ref=z_ref_0
         end if

         z_ref=(z_ref-wave_par%z_0)/wave_par%d_z +1
         i_z_ref=int(z_ref)
         w_z=z_ref-i_z_ref

         mx_ref=(mx_ref-wave_par%mx_0)/wave_par%d_mx+1
         i_mx_ref=int(mx_ref)
         w_mx=mx_ref-i_mx_ref

!write(0,*)'i_z_ref=',i_z_ref
!write(0,*)'i_mx_ref=',i_mx_ref
!write(0,*)'w_z=',w_z

         if ((i_z_ref >=1) .and.(i_z_ref < wave_par%n_z) .and. &
             (i_mx_ref >=1) .and.(i_mx_ref < wave_par%n_mx)) then
         !{

           wave_block(i_z_ref  ,i_mx_ref  , 1,1,1,wave_par%i_t_step(1))= &
           wave_block(i_z_ref  ,i_mx_ref  , 1,1,1,wave_par%i_t_step(1))+ &
              wavelet(i_t_wavelet,1,1,1)* &
              image_mem(i_z,i_mx_image,i_hx,1)*(1.-w_z)*(1.-w_mx)
  
           wave_block(i_z_ref+1,i_mx_ref  , 1,1,1,wave_par%i_t_step(1))= &
           wave_block(i_z_ref+1,i_mx_ref  , 1,1,1,wave_par%i_t_step(1))+ &
              wavelet(i_t_wavelet,1,1,1)* &
              image_mem(i_z,i_mx_image,i_hx,1)*(w_z)*(1.-w_mx)
  
           wave_block(i_z_ref  ,i_mx_ref+1, 1,1,1,wave_par%i_t_step(1))= &
           wave_block(i_z_ref  ,i_mx_ref+1, 1,1,1,wave_par%i_t_step(1))+ &
              wavelet(i_t_wavelet,1,1,1)* &
              image_mem(i_z,i_mx_image,i_hx,1)*(1.-w_z)*(w_mx)
  
           wave_block(i_z_ref+1,i_mx_ref+1, 1,1,1,wave_par%i_t_step(1))= &
           wave_block(i_z_ref+1,i_mx_ref+1, 1,1,1,wave_par%i_t_step(1))+ &
              wavelet(i_t_wavelet,1,1,1)* &
              image_mem(i_z,i_mx_image,i_hx,1)*(w_z)*(w_mx)

!         wave_block(wave_par%i_t_step(1),i_z+i_z_0_image,i_mx_wave+i_hx_shift, &
!                             1,1,1)= &
!         wave_block(wave_par%i_t_step(1),i_z+i_z_0_image,i_mx_wave+i_hx_shift, &
!                             1,1,1) + &
!         image_mem(i_z,i_mx_image,i_hx,1)

         !}
         end if

!!write(0,*)'#2 after'
       !}
       end do
!!$omp end parallel
      end do
    end do

   !}
   end if

    return	
  end subroutine add_image

  !-----------------------------------------------------------------------------


  !-----------------------------------------------------------------------------
  subroutine clean_io_image()

    integer ierr

    if(allocated(image_io)) then
       deallocate (image_io,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating image_io')
    end if
    if (write_image_hz) then
      if(allocated(image_io_hz)) then
       deallocate (image_io_hz,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating image_io_hz')
      end if
    end if

    return
  end subroutine clean_io_image


end module image_mod
