subroutine param_gen(gen_par)
  use genrev3d_types
  use sep

  implicit none
  type (genrev3d_param_type) gen_par

  !integer, intrinsic ::  mp_numthreads
!  integer ::  mp_numthreads
  integer ::  omp_get_max_threads
  integer, external :: auxpar,getch,fetch,auxin,sreed,sseek,auxout

  integer n_threads_env

!write(0,*)'#21'
!pause 

  call from_param('source_type',gen_par%source_type,trim(gen_par%source_type))

  call from_param('oper_type',gen_par%oper_type,trim(gen_par%oper_type))

  call from_param('domain_range',gen_par%domain_range,trim(gen_par%domain_range))
  call from_param('verb',gen_par%verb,gen_par%verb)
  call from_param('debug1',gen_par%debug1,gen_par%debug1)
  call from_param('debug2',gen_par%debug2,gen_par%debug2)

  call from_param('timers_on',gen_par%timers_on,gen_par%timers_on)
  call from_param('image_zero_offset',gen_par%image_zero_offset,gen_par%image_zero_offset)
  call from_param('image_zero_time',gen_par%image_zero_time,gen_par%image_zero_time)
  call from_param('surf_offset_cig',gen_par%surf_offset_cig,gen_par%surf_offset_cig)
  call from_param('write_image_progress_with_shots',gen_par%write_image_progress_with_shots,gen_par%write_image_progress_with_shots)

  call from_param('first_adj',gen_par%first_adj,gen_par%first_adj)
  call from_param('first_add',gen_par%first_add,gen_par%first_add)

  call from_param('check_stability',gen_par%check_stability,gen_par%check_stability)
  call from_param('layered_mig_vel',gen_par%layered_mig_vel,gen_par%layered_mig_vel)
  call from_param('skip_source_prop',gen_par%skip_source_prop,gen_par%skip_source_prop)
  call from_param('apply_bound',gen_par%apply_bound,gen_par%apply_bound)

  gen_par%inject_data=.false.
  call from_param('inject_data',gen_par%inject_data,gen_par%inject_data)

  call from_param('n_threads',gen_par%n_threads,gen_par%n_threads)
  call from_param('max_memory',gen_par%max_memory,gen_par%max_memory)

  call from_param('n_data_dim',gen_par%n_data_dim,gen_par%n_data_dim)
  call from_param('slow_const',gen_par%slow_const,gen_par%slow_const)

  call from_param('sou_depth',gen_par%sou_depth,gen_par%sou_depth)
  call from_param('rec_depth',gen_par%rec_depth,gen_par%rec_depth)

  call from_param('refl_dip',gen_par%refl_dip,gen_par%refl_dip)
  call from_param('deep_refl_depth',gen_par%deep_refl_depth,gen_par%deep_refl_depth)
  if(gen_par%deep_refl_depth /= deep_refl_depth_init) then
    call from_param('deep_refl_dip',gen_par%deep_refl_dip,gen_par%deep_refl_dip)
    call from_param('deep_refl_z0',gen_par%deep_refl_z0)
  end if
  if((gen_par%refl_dip /= refl_dip_init) .or. &
     (gen_par%deep_refl_dip /= deep_refl_dip_init)) then
    call from_param('transf_geol_offset',gen_par%transf_geol_offset,gen_par%transf_geol_offset)
    call from_param('refl_z0',gen_par%refl_z0)
  end if


  ! checking parameters
  n_threads_env=omp_get_max_threads()
  if(gen_par%n_threads == n_threads_init) then
     gen_par%n_threads=n_threads_env
  else
     if(n_threads_env > gen_par%n_threads) then
        call omp_set_num_threads(gen_par%n_threads)
        write(0,*)'WARNING: overwriting environment variable'
        write(0,*)'gen_par-n_threads=',gen_par%n_threads
     end if
     if(n_threads_env < gen_par%n_threads) then
        write(0,*)'gen_par-n_threads=',gen_par%n_threads
        write(0,*)'OMP_NUM_THREADS=',n_threads_env
        call seperr('ERROR: Asked for more threads than the environment variable allows')
     end if
  end if
  if(gen_par%debug1)write(0,*)'gen_par-n_threads=',gen_par%n_threads
  call to_history('n_threads',gen_par%n_threads)

  !converting source_type to standard strings
  select case (gen_par%source_type(1:6))
  case ('point_')
     gen_par%source_type(1:12)='point_source'
  case ('global')
     gen_par%source_type(1:13)='global_source'
  case default
     write(0,*)'gen_par-source_type=',trim(gen_par%source_type)
     call seperr('Not supported source_type')
  end select

  !coverting domain_range to standard strings
  select case (gen_par%domain_range(1:8))
  case ('image_da')
     gen_par%domain_range(1:10)='image_data'
    if(gen_par%verb) then
      write(0,*)'gen_par-domain_range=',trim(gen_par%domain_range(1:10))
    end if
  case ('lin_imag')
     gen_par%domain_range(1:14)='lin_image_data'
    if(gen_par%verb) then
      write(0,*)'gen_par-domain_range=',trim(gen_par%domain_range(1:14))
    end if
  case ('image_gl')
     gen_par%domain_range(1:17)='image_global_data'
    if(gen_par%verb) then
      write(0,*)'gen_par-domain_range=',trim(gen_par%domain_range(1:17))
    end if
  case ('vel_data')
     gen_par%domain_range(1:8)='vel_data'
    if(gen_par%verb) then
      write(0,*)'gen_par-domain_range=',trim(gen_par%domain_range(1:8))
    end if
  case default
     write(0,*)'gen_par-domain_range=',trim(gen_par%domain_range)
     call seperr('Not supported domain_range')
  end select

  return
end subroutine param_gen


subroutine param_data(gen_par,data_par)
  use genrev3d_types
  use data_types
  use sep

  implicit none

  type (genrev3d_param_type) gen_par
  type (data_param_type) data_par

  logical write_down_data,write_down_ddata
  character(len=128) :: tag_name
  integer, external :: auxpar,getch,fetch,auxin,sreed,sseek,auxout,auxinout
  integer ierr
  integer n1,n2,n3,n4,n5,n,esize
  real d1,d2,d3,d4,d5,o1,o2,o3,o4,o5,d,o,z_data_0

  data_par%esize=4

  call from_param('tag_data',data_par%tag,'Data')
  call from_param(trim(data_par%tag),tag_name,'Data not found')

  ! First deal with tag Data

  select case (gen_par%domain_range(1:8))
  !{
  case ('image_da')
  !{

   if (.not. gen_par%first_adj) then
   !{
      ierr=auxinout(trim(data_par%tag))
      data_par%file=DATA_OUTPUT

      call from_param('n_t',data_par%n_t)
      call from_param('n_sx',data_par%n_sx)
      call from_param('n_sy',data_par%n_sy)
      call from_param('n_off_x',data_par%n_off_x)

      call from_param('d_t',data_par%d_t)
      call from_param('d_sx',data_par%d_sx)
      call from_param('d_sy',data_par%d_sy)
      call from_param('d_off_x',data_par%d_off_x)

      call from_param('t_0',data_par%t_0,0.)
      call from_param('sx_0',data_par%sx_0,0.)
      call from_param('sy_0',data_par%sy_0,0.)
      call from_param('off_x_0',data_par%off_x_0,0.)

      if(gen_par%n_data_dim >4) then 
         call from_param('n_off_y',data_par%n_off_y)
         call from_param('d_off_y',data_par%d_off_y)
         call from_param('off_y_0',data_par%off_y_0,0.)
      else
         data_par%n_off_y=1
         call from_param('d_off_y',data_par%d_off_y,1.)
         call from_param('off_y_0',data_par%off_y_0,0.)
      end if

      call to_history("title",trim(data_par%tag),trim(data_par%tag))

      call to_history("esize",data_par%esize,trim(data_par%tag))
      call to_history("n1",data_par%n_off_x,trim(data_par%tag))
      call to_history("d1",data_par%d_off_x,trim(data_par%tag))
      call to_history("o1",data_par%off_x_0,trim(data_par%tag))

      if(gen_par%n_data_dim >4) then 
         call to_history("n2",data_par%n_off_y,trim(data_par%tag))
         call to_history("n3",data_par%n_t,trim(data_par%tag))
         call to_history("n4",data_par%n_sx,trim(data_par%tag))
         call to_history("n5",data_par%n_sy,trim(data_par%tag))

         call to_history("o2",data_par%off_y_0,trim(data_par%tag))
         call to_history("o3",data_par%t_0,trim(data_par%tag))
         call to_history("o4",data_par%sx_0,trim(data_par%tag))
         call to_history("o5",data_par%sy_0,trim(data_par%tag))

         call to_history("d2",data_par%d_off_y,trim(data_par%tag))
         call to_history("d3",data_par%d_t,trim(data_par%tag))
         call to_history("d4",data_par%d_sx,trim(data_par%tag))
         call to_history("n5",data_par%d_sy,trim(data_par%tag))
      else
         call to_history("n2",data_par%n_t,trim(data_par%tag))
         call to_history("n3",data_par%n_sx,trim(data_par%tag))
         call to_history("n4",data_par%n_sy,trim(data_par%tag))

         call to_history("d2",data_par%d_t,trim(data_par%tag))
         call to_history("d3",data_par%d_sx,trim(data_par%tag))
         call to_history("d4",data_par%d_sy,trim(data_par%tag))

         call to_history("o2",data_par%t_0,trim(data_par%tag))
         call to_history("o3",data_par%sx_0,trim(data_par%tag))
         call to_history("o4",data_par%sy_0,trim(data_par%tag))
      end if
      call to_history("z_data_0",gen_par%rec_depth,trim(data_par%tag))
   !}
   else
   !{
     ierr=auxinout(trim(data_par%tag))
     data_par%file=DATA_INPUT

     call from_aux(trim(data_par%tag),"esize",esize)
     if(esize /= data_par%esize) call seperr('Data must be real')

     if (gen_par%source_type(1:6) == "point_") then
     !{

       call from_aux(trim(data_par%tag),"n1",data_par%n_off_x)
       call to_history("From Data: n1",data_par%n_off_x)
       call from_aux(trim(data_par%tag),"o1",data_par%off_x_0,0.)
       call to_history("From Data: o1",data_par%off_x_0)
       call from_aux(trim(data_par%tag),"d1",data_par%d_off_x)
       call to_history("From Data: d1",data_par%d_off_x)
  
       if(gen_par%n_data_dim >4) then 
         call from_aux(trim(data_par%tag),"n2",data_par%n_off_y)
         call to_history("From Data: n2",data_par%n_off_y)
         call from_aux(trim(data_par%tag),"n3",data_par%n_t)
         call to_history("From Data: n3",data_par%n_t)
         call from_aux(trim(data_par%tag),"n4",data_par%n_sx)
         call to_history("From Data: n4",data_par%n_sx)
         call from_aux(trim(data_par%tag),"n5",data_par%n_sy)
         call to_history("From Data: n5",data_par%n_sy)
  
         call from_aux(trim(data_par%tag),"d2",data_par%d_off_y)
         call to_history("From Data: d2",data_par%d_off_y)
         call from_aux(trim(data_par%tag),"d3",data_par%d_t)
         call to_history("From Data: d3",data_par%d_t)
         call from_aux(trim(data_par%tag),"d4",data_par%d_sx)
         call to_history("From Data: d4",data_par%d_sx)
         call from_aux(trim(data_par%tag),"d5",data_par%d_sy)
         call to_history("From Data: d5",data_par%d_sy)
  
         call from_aux(trim(data_par%tag),"o2",data_par%off_y_0)
         call to_history("From Data: o2",data_par%off_y_0)
         call from_aux(trim(data_par%tag),"o3",data_par%t_0)
         call to_history("From Data: o3",data_par%t_0)
         call from_aux(trim(data_par%tag),"o4",data_par%sx_0)
         call to_history("From Data: o4",data_par%sx_0)
         call from_aux(trim(data_par%tag),"o5",data_par%sy_0)
         call to_history("From Data: o5",data_par%sy_0)
       else
         call from_aux(trim(data_par%tag),"n2",data_par%n_t)
         call to_history("From Data: n2",data_par%n_t)
         call from_aux(trim(data_par%tag),"n3",data_par%n_sx)
         call to_history("From Data: n3",data_par%n_sx)
         call from_aux(trim(data_par%tag),"n4",data_par%n_sy)
         call to_history("From Data: n4",data_par%n_sy)
  
         call from_aux(trim(data_par%tag),"d2",data_par%d_t)
         call to_history("From Data: d2",data_par%d_t)
         call from_aux(trim(data_par%tag),"d3",data_par%d_sx)
         call to_history("From Data: d3",data_par%d_sx)
         call from_aux(trim(data_par%tag),"d4",data_par%d_sy)
         call to_history("From Data: d4",data_par%d_sy)
  
         call from_aux(trim(data_par%tag),"o2",data_par%t_0)
         call to_history("From Data: o2",data_par%t_0)
         call from_aux(trim(data_par%tag),"o3",data_par%sx_0)
         call to_history("From Data: o3",data_par%sx_0)
         call from_aux(trim(data_par%tag),"o4",data_par%sy_0)
         call to_history("From Data: o4",data_par%sy_0)
  
         data_par%n_off_y=1
         data_par%d_off_y=1.
         data_par%off_y_0=0.
  
       end if

       call from_aux(trim(data_par%tag),"z_data_0",data_par%z_0)
       call to_history("From Data: z_data_0",data_par%z_0)
       if(gen_par%rec_depth /= rec_depth_init) then
         if(gen_par%debug1) write(0,*)'ATTENTION: overwriting receiver depth'
         if(gen_par%debug1) write(0,*)'From data file receiver depth=',data_par%z_0
         if(gen_par%debug1) write(0,*)'From parameter file receiver depth=',gen_par%rec_depth
         data_par%z_0=gen_par%rec_depth 
         call to_history("From par: z_data_0",data_par%z_0)
       end if
     !}
     else
     !{


       call from_aux(trim(data_par%tag),"n1",data_par%n_mx)
       call to_history("From Data: n1",data_par%n_mx)
       call from_aux(trim(data_par%tag),"o1",data_par%mx_0)
       call to_history("From Data: o1",data_par%mx_0)
       call from_aux(trim(data_par%tag),"d1",data_par%d_mx)
       call to_history("From Data: d1",data_par%d_mx)
  
       if(gen_par%n_data_dim >4) then 
         call from_aux(trim(data_par%tag),"n2",data_par%n_my)
         call to_history("From Data: n2",data_par%n_my)
         call from_aux(trim(data_par%tag),"n3",data_par%n_t)
         call to_history("From Data: n3",data_par%n_t)
         call from_aux(trim(data_par%tag),"n4",data_par%n_s1_par)
         call to_history("From Data: n4",data_par%n_s1_par)
         call from_aux(trim(data_par%tag),"n5",data_par%n_s2_par)
         call to_history("From Data: n5",data_par%n_s2_par)
  
         call from_aux(trim(data_par%tag),"d2",data_par%d_my)
         call to_history("From Data: d2",data_par%d_my)
         call from_aux(trim(data_par%tag),"d3",data_par%d_t)
         call to_history("From Data: d3",data_par%d_t)
         call from_aux(trim(data_par%tag),"d4",data_par%d_s1_par)
         call to_history("From Data: d4",data_par%d_s1_par)
         call from_aux(trim(data_par%tag),"d5",data_par%d_s2_par)
         call to_history("From Data: d5",data_par%d_s2_par)
  
         call from_aux(trim(data_par%tag),"o2",data_par%my_0)
         call to_history("From Data: o2",data_par%my_0)
         call from_aux(trim(data_par%tag),"o3",data_par%t_0)
         call to_history("From Data: o3",data_par%t_0)
         call from_aux(trim(data_par%tag),"o4",data_par%s1_par_0)
         call to_history("From Data: o4",data_par%s1_par_0)
         call from_aux(trim(data_par%tag),"o5",data_par%s2_par_0)
         call to_history("From Data: o5",data_par%s2_par_0)
       else
         call from_aux(trim(data_par%tag),"n2",data_par%n_t)
         call to_history("From Data: n2",data_par%n_t)
         call from_aux(trim(data_par%tag),"n3",data_par%n_s1_par)
         call to_history("From Data: n3",data_par%n_s1_par)
         call from_aux(trim(data_par%tag),"n4",data_par%n_s2_par)
         call to_history("From Data: n4",data_par%n_s2_par)
  
         call from_aux(trim(data_par%tag),"d2",data_par%d_t)
         call to_history("From Data: d2",data_par%d_t)
         call from_aux(trim(data_par%tag),"d3",data_par%d_s1_par)
         call to_history("From Data: d3",data_par%d_s1_par)
         call from_aux(trim(data_par%tag),"d4",data_par%d_s2_par)
         call to_history("From Data: d4",data_par%d_s2_par)
  
         call from_aux(trim(data_par%tag),"o2",data_par%t_0)
         call to_history("From Data: o2",data_par%t_0)
         call from_aux(trim(data_par%tag),"o3",data_par%s1_par_0)
         call to_history("From Data: o3",data_par%s1_par_0)
         call from_aux(trim(data_par%tag),"o4",data_par%s2_par_0)
         call to_history("From Data: o4",data_par%s2_par_0)
  
         data_par%n_my=1
         data_par%d_my=1.
         data_par%my_0=0.
       end if

       call from_aux(trim(data_par%tag),"z_data_0",data_par%z_0)
       call to_history("From Data: z_data_0",data_par%z_0)
       if(gen_par%rec_depth /= rec_depth_init) then
         if(gen_par%debug1) write(0,*)'ATTENTION: overwriting receiver depth'
         if(gen_par%debug1) write(0,*)'From data file receiver depth=',data_par%z_0
         if(gen_par%debug1) write(0,*)'From parameter file receiver depth=',gen_par%rec_depth
         data_par%z_0=gen_par%rec_depth 
         call to_history("From par: z_data_0",data_par%z_0)
       end if


       call from_param('tag_data_sou',data_par%tag_sou,'Data_sou')
       call from_param(trim(data_par%tag_sou),tag_name,'Data_sou not found')

       ierr=auxinout(trim(data_par%tag_sou))

       call from_aux(trim(data_par%tag_sou),"esize",esize)
       if(esize /= data_par%esize) call seperr('Data_sou must be real')

       call from_aux(trim(data_par%tag_sou),"n1",n)
       call to_history("From Data_sou: n1",n)
       if(n /= data_par%n_mx) call seperr('Data_sou n_mx not consintent with Data n_mx')
       call from_aux(trim(data_par%tag_sou),"o1",o)
       call to_history("From Data_sou: o1",o)
       if(o /= data_par%mx_0) call seperr('Data_sou mx_0 not consintent with Data mx_0')
       call from_aux(trim(data_par%tag_sou),"d1",d)
       call to_history("From Data_sou: d1",d)
       if(d /= data_par%d_mx) call seperr('Data_sou d_mx not consintent with Data d_mx')
  
       if(gen_par%n_data_dim >4) then 

         call from_aux(trim(data_par%tag_sou),"n2",n)
         call to_history("From Data_sou: n2",n)
         if(n /= data_par%n_my) call seperr('Data_sou n_my not consintent with Data n_my')
         call from_aux(trim(data_par%tag_sou),"n3",n)
         call to_history("From Data_sou: n3",n)
         if(n /= data_par%n_t) call seperr('Data_sou n_t not consintent with Data n_t')
         call from_aux(trim(data_par%tag_sou),"n4",n)
         call to_history("From Data_sou: n4",n)
         if(n /= data_par%n_s1_par) call seperr('Data_sou n_s1_par not consintent with Data n_s1_par')
         call from_aux(trim(data_par%tag_sou),"n5",n)
         call to_history("From Data_sou: n5",n)
         if(n /= data_par%n_s2_par) call seperr('Data_sou n_s2_par not consintent with Data n_s2_par')

         call from_aux(trim(data_par%tag_sou),"d2",d)
         call to_history("From Data_sou: d2",d)
         if(d /= data_par%d_my) call seperr('Data_sou d_my not consintent with Data d_my')
         call from_aux(trim(data_par%tag_sou),"d3",d)
         call to_history("From Data_sou: d3",d)
         if(d /= data_par%d_t) call seperr('Data_sou d_t not consintent with Data d_t')
         call from_aux(trim(data_par%tag_sou),"d4",d)
         call to_history("From Data_sou: d4",d)
         if(d /= data_par%d_s1_par) call seperr('Data_sou d_s1_par not consintent with Data d_s1_par')
         call from_aux(trim(data_par%tag_sou),"d5",d)
         call to_history("From Data_sou: d5",d)
         if(d /= data_par%d_s2_par) call seperr('Data_sou d_s2_par not consintent with Data d_s2_par')

         call from_aux(trim(data_par%tag_sou),"o2",o)
         call to_history("From Data_sou: o2",o)
         if(o /= data_par%my_0) call seperr('Data_sou my_0 not consintent with Data my_0')
         call from_aux(trim(data_par%tag_sou),"o3",data_par%t_sou_0)
         call to_history("From Data_sou: o3",data_par%t_sou_0)
!         if(o /= data_par%t_0) call seperr('Data_sou t_0 not consintent with Data t_0')
         call from_aux(trim(data_par%tag_sou),"o4",o)
         call to_history("From Data_sou: o4",o)
         if(o /= data_par%s1_par_0) call seperr('Data_sou s1_par_0 not consintent with Data s1_par_0')
         call from_aux(trim(data_par%tag_sou),"o5",o)
         call to_history("From Data_sou: o5",o)
         if(o /= data_par%s2_par_0) call seperr('Data_sou s2_par_0 not consintent with Data s2_par_0')
  
       else

         call from_aux(trim(data_par%tag_sou),"n2",n)
         call to_history("From Data_sou: n2",n)
         if(n /= data_par%n_t) call seperr('Data_sou n_t not consintent with Data n_t')
         call from_aux(trim(data_par%tag_sou),"n3",n)
         call to_history("From Data_sou: n3",n)
         if(n /= data_par%n_s1_par) call seperr('Data_sou n_s1_par not consintent with Data n_s1_par')
         call from_aux(trim(data_par%tag_sou),"n4",n)
         call to_history("From Data_sou: n4",n)
         if(n /= data_par%n_s2_par) call seperr('Data_sou n_s2_par not consintent with Data n_s2_par')

         call from_aux(trim(data_par%tag_sou),"d2",d)
         call to_history("From Data_sou: d2",d)
         if(d /= data_par%d_t) call seperr('Data_sou d_t not consintent with Data d_t')
         call from_aux(trim(data_par%tag_sou),"d3",d)
         call to_history("From Data_sou: d3",d)
         if(d /= data_par%d_s1_par) call seperr('Data_sou d_s1_par not consintent with Data d_s1_par')
         call from_aux(trim(data_par%tag_sou),"d4",d)
         call to_history("From Data_sou: d4",d)
         if(d /= data_par%d_s2_par) call seperr('Data_sou d_s2_par not consintent with Data d_s2_par')

         call from_aux(trim(data_par%tag_sou),"o2",data_par%t_sou_0)
         call to_history("From Data_sou: o2",data_par%t_sou_0)
!         if(o /= data_par%t_0) call seperr('Data_sou t_0 not consintent with Data t_0')
         call from_aux(trim(data_par%tag_sou),"o3",o)
         call to_history("From Data_sou: o3",o)
         if(o /= data_par%s1_par_0) call seperr('Data_sou s1_par_0 not consintent with Data s1_par_0')
         call from_aux(trim(data_par%tag_sou),"o4",o)
         call to_history("From Data_sou: o4",o)
         if(o /= data_par%s2_par_0) call seperr('Data_sou s2_par_0 not consintent with Data s2_par_0')
  
       end if

       call from_aux(trim(data_par%tag_sou),"z_data_0",z_data_0)
       call to_history("From Data_sou: z_data_0",z_data_0)
       if(z_data_0 /= data_par%z_0) call seperr('Data_sou z_data_0 not consintent with Data z_data_0')
       data_par%z_0_sou=z_data_0

       if(gen_par%sou_depth /= sou_depth_init) then
         if(gen_par%debug1) write(0,*)'ATTENTION: overwriting source depth'
         if(gen_par%debug1) write(0,*)'From data file source depth=',data_par%z_0_sou
         if(gen_par%debug1) write(0,*)'From parameter file source depth=',gen_par%sou_depth
         data_par%z_0_sou=gen_par%sou_depth
         call to_history("From par: z_data_0_sou",data_par%z_0_sou)
       end if
  
     !}
     end if
   !}
   end if
  !}

case ('lin_imag')
  !{
   if (.not. gen_par%first_adj) then
   !{
      ierr=auxinout(trim(data_par%tag))
      data_par%file=DATA_OUTPUT

      call from_param('n_t',data_par%n_t)
      call from_param('n_sx',data_par%n_sx)
      call from_param('n_sy',data_par%n_sy)
      call from_param('n_off_x',data_par%n_off_x)

      call from_param('d_t',data_par%d_t)
      call from_param('d_sx',data_par%d_sx)
      call from_param('d_sy',data_par%d_sy)
      call from_param('d_off_x',data_par%d_off_x)

      call from_param('t_0',data_par%t_0,0.)
      call from_param('sx_0',data_par%sx_0,0.)
      call from_param('sy_0',data_par%sy_0,0.)
      call from_param('off_x_0',data_par%off_x_0,0.)

      if(gen_par%n_data_dim >4) then 
         call from_param('n_off_y',data_par%n_off_y)
         call from_param('d_off_y',data_par%d_off_y)
         call from_param('off_y_0',data_par%off_y_0,0.)
      else
         data_par%n_off_y=1
         call from_param('d_off_y',data_par%d_off_y,1.)
         call from_param('off_y_0',data_par%off_y_0,0.)
      end if

      call to_history("title",trim(data_par%tag),trim(data_par%tag))

      call to_history("esize",data_par%esize,trim(data_par%tag))
      call to_history("n1",data_par%n_off_x,trim(data_par%tag))
      call to_history("d1",data_par%d_off_x,trim(data_par%tag))
      call to_history("o1",data_par%off_x_0,trim(data_par%tag))

      if(gen_par%n_data_dim >4) then 
         call to_history("n2",data_par%n_off_y,trim(data_par%tag))
         call to_history("n3",data_par%n_t,trim(data_par%tag))
         call to_history("n4",data_par%n_sx,trim(data_par%tag))
         call to_history("n5",data_par%n_sy,trim(data_par%tag))

         call to_history("o2",data_par%off_y_0,trim(data_par%tag))
         call to_history("o3",data_par%t_0,trim(data_par%tag))
         call to_history("o4",data_par%sx_0,trim(data_par%tag))
         call to_history("o5",data_par%sy_0,trim(data_par%tag))

         call to_history("d2",data_par%d_off_y,trim(data_par%tag))
         call to_history("d3",data_par%d_t,trim(data_par%tag))
         call to_history("d4",data_par%d_sx,trim(data_par%tag))
         call to_history("n5",data_par%d_sy,trim(data_par%tag))
      else
         call to_history("n2",data_par%n_t,trim(data_par%tag))
         call to_history("n3",data_par%n_sx,trim(data_par%tag))
         call to_history("n4",data_par%n_sy,trim(data_par%tag))

         call to_history("d2",data_par%d_t,trim(data_par%tag))
         call to_history("d3",data_par%d_sx,trim(data_par%tag))
         call to_history("d4",data_par%d_sy,trim(data_par%tag))

         call to_history("o2",data_par%t_0,trim(data_par%tag))
         call to_history("o3",data_par%sx_0,trim(data_par%tag))
         call to_history("o4",data_par%sy_0,trim(data_par%tag))
      end if
      call to_history("z_data_0",gen_par%rec_depth,trim(data_par%tag))
   !}
   else
   !{
     ierr=auxinout(trim(data_par%tag))
     data_par%file=DATA_INPUT

     call from_aux(trim(data_par%tag),"esize",esize)
     if(esize /= data_par%esize) call seperr('Data must be real')

     call from_param('n_off_x_taper',data_par%n_off_x_taper,-1)
     call from_param('off_x_taper_stdev',data_par%off_x_taper_stdev,0.)
     call from_param('off_x_taper_min',data_par%off_x_taper_min,0.)

     if (gen_par%source_type(1:6) == "point_") then
     !{

       call from_aux(trim(data_par%tag),"n1",data_par%n_off_x)
       call to_history("From Data: n1",data_par%n_off_x)
       call from_aux(trim(data_par%tag),"o1",data_par%off_x_0,0.)
       call to_history("From Data: o1",data_par%off_x_0)
       call from_aux(trim(data_par%tag),"d1",data_par%d_off_x)
       call to_history("From Data: d1",data_par%d_off_x)
  
       if(gen_par%n_data_dim >4) then 
         call from_aux(trim(data_par%tag),"n2",data_par%n_off_y)
         call to_history("From Data: n2",data_par%n_off_y)
         call from_aux(trim(data_par%tag),"n3",data_par%n_t)
         call to_history("From Data: n3",data_par%n_t)
         call from_aux(trim(data_par%tag),"n4",data_par%n_sx)
         call to_history("From Data: n4",data_par%n_sx)
         call from_aux(trim(data_par%tag),"n5",data_par%n_sy)
         call to_history("From Data: n5",data_par%n_sy)
  
         call from_aux(trim(data_par%tag),"d2",data_par%d_off_y)
         call to_history("From Data: d2",data_par%d_off_y)
         call from_aux(trim(data_par%tag),"d3",data_par%d_t)
         call to_history("From Data: d3",data_par%d_t)
         call from_aux(trim(data_par%tag),"d4",data_par%d_sx)
         call to_history("From Data: d4",data_par%d_sx)
         call from_aux(trim(data_par%tag),"d5",data_par%d_sy)
         call to_history("From Data: d5",data_par%d_sy)
  
         call from_aux(trim(data_par%tag),"o2",data_par%off_y_0)
         call to_history("From Data: o2",data_par%off_y_0)
         call from_aux(trim(data_par%tag),"o3",data_par%t_0)
         call to_history("From Data: o3",data_par%t_0)
         call from_aux(trim(data_par%tag),"o4",data_par%sx_0)
         call to_history("From Data: o4",data_par%sx_0)
         call from_aux(trim(data_par%tag),"o5",data_par%sy_0)
         call to_history("From Data: o5",data_par%sy_0)
       else
         call from_aux(trim(data_par%tag),"n2",data_par%n_t)
         call to_history("From Data: n2",data_par%n_t)
         call from_aux(trim(data_par%tag),"n3",data_par%n_sx)
         call to_history("From Data: n3",data_par%n_sx)
         call from_aux(trim(data_par%tag),"n4",data_par%n_sy)
         call to_history("From Data: n4",data_par%n_sy)
  
         call from_aux(trim(data_par%tag),"d2",data_par%d_t)
         call to_history("From Data: d2",data_par%d_t)
         call from_aux(trim(data_par%tag),"d3",data_par%d_sx)
         call to_history("From Data: d3",data_par%d_sx)
         call from_aux(trim(data_par%tag),"d4",data_par%d_sy)
         call to_history("From Data: d4",data_par%d_sy)
  
         call from_aux(trim(data_par%tag),"o2",data_par%t_0)
         call to_history("From Data: o2",data_par%t_0)
         call from_aux(trim(data_par%tag),"o3",data_par%sx_0)
         call to_history("From Data: o3",data_par%sx_0)
         call from_aux(trim(data_par%tag),"o4",data_par%sy_0)
         call to_history("From Data: o4",data_par%sy_0)
  
         data_par%n_off_y=1
         data_par%d_off_y=1.
         data_par%off_y_0=0.
  
       end if

       call from_aux(trim(data_par%tag),"z_data_0",data_par%z_0)
       call to_history("From Data: z_data_0",data_par%z_0)
       if(gen_par%rec_depth /= rec_depth_init) then
         if(gen_par%debug1) write(0,*)'ATTENTION: overwriting receiver depth'
         if(gen_par%debug1) write(0,*)'From data file receiver depth=',data_par%z_0
         if(gen_par%debug1) write(0,*)'From parameter file receiver depth=',gen_par%rec_depth
         data_par%z_0=gen_par%rec_depth 
         call to_history("From par: z_data_0",data_par%z_0)
       end if
     !}
     else
     !{
       call from_aux(trim(data_par%tag),"n1",data_par%n_mx)
       call to_history("From Data: n1",data_par%n_mx)
       call from_aux(trim(data_par%tag),"o1",data_par%mx_0)
       call to_history("From Data: o1",data_par%mx_0)
       call from_aux(trim(data_par%tag),"d1",data_par%d_mx)
       call to_history("From Data: d1",data_par%d_mx)
  
       if(gen_par%n_data_dim >4) then 
         call from_aux(trim(data_par%tag),"n2",data_par%n_my)
         call to_history("From Data: n2",data_par%n_my)
         call from_aux(trim(data_par%tag),"n3",data_par%n_t)
         call to_history("From Data: n3",data_par%n_t)
         call from_aux(trim(data_par%tag),"n4",data_par%n_s1_par)
         call to_history("From Data: n4",data_par%n_s1_par)
         call from_aux(trim(data_par%tag),"n5",data_par%n_s2_par)
         call to_history("From Data: n5",data_par%n_s2_par)
  
         call from_aux(trim(data_par%tag),"d2",data_par%d_my)
         call to_history("From Data: d2",data_par%d_my)
         call from_aux(trim(data_par%tag),"d3",data_par%d_t)
         call to_history("From Data: d3",data_par%d_t)
         call from_aux(trim(data_par%tag),"d4",data_par%d_s1_par)
         call to_history("From Data: d4",data_par%d_s1_par)
         call from_aux(trim(data_par%tag),"d5",data_par%d_s2_par)
         call to_history("From Data: d5",data_par%d_s2_par)
  
         call from_aux(trim(data_par%tag),"o2",data_par%my_0)
         call to_history("From Data: o2",data_par%my_0)
         call from_aux(trim(data_par%tag),"o3",data_par%t_0)
         call to_history("From Data: o3",data_par%t_0)
         call from_aux(trim(data_par%tag),"o4",data_par%s1_par_0)
         call to_history("From Data: o4",data_par%s1_par_0)
         call from_aux(trim(data_par%tag),"o5",data_par%s2_par_0)
         call to_history("From Data: o5",data_par%s2_par_0)
       else
         call from_aux(trim(data_par%tag),"n2",data_par%n_t)
         call to_history("From Data: n2",data_par%n_t)
         call from_aux(trim(data_par%tag),"n3",data_par%n_s1_par)
         call to_history("From Data: n3",data_par%n_s1_par)
         call from_aux(trim(data_par%tag),"n4",data_par%n_s2_par)
         call to_history("From Data: n4",data_par%n_s2_par)
  
         call from_aux(trim(data_par%tag),"d2",data_par%d_t)
         call to_history("From Data: d2",data_par%d_t)
         call from_aux(trim(data_par%tag),"d3",data_par%d_s1_par)
         call to_history("From Data: d3",data_par%d_s1_par)
         call from_aux(trim(data_par%tag),"d4",data_par%d_s2_par)
         call to_history("From Data: d4",data_par%d_s2_par)
  
         call from_aux(trim(data_par%tag),"o2",data_par%t_0)
         call to_history("From Data: o2",data_par%t_0)
         call from_aux(trim(data_par%tag),"o3",data_par%s1_par_0)
         call to_history("From Data: o3",data_par%s1_par_0)
         call from_aux(trim(data_par%tag),"o4",data_par%s2_par_0)
         call to_history("From Data: o4",data_par%s2_par_0)
  
         data_par%n_my=1
         data_par%d_my=1.
         data_par%my_0=0.
       end if

       call from_aux(trim(data_par%tag),"z_data_0",data_par%z_0)
       call to_history("From Data: z_data_0",data_par%z_0)
       if(gen_par%rec_depth /= rec_depth_init) then
         if(gen_par%debug1) write(0,*)'ATTENTION: overwriting receiver depth'
         if(gen_par%debug1) write(0,*)'From data file receiver depth=',data_par%z_0
         if(gen_par%debug1) write(0,*)'From parameter file receiver depth=',gen_par%rec_depth
         data_par%z_0=gen_par%rec_depth 
         call to_history("From par: z_data_0",data_par%z_0)
       end if


       call from_param('tag_data_sou',data_par%tag_sou,'Data_sou')
       call from_param(trim(data_par%tag_sou),tag_name,'Data_sou not found')

       ierr=auxinout(trim(data_par%tag_sou))

       call from_aux(trim(data_par%tag_sou),"esize",esize)
       if(esize /= data_par%esize) call seperr('Data_sou must be real')

       call from_aux(trim(data_par%tag_sou),"n1",n)
       call to_history("From Data_sou: n1",n)
       if(n /= data_par%n_mx) call seperr('Data_sou n_mx not consintent with Data n_mx')
       call from_aux(trim(data_par%tag_sou),"o1",o)
       call to_history("From Data_sou: o1",o)
       if(o /= data_par%mx_0) call seperr('Data_sou mx_0 not consintent with Data mx_0')
       call from_aux(trim(data_par%tag_sou),"d1",d)
       call to_history("From Data_sou: d1",d)
       if(d /= data_par%d_mx) call seperr('Data_sou d_mx not consintent with Data d_mx')
  
       if(gen_par%n_data_dim >4) then 

         call from_aux(trim(data_par%tag_sou),"n2",n)
         call to_history("From Data_sou: n2",n)
         if(n /= data_par%n_my) call seperr('Data_sou n_my not consintent with Data n_my')
         call from_aux(trim(data_par%tag_sou),"n3",n)
         call to_history("From Data_sou: n3",n)
         if(n /= data_par%n_t) call seperr('Data_sou n_t not consintent with Data n_t')
         call from_aux(trim(data_par%tag_sou),"n4",n)
         call to_history("From Data_sou: n4",n)
         if(n /= data_par%n_s1_par) call seperr('Data_sou n_s1_par not consintent with Data n_s1_par')
         call from_aux(trim(data_par%tag_sou),"n5",n)
         call to_history("From Data_sou: n5",n)
         if(n /= data_par%n_s2_par) call seperr('Data_sou n_s2_par not consintent with Data n_s2_par')

         call from_aux(trim(data_par%tag_sou),"d2",d)
         call to_history("From Data_sou: d2",d)
         if(d /= data_par%d_my) call seperr('Data_sou d_my not consintent with Data d_my')
         call from_aux(trim(data_par%tag_sou),"d3",d)
         call to_history("From Data_sou: d3",d)
         if(d /= data_par%d_t) call seperr('Data_sou d_t not consintent with Data d_t')
         call from_aux(trim(data_par%tag_sou),"d4",d)
         call to_history("From Data_sou: d4",d)
         if(d /= data_par%d_s1_par) call seperr('Data_sou d_s1_par not consintent with Data d_s1_par')
         call from_aux(trim(data_par%tag_sou),"d5",d)
         call to_history("From Data_sou: d5",d)
         if(d /= data_par%d_s2_par) call seperr('Data_sou d_s2_par not consintent with Data d_s2_par')

         call from_aux(trim(data_par%tag_sou),"o2",o)
         call to_history("From Data_sou: o2",o)
         if(o /= data_par%my_0) call seperr('Data_sou my_0 not consintent with Data my_0')
         call from_aux(trim(data_par%tag_sou),"o3",data_par%t_sou_0)
         call to_history("From Data_sou: o3",data_par%t_sou_0)
!         if(o /= data_par%t_0) call seperr('Data_sou t_0 not consintent with Data t_0')
         call from_aux(trim(data_par%tag_sou),"o4",o)
         call to_history("From Data_sou: o4",o)
         if(o /= data_par%s1_par_0) call seperr('Data_sou s1_par_0 not consintent with Data s1_par_0')
         call from_aux(trim(data_par%tag_sou),"o5",o)
         call to_history("From Data_sou: o5",o)
         if(o /= data_par%s2_par_0) call seperr('Data_sou s2_par_0 not consintent with Data s2_par_0')
  
       else

         call from_aux(trim(data_par%tag_sou),"n2",n)
         call to_history("From Data_sou: n2",n)
         if(n /= data_par%n_t) call seperr('Data_sou n_t not consintent with Data n_t')
         call from_aux(trim(data_par%tag_sou),"n3",n)
         call to_history("From Data_sou: n3",n)
         if(n /= data_par%n_s1_par) call seperr('Data_sou n_s1_par not consintent with Data n_s1_par')
         call from_aux(trim(data_par%tag_sou),"n4",n)
         call to_history("From Data_sou: n4",n)
         if(n /= data_par%n_s2_par) call seperr('Data_sou n_s2_par not consintent with Data n_s2_par')

         call from_aux(trim(data_par%tag_sou),"d2",d)
         call to_history("From Data_sou: d2",d)
         if(d /= data_par%d_t) call seperr('Data_sou d_t not consintent with Data d_t')
         call from_aux(trim(data_par%tag_sou),"d3",d)
         call to_history("From Data_sou: d3",d)
         if(d /= data_par%d_s1_par) call seperr('Data_sou d_s1_par not consintent with Data d_s1_par')
         call from_aux(trim(data_par%tag_sou),"d4",d)
         call to_history("From Data_sou: d4",d)
         if(d /= data_par%d_s2_par) call seperr('Data_sou d_s2_par not consintent with Data d_s2_par')

         call from_aux(trim(data_par%tag_sou),"o2",data_par%t_sou_0)
         call to_history("From Data_sou: o2",data_par%t_sou_0)
!         if(o /= data_par%t_0) call seperr('Data_sou t_0 not consintent with Data t_0')
         call from_aux(trim(data_par%tag_sou),"o3",o)
         call to_history("From Data_sou: o3",o)
         if(o /= data_par%s1_par_0) call seperr('Data_sou s1_par_0 not consintent with Data s1_par_0')
         call from_aux(trim(data_par%tag_sou),"o4",o)
         call to_history("From Data_sou: o4",o)
         if(o /= data_par%s2_par_0) call seperr('Data_sou s2_par_0 not consintent with Data s2_par_0')
  
       end if

       call from_aux(trim(data_par%tag_sou),"z_data_0",z_data_0)
       call to_history("From Data_sou: z_data_0",z_data_0)
       if(z_data_0 /= data_par%z_0) call seperr('Data_sou z_data_0 not consintent with Data z_data_0')
       data_par%z_0_sou=z_data_0

       if(gen_par%sou_depth /= sou_depth_init) then
         if(gen_par%debug1) write(0,*)'ATTENTION: overwriting source depth'
         if(gen_par%debug1) write(0,*)'From data file source depth=',data_par%z_0_sou
         if(gen_par%debug1) write(0,*)'From parameter file source depth=',gen_par%sou_depth
         data_par%z_0_sou=gen_par%sou_depth
         call to_history("From par: z_data_0_sou",data_par%z_0_sou)
       end if
  
     !}
     end if
   !}
   end if

  !}

  case ('image_gl')
  !{

    ierr=auxinout(trim(data_par%tag))
    data_par%file=DATA_OUTPUT

    call from_param('n_t',data_par%n_t)
    call from_param('n_mx_data',data_par%n_mx)
    call from_param('n_s1_par',data_par%n_s1_par)
    call from_param('n_s2_par',data_par%n_s2_par)

    call from_param('d_t',data_par%d_t)
    data_par%d_t_sou=data_par%d_t
    call from_param('d_mx_data',data_par%d_mx)
    call from_param('d_s1_par',data_par%d_s1_par,1.)
    call from_param('d_s2_par',data_par%d_s2_par,1.)

    call from_param('t_0',data_par%t_0,0.)
    data_par%t_sou_0=(data_par%n_t-1)*(-data_par%d_t_sou)
    call from_param('mx_0_data',data_par%mx_0)
    call from_param('s1_par_0',data_par%s1_par_0,0.)
    call from_param('s2_par_0',data_par%s2_par_0,0.)

    if(gen_par%n_data_dim >4) then 
      call from_param('n_my_data',data_par%n_my)
      call from_param('d_my_data',data_par%d_my)
      call from_param('my_0_data',data_par%my_0)
    else
       data_par%n_my=1
       call from_param('d_my',data_par%d_my,1.)
       call from_param('my_0',data_par%my_0,0.)
    end if

    call to_history("title",trim(data_par%tag),trim(data_par%tag))

    call to_history("esize",data_par%esize,trim(data_par%tag))
    call to_history("n1",data_par%n_mx,trim(data_par%tag))
    call to_history("d1",data_par%d_mx,trim(data_par%tag))
    call to_history("o1",data_par%mx_0,trim(data_par%tag))

    if(gen_par%n_data_dim >4) then 
       call to_history("n2",data_par%n_my,trim(data_par%tag))
       call to_history("n3",data_par%n_t,trim(data_par%tag))
       call to_history("n4",data_par%n_s1_par,trim(data_par%tag))
       call to_history("n5",data_par%n_s2_par,trim(data_par%tag))

       call to_history("o2",data_par%my_0,trim(data_par%tag))
       call to_history("o3",data_par%t_0,trim(data_par%tag))
       call to_history("o4",data_par%s1_par_0,trim(data_par%tag))
       call to_history("o5",data_par%s2_par_0,trim(data_par%tag))

       call to_history("d2",data_par%d_my,trim(data_par%tag))
       call to_history("d3",data_par%d_t,trim(data_par%tag))
       call to_history("d4",data_par%d_s1_par,trim(data_par%tag))
       call to_history("n5",data_par%d_s2_par,trim(data_par%tag))
    else
       call to_history("n2",data_par%n_t,trim(data_par%tag))
       call to_history("n3",data_par%n_s1_par,trim(data_par%tag))
       call to_history("n4",data_par%n_s2_par,trim(data_par%tag))

       call to_history("d2",data_par%d_t,trim(data_par%tag))
       call to_history("d3",data_par%d_s1_par,trim(data_par%tag))
       call to_history("d4",data_par%d_s2_par,trim(data_par%tag))

       call to_history("o2",data_par%t_0,trim(data_par%tag))
       call to_history("o3",data_par%s1_par_0,trim(data_par%tag))
       call to_history("o4",data_par%s2_par_0,trim(data_par%tag))
    end if
    call to_history("z_data_0",gen_par%rec_depth,trim(data_par%tag))

    call from_param('tag_data_sou',data_par%tag_sou,'Data_sou')
    call from_param(trim(data_par%tag_sou),tag_name,'Data_sou not found')

    ierr=auxinout(trim(data_par%tag_sou))

    call to_history("title",trim(data_par%tag_sou),trim(data_par%tag_sou))

    call to_history("esize",data_par%esize,trim(data_par%tag_sou))
    call to_history("n1",data_par%n_mx,trim(data_par%tag_sou))
    call to_history("d1",data_par%d_mx,trim(data_par%tag_sou))
    call to_history("o1",data_par%mx_0,trim(data_par%tag_sou))

    if(gen_par%n_data_dim >4) then 
       call to_history("n2",data_par%n_my,trim(data_par%tag_sou))
       call to_history("n3",data_par%n_t,trim(data_par%tag_sou))
       call to_history("n4",data_par%n_s1_par,trim(data_par%tag_sou))
       call to_history("n5",data_par%n_s2_par,trim(data_par%tag_sou))

       call to_history("o2",data_par%my_0,trim(data_par%tag_sou))
       call to_history("o3",data_par%t_sou_0,trim(data_par%tag_sou))
       call to_history("o4",data_par%s1_par_0,trim(data_par%tag_sou))
       call to_history("o5",data_par%s2_par_0,trim(data_par%tag_sou))

       call to_history("d2",data_par%d_my,trim(data_par%tag_sou))
       call to_history("d3",data_par%d_t_sou,trim(data_par%tag_sou))
       call to_history("d4",data_par%d_s1_par,trim(data_par%tag_sou))
       call to_history("n5",data_par%d_s2_par,trim(data_par%tag_sou))
    else
       call to_history("n2",data_par%n_t,trim(data_par%tag_sou))
       call to_history("n3",data_par%n_s1_par,trim(data_par%tag_sou))
       call to_history("n4",data_par%n_s2_par,trim(data_par%tag_sou))

       call to_history("d2",data_par%d_t_sou,trim(data_par%tag_sou))
       call to_history("d3",data_par%d_s1_par,trim(data_par%tag_sou))
       call to_history("d4",data_par%d_s2_par,trim(data_par%tag_sou))

       call to_history("o2",data_par%t_sou_0,trim(data_par%tag_sou))
       call to_history("o3",data_par%s1_par_0,trim(data_par%tag_sou))
       call to_history("o4",data_par%s2_par_0,trim(data_par%tag_sou))
    end if
    call to_history("z_data_0",gen_par%sou_depth,trim(data_par%tag_sou))

  !}

case ('vel_data')
  !{
    ierr=auxinout(trim(data_par%tag))
    data_par%file=DATA_OUTPUT

    call from_param('n_t',data_par%n_t)
    call from_param('n_sx',data_par%n_sx)
    call from_param('n_sy',data_par%n_sy)
    call from_param('n_off_x',data_par%n_off_x)

    call from_param('d_t',data_par%d_t)
    call from_param('d_sx',data_par%d_sx)
    call from_param('d_sy',data_par%d_sy)
    call from_param('d_off_x',data_par%d_off_x)

    call from_param('t_0',data_par%t_0,0.)
    call from_param('sx_0',data_par%sx_0,0.)
    call from_param('sy_0',data_par%sy_0,0.)
    call from_param('off_x_0',data_par%off_x_0,0.)

    if(gen_par%n_data_dim >4) then 
       call from_param('n_off_y',data_par%n_off_y)
       call from_param('d_off_y',data_par%d_off_y)
       call from_param('off_y_0',data_par%off_y_0,0.)
    else
       data_par%n_off_y=1
       call from_param('d_off_y',data_par%d_off_y,1.)
       call from_param('off_y_0',data_par%off_y_0,0.)
    end if

    call to_history("title",trim(data_par%tag),trim(data_par%tag))
    call to_history("esize",data_par%esize,trim(data_par%tag))
    call to_history("n1",data_par%n_off_x,trim(data_par%tag))
    call to_history("d1",data_par%d_off_x,trim(data_par%tag))
    call to_history("o1",data_par%off_x_0,trim(data_par%tag))

    if(gen_par%n_data_dim >4) then 
       call to_history("n2",data_par%n_off_y,trim(data_par%tag))
       call to_history("n3",data_par%n_t,trim(data_par%tag))
       call to_history("n4",data_par%n_sx,trim(data_par%tag))
       call to_history("n5",data_par%n_sy,trim(data_par%tag))

       call to_history("o2",data_par%off_y_0,trim(data_par%tag))
       call to_history("o3",data_par%t_0,trim(data_par%tag))
       call to_history("o4",data_par%sx_0,trim(data_par%tag))
       call to_history("o5",data_par%sy_0,trim(data_par%tag))

       call to_history("d2",data_par%d_off_y,trim(data_par%tag))
       call to_history("d3",data_par%d_t,trim(data_par%tag))
       call to_history("d4",data_par%d_sx,trim(data_par%tag))
       call to_history("n5",data_par%d_sy,trim(data_par%tag))
    else
       call to_history("n2",data_par%n_t,trim(data_par%tag))
       call to_history("n3",data_par%n_sx,trim(data_par%tag))
       call to_history("n4",data_par%n_sy,trim(data_par%tag))

       call to_history("d2",data_par%d_t,trim(data_par%tag))
       call to_history("d3",data_par%d_sx,trim(data_par%tag))
       call to_history("d4",data_par%d_sy,trim(data_par%tag))

       call to_history("o2",data_par%t_0,trim(data_par%tag))
       call to_history("o3",data_par%sx_0,trim(data_par%tag))
       call to_history("o4",data_par%sy_0,trim(data_par%tag))
    end if
    call to_history("z_data_0",gen_par%rec_depth,trim(data_par%tag))
  !}

  case default
     write(0,*)'gen_par-domain_range=',trim(gen_par%domain_range)
     call seperr('Not supported domain_range')
  !}
  end select

   return
end subroutine param_data


subroutine param_image(gen_par,data_par,image_par,wavelet_par)
  use sep
  use genrev3d_types
  use data_types
  use image_types
  use wavelet_types

  implicit none

  character(len=128) :: tag_name,tag_hz_name
  type (genrev3d_param_type) gen_par
  type (data_param_type) data_par
  type (image_param_type) image_par
  type (wavelet_param_type) wavelet_par
  integer, external :: auxpar,getch,fetch,auxin,sreed,sseek,auxout,auxinout


  character(len=128) :: tag_image
  logical write_image
  integer esize,ierr,found_image,found_image_sou,found_image_rec
  integer n1_image,n2_image,n3_image,n4_image,n5_image,n6_image,n7_image
  real o1_image,o2_image,o3_image,o4_image,o5_image,o6_image,o7_image
  real d1_image,d2_image,d3_image,d4_image,d5_image,d6_image,d7_image
  real t_0_lag_image,epsi_image_d_t

  epsi_image_d_t=.001*data_par%d_t

  call from_param('read_image_progress',image_par%read_image_progress,.false.)
  call from_param('write_image_progress',image_par%write_image_progress,.false.)
  call from_param('write_image_hz',image_par%write_image_hz,.false.)
  call from_param('image_memory',image_par%image_memory,.true.)

  image_par%esize=4

  ! image parameters
  call from_param('tag_image',image_par%tag,'Image')
  call from_param(trim(image_par%tag),tag_name,'Image not found')

  if(image_par%write_image_hz) then
    call from_param('tag_image_hz',image_par%tag_hz,'Image_hz')
    call from_param(trim(image_par%tag_hz),tag_hz_name,'Image_hz not found')
  end if

  ! find out what to do with image file

  select case (gen_par%domain_range(1:8))
  case ('image_da','lin_imag')
    if (gen_par%first_adj) then
      call from_param('write_image',write_image,.true.)
      if(write_image) then
         image_par%file=IMAGE_OUTPUT
      else
         image_par%file=IMAGE_NO_FILE
      end if
    else
      image_par%file=IMAGE_INPUT
    end if
  case ('image_gl')
    if (gen_par%first_adj) then
     call seperr('domain_range = image_global_data and adj=true')
    else
      image_par%file=IMAGE_INPUT
    end if
  case ('vel_data')
     image_par%image_memory=.false.
     image_par%file=IMAGE_NO_FILE
  case default
     write(0,*)'gen_par-domain_range=',trim(gen_par%domain_range)
     call seperr('Not supported domain_range')
  end select

  if(gen_par%debug1) write(0,*)'image_file=',image_par%file

  if (image_par%file == IMAGE_INPUT) then
  !{

    call from_param('tag_image_sou',image_par%tag_sou,'Image_sou')
    call from_param(trim(image_par%tag_sou),tag_name,'Image_sou not found')
    call from_param('tag_image_rec',image_par%tag_rec,'Image_rec')
    call from_param(trim(image_par%tag_rec),tag_name,'Image_rec not found')
  
    found_image=auxin(trim(image_par%tag))
    found_image_sou=auxin(trim(image_par%tag_sou))
    found_image_rec=auxin(trim(image_par%tag_rec))
  
    write(0,*)'found_image=',found_image
    write(0,*)'found_image_sou=',found_image_sou
    write(0,*)'found_image_rec=',found_image_rec
     
    if((found_image <0).and.(found_image_sou <0).and.(found_image_rec <0)) &
             then
      call seperr('Need at least one Image file')
    end if
    if((found_image >0).and.((found_image_sou >0).or.(found_image_rec >0))) &
             then
      call seperr('Either provide Image or Image_sou/Image_rec')
    end if
    if((found_image_sou >0) .and. (found_image_rec <0)) then
      call seperr('Image_sou present and Image_rec missing')
    end if
    if((found_image_sou <0) .and. (found_image_rec >0)) then
      call seperr('Image_rec present and Image_sou missing')
    end if
    image_par%two_images= ((found_image_sou >0) .and. (found_image_rec >0)) 
    write(0,*)'two_images=',image_par%two_images
        
    if(image_par%two_images) then
      tag_image=trim(image_par%tag_sou)
      image_par%tag=trim(image_par%tag_sou)
  write(0,*)'image_par-tag=',image_par%tag
    else
  write(0,*)'Only one image'
      tag_image=trim(image_par%tag)
    end if

   call from_aux(trim(tag_image),"esize",esize)
   if(esize /= image_par%esize) call seperr('Image of wrong type')

   call from_aux(trim(tag_image),"n1",image_par%n_z)
   call from_aux(trim(tag_image),"d1",image_par%d_z)
   call from_aux(trim(tag_image),"o1",image_par%z_0)
   call from_aux(trim(tag_image),"n2",image_par%n_mx)
   call from_aux(trim(tag_image),"d2",image_par%d_mx)
   call from_aux(trim(tag_image),"o2",image_par%mx_0)

   call to_history("From Image: n1",image_par%n_z)
   call to_history("From Image: o1",image_par%z_0)
   call to_history("From Image: d1",image_par%d_z)
   call to_history("From Image: n2",image_par%n_mx)
   call to_history("From Image: o2",image_par%mx_0)
   call to_history("From Image: d2",image_par%d_mx)

!     image_par%d_t=data_par%d_t

   if(gen_par%n_data_dim >4) then 
   !{

     call from_aux(trim(tag_image),"n3",image_par%n_my)
     call from_aux(trim(tag_image),"d3",image_par%d_my)
     call from_aux(trim(tag_image),"o3",image_par%my_0)
     call to_history("From Image: n3",image_par%n_my)
     call to_history("From Image: o3",image_par%my_0)
     call to_history("From Image: d3",image_par%d_my)

     if(.not.gen_par%image_zero_offset) then
     !{
       if(gen_par%surf_offset_cig) then
       !{
         call from_aux(trim(tag_image),"n4",image_par%n_sx)
         call from_aux(trim(tag_image),"n5",image_par%n_sy)
         call from_aux(trim(tag_image),"d4",image_par%d_sx)
         call from_aux(trim(tag_image),"d5",image_par%d_sy)
         call from_aux(trim(tag_image),"o4",image_par%sx_0)
         call from_aux(trim(tag_image),"o5",image_par%sy_0)
         image_par%n_hx=1
         image_par%hx_0=0.
         image_par%d_hx=1.
         image_par%n_hy=1
         image_par%hy_0=0.
         image_par%d_hy=1.
         call to_history("From Image: n4",image_par%n_sx)
         call to_history("From Image: o4",image_par%sx_0)
         call to_history("From Image: d4",image_par%d_sx)
         call to_history("From Image: n5",image_par%n_sy)
         call to_history("From Image: o5",image_par%sy_0)
         call to_history("From Image: d5",image_par%d_sy)
       !}
       else
       !{
         call from_aux(trim(tag_image),"n4",image_par%n_hx)
         call from_aux(trim(tag_image),"n5",image_par%n_hy)
         call from_aux(trim(tag_image),"d4",image_par%d_hx)
         call from_aux(trim(tag_image),"d5",image_par%d_hy)
         call from_aux(trim(tag_image),"o4",image_par%hx_0)
         call from_aux(trim(tag_image),"o5",image_par%hy_0)

         call to_history("From Image: n4",image_par%n_hx)
         call to_history("From Image: o4",image_par%hx_0)
         call to_history("From Image: d4",image_par%d_hx)
         call to_history("From Image: n5",image_par%n_hy)
         call to_history("From Image: o5",image_par%hy_0)
         call to_history("From Image: d5",image_par%d_hy)
       !}
       end if
     !}
     end if

     if(image_par%read_image_progress) then
     !{
         call from_aux(trim(tag_image),"n6",image_par%n_t_lag)
         call from_aux(trim(tag_image),"n7",image_par%n_t)
         call from_aux(trim(tag_image),"d6",image_par%d_t_lag)
         call from_aux(trim(tag_image),"d7",image_par%d_t)
         call from_aux(trim(tag_image),"o6",image_par%t_0_lag)
         call from_aux(trim(tag_image),"o7",image_par%t_0)

         call to_history("From Image: n6",image_par%n_t_lag)
         call to_history("From Image: o6",image_par%t_0_lag)
         call to_history("From Image: d6",image_par%d_t_lag)
         call to_history("From Image: n7",image_par%n_t)
         call to_history("From Image: o7",image_par%t_0)
         call to_history("From Image: d7",image_par%d_t)
     !}
     else
     !{
         call from_aux(trim(tag_image),"n6",image_par%n_t_lag,1)
         call from_aux(trim(tag_image),"d6",image_par%d_t_lag,1.)
         call from_aux(trim(tag_image),"o6",image_par%t_0_lag,0.)
         call to_history("From Image: n6",image_par%n_t_lag)
         call to_history("From Image: o6",image_par%t_0_lag)
         call to_history("From Image: d6",image_par%d_t_lag)

!         call from_param('n_t_image',image_par%n_t,data_par%n_t)
         call from_param('d_t_image',image_par%d_t,data_par%d_t)
         call from_param('t_0_image',image_par%t_0,data_par%t_0)
         image_par%t_max=data_par%t_0+(data_par%n_t-1)*data_par%d_t
         call from_param('t_max_image',image_par%t_max,image_par%t_max)
         image_par%n_t= &
           ceiling((image_par%t_max-image_par%t_0+epsi_image_d_t)/image_par%d_t)
     !}
     end if
   !}
   else
   !{
     if(.not.gen_par%image_zero_offset) then
     !{
       if(gen_par%surf_offset_cig) then
       !{
         call from_aux(trim(tag_image),"n3",image_par%n_sx)
         call from_aux(trim(tag_image),"n4",image_par%n_sy)
         call from_aux(trim(tag_image),"d3",image_par%d_sx)
         call from_aux(trim(tag_image),"d4",image_par%d_sy)
         call from_aux(trim(tag_image),"o3",image_par%sx_0)
         call from_aux(trim(tag_image),"o4",image_par%sy_0)
         image_par%n_hx=1
         image_par%hx_0=0.
         image_par%d_hx=1.
         image_par%n_hy=1
         image_par%hy_0=0.
         image_par%d_hy=1.
         call to_history("From Image: n3",image_par%n_sx)
         call to_history("From Image: o3",image_par%sx_0)
         call to_history("From Image: d3",image_par%d_sx)
         call to_history("From Image: n4",image_par%n_sy)
         call to_history("From Image: o4",image_par%sy_0)
         call to_history("From Image: d4",image_par%d_sy)
       !}
       else
       !{
         call from_aux(trim(tag_image),"n3",image_par%n_hx)
         call from_aux(trim(tag_image),"d3",image_par%d_hx)
         call from_aux(trim(tag_image),"o3",image_par%hx_0)
         call to_history("From Image: n3",image_par%n_hx)
         call to_history("From Image: o3",image_par%hx_0)
         call to_history("From Image: d3",image_par%d_hx)

         image_par%n_my=1
         image_par%my_0=0.
         image_par%d_my=1.
         image_par%n_hy=1
         image_par%hy_0=0.
         image_par%d_hy=1.

         call to_history("From Image: n4",image_par%n_my)
         call to_history("From Image: o4",image_par%my_0)
         call to_history("From Image: d4",image_par%d_my)
         call to_history("From Image: n5",image_par%n_hy)
         call to_history("From Image: o5",image_par%hy_0)
         call to_history("From Image: d5",image_par%d_hy)
       !}
       end if
     !}
     else
     !{
       image_par%n_hx=1
       image_par%hx_0=0.
       image_par%n_hy=1
       image_par%hy_0=0.
     !}
     end if
  
     if(image_par%read_image_progress) then
     !{
         call from_aux(trim(tag_image),"n6",image_par%n_t_lag)
         call from_aux(trim(tag_image),"n7",image_par%n_t)
         call from_aux(trim(tag_image),"d6",image_par%d_t_lag)
         call from_aux(trim(tag_image),"d7",image_par%d_t)
         call from_aux(trim(tag_image),"o6",image_par%t_0_lag)
         call from_aux(trim(tag_image),"o7",image_par%t_0)

         call to_history("From Image: n6",image_par%n_t_lag)
         call to_history("From Image: o6",image_par%t_0_lag)
         call to_history("From Image: d6",image_par%d_t_lag)
         call to_history("From Image: n7",image_par%n_t)
         call to_history("From Image: o7",image_par%t_0)
         call to_history("From Image: d7",image_par%d_t)

     !}
     else
     !{
         call from_aux(trim(tag_image),"n6",image_par%n_t_lag,1)
         call from_aux(trim(tag_image),"d6",image_par%d_t_lag,1.)
         call from_aux(trim(tag_image),"o6",image_par%t_0_lag,0.)
         call to_history("From Image: n6",image_par%n_t_lag)
         call to_history("From Image: o6",image_par%t_0_lag)
         call to_history("From Image: d6",image_par%d_t_lag)

!         call from_param('n_t_image',image_par%n_t,data_par%n_t)
         call from_param('d_t_image',image_par%d_t,data_par%d_t)
         call from_param('t_0_image',image_par%t_0,data_par%t_0)
         image_par%t_max=data_par%t_0+(data_par%n_t-1)*data_par%d_t
         call from_param('t_max_image',image_par%t_max,image_par%t_max)
         image_par%n_t= &
           ceiling((image_par%t_max-image_par%t_0+epsi_image_d_t)/image_par%d_t)
     !}
     end if
   !}
  end if
  !}
  else
  !{
    if(image_par%file == IMAGE_OUTPUT) then
    !{
       tag_image=trim(image_par%tag)

       call from_param('n_z_image',image_par%n_z)
       call from_param('z_image_0',image_par%z_0,0.)
       call from_param('d_z_image',image_par%d_z)

        call from_param('d_t_image',image_par%d_t,data_par%d_t)
        if(gen_par%image_zero_time) then
        !{
          call from_param('t_0_image',image_par%t_0,0.)
          if(image_par%t_0 < -((data_par%n_t-1)*data_par%d_t + data_par%t_0)) then
            call seperr("t_0_image before first (negative) time sample of data")
          end if
          image_par%n_t=2*nint(-image_par%t_0/image_par%d_t)+1
          call from_param('n_t_image',image_par%n_t,image_par%n_t)
!          image_par%t_max=data_par%t_0+(data_par%n_t-1)*data_par%d_t
!          call from_param('t_max_image',image_par%t_max,image_par%t_max)
!          image_par%n_t=nint((image_par%t_max-image_par%t_0)/image_par%d_t)+1
        !}
        else
        !{
          call from_param('t_0_image',image_par%t_0,data_par%t_0)
!          image_par%n_t= &
!    (data_par%t_0+(data_par%n_t-1)*data_par%d_t-image_par%t_0)/image_par%d_t + 1
!          call from_param('n_t_image',image_par%n_t,image_par%n_t)
          image_par%t_max=data_par%t_0+(data_par%n_t-1)*data_par%d_t
          call from_param('t_max_image',image_par%t_max,image_par%t_max)
          image_par%n_t= &
           ceiling((image_par%t_max-image_par%t_0+epsi_image_d_t)/image_par%d_t)
write(0,*)'n_t_image=',image_par%n_t
        !}
        end if
!                      image_par%d_t + 2

        call from_param('n_mx_image',image_par%n_mx)
        call from_param('mx_image_0',image_par%mx_0)
        call from_param('d_mx_image',image_par%d_mx)

        call from_param('n_my_image',image_par%n_my)
        call from_param('my_image_0',image_par%my_0)
        call from_param('d_my_image',image_par%d_my)

        call from_param('n_t_lag_image',image_par%n_t_lag,1)
        call from_param('d_t_lag_image',image_par%d_t_lag,image_par%d_t)

        t_0_lag_image=-image_par%d_t_lag*(image_par%n_t_lag-1)/2
        call from_param('t_0_lag_image',image_par%t_0_lag,t_0_lag_image)

        if(gen_par%image_zero_offset) then
        !{
          if(gen_par%surf_offset_cig .or. &
             gen_par%write_image_progress_with_shots) then
          !{
            image_par%n_sx=data_par%n_sx
            image_par%n_sy=data_par%n_sy
            image_par%sx_0=data_par%sx_0
            image_par%sy_0=data_par%sy_0
            image_par%d_sx=data_par%d_sx
            image_par%d_sy=data_par%d_sy

            n2_image=image_par%n_mx
            o2_image=image_par%mx_0
            d2_image=image_par%d_mx

            n3_image=image_par%n_my
            o3_image=image_par%my_0
            d3_image=image_par%d_my

            n4_image=image_par%n_sx
            o4_image=image_par%sx_0
            d4_image=image_par%d_sx
    
            n5_image=image_par%n_sy
            o5_image=image_par%sy_0
            d5_image=image_par%d_sy

            image_par%n_hx=1
            image_par%hx_0=0.
            image_par%d_hx=1.
            image_par%n_hy=1
            image_par%hy_0=0.
            image_par%d_hy=1.
          !}
          else
          !{
            image_par%n_hx=1
            image_par%hx_0=0.
            image_par%d_hx=1.
            image_par%n_hy=1
            image_par%hy_0=0.
            image_par%d_hy=1.

            n2_image=image_par%n_mx
            o2_image=image_par%mx_0
            d2_image=image_par%d_mx

            n3_image=image_par%n_my
            o3_image=image_par%my_0
            d3_image=image_par%d_my

            n4_image=image_par%n_hx
            o4_image=image_par%hx_0
            d4_image=image_par%d_hx

            n5_image=image_par%n_hy
            o5_image=image_par%hy_0
            d5_image=image_par%d_hy
          !}
          end if
        !}
        else
        !{
          call from_param('n_hx_image',image_par%n_hx)
          call from_param('d_hx_image',image_par%d_hx)
          image_par%hx_0=-int((image_par%n_hx-1)/2)*image_par%d_hx
          call from_param('hx_image_0',image_par%hx_0,image_par%hx_0)
  
          call from_param('n_hy_image',image_par%n_hy,1)
          call from_param('d_hy_image',image_par%d_hy,1.)
          image_par%hy_0=-int((image_par%n_hy-1)/2)*image_par%d_hy
          call from_param('hy_image_0',image_par%hy_0,image_par%hy_0)

          if(gen_par%write_image_progress_with_shots) then
          !{
            image_par%n_sx=data_par%n_sx
            image_par%n_sy=data_par%n_sy
            image_par%sx_0=data_par%sx_0
            image_par%sy_0=data_par%sy_0
            image_par%d_sx=data_par%d_sx
            image_par%d_sy=data_par%d_sy

            n2_image=image_par%n_mx
            o2_image=image_par%mx_0
            d2_image=image_par%d_mx
  
            n3_image=image_par%n_hx
            o3_image=image_par%hx_0
            d3_image=image_par%d_hx
    
            n4_image=image_par%n_my
            o4_image=image_par%my_0
            d4_image=image_par%d_my
    
            n5_image=image_par%n_hy
            o5_image=image_par%hy_0
            d5_image=image_par%d_hy

            n6_image=image_par%n_sx
            o6_image=image_par%sx_0
            d6_image=image_par%d_sx
    
            n7_image=image_par%n_sy
            o7_image=image_par%sy_0
            d7_image=image_par%d_sx
          !}
          else
          !{
            n2_image=image_par%n_mx
            o2_image=image_par%mx_0
            d2_image=image_par%d_mx
  
            n3_image=image_par%n_hx
            o3_image=image_par%hx_0
            d3_image=image_par%d_hx
    
            n4_image=image_par%n_my
            o4_image=image_par%my_0
            d4_image=image_par%d_my
    
            n5_image=image_par%n_hy
            o5_image=image_par%hy_0
            d5_image=image_par%d_hy
          !}
          end if

        !}
        end if

        n1_image=image_par%n_z
        o1_image=image_par%z_0
        d1_image=image_par%d_z

        if(image_par%write_image_progress) then
        !{
          n6_image=image_par%n_t_lag
          o6_image=image_par%t_0_lag
          d6_image=image_par%d_t_lag

          n7_image=image_par%n_t
!          o7_image=(rev_extrap_par%t_0_second+wavelet_par%t_delay)
          o7_image=image_par%t_0+(image_par%n_t-1)*image_par%d_t
          d7_image=-image_par%d_t
        !}
        else
        !{
          if(gen_par%write_image_progress_with_shots) then
          !{
          !}
          else
          !{
            n6_image=image_par%n_t_lag
            o6_image=image_par%t_0_lag
            d6_image=image_par%d_t_lag
  
            n7_image=1
            o7_image=0.
            d7_image=1.
          !}
          end if
        !}
        end if

        ierr=auxinout(trim(tag_image))

        call to_history("esize",image_par%esize,trim(tag_image))
        call to_history("n1",n1_image,trim(tag_image))
        call to_history("n2",n2_image,trim(tag_image))
        call to_history("n3",n3_image,trim(tag_image))
        call to_history("n4",n4_image,trim(tag_image))
        call to_history("n5",n5_image,trim(tag_image))
        call to_history("n6",n6_image,trim(tag_image))
        call to_history("n7",n7_image,trim(tag_image))

        call to_history("d1",d1_image,trim(tag_image))
        call to_history("d2",d2_image,trim(tag_image))
        call to_history("d3",d3_image,trim(tag_image))
        call to_history("d4",d4_image,trim(tag_image))
        call to_history("d5",d5_image,trim(tag_image))
        call to_history("d6",d6_image,trim(tag_image))
        call to_history("d7",d7_image,trim(tag_image))

        call to_history("o1",o1_image,trim(tag_image))
        call to_history("o2",o2_image,trim(tag_image))
        call to_history("o3",o3_image,trim(tag_image))
        call to_history("o4",o4_image,trim(tag_image))
        call to_history("o5",o5_image,trim(tag_image))
        call to_history("o6",o6_image,trim(tag_image))
        call to_history("o7",o7_image,trim(tag_image))

!       Guojian
        if(image_par%write_image_hz) then
          image_par%n_hz=image_par%n_hx
          image_par%hz_0=image_par%hx_0
          image_par%d_hz=image_par%d_hx

          ierr=auxinout(trim(image_par%tag_hz))
  
          call to_history("esize",image_par%esize,trim(image_par%tag_hz))
          call to_history("n1",n1_image,trim(image_par%tag_hz))
          call to_history("n2",n2_image,trim(image_par%tag_hz))
          call to_history("n3",n3_image,trim(image_par%tag_hz))
          call to_history("n4",n4_image,trim(image_par%tag_hz))
          call to_history("n5",n5_image,trim(image_par%tag_hz))
          call to_history("n6",n6_image,trim(image_par%tag_hz))
          call to_history("n7",n7_image,trim(image_par%tag_hz))
  
          call to_history("d1",d1_image,trim(image_par%tag_hz))
          call to_history("d2",d2_image,trim(image_par%tag_hz))
          call to_history("d3",d3_image,trim(image_par%tag_hz))
          call to_history("d4",d4_image,trim(image_par%tag_hz))
          call to_history("d5",d5_image,trim(image_par%tag_hz))
          call to_history("d6",d6_image,trim(image_par%tag_hz))
          call to_history("d7",d7_image,trim(image_par%tag_hz))
  
          call to_history("o1",o1_image,trim(image_par%tag_hz))
          call to_history("o2",o2_image,trim(image_par%tag_hz))
          call to_history("o3",o3_image,trim(image_par%tag_hz))
          call to_history("o4",o4_image,trim(image_par%tag_hz))
          call to_history("o5",o5_image,trim(image_par%tag_hz))
          call to_history("o6",o6_image,trim(image_par%tag_hz))
          call to_history("o7",o7_image,trim(image_par%tag_hz))
        end if

!       end Guojian

    !}
    end if
  !}
  end if

   return
end subroutine param_image

subroutine param_extrap(gen_par,data_par,image_par,wavelet_par,rev_extrap_par)

  use sep
  use genrev3d_types
  use data_types
  use image_types
  use wavelet_types
  use rev_extrap_types

  implicit none

  type (genrev3d_param_type) gen_par
  type (data_param_type) data_par
  type (image_param_type) image_par
  type (wavelet_param_type) wavelet_par
  type (rev_extrap_param_type) rev_extrap_par

  integer n_t_delay,n_t_prolong_first,n_t_prolong_second,n_t,n_t_extrap
  real t_0_extrap,d_t,t_0

  ! rev_extrap parameters

  n_t_delay=wavelet_par%t_delay/wavelet_par%d_t
  write(0,*)'n_t_delay=',n_t_delay

  select case (gen_par%domain_range(1:8))
  case ('image_da')
  !{
    if (gen_par%source_type(1:6) == "point_") then
    !{
      if (gen_par%first_adj) then
      !{
        call from_param('n_t_extrap',n_t,data_par%n_t)
        rev_extrap_par%n_t_first=n_t+n_t_delay
        rev_extrap_par%n_t_second=nint(image_par%n_t*image_par%d_t/data_par%d_t)
      
        call from_param('d_t_extrap',d_t,-data_par%d_t)
        rev_extrap_par%d_t_first=-d_t
        rev_extrap_par%d_t_second=d_t
        rev_extrap_par%d_t=abs(d_t)

    
        t_0_extrap=(rev_extrap_par%n_t_second-1)*(-rev_extrap_par%d_t_second)+ &
                    image_par%t_0
        call from_param('t_0_extrap',t_0,t_0_extrap)
        rev_extrap_par%t_0_first=-wavelet_par%t_delay
        rev_extrap_par%t_0_second=t_0
      !}
      else
      !{
        call from_param('n_t_sou_extrap',rev_extrap_par%n_t_first, &
                       data_par%n_t)
!                       nint(image_par%n_t*image_par%d_t/data_par%d_t))
        call from_param('n_t_extrap',rev_extrap_par%n_t_second, &
                       data_par%n_t)

        call from_param('d_t_sou_extrap',rev_extrap_par%d_t_first,image_par%d_t)
        call from_param('d_t_extrap',rev_extrap_par%d_t_second,-data_par%d_t)
        rev_extrap_par%d_t=abs(data_par%d_t)

        t_0_extrap=image_par%t_0
        call from_param('t_0_sou_extrap',rev_extrap_par%t_0_first,t_0_extrap)
    
        t_0_extrap=(rev_extrap_par%n_t_second-1)*(rev_extrap_par%d_t_second) + &
                    image_par%t_0
        call from_param('t_0_extrap',rev_extrap_par%t_0_second,t_0_extrap)
!       call seperr("Chek it because this has not debugged")
      !}
      end if
    !}
    else
    !{
      if (gen_par%first_adj) then
      !{
        n_t_prolong_first=nint((image_par%t_0 + (image_par%n_t-1)*image_par%d_t)/ &
                             data_par%d_t)
        n_t_prolong_second=nint(-image_par%t_0/data_par%d_t)
        write(0,*)'n_t_prolong_first=',n_t_prolong_first
        write(0,*)'n_t_prolong_second=',n_t_prolong_second
        call from_param('n_t_sou_extrap',rev_extrap_par%n_t_first, &
                         data_par%n_t+n_t_prolong_first)
        call from_param('n_t_extrap',rev_extrap_par%n_t_second, &
                       data_par%n_t+n_t_prolong_second)
  
        call from_param('d_t_sou_extrap',rev_extrap_par%d_t_first,data_par%d_t)
        call from_param('d_t_extrap',rev_extrap_par%d_t_second,-data_par%d_t)
        rev_extrap_par%d_t=abs(rev_extrap_par%d_t_first)
  
  !    t_0_extrap=data_par%t_sou_0
        t_0_extrap=data_par%t_sou_0
        call from_param('t_0_sou_extrap',rev_extrap_par%t_0_first,t_0_extrap)
    
        t_0_extrap=(data_par%n_t-1)*(data_par%d_t) + data_par%t_0
        call from_param('t_0_extrap',rev_extrap_par%t_0_second,t_0_extrap)

      !}
      end if
    !}
    end if

  !}
  case ('lin_imag')
  !{
    if (gen_par%source_type(1:6) == "point_") then
    !{
      if (gen_par%first_adj) then
      !{
        call from_param('n_t_extrap',n_t, &
                     nint(image_par%n_t*image_par%d_t/data_par%d_t))
        rev_extrap_par%n_t_first=n_t+n_t_delay
        rev_extrap_par%n_t_second=n_t

        call from_param('d_t_extrap',d_t,-data_par%d_t)
        rev_extrap_par%d_t_first=-d_t
        rev_extrap_par%d_t_second=d_t
        rev_extrap_par%d_t=abs(d_t)

      
        t_0_extrap=(rev_extrap_par%n_t_second-1)*(-rev_extrap_par%d_t_second)+ &
                    image_par%t_0
        call from_param('t_0_extrap',t_0,t_0_extrap)
        rev_extrap_par%t_0_first=-wavelet_par%t_delay+rev_extrap_par%d_t_first
        rev_extrap_par%t_0_second=t_0

        if(gen_par%debug2) write(0,*)'t_0_second=',rev_extrap_par%t_0_second
      !}
      else
      !{
        call from_param('n_t_sou_extrap',rev_extrap_par%n_t_first, &
                       data_par%n_t+n_t_delay)
        call from_param('d_t_sou_extrap',rev_extrap_par%d_t_first,data_par%d_t)
        t_0_extrap=data_par%t_0-wavelet_par%t_delay+rev_extrap_par%d_t_first
        call from_param('t_0_sou_extrap',rev_extrap_par%t_0_first,t_0_extrap)
    
        call from_param('n_t_extrap',rev_extrap_par%n_t_second, &
                       data_par%n_t)
        call from_param('d_t_extrap',rev_extrap_par%d_t_second,data_par%d_t)
        rev_extrap_par%d_t=abs(data_par%d_t)
        t_0_extrap=data_par%t_0
        call from_param('t_0_extrap',rev_extrap_par%t_0_second,t_0_extrap)
!       call seperr("Chek it because this has not debugged")
      !}
      end if
    !}
    else
    !{
       call seperr("For the moment it must be point_source")
    !}
    end if
  !}
  case ('image_gl')
  !{
    call from_param('n_t_sou_extrap',rev_extrap_par%n_t_first,data_par%n_t+n_t_delay)
    call from_param('n_t_extrap',rev_extrap_par%n_t_second,data_par%n_t+n_t_delay)
  
    call from_param('d_t_sou_extrap',rev_extrap_par%d_t_first,-data_par%d_t)
    call from_param('d_t_extrap',rev_extrap_par%d_t_second,data_par%d_t)
    rev_extrap_par%d_t=abs(data_par%d_t)

!    t_0_extrap=data_par%t_0+(rev_extrap_par%n_t_first-1)*rev_extrap_par%d_t_first
    t_0_extrap=data_par%t_0+wavelet_par%t_delay
    call from_param('t_0_sou_extrap',rev_extrap_par%t_0_first,t_0_extrap)

    t_0_extrap=data_par%t_0-wavelet_par%t_delay
    call from_param('t_0_extrap',rev_extrap_par%t_0_second,t_0_extrap)

  !}
  case ('vel_data')
  !{
    call from_param('n_t_extrap',n_t,data_par%n_t+n_t_delay)
    rev_extrap_par%n_t_first=0
    rev_extrap_par%n_t_second=n_t

    call from_param('d_t_extrap',d_t,data_par%d_t)
    rev_extrap_par%d_t_first=0.
    rev_extrap_par%d_t_second=d_t
    rev_extrap_par%d_t=abs(d_t)

    call from_param('t_0_extrap',t_0,data_par%t_0-wavelet_par%t_delay)
    rev_extrap_par%t_0_first=0.
    rev_extrap_par%t_0_second=t_0

  !}
  case default
     write(0,*)'gen_par-domain_range=',trim(gen_par%domain_range)
     call seperr('Not supported domain_range')
  end select


  rev_extrap_par%n_threads=gen_par%n_threads

  return
end subroutine param_extrap

subroutine param_wave(gen_par,data_par,rev_extrap_par,image_par,wavelet_par, &
                      wave_par)
  use sep
  use genrev3d_types
  use data_types
  use rev_extrap_types
  use image_types
  use wavelet_types
  use wave_types

  implicit none

  type (genrev3d_param_type) gen_par
  type (data_param_type) data_par
  type (rev_extrap_param_type) rev_extrap_par
  type (image_param_type) image_par
  type (wavelet_param_type) wavelet_par
  type (wave_param_type) wave_par

  logical write_wavefield,write_wavefield_back
  character(len=128) :: tag_name
  integer, external :: auxpar,getch,fetch,auxin,sreed,sseek,auxout,auxinout
  integer esize, ierr
  integer n1,n2,n3,n4,n5,n6,n_t_snap
  integer source_n_t,n_t_delay
  real t_extrap,source_d_t,source_t_0
  real o1,o2,o3,o4,o5,o6
  real d1,d2,d3,d4,d5,d6
  real pi

  wave_par%esize=4

  n_t_delay=wavelet_par%t_delay/wavelet_par%d_t

  call from_param('write_wavefield', write_wavefield,.false.)
  call from_param('write_wavefield_back', write_wavefield_back,.true.)

!  call from_param('n_z',wave_par%n_z,image_par%n_z)
!  call from_param('z_0',wave_par%z_0,image_par%z_0)
!  call from_param('d_z',wave_par%d_z,image_par%d_z)

  call from_param('n_z',wave_par%n_z)
  call from_param('z_0',wave_par%z_0,image_par%z_0)
  call from_param('d_z',wave_par%d_z,image_par%d_z)

  if (gen_par%source_type(1:6) == "point_") then
  !{
    call from_param('n_off_x_pad',wave_par%n_off_x,-integer_init)
    if((wave_par%n_off_x > 0).and. (wave_par%n_off_x > data_par%n_off_x)) then
       wave_par%n_off_x=wave_par%n_off_x
       wave_par%n_mx=wave_par%n_off_x
       call from_param('off_x_0_pad',wave_par%off_x_0,-real_init)
       if(wave_par%off_x_0 == -real_init) wave_par%off_x_0=data_par%off_x_0
    else
       wave_par%n_off_x=data_par%n_off_x
       wave_par%n_mx=data_par%n_off_x
       wave_par%off_x_0=data_par%off_x_0
       call to_history("n_off_x_pad",wave_par%n_off_x)
       call to_history("off_x_0_pad",wave_par%off_x_0)
    end if
    wave_par%d_off_x=data_par%d_off_x
    wave_par%d_mx=data_par%d_off_x
  
    call from_param('n_off_y_pad',wave_par%n_off_y,-integer_init)
    if((wave_par%n_off_y > 0).and. (wave_par%n_off_y > data_par%n_off_y)) then
       wave_par%n_off_y=wave_par%n_off_y
       wave_par%n_my=wave_par%n_off_y
       call from_param('off_y_0_pad',wave_par%off_y_0,-real_init)
       if(wave_par%off_y_0 == -real_init) wave_par%off_y_0=data_par%off_y_0
    else
       wave_par%n_off_y=data_par%n_off_y
       wave_par%n_my=data_par%n_off_y
       wave_par%off_y_0=data_par%off_y_0
       call to_history("n_off_y_pad",wave_par%n_off_y)
       call to_history("off_y_0_pad",wave_par%off_y_0)
    end if
    wave_par%d_off_y=data_par%d_off_y
    wave_par%d_my=data_par%d_off_y
  
    wave_par%n_sx=data_par%n_sx
    wave_par%d_sx=data_par%d_sx
    wave_par%sx_0=data_par%sx_0
    wave_par%n_sy=data_par%n_sy
    wave_par%d_sy=data_par%d_sy
    wave_par%sy_0=data_par%sy_0

    wave_par%n_t_back=rev_extrap_par%n_t_first-n_t_delay
    if(gen_par%debug2) write(0,*)'n_t_back=',wave_par%n_t_back
    wave_par%d_t_back=rev_extrap_par%d_t_first
    wave_par%t_back_0=image_par%t_0
    if(gen_par%debug2) write(0,*)'t_back_0=',wave_par%t_back_0

    n1=wave_par%n_z
    d1=wave_par%d_z
    o1=wave_par%z_0

    n2=wave_par%n_off_x
    d2=wave_par%d_off_x
    o2=wave_par%off_x_0

    if(gen_par%n_data_dim >4) then 
    !{
      n3=wave_par%n_off_y
      d3=wave_par%d_off_y
      o3=wave_par%off_y_0

      n4=wave_par%n_t_back
      d4=wave_par%d_t_back
      o4=wave_par%t_back_0

!      n5=data_par%n_sx
!      d5=data_par%d_sx
!      o5=data_par%sx_0
!
!      n6=data_par%n_sy
!      d6=data_par%d_sy
!      o6=data_par%sy_0

      n5=1
      d5=1.
      o5=0.

      n6=1
      d6=1.
      o6=0.

    !}
    else
    !{
      n3=wave_par%n_t_back
      d3=wave_par%d_t_back
      o3=wave_par%t_back_0

!      n4=data_par%n_sx
!      d4=data_par%d_sx
!      o4=data_par%sx_0

      n4=1
      d4=1.
      o4=0.

      n5=1
      d5=1.
      o5=0.

      n6=1
      d6=1.
      o6=0.

    !}
    end if
  
    if(data_par%file == DATA_OUTPUT) then
      wave_par%i_z_sou=(gen_par%sou_depth-wave_par%z_0)/wave_par%d_z+1
      wave_par%i_z_rec=(gen_par%rec_depth-wave_par%z_0)/wave_par%d_z+1
    else
      wave_par%i_z_sou=(gen_par%sou_depth-wave_par%z_0)/wave_par%d_z+1
      wave_par%i_z_rec=(data_par%z_0-wave_par%z_0)/wave_par%d_z+1
    end if
  
    if(gen_par%verb) then
      write(0,*)'i_z_sou =',wave_par%i_z_sou
      write(0,*)'i_z_rec =',wave_par%i_z_rec
    end if

  !}
  else
  !{
    call from_param('n_mx_pad',wave_par%n_mx,-integer_init)
    if((wave_par%n_mx > 0).and. (wave_par%n_mx > data_par%n_mx)) then
       wave_par%n_mx=wave_par%n_mx
       call from_param('mx_0_pad',wave_par%mx_0,-real_init)
       if(wave_par%mx_0 == -real_init) wave_par%mx_0=data_par%mx_0
    else
       wave_par%n_mx=data_par%n_mx
       wave_par%mx_0=data_par%mx_0
       call to_history("n_mx_pad",wave_par%n_mx)
       call to_history("mx_0_pad",wave_par%mx_0)
    end if
    wave_par%d_mx=data_par%d_mx

    call from_param('n_my_pad',wave_par%n_my,-integer_init)
    if((wave_par%n_mx > 0).and. (wave_par%n_mx > data_par%n_mx)) then
       wave_par%n_my=wave_par%n_my
       call from_param('my_0_pad',wave_par%my_0,-real_init)
       if(wave_par%my_0 == -real_init) wave_par%my_0=data_par%my_0
    else
       wave_par%n_my=data_par%n_my
       wave_par%my_0=data_par%my_0
       call to_history("n_my_pad",wave_par%n_my)
       call to_history("my_0_pad",wave_par%my_0)
    end if
    wave_par%d_my=data_par%d_my

    wave_par%n_s1_par=data_par%n_s1_par
    wave_par%n_s2_par=data_par%n_s2_par

    if(gen_par%image_zero_time) then
      wave_par%n_t_back=image_par%n_t
      wave_par%d_t_back=image_par%d_t
      wave_par%t_back_0=image_par%t_0
    else
      wave_par%n_t_back=rev_extrap_par%n_t_first
      wave_par%d_t_back=rev_extrap_par%d_t_first
      wave_par%t_back_0=rev_extrap_par%t_0_first
!      wave_par%t_back_0=image_par%t_0-rev_extrap_par%t_0_first
    end if

    n1=wave_par%n_z
    d1=wave_par%d_z
    o1=wave_par%z_0

    n2=wave_par%n_mx
    d2=wave_par%d_mx
    o2=wave_par%mx_0

    if(gen_par%n_data_dim >4) then 
      n3=wave_par%n_my
      d3=wave_par%d_my
      o3=wave_par%my_0

      n4=wave_par%n_t_back
      d4=wave_par%d_t_back
      o4=wave_par%t_back_0

      n5=data_par%n_s1_par
      d5=data_par%d_s1_par
      o5=data_par%s1_par_0

      n6=data_par%n_s2_par
      d6=data_par%d_s2_par
      o6=data_par%s2_par_0
    else
      n3=wave_par%n_t_back
      d3=wave_par%d_t_back
      o3=wave_par%t_back_0

      n4=data_par%n_s1_par
      d4=data_par%d_s1_par
      o4=data_par%s1_par_0

      n5=1
      d5=1.
      o5=0.

      n6=1
      d6=1.
      o6=0.
    end if

    if(data_par%file == DATA_OUTPUT) then
      wave_par%i_z_sou=(gen_par%sou_depth-wave_par%z_0)/wave_par%d_z+1
      wave_par%i_z_rec=(gen_par%rec_depth-wave_par%z_0)/wave_par%d_z+1
    else
      wave_par%i_z_sou=(data_par%z_0_sou-wave_par%z_0)/wave_par%d_z+1
      wave_par%i_z_rec=(data_par%z_0-wave_par%z_0)/wave_par%d_z+1
    end if
  
     write(0,*)'i_z_sou =',wave_par%i_z_sou
     write(0,*)'i_z_rec =',wave_par%i_z_rec

  !}
  end if

  if(write_wavefield_back) then
  !{
    wave_par%file_back=WAVE_OUTPUT

    call from_param('tag_wavefield_back',wave_par%tag_back,'Wavefield_back')

    call from_param(trim(wave_par%tag_back),tag_name,'Wavefield_back not found')
    ierr=auxinout(trim(wave_par%tag_back))

    if((.not. gen_par%skip_source_prop)) then
    !{

      call to_history("title",trim(wave_par%tag_back),trim(wave_par%tag_back))
  
      call to_history("esize",data_par%esize,trim(wave_par%tag_back))
  
      call to_history("n1",n1,trim(wave_par%tag_back))
      call to_history("d1",d1,trim(wave_par%tag_back))
      call to_history("o1",o1,trim(wave_par%tag_back))
  
      call to_history("n2",n2,trim(wave_par%tag_back))
      call to_history("d2",d2,trim(wave_par%tag_back))
      call to_history("o2",o2,trim(wave_par%tag_back))
  
      call to_history("n3",n3,trim(wave_par%tag_back))
      call to_history("d3",d3,trim(wave_par%tag_back))
      call to_history("o3",o3,trim(wave_par%tag_back))
  
      call to_history("n4",n4,trim(wave_par%tag_back))
      call to_history("d4",d4,trim(wave_par%tag_back))
      call to_history("o4",o4,trim(wave_par%tag_back))
  
      call to_history("n5",n5,trim(wave_par%tag_back))
      call to_history("d5",d5,trim(wave_par%tag_back))
      call to_history("o5",o5,trim(wave_par%tag_back))
  
      call to_history("n6",n6,trim(wave_par%tag_back))
      call to_history("d6",d6,trim(wave_par%tag_back))
      call to_history("o6",o6,trim(wave_par%tag_back))
  
    !}
    end if
  !}
  else
  !{
    wave_par%file_back=WAVE_NO_FILE
  !}
  end if

  ! wavefield parameters
  if(write_wavefield) then
    !{
    wave_par%file=WAVE_OUTPUT
    call from_param('d_t_snap',wave_par%d_t,rev_extrap_par%d_t_second)

    select case (gen_par%domain_range(1:8))
    case ('image_da')
      if (gen_par%first_adj) then
        call from_param('t_0_snap',wave_par%t_0,rev_extrap_par%t_0_second)
        n_t_snap=-(wave_par%t_0)/wave_par%d_t+1
        call from_param('n_t_snap',wave_par%n_t,n_t_snap)
      else
        call from_param('t_0_snap',wave_par%t_0,rev_extrap_par%t_0_second)
        n_t_snap=int(( &
          (rev_extrap_par%n_t_second-1)*rev_extrap_par%d_t_second - &
             rev_extrap_par%t_0_second)/wave_par%d_t)+1
        call from_param('n_t_snap',wave_par%n_t,n_t_snap)
!        write(0,*)'n_t_second=',rev_extrap_par%n_t_second
!        write(0,*)'d_t_second=',rev_extrap_par%d_t_second
!        write(0,*)'t_0_second=',rev_extrap_par%t_0_second
!        write(0,*)'wave dt=',wave_par%d_t
!        write(0,*)'temp=',&
!	(((rev_extrap_par%n_t_second-1)*rev_extrap_par%d_t_second - &
!             rev_extrap_par%t_0_second)/wave_par%d_t)+1
      end if
      call from_param('tag_wavefield_sou',wave_par%tag_first,'Wavefield_sou')
      call from_param(trim(wave_par%tag_first),tag_name, &
                     'Wavefield_sou not found')
      ierr=auxinout(trim(wave_par%tag_first))

      call from_param('tag_wavefield_rec',wave_par%tag_second,'Wavefield_rec')
      call from_param(trim(wave_par%tag_second),tag_name, &
                     'Wavefield_rec not found')
      ierr=auxinout(trim(wave_par%tag_second))

    case ('lin_imag')
      if (gen_par%first_adj) then
        call from_param('t_0_snap',wave_par%t_0,rev_extrap_par%t_0_second)
!        n_t_snap=-(wave_par%t_0)/wave_par%d_t+1
        n_t_snap=int(( &
          (rev_extrap_par%n_t_second-1)*rev_extrap_par%d_t_second)/ &
           wave_par%d_t)+1
        call from_param('n_t_snap',wave_par%n_t,n_t_snap)
      else
        call from_param('t_0_snap',wave_par%t_0, &
                  rev_extrap_par%t_0_second+rev_extrap_par%d_t_second)
        n_t_snap=int(( &
          (rev_extrap_par%n_t_second-1)*rev_extrap_par%d_t_second - &
           rev_extrap_par%t_0_second-rev_extrap_par%d_t_second)/wave_par%d_t)+1
        call from_param('n_t_snap',wave_par%n_t,n_t_snap)
      end if
      call from_param('tag_wavefield_sou',wave_par%tag_first,'Wavefield_sou')
      call from_param(trim(wave_par%tag_first),tag_name, &
                     'Wavefield_sou not found')
      ierr=auxinout(trim(wave_par%tag_first))

      call from_param('tag_wavefield_rec',wave_par%tag_second,'Wavefield_rec')
      call from_param(trim(wave_par%tag_second),tag_name, &
                     'Wavefield_rec not found')
      ierr=auxinout(trim(wave_par%tag_second))

    case ('image_gl')
      if (gen_par%first_adj) then
        call from_param('t_0_snap',wave_par%t_0,rev_extrap_par%t_0_second)
        n_t_snap=-(wave_par%t_0)/wave_par%d_t+1
        call from_param('n_t_snap',wave_par%n_t,n_t_snap)
      else
        call from_param('t_0_snap',wave_par%t_0,rev_extrap_par%t_0_second)
!        n_t_snap=int(( &
!          rev_extrap_par%n_t_second*rev_extrap_par%d_t_second+rev_extrap_par%t_0_second)/ &
!                           wave_par%d_t)+1
        n_t_snap=int(( &
!          rev_extrap_par%n_t_second*rev_extrap_par%d_t_second)/ &
          rev_extrap_par%n_t_second*rev_extrap_par%d_t_second+rev_extrap_par%t_0_second)/ &
                           wave_par%d_t)+1
        call from_param('n_t_snap',wave_par%n_t,n_t_snap)
      end if
      call from_param('tag_wavefield_sou',wave_par%tag_first,'Wavefield_sou')
      call from_param(trim(wave_par%tag_first),tag_name,'Wavefield_sou not found')
      ierr=auxinout(trim(wave_par%tag_first))

      call from_param('tag_wavefield_rec',wave_par%tag_second,'Wavefield_rec')
      call from_param(trim(wave_par%tag_second),tag_name,'Wavefield_rec not found')
      ierr=auxinout(trim(wave_par%tag_second))

    case ('vel_data')
      call from_param('t_0_snap',wave_par%t_0,rev_extrap_par%t_0_second+wavelet_par%t_delay)
      n_t_snap=int(( &
        rev_extrap_par%n_t_second*rev_extrap_par%d_t_second+rev_extrap_par%t_0_second)/ &
                         wave_par%d_t)+1
      call from_param('n_t_snap',wave_par%n_t,n_t_snap)
      call from_param('tag_wavefield_sou',wave_par%tag_second,'Wavefield_sou')
      call from_param(trim(wave_par%tag_second),tag_name,'Wavefield_sou not found')
      ierr=auxinout(trim(wave_par%tag_second))
    case default
       write(0,*)'gen_par-domain_range=',trim(gen_par%domain_range)
       call seperr('Not supported domain_range')
    end select

    if(gen_par%n_data_dim >4) then 
!      n4=wave_par%n_t_back
!      o4=wave_par%t_0_back
!      d4=wave_par%d_t_back
      n4=wave_par%n_t
      o4=wave_par%t_0
      d4=wave_par%d_t
    else
!      n3=wave_par%n_t_back
!      o3=wave_par%t_back_0
!      d3=wave_par%d_t_back
      n3=wave_par%n_t
      o3=wave_par%t_0
      d3=wave_par%d_t
    end if

    call to_history("title",trim(wave_par%tag_second),trim(wave_par%tag_second))
    call to_history("esize",data_par%esize,trim(wave_par%tag_second))

    call to_history("n1",n1,trim(wave_par%tag_second))
    call to_history("d1",d1,trim(wave_par%tag_second))
    call to_history("o1",o1,trim(wave_par%tag_second))

    call to_history("n2",n2,trim(wave_par%tag_second))
    call to_history("d2",d2,trim(wave_par%tag_second))
    call to_history("o2",o2,trim(wave_par%tag_second))

    call to_history("n3",n3,trim(wave_par%tag_second))
    call to_history("d3",d3,trim(wave_par%tag_second))
    call to_history("o3",o3,trim(wave_par%tag_second))

    call to_history("n4",n4,trim(wave_par%tag_second))
    call to_history("d4",d4,trim(wave_par%tag_second))
    call to_history("o4",o4,trim(wave_par%tag_second))

    call to_history("n5",n5,trim(wave_par%tag_second))
    call to_history("d5",d5,trim(wave_par%tag_second))
    call to_history("o5",o5,trim(wave_par%tag_second))

    call to_history("n6",n6,trim(wave_par%tag_second))
    call to_history("d6",d6,trim(wave_par%tag_second))
    call to_history("o6",o6,trim(wave_par%tag_second))

    select case (gen_par%domain_range(1:8))
    case ('image_da','lin_imag')
      source_n_t=wave_par%n_t
      source_t_0=rev_extrap_par%t_0_first+wavelet_par%t_delay
      if (gen_par%first_adj) then
        source_d_t=-wave_par%d_t
      else
        source_d_t=wave_par%d_t
      end if
    case ('image_gl')
      source_n_t=wave_par%n_t
      source_t_0=rev_extrap_par%t_0_first+wavelet_par%t_delay
      if (gen_par%first_adj) then
!          source_d_t=-wave_par%d_t
      else
        source_d_t=-wave_par%d_t
      end if
    case ('vel_data')
    case default
       write(0,*)'gen_par-domain_range=',trim(gen_par%domain_range)
       call seperr('Not supported domain_range')
    end select

    if((.not. gen_par%skip_source_prop) .and. &
       (gen_par%domain_range(1:8) /= 'vel_data')) then
    !{
write(0,*)'source_n_t=',source_n_t
write(0,*)'source_d_t=',source_d_t
write(0,*)'source_t_0=',source_t_0

      if(gen_par%n_data_dim >4) then 
        n4=source_n_t
        d4=source_d_t
        o4=source_t_0
      else
        n3=source_n_t
        d3=source_d_t
        o3=source_t_0
      end if
    
      call to_history("title",trim(wave_par%tag_first),trim(wave_par%tag_first))
  
      call to_history("esize",data_par%esize,trim(wave_par%tag_first))
  
      call to_history("n1",n1,trim(wave_par%tag_first))
      call to_history("d1",d1,trim(wave_par%tag_first))
      call to_history("o1",o1,trim(wave_par%tag_first))
  
      call to_history("n2",n2,trim(wave_par%tag_first))
      call to_history("d2",d2,trim(wave_par%tag_first))
      call to_history("o2",o2,trim(wave_par%tag_first))
  
      call to_history("n3",n3,trim(wave_par%tag_first))
      call to_history("d3",d3,trim(wave_par%tag_first))
      call to_history("o3",o3,trim(wave_par%tag_first))
  
      call to_history("n4",n4,trim(wave_par%tag_first))
      call to_history("d4",d4,trim(wave_par%tag_first))
      call to_history("o4",o4,trim(wave_par%tag_first))
  
      call to_history("n5",n5,trim(wave_par%tag_first))
      call to_history("d5",d5,trim(wave_par%tag_first))
      call to_history("o5",o5,trim(wave_par%tag_first))
  
      call to_history("n6",n6,trim(wave_par%tag_first))
      call to_history("d6",d6,trim(wave_par%tag_first))
      call to_history("o6",o6,trim(wave_par%tag_first))
    !}
    end if 
  !}
  else 
    wave_par%file=WAVE_NO_FILE
  end if

  if(gen_par%debug1) write(0,*)'wavefield_file=',wave_par%file

! kept for backward compatibility
  wave_par%n_hx=1
  wave_par%d_hx=1.
  wave_par%hx_0=0.
  wave_par%n_hy=1
  wave_par%d_hy=1.
  wave_par%hy_0=0.

  pi=acos(-1.)

  wave_par%kz_0=0.
  wave_par%d_kz=2.*pi/(wave_par%d_z*FLOAT(wave_par%n_z))

  wave_par%kmx_0=0.
  wave_par%kmy_0=0.
  wave_par%d_kmx=2.*pi/(wave_par%d_mx*FLOAT(wave_par%n_mx))
  wave_par%d_kmy=2.*pi/(wave_par%d_my*FLOAT(wave_par%n_my))

  if(gen_par%debug1) write(0,*)'wave_par-n_my=',wave_par%n_my
  if(gen_par%debug1) write(0,*)'wave_par-d_my=',wave_par%d_my
  if(gen_par%debug1) write(0,*)'wave_par-d_kmy=',wave_par%d_kmy

  return
end subroutine param_wave

subroutine param_slow(gen_par,wave_par,image_par,slow_par)
  use sep
  use genrev3d_types
  use wave_types
  use image_types
  use slow_types

  implicit none

  type (genrev3d_param_type) gen_par
  type (wave_param_type) wave_par
  type (image_param_type) image_par
  type (slow_param_type) slow_par
  character(len=128) :: tag_name
  integer esize
  integer, external :: auxpar,getch,fetch,auxin,sreed,sseek,auxout

  ! slow parameters
  call from_param('tag_slow',slow_par%tag,'Slow')
  call from_param(trim(slow_par%tag),tag_name,'Slow not found')
  slow_par%tag=trim(slow_par%tag)

   slow_par%esize=4

   slow_par%n_comp=1 
  
   if (auxin(trim(slow_par%tag)) .gt. 0) then
   !{
     slow_par%file=SLOW_INPUT
     call from_param('slowness',slow_par%slowness,.true.)
     call from_param('extrapolate_slow',slow_par%extrapolate_slow,.false.)
     call from_param('slow_scale',slow_par%scale,1.)

     call from_aux(trim(slow_par%tag),"esize",esize)

     if(esize /= slow_par%esize) call seperr('Slow must be real')

     call from_aux(trim(slow_par%tag),"n1",slow_par%n_z,1)
     call from_aux(trim(slow_par%tag),"o1",slow_par%z_0,0.)
     call from_aux(trim(slow_par%tag),"d1",slow_par%d_z,1.)

     call to_history("From Slow: n1",slow_par%n_z)
     call to_history("From Slow: o1",slow_par%z_0)
     call to_history("From Slow: d1",slow_par%d_z)

     if(wave_par%n_z > slow_par%n_z) call seperr('n1 in Slow < Wave')
     if(wave_par%z_0 /= slow_par%z_0) call seperr('o1 in Slow /= Wave')
     if(wave_par%d_z /= slow_par%d_z) call seperr('d1 in Slow /= Wave')

     call from_aux(trim(slow_par%tag),"n2",slow_par%n_mx)
     call from_aux(trim(slow_par%tag),"o2",slow_par%mx_0)
     call from_aux(trim(slow_par%tag),"d2",slow_par%d_mx)

     call to_history("From Slow: n2",slow_par%n_mx)
     call to_history("From Slow: o2",slow_par%mx_0)
     call to_history("From Slow: d2",slow_par%d_mx)

!     if(wave_par%n_mx /= slow_par%n_mx) call seperr('n2 in Slow /= Wave')
!     if(wave_par%mx_0 /= slow_par%mx_0) call seperr('o2 in Slow /= Wave')
!     if(wave_par%d_mx /= slow_par%d_mx) call seperr('d2 in Slow /= Wave')

     call from_aux(trim(slow_par%tag),"n3",slow_par%n_my)
     call from_aux(trim(slow_par%tag),"o3",slow_par%my_0)
     call from_aux(trim(slow_par%tag),"d3",slow_par%d_my)

     call to_history("From Slow: n3",slow_par%n_my)
     call to_history("From Slow: o3",slow_par%my_0)
     call to_history("From Slow: d3",slow_par%d_my)

!     if(wave_par%n_my /= slow_par%n_my) call seperr('n3 in Slow /= Wave')
!     if(wave_par%my_0 /= slow_par%my_0) call seperr('o3 in Slow /= Wave')
!     if(wave_par%d_my /= slow_par%d_my) call seperr('d3 in Slow /= Wave')

     if (gen_par%domain_range(1:8) == 'lin_imag') then
     !{
       slow_par%n_hx = image_par%n_hx
       slow_par%d_hx = image_par%d_hx
       slow_par%hx_0 = image_par%hx_0
       slow_par%n_hy = image_par%n_hy
       slow_par%d_hy = image_par%d_hy
       slow_par%hy_0 = image_par%hy_0

       call to_history("From Par: n4",slow_par%n_hx)
       call to_history("From Par o4",slow_par%hx_0)
       call to_history("From Par d4",slow_par%d_hx)
       call to_history("From Par: n5",slow_par%n_hy)
       call to_history("From Par o5",slow_par%hy_0)
       call to_history("From Par d5",slow_par%d_hy)
     !}
     end if

   !}
   else   
   !}
     slow_par%file=SLOW_NO_FILE
   !}
   end if

   return
end subroutine param_slow

subroutine param_wavelet(gen_par,wavelet_par)
  use sep
  use genrev3d_types
  use wavelet_types

  implicit none

  type (genrev3d_param_type) gen_par
  type (wavelet_param_type) wavelet_par

  character(len=128) :: tag_name
  integer esize
  integer, external :: auxpar,getch,fetch,auxin,sreed,sseek,auxout

  if ((gen_par%source_type(1:6) == 'global') .and. gen_par%first_adj) then
  !{
    wavelet_par%n_t=0
    wavelet_par%d_t=1
    wavelet_par%t_0=1
    wavelet_par%t_delay=0.
    wavelet_par%file=WAVELET_NO_FILE
  !}
  else
  !{
    ! wavelet parameters
    call from_param('tag_wavelet',wavelet_par%tag,'Wavelet')
    call from_param(trim(wavelet_par%tag),tag_name,'Wavelet not found')
    wavelet_par%tag=trim(wavelet_par%tag)
  
    wavelet_par%esize=4
  
    wavelet_par%file=WAVELET_INPUT
  
     call from_aux(trim(wavelet_par%tag),"n1",wavelet_par%n_t,1)
     call from_aux(trim(wavelet_par%tag),"o1",wavelet_par%t_0,0.)
     call from_aux(trim(wavelet_par%tag),"d1",wavelet_par%d_t,1.)

     call from_aux(trim(wavelet_par%tag),"n2",wavelet_par%n_z,1)
     call from_aux(trim(wavelet_par%tag),"o2",wavelet_par%z_0,0.)
     call from_aux(trim(wavelet_par%tag),"d2",wavelet_par%d_z,1.)
  
     call from_aux(trim(wavelet_par%tag),"n3",wavelet_par%n_off_x,1)
     call from_aux(trim(wavelet_par%tag),"o3",wavelet_par%off_x_0,0.)
     call from_aux(trim(wavelet_par%tag),"d3",wavelet_par%d_off_x,1.)
  
     call from_aux(trim(wavelet_par%tag),"n4",wavelet_par%n_off_y,1)
     call from_aux(trim(wavelet_par%tag),"o4",wavelet_par%off_y_0,0.)
     call from_aux(trim(wavelet_par%tag),"d4",wavelet_par%d_off_y,1.)
  
     call from_aux(trim(wavelet_par%tag),"tdelay",wavelet_par%t_delay,0.)
  
     call to_history("From Wavelet: n1",wavelet_par%n_t)
     call to_history("From Wavelet: o1",wavelet_par%t_0)
     call to_history("From Wavelet: d1",wavelet_par%d_t)

     call to_history("From Wavelet: n2",wavelet_par%n_z)
     call to_history("From Wavelet: o2",wavelet_par%z_0)
     call to_history("From Wavelet: d2",wavelet_par%d_z)
  
     call to_history("From Wavelet: n3",wavelet_par%n_off_x)
     call to_history("From Wavelet: o3",wavelet_par%off_x_0)
     call to_history("From Wavelet: d3",wavelet_par%d_off_x)
  
     call to_history("From Wavelet: n4",wavelet_par%n_off_y)
     call to_history("From Wavelet: o4",wavelet_par%off_y_0)
     call to_history("From Wavelet: d4",wavelet_par%d_off_y)
  
     call to_history("From Wavelet: tdelay",wavelet_par%t_delay)
  !}
  end if
  if(mod(wavelet_par%n_z+1,2) /= 0) call seperr("n_z of wavelet myust be odd")
  
  return
end subroutine param_wavelet

subroutine param_bound(gen_par,bound_par)
  use genrev3d_types
  use bound_types
  use sep

  type (genrev3d_param_type) gen_par
  type (bound_param_type) bound_par

  call from_param('n_z_bound',bound_par%n_z_bound,0)
  call from_param('n_mx_bound',bound_par%n_mx_bound,0)
  call from_param('n_my_bound',bound_par%n_my_bound,0)
  call from_param('n_hx_bound',bound_par%n_hx_bound,0)
  call from_param('n_hy_bound',bound_par%n_hy_bound,0)

  call from_param('z_bound_stdev',bound_par%z_bound_stdev,0.)
  call from_param('mx_bound_stdev',bound_par%mx_bound_stdev,0.)
  call from_param('my_bound_stdev',bound_par%my_bound_stdev,0.)
  call from_param('hx_bound_stdev',bound_par%hx_bound_stdev,0.)
  call from_param('hy_bound_stdev',bound_par%hy_bound_stdev,0.)

  call from_param('z_bound_min',bound_par%z_bound_min,0.)
  call from_param('mx_bound_min',bound_par%mx_bound_min,0.)
  call from_param('my_bound_min',bound_par%my_bound_min,0.)
  call from_param('hx_bound_min',bound_par%hx_bound_min,0.)
  call from_param('hy_bound_min',bound_par%hy_bound_min,0.)

  bound_par%bound_z=(bound_par%n_z_bound /= 0)
  bound_par%bound_mx=(bound_par%n_mx_bound /= 0)
  bound_par%bound_my=(bound_par%n_my_bound /= 0)
  bound_par%bound_hx=(bound_par%n_hx_bound /= 0)
  bound_par%bound_hy=(bound_par%n_hy_bound /= 0)

  if(gen_par%debug2) write(0,*)'bound_z',bound_par%bound_z
  if(gen_par%debug2) write(0,*)'bound_mx',bound_par%bound_mx
  if(gen_par%debug2) write(0,*)'bound_my',bound_par%bound_my
  if(gen_par%debug2) write(0,*)'bound_hx',bound_par%bound_hx
  if(gen_par%debug2) write(0,*)'bound_hy',bound_par%bound_hy

  bound_par%bound_any= bound_par%bound_z .or. &
                       bound_par%bound_mx .or. bound_par%bound_my .or. &
                       bound_par%bound_hx .or. bound_par%bound_hy
  if(gen_par%debug2) write(0,*)'bound_any',bound_par%bound_any

  gen_par%apply_bound=bound_par%bound_any
  if(gen_par%debug2) write(0,*)'apply_bound',gen_par%apply_bound

end subroutine param_bound

subroutine param_check(gen_par,data_par,wave_par,wavelet_par,slow_par, &
                       rev_extrap_par,image_par)
  use sep
  use genrev3d_types
  use data_types
  use wave_types
  use wavelet_types
  use slow_types
  use rev_extrap_types
  use image_types

  implicit none

  type (genrev3d_param_type) gen_par
  type (data_param_type) data_par
  type (wave_param_type) wave_par
  type (wavelet_param_type) wavelet_par
  type (slow_param_type) slow_par
  type (rev_extrap_param_type) rev_extrap_par
  type (image_param_type) image_par
  

  ! Check parameters


  ! first check wether implemented
  select case (gen_par%domain_range(1:8))
  case ('image_da')
    if (gen_par%source_type(1:6) == "point_") then
    else
     if (gen_par%first_adj) then
     else
       call seperr('domain_range=image_data and source_type=point_source cannot  run forward')
     end if
    end if
  case ('lin_imag')
  case ('image_gl')
     if (gen_par%first_adj) then
       call seperr('domain_range=image_global_data can only run forward')
     end if
  case ('vel_data')
     if (gen_par%first_adj) then
       call seperr('domain_range=vel_data can only run forward')
     end if
  case default
     write(0,*)'gen_par-domain_range=',trim(gen_par%domain_range)
     call seperr('Not supported domain_range')
  end select

  select case (gen_par%oper_type(1:2))
  case ('co')
     gen_par%oper_type(1:10)='const_slow'
     if (slow_par%file /= SLOW_NO_FILE) then
        write(0,*)'WARNING: Does not need a Slow file'
        !      call seperr('Does not need a Slow file')
     end if
     if (wave_par%file == WAVE_INPUT) then
        call seperr('It should not try to read Wavefield file')
     end if
  case ('fu')
     gen_par%oper_type(1:5)='full'
  case default
     write(0,*)'gen_par-oper_type=',trim(gen_par%oper_type)
     call seperr('Not supported oper_type')
  end select

  !check that grids are congruent when source_type is point_source

  select case (gen_par%domain_range(1:8))
  case ('image_da')
  !{
    if (gen_par%source_type(1:6) == "point_") then
      if(wavelet_par%d_t /= rev_extrap_par%d_t_first) then
        write(0,*)'wavelet_par-d_t=',wavelet_par%d_t
        write(0,*)'rev_extrap_par-d_t=',rev_extrap_par%d_t_first
        call seperr("wavelet_par-d_t /= rev_extrap_par-d_t")
      end if
    end if
    if(image_par%d_t /= rev_extrap_par%d_t_first) then
      write(0,*)'image_par-d_t=',image_par%d_t
      write(0,*)'rev_extrap_par-d_t=',rev_extrap_par%d_t_first
      call seperr("image_par-d_t /= rev_extrap_par-d_t")
    end if
    if(amod(image_par%t_0,image_par%d_t) /= 0. ) then
      write(0,*)'image_par-t_0=',image_par%t_0
      write(0,*)'image_par-d_t=',image_par%d_t
      write(0,*) 'mod=',amod(image_par%t_0,image_par%d_t)
!        call seperr("image_par-t_0 not multiple of image_par-d_t")
    end if
    if(image_par%n_t_lag /= 1) then
      if(image_par%d_t /= image_par%d_t_lag) then
        write(0,*)'image_par-d_t=',image_par%d_t
        write(0,*)'image_par-d_t_lag=',image_par%d_t_lag
        call seperr("image_par-d_t /= image_par-d_t_lag")
      end if
    end if

    if(wave_par%d_mx /= image_par%d_mx) then
      write(0,*)'wave_par-d_mx=',wave_par%d_mx
      write(0,*)'image_par-d_mx=',image_par%d_mx
      call seperr("wave-d_mx /=  image-d_mx")
    end if

  !}
  case ('lin_imag')
  !{
    if(wavelet_par%d_t /= rev_extrap_par%d_t_first) then
      write(0,*)'wavelet_par-d_t=',wavelet_par%d_t
      write(0,*)'rev_extrap_par-d_t=',rev_extrap_par%d_t_first
      call seperr("wavelet_par-d_t /= rev_extrap_par-d_t")
    end if
    if(.not. image_par%image_memory) then
      call seperr("With lin_image_data image must be in memory")
    end if
    if(2*data_par%n_off_x_taper >= data_par%n_off_x ) then
      call seperr("Width of data taper exceeds width of data array")
    end if
  !}
  case ('image_gl')
  !{
    if(image_par%d_t /= rev_extrap_par%d_t) then
      write(0,*)'image_par-d_t=',image_par%d_t
      write(0,*)'rev_extrap_par-d_t=',rev_extrap_par%d_t
      call seperr("image_par-d_t /= rev_extrap_par-d_t")
    end if
    if(amod(image_par%t_0,image_par%d_t) /= 0. ) then
      write(0,*)'image_par-t_0=',image_par%t_0
      write(0,*)'image_par-d_t=',image_par%d_t
      write(0,*) 'mod=',amod(image_par%t_0,image_par%d_t)
!        call seperr("image_par-t_0 not multiple of image_par-d_t")
    end if
    if(image_par%n_t_lag /= 1) then
      if(image_par%d_t /= image_par%d_t_lag) then
        write(0,*)'image_par-d_t=',image_par%d_t
        write(0,*)'image_par-d_t_lag=',image_par%d_t_lag
        call seperr("image_par-d_t /= image_par-d_t_lag")
      end if
    end if

    if (.not. gen_par%first_adj) then
      if(wavelet_par%d_t /= rev_extrap_par%d_t) then
        write(0,*)'wavelet_par-d_t=',wavelet_par%d_t
        write(0,*)'rev_extrap_par-d_t=',rev_extrap_par%d_t
        call seperr("wavelet_par-d_t /= rev_extrap_par-d_t")
      end if
    end if
  
    if(wave_par%d_mx /= image_par%d_mx) then
      write(0,*)'wave_par-d_mx=',wave_par%d_mx
      write(0,*)'image_par-d_mx=',image_par%d_mx
      call seperr("wave-d_mx /=  image-d_mx")
    end if

    if(mod(data_par%off_x_0,data_par%d_mx) /= 0) then
      write(0,*)'data_par-off_x_0=',data_par%off_x_0
      write(0,*)'wave_par-d_mx=',wave_par%d_mx
      call seperr("off_x_0 not multiple of d_mx")
    end if

  !}
  case ('vel_data')
  !{
    if(wavelet_par%d_t /= rev_extrap_par%d_t_second) then
      write(0,*)'wavelet_par-d_t=',wavelet_par%d_t
      write(0,*)'rev_extrap_par-d_t=',rev_extrap_par%d_t_second
      call seperr("wavelet_par-d_t /= rev_extrap_par-d_t")
    end if
  !}
  case default
     write(0,*)'gen_par-domain_range=',trim(gen_par%domain_range)
     call seperr('param_check: Not supported domain_range')
  end select

  if(wave_par%off_x_0 >0) then
    write(0,*)'wave_par-off_x_0=',wave_par%off_x_0
    call seperr("off_x_0 >0")
  end if

  if(mod(wave_par%off_x_0,wave_par%d_mx) /= 0) then
    write(0,*)'wave_par-off_x_0=',wave_par%off_x_0
    write(0,*)'wave_par-d_mx=',wave_par%d_mx
    call seperr("off_x_0_pad not multiple of d_mx")
  end if
  if(wave_par%off_x_0 > data_par%off_x_0) then
    write(0,*)'wave_par-off_x_0=',wave_par%off_x_0
    write(0,*)'data_par-off_x_0=',data_par%off_x_0
    call seperr("off_x_0_pad > off_x_0")
  end if

  if(.not. gen_par%image_zero_offset) then
    if(wave_par%d_mx /= image_par%d_hx) then
      write(0,*)'wave_par-d_mx=',wave_par%d_mx
      write(0,*)'image_par-d_hx=',image_par%d_hx
      call seperr("d_mx /=  d_hx")
    end if
  end if

  if(wave_par%n_my /= 1) then
  !{
    if(amod(wave_par%d_sy,wave_par%d_my) /= 0) then
      write(0,*)'wave_par-d_sy=',wave_par%d_sy
      write(0,*)'wave_par-d_my=',wave_par%d_my
      call seperr("d_sy not multiple of d_my")
    end if
    if(wave_par%off_y_0 >0) then
      write(0,*)'wave_par-off_y_0=',wave_par%off_y_0
      call seperr("off_y_0 >0")
    end if
    if(mod(wave_par%off_y_0,wave_par%d_my) /= 0) then
      write(0,*)'wave_par-d_my=',wave_par%d_my
      write(0,*)'wave_par-off_y_0=',wave_par%off_y_0
      call seperr("off_y_0_pad not multiple of d_my")
    end if
    if(mod(data_par%off_y_0,data_par%d_my) /= 0) then
      write(0,*)'data_par-off_y_0=',data_par%off_y_0
      write(0,*)'wave_par-d_my=',wave_par%d_my
      call seperr("off_y_0 not multiple of d_my")
    end if
    if(wave_par%off_y_0 > data_par%off_y_0) then
      write(0,*)'wave_par-off_y_0=',wave_par%off_y_0
      write(0,*)'data_par-off_y_0=',data_par%off_y_0
      call seperr("off_y_0_pad > off_y_0")
    end if
    if(wave_par%d_my /= image_par%d_my) then
      write(0,*)'wave_par-d_my=',wave_par%d_my
      write(0,*)'image_par-d_my=',image_par%d_my
      call seperr("d_my /=  d_my")
    end if

    if(.not. gen_par%image_zero_offset) then
      if(wave_par%d_my /= image_par%d_hy) then
        write(0,*)'wave_par-d_my=',wave_par%d_my
        write(0,*)'image_par-d_hy=',image_par%d_hy
        call seperr("d_my /= d_hy")
      end if
    end if

  !}
  end if


  if(gen_par%image_zero_offset .and. image_par%write_image_hz) then
     call seperr("image_zero_offset unconsinstent with write_image_hz")
  end if

  if(gen_par%surf_offset_cig .and.  &
     ((.not. image_par%image_memory) .or. &
      (.not. gen_par%image_zero_offset) &
     ) &
    ) then
     call seperr("surf_offset_cig only with image_zero_offset and image_memory")
  end if

  if(gen_par%write_image_progress_with_shots .and. &
     (image_par%write_image_progress .or. image_par%n_t_lag /= 1)) then
     call seperr("write_image_progress_with_shots only with simple imaging ")
  end if


  return
end subroutine param_check

