!*$*OPTIMIZE(5)
!*$*ROUNDOFF(3)
!*$*SCALAROPTIMIZE(3)

module rev_extrap_gen_lin_imag_mod
  use genrev3d_types
  use data_types
  use image_types
  use slow_types
  use wave_types
  use wavelet_types
  use rev_extrap_types

  use wave_mod
  use wavelet_mod
  use image_mod
  use data_mod
  use slow_mod

  use timers_mod

!  use down_gen_ser_mod
!  use image_mod
!  use filter_data_mod

  implicit none
contains
  function rev_extrap_gen_lin_imag( &
    gen_par,data_par,image_par,rev_extrap_par,slow_par,wave_par,wavelet_par, &
    rev_extrap_oper,rev_der2time_oper,rev_scatter_oper, &
    slow_block,&
    dslow_block,&
    force_block,&
    wave_block,&
    wave_back_block,&
    wavelet, &
    image_mem, &
    image_mem_hz) &
  result(stat)

  interface 
    function rev_extrap_oper( &
                             rev_extrap_par, &
                             wave_par, &
                             force_block, &
                             wave_block, slow_block) &
                             result(stat)
      use rev_extrap_types, only: rev_extrap_param_type
      use data_types, only: data_param_type
      use wave_types, only: wave_param_type
      integer                                   :: stat
      type (rev_extrap_param_type)              :: rev_extrap_par
      type (wave_param_type)                    :: wave_par
      real,  dimension (:,:,:,:)                :: force_block
      real,  dimension (:,:,:,:,:,:)            :: wave_block
      real, optional,  dimension (:,:,:,:,:,:)  :: slow_block
    end function rev_extrap_oper

    function rev_der2time_oper (&
          rev_extrap_par, &
          wave_par, &
          wave_block, &
          wave_back_block, &
          slow_block) &
        result(stat)
      use rev_extrap_types, only: rev_extrap_param_type
      use wave_types, only: wave_param_type
      integer                                   :: stat
      type (rev_extrap_param_type)              :: rev_extrap_par
      type (wave_param_type)                    :: wave_par
      real,  dimension (:,:,:,:,:,:)            :: wave_block
      real,  dimension (:,:,:,:,:,:)            :: wave_back_block
      real,  dimension (:,:,:,:,:,:)            :: slow_block
    end function rev_der2time_oper
  
    function rev_scatter_oper( &
        rev_extrap_par, &
        wave_par, &
        slow_par, &
        wave_block,wave_back_block, &
        slow_block,dslow_block) &
      result(stat)
      use rev_extrap_types, only: rev_extrap_param_type
      use wave_types, only: wave_param_type
      use slow_types, only: slow_param_type
      integer                                   :: stat
      type (rev_extrap_param_type)              :: rev_extrap_par
      type (wave_param_type)                    :: wave_par
      type (slow_param_type)                    :: slow_par
      real,  dimension (:,:,:,:,:,:)            :: wave_block
      real,  dimension (:,:,:,:,:,:)            :: wave_back_block
      real, optional,  dimension (:,:,:,:,:,:)  :: slow_block
      real, optional,  dimension (:,:,:,:,:,:)  :: dslow_block
    end function rev_scatter_oper
  
  end interface

    integer stat
    type(genrev3d_param_type) :: gen_par
    type (data_param_type) :: data_par
    type (image_param_type) :: image_par
    type (rev_extrap_param_type) :: rev_extrap_par
    type (slow_param_type) :: slow_par
    type (wave_param_type) :: wave_par
    type (wavelet_param_type) :: wavelet_par

    optional rev_scatter_oper


    integer stat_par(gen_par%n_threads)
    real, dimension    (:,:,:,:,:,:) :: slow_block
    real, dimension    (:,:,:,:,:,:) :: dslow_block
    real, dimension (:,:,:,:) :: force_block
    real, dimension (:,:,:,:,:,:) :: wave_block
    real, dimension (:,:,:,:,:,:) :: wave_back_block
    real, dimension (:,:,:,:) :: wavelet
    real, dimension (:,:,:,:) :: image_mem,image_mem_hz 

    integer, external :: omp_get_thread_num

    integer ierr,i_t_snap
    integer rat_t_snap_wave_second,rat_t_snap_image
    integer rat_t_snap_data
    integer rat_t_snap_wave_first
    integer rat_t_snap_wave_back_first,rat_t_image
    integer min_it_snap_wave_second,min_it_snap_image
    integer min_it_snap_data
    integer min_it_snap_wave_first
    integer min_it_snap_wave_back_first,min_it_image
    integer max_it_snap_wave_second,max_it_snap_image
    integer max_it_snap_data
    integer max_it_snap_wave_first
    integer max_it_snap_wave_back_first,max_it_image
    integer n_s1_par,n_s2_par,i_s1_par,i_s2_par,i_t_extrap
    integer i_t_extrap_second_beg,i_t_extrap_second_end,i_t_extrap_second_step

    real t_extrap,epsi_image_d_t,epsi_wave_d_t,epsi_data_d_t
    real s1_par_0,s2_par_0,d_s1_par,d_s2_par
    real norm_wave,courant_numb

    stat_par=0 ! in case that we never call the inner routines

    epsi_image_d_t=.01*image_par%d_t
    epsi_wave_d_t=.01*wave_par%d_t
    epsi_data_d_t=.01*data_par%d_t

    ! read appropriate wavelet files
    if(wavelet_par%file == WAVELET_INPUT) then
      if(gen_par%debug1) write(0,*)'Read Wavelet'
      call read_wavelet(wavelet_par%tag,wavelet_par,wavelet)
    end if
    call snap('Wave-lin.H',wavelet_par%n_t,1,wavelet)

    n_s1_par=data_par%n_sx
    n_s2_par=data_par%n_sy
    d_s1_par=data_par%d_sx
    d_s2_par=data_par%d_sy
    s1_par_0=data_par%sx_0
    s2_par_0=data_par%sy_0

    min_it_snap_wave_first=nint((data_par%t_0-rev_extrap_par%t_0_first)/ &
                              rev_extrap_par%d_t_first)
    max_it_snap_wave_first=rev_extrap_par%n_t_first-1
    rat_t_snap_wave_first=nint(abs(wave_par%d_t)/rev_extrap_par%d_t_first)
    if(rat_t_snap_wave_first == 0) rat_t_snap_wave_first=HUGE(1)
    if(gen_par%debug2) &
      write(0,*)'min_it_snap_wave_first=',min_it_snap_wave_first
    if(gen_par%debug2) &
      write(0,*)'max_it_snap_wave_first=',max_it_snap_wave_first
    if(gen_par%debug2) &
      write(0,*)'rat_t_snap_wave_first=',rat_t_snap_wave_first

! ATTENTION same expression used in output_wave
    min_it_snap_wave_back_first=&
               nint((wave_par%t_back_0-rev_extrap_par%t_0_first)/ & 
                     rev_extrap_par%d_t_first)+1
    max_it_snap_wave_back_first=rev_extrap_par%n_t_first
    rat_t_snap_wave_back_first=nint(wave_par%d_t_back/rev_extrap_par%d_t_first)
    if(rat_t_snap_wave_back_first == 0) rat_t_snap_wave_back_first=HUGE(1)
    if(gen_par%debug2) &
      write(0,*)'min_it_snap_wave_back_first=',min_it_snap_wave_back_first
    if(gen_par%debug2) &
      write(0,*)'max_it_snap_wave_back_first=',max_it_snap_wave_back_first
    if(gen_par%debug2) &
      write(0,*)'rat_t_snap_wave_back_first=',rat_t_snap_wave_back_first

    if (rev_extrap_par%adj) then
    !{
      i_t_extrap_second_beg=1
      i_t_extrap_second_end=rev_extrap_par%n_t_second
      i_t_extrap_second_step=1

      min_it_snap_wave_second=nint((wave_par%t_0-rev_extrap_par%t_0_second)/ &
                            rev_extrap_par%d_t_second)+1
      max_it_snap_wave_second=rev_extrap_par%n_t_second
      rat_t_snap_wave_second=nint(wave_par%d_t/rev_extrap_par%d_t_second)
      if(rat_t_snap_wave_second == 0) rat_t_snap_wave_second=HUGE(1)
      if(gen_par%debug2) &
        write(0,*)'min_it_snap_wave_second=',min_it_snap_wave_second
      if(gen_par%debug2) &
        write(0,*)'max_it_snap_wave_second=',max_it_snap_wave_second
      if(gen_par%debug2) &
        write(0,*)'rat_t_snap_wave_second=',rat_t_snap_wave_second

      if(image_par%write_image_progress) then
        min_it_snap_image=nint((wave_par%t_0-rev_extrap_par%t_0_second)/ &
                              rev_extrap_par%d_t_second)+1
        max_it_snap_image=rev_extrap_par%n_t_second
        rat_t_snap_image=nint(wave_par%d_t/rev_extrap_par%d_t_second)
        if(rat_t_snap_image == 0) rat_t_snap_image=HUGE(1)
      else
        min_it_snap_image=i_t_extrap_second_end
        max_it_snap_image=i_t_extrap_second_end
        rat_t_snap_image=1
      end if
      if(gen_par%debug2) &
        write(0,*)'min_it_snap_image=',min_it_snap_image
      if(gen_par%debug2) &
        write(0,*)'max_it_snap_image=',max_it_snap_image
      if(gen_par%debug2) &
        write(0,*)'rat_t_snap_image=',rat_t_snap_image


      min_it_image=&
         nint(((image_par%t_0+(image_par%n_t-1)*image_par%d_t)-&
                rev_extrap_par%t_0_second)/rev_extrap_par%d_t_second)+1
      max_it_image=nint((image_par%t_0-&
                rev_extrap_par%t_0_second)/rev_extrap_par%d_t_second)+1
      rat_t_image= &
                nint(-image_par%d_t/rev_extrap_par%d_t_second)

!      min_it_image=&
!         nint(((wave_par%t_back_0+(wave_par%n_t_back-1)*wave_par%d_t_back)-&
!                rev_extrap_par%t_0_second)/rev_extrap_par%d_t_second)+1
!      max_it_image=nint((wave_par%t_back_0-&
!                rev_extrap_par%t_0_second)/rev_extrap_par%d_t_second)+1
!      rat_t_image= &
!                nint(-wave_par%d_t_back/rev_extrap_par%d_t_second)

      if(rat_t_image == 0) rat_t_image=-1
      if(gen_par%debug2) & 
        write(0,*)'min_it_image=',min_it_image
      if(gen_par%debug2) & 
        write(0,*)'max_it_image=',max_it_image
      if(gen_par%debug2) & 
        write(0,*)'rat_t_image=',rat_t_image

      min_it_snap_data=&
         nint(((data_par%t_0+(data_par%n_t-1)*data_par%d_t)-&
                rev_extrap_par%t_0_second)/rev_extrap_par%d_t_second)+1
      max_it_snap_data=nint((data_par%t_0-&
                rev_extrap_par%t_0_second)/rev_extrap_par%d_t_second)+1
      rat_t_snap_data=nint(-data_par%d_t/rev_extrap_par%d_t_second)
  
      if(rat_t_snap_data == 0) rat_t_snap_data=-1
      if(gen_par%debug2) write(0,*)'min_it_snap_data=',min_it_snap_data
      if(gen_par%debug2) write(0,*)'max_it_snap_data=',max_it_snap_data
      if(gen_par%debug2) write(0,*)'rat_t_snap_data=',rat_t_snap_data
    !}
    else
    !{
      i_t_extrap_second_beg=1
      i_t_extrap_second_end=rev_extrap_par%n_t_second
      i_t_extrap_second_step=1

      min_it_snap_wave_second=nint((wave_par%t_0-rev_extrap_par%t_0_second)/ &
                            rev_extrap_par%d_t_second)
      max_it_snap_wave_second=rev_extrap_par%n_t_second
      rat_t_snap_wave_second=nint(wave_par%d_t/rev_extrap_par%d_t_second)
      if(rat_t_snap_wave_second == 0) rat_t_snap_wave_second=HUGE(1)
      if(gen_par%debug2) &
        write(0,*)'min_it_snap_wave_second=',min_it_snap_wave_second
      if(gen_par%debug2) &
        write(0,*)'max_it_snap_wave_second=',max_it_snap_wave_second
      if(gen_par%debug2) &
        write(0,*)'rat_t_snap_wave_second=',rat_t_snap_wave_second

      min_it_snap_image=-1
      max_it_snap_image=-1
      rat_t_snap_image=HUGE(-1)

      min_it_image=&
         nint((image_par%t_0-&
                rev_extrap_par%t_0_second)/rev_extrap_par%d_t_second)+1
      max_it_image= &
         nint(((image_par%t_0+(image_par%n_t-1)*image_par%d_t)-&
                rev_extrap_par%t_0_second)/rev_extrap_par%d_t_second)+1
      rat_t_image= &
                nint(image_par%d_t/rev_extrap_par%d_t_second)
  
!      min_it_image=nint((wave_par%t_back_0-&
!                rev_extrap_par%t_0_second)/rev_extrap_par%d_t_second)+1
!      max_it_image=&
!         nint(((wave_par%t_back_0+(wave_par%n_t_back-1)*wave_par%d_t_back)-&
!                rev_extrap_par%t_0_second)/rev_extrap_par%d_t_second)+1
!      rat_t_image= &
!                nint(wave_par%d_t_back/rev_extrap_par%d_t_second)

      if(rat_t_image == 0) rat_t_image=-1
      if(gen_par%debug2) & 
        write(0,*)'min_it_image=',min_it_image
      if(gen_par%debug2) & 
        write(0,*)'max_it_image=',max_it_image
      if(gen_par%debug2) & 
        write(0,*)'rat_t_image=',rat_t_image

      min_it_snap_data=nint((data_par%t_0-&
                rev_extrap_par%t_0_second)/rev_extrap_par%d_t_second)+1
      max_it_snap_data=&
         nint(((data_par%t_0+(data_par%n_t-1)*data_par%d_t)-&
                rev_extrap_par%t_0_second)/rev_extrap_par%d_t_second)+1
      rat_t_snap_data=nint(data_par%d_t/rev_extrap_par%d_t_second)
  
      if(rat_t_snap_data == 0) rat_t_snap_data=-1
      if(gen_par%debug2) write(0,*)'min_it_snap_data=',min_it_snap_data
      if(gen_par%debug2) write(0,*)'max_it_snap_data=',max_it_snap_data
      if(gen_par%debug2) write(0,*)'rat_t_snap_data=',rat_t_snap_data
    !}
    end if

  if (image_par%file == IMAGE_INPUT) then
  !{
    if(gen_par%debug1) write(0,*)'Input Image'
    call input_image_lin(gen_par,image_par,wave_par, dslow_block)
  !}
  end if

  do_s2_par : do i_s2_par=1,n_s2_par
  !{
  do_s1_par : do i_s1_par=1,n_s1_par
  !{
!    if(gen_par%surf_offset_cig) dslow_block=0
    dslow_block=0

    rev_extrap_par%s1_par=s1_par_0 + (i_s1_par-1)*d_s1_par
    rev_extrap_par%s2_par=s2_par_0 + (i_s2_par-1)*d_s2_par 

    rev_extrap_par%i_s1_par=i_s1_par
    rev_extrap_par%i_s2_par=i_s2_par

    data_par%i_s1_par=i_s1_par
    data_par%i_s2_par=i_s2_par
    rev_extrap_par%last_shot=((i_s1_par == n_s1_par) .and. & 
                              (i_s2_par == n_s2_par))

    if (gen_par%source_type(1:6) == "point_") then
      if(gen_par%verb) then
        write(0,*)'i_s1_par=',i_s1_par,'at location x=',rev_extrap_par%s1_par
        write(0,*)'i_s2_par=',i_s2_par,'at location y=',rev_extrap_par%s2_par
      end if

      wave_par%mx_0=rev_extrap_par%s1_par + wave_par%off_x_0
      wave_par%my_0=rev_extrap_par%s2_par + wave_par%off_y_0

      if(gen_par%debug2) write(0,*)'wave_par-mx_0=',wave_par%mx_0
      if(gen_par%debug2) write(0,*)'wave_par-my_0=',wave_par%my_0

    else
      if(gen_par%verb) then
        write(0,*)'i_s1_par=',i_s1_par,'with value=',rev_extrap_par%s1_par
        write(0,*)'i_s2_par=',i_s2_par,'with value=',rev_extrap_par%s2_par
      end if
    end if

    wave_block=0.
    wave_back_block=0.
    force_block=0

    ! read appropriate veocity files
    if(slow_par%file == SLOW_INPUT) then
       if(gen_par%debug1) write(0,*)'Read Slow'
         call SEP_timer_start(11)
         call read_slow(gen_par,slow_par%tag, wave_par,slow_par, slow_block)
         call SEP_timer_stop(11)

    else
      if(gen_par%debug1) write(0,*)'DO NOT read Slow'
      slow_par%slow_min=gen_par%slow_const
    end if

    courant_numb=max(abs(rev_extrap_par%d_t_first), &
                     abs(rev_extrap_par%d_t_second))/ &
              (slow_par%slow_min*min(wave_par%d_mx,wave_par%d_z))
    if(gen_par%verb) write(0,*)'Courant number=',courant_numb, &
        'Maximum allowed is',gen_par%max_courant_numb

    if(gen_par%check_stability) then
      if(courant_numb > gen_par%max_courant_numb) then
        write(0,*)'Courant number=',courant_numb, &
        'larger than maximum allowed of',gen_par%max_courant_numb
        call seperr('Exceeded CFL stability condition')
      end if
    end if


    if((.not. gen_par%skip_source_prop) .and. &
       (((i_s1_par==1).and.(i_s2_par==1)).or. &
         (.not. gen_par%layered_mig_vel)) &
      ) then
    !{

      if(gen_par%verb) write(0,*)'Start first time loop'

      rev_extrap_par%sou_rec_wave=SOU_WAVE

!      rev_extrap_par%i_t_extrap=0
!      rev_extrap_par%t= rev_extrap_par%t_0_first+ &
!                          (rev_extrap_par%i_t_extrap)*rev_extrap_par%d_t_first
!
!      if(gen_par%verb) then 
!        write(0,*)'Source wavefield: Doing time step #', &
!                      rev_extrap_par%i_t_extrap
!        write(0,*)'Source wavefield: time=',rev_extrap_par%t
!      end if

! determine wether to write out wavefield
!      rev_extrap_par%snap_wave_first_step= &
! (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_first,rat_t_snap_wave_first) &
!       ==0).and.&
!       (rev_extrap_par%i_t_extrap >= min_it_snap_wave_first).and.&
!       (rev_extrap_par%i_t_extrap <= max_it_snap_wave_first)
!      if(gen_par%verb) write(0,*)'snap_wave_first_step=',rev_extrap_par%snap_wave_first_step
!      if(gen_par%verb) write(0,*)'i_t_extrap=',rev_extrap_par%i_t_extrap
!
!
!      rev_extrap_par%wave_back_step= &
! (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_back,rat_t_snap_wave_back)==0)&
!   .and.&
!       (rev_extrap_par%i_t_extrap >= min_it_snap_wave_back).and.&
!       (rev_extrap_par%i_t_extrap <= max_it_snap_wave_back)
!      if(gen_par%verb) write(0,*)'wave_back_step=',rev_extrap_par%wave_back_step
!
      rev_extrap_par%t= rev_extrap_par%t_0_first - rev_extrap_par%d_t_first

      if(wavelet_par%file == WAVELET_INPUT) then
        force_block=0
        call add_wavelet(gen_par,wave_par,wavelet_par,rev_extrap_par, &
                           force_block,wavelet)
      end if

!      if(wave_par%file == WAVE_OUTPUT) then
!        if(rev_extrap_par%snap_wave_first_step) then
!          if(gen_par%debug1) write(0,*)'Output Wavefield_sou'
!          call output_wave(wave_par%tag_first,gen_par,wave_par, &
!                     rev_extrap_par,wave_block)
!        end if
!      end if
!
!      if(wave_par%file_back == WAVE_OUTPUT) then
!        if(rev_extrap_par%wave_back_step) then
!          stat= rev_der2time_oper (&
!            rev_extrap_par=rev_extrap_par, &
!            wave_par=wave_par, &
!            wave_block=wave_block, &
!            wave_back_block=wave_back_block)
!           if(gen_par%debug1) write(0,*)'Output Der2Time Wavefield_back'
!           call output_wave(wave_par%tag_back,gen_par,wave_par, &
!                    rev_extrap_par,wave_back_block)
!        end if
!      end if

      do i_t_extrap=1,rev_extrap_par%n_t_first
      !{

        rev_extrap_par%i_t_extrap=i_t_extrap

        rev_extrap_par%t= rev_extrap_par%t_0_first+ &
                          (rev_extrap_par%i_t_extrap-1)*rev_extrap_par%d_t_first

        if(gen_par%debug1) then 
          write(0,*)'Source wavefield: Doing time step #', &
                        rev_extrap_par%i_t_extrap
          write(0,*)'Source wavefield: time=',rev_extrap_par%t
        end if

        if(gen_par%check_stability) then
         !{
           norm_wave=sqrt(sum(wave_block**2))
           write(0,*)'norm wavefield at time=',rev_extrap_par%t,norm_wave
         !}
        end if

! determine wether to write out wavefield
        rev_extrap_par%snap_wave_first_step= &
          (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_first, &
           rat_t_snap_wave_first)==0).and.&
          (rev_extrap_par%i_t_extrap >= min_it_snap_wave_first).and.&
          (rev_extrap_par%i_t_extrap <= max_it_snap_wave_first)
        if(gen_par%debug2) write(0,*)'snap_wave_first_step=',rev_extrap_par%snap_wave_first_step
!        if(gen_par%verb) write(0,*)'i_t_extrap=',rev_extrap_par%i_t_extrap
!        if(gen_par%verb) write(0,*)'rat_t_snap_wave_first=',rat_t_snap_wave_first
!        if(gen_par%verb) write(0,*)'min_it_snap_wave_first=',min_it_snap_wave_first
!        if(gen_par%verb) write(0,*)'max_it_snap_wave_first=',max_it_snap_wave_first
  
        rev_extrap_par%wave_back_first_step= &
          (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_back_first, &
          rat_t_snap_wave_back_first)==0).and.&
          (rev_extrap_par%i_t_extrap >= min_it_snap_wave_back_first).and.&
          (rev_extrap_par%i_t_extrap <= max_it_snap_wave_back_first)
        if(gen_par%debug2)&
           write(0,*)'wave_back_first_step=',rev_extrap_par%wave_back_first_step
  
!        rev_extrap_par%snap_image_step= &
!          (mod(rev_extrap_par%i_t_extrap-min_it_snap_image,rat_t_snap_image)==0).and.&
!          (rev_extrap_par%i_t_extrap >= min_it_snap_image).and.&
!          (rev_extrap_par%i_t_extrap <= max_it_snap_image)
!        if(gen_par%debug1) write(0,*)'snap_image_step=',rev_extrap_par%snap_image_step

  ! extrapolation step
  
        select case (gen_par%oper_type(1:2))
        case ('co')
          if(gen_par%debug2) write(0,*)'oper_type = Constant'
            stat= rev_extrap_oper( &
              rev_extrap_par=rev_extrap_par, &
              wave_par=wave_par, &
              force_block=force_block, &
              wave_block=wave_block &
              )
        case ('fu')
          if(gen_par%debug2) write(0,*)'oper_type = full'
            stat= rev_extrap_oper( &
              rev_extrap_par=rev_extrap_par, &
              wave_par=wave_par, &
              force_block=force_block, &
              wave_block=wave_block, &
              slow_block=slow_block)
        case default
           write(0,*)'gen_par-oper_type=',trim(gen_par%oper_type)
           call seperr('Not supported oper_type')
        end select

        if(wave_par%file == WAVE_OUTPUT) then
        !{
          if(rev_extrap_par%snap_wave_first_step) then
            if(gen_par%debug2) write(0,*)'Output Wavefield_sou'
            call output_wave(wave_par%tag_first,gen_par,wave_par, &
                       rev_extrap_par,wave_block)
          end if
        !}
        end if
  
        if(wave_par%file_back == WAVE_OUTPUT) then
        !{
          if(rev_extrap_par%wave_back_first_step) then
            if(gen_par%debug1)write(0,*)'Compute Der2Time Wavefield_back'
            stat= rev_der2time_oper (&
              rev_extrap_par=rev_extrap_par, &
              wave_par=wave_par, &
              wave_block=wave_block, &
              wave_back_block=wave_back_block, &
              slow_block=slow_block)
            if(gen_par%debug1)write(0,*)'Output avefield_back'
            call output_wave(wave_par%tag_back,gen_par,wave_par, &
                rev_extrap_par,wave_back_block)

          end if
        !}
        end if

        if(wavelet_par%file == WAVELET_INPUT) then
          force_block=0
          if(gen_par%debug1)write(0,*)'Add wavelet'
          call add_wavelet(gen_par,wave_par,wavelet_par,rev_extrap_par, &
                             force_block,wavelet)
        end if

      !}
      end do
    !}
    end if


! now propagate recorded wavefield and image or forwar scattered wavefiled
    if(gen_par%verb) write(0,*)'Start second time loop'

    wave_block=0.
    wave_back_block=0.
    force_block=0
    rev_extrap_par%sou_rec_wave=REC_WAVE

!    rev_extrap_par%i_t_extrap=0
!    rev_extrap_par%t= rev_extrap_par%t_0_second
!
!    if(gen_par%verb) then 
!      write(0,*)'Receiver wavefield: Doing time step #', &
!                    rev_extrap_par%i_t_extrap
!      write(0,*)'Receiver wavefield: time=',rev_extrap_par%t
!    end if


!      min_it_snap_image=-1
!
!      rev_extrap_par%wave_back_step= &
! (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_back,rat_t_snap_wave_back)==0)&
!   .and.&
!       (rev_extrap_par%i_t_extrap >= min_it_snap_wave_back).and.&
!       (rev_extrap_par%i_t_extrap <= max_it_snap_wave_back)
!      if(gen_par%verb) &
!        write(0,*)'wave_back_step=',rev_extrap_par%wave_back_step
!
!    rev_extrap_par%wave_second_step= &
!      (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_second,rat_t_snap_wave_second) == 0).and.&
!      (rev_extrap_par%i_t_extrap >= min_it_snap_wave_second).and.&
!      (rev_extrap_par%i_t_extrap <= max_it_snap_wave_second)
!    if(gen_par%verb) &
!        write(0,*)'wave_second_step=',rev_extrap_par%wave_second_step

!    rev_extrap_par%snap_image_step= &
!      (mod(rev_extrap_par%i_t_extrap-min_it_snap_image,rat_t_snap_image)==0).and.&
!      (rev_extrap_par%i_t_extrap >= min_it_snap_image).and.&
!      (rev_extrap_par%i_t_extrap <= max_it_snap_image)
!    if(gen_par%verb) write(0,*)'snap_image_step=',rev_extrap_par%snap_image_step
!
!    rev_extrap_par%snap_data_step= &
!      (mod(rev_extrap_par%i_t_extrap-min_it_snap_data,rat_t_snap_data)==0).and.&
!      (rev_extrap_par%i_t_extrap >= min_it_snap_data).and.&
!      (rev_extrap_par%i_t_extrap <= max_it_snap_data)
!    if(gen_par%verb) write(0,*)'snap_data_step=',rev_extrap_par%snap_data_step

!    if(rev_extrap_par%wave_back_step) then
!      call input_wave(wave_par%tag_back,gen_par,wave_par,rev_extrap_par,& 
!               wave_back_block)
!      stat=rev_scatter_oper(rev_extrap_par,wave_par, &
!         wave_block,wave_back_block,slow_block,dslow_block) 
!    end if

!    if(wave_par%file == WAVE_OUTPUT) then
!      if(rev_extrap_par%wave_second_step) then
!        if(gen_par%debug1) write(0,*)'Output Wavefield_rec'
!        call output_wave(wave_par%tag_second,gen_par,wave_par,rev_extrap_par, &
!                          wave_block)
!      end if
!    end if
  
!    if(data_par%file == DATA_OUTPUT) then
!      if(rev_extrap_par%snap_data_step) then
!        if(gen_par%debug1) write(0,*)'Output Data'
!        call output_data(data_par%tag,gen_par,data_par,wave_par,rev_extrap_par,&
!                         wave_block)
!      end if
!    end if



!    !*$*assert do (serial) 
    do i_t_extrap= &
      i_t_extrap_second_beg,i_t_extrap_second_end,i_t_extrap_second_step
    !{
      rev_extrap_par%i_t_extrap=i_t_extrap
      rev_extrap_par%t= rev_extrap_par%t_0_second+ &
         (rev_extrap_par%i_t_extrap-1)*rev_extrap_par%d_t_second

      if(gen_par%debug1) then 
        write(0,*)'Receiver (source) wavefield: Doing time step #', &
                      rev_extrap_par%i_t_extrap
        write(0,*)'Receiver (source) wavefield: time=',rev_extrap_par%t
      end if

      rev_extrap_par%image_step= &
 (mod(rev_extrap_par%i_t_extrap-min_it_image,rat_t_image)==0)&
   .and.&
       (rev_extrap_par%i_t_extrap >= min_it_image).and.&
       (rev_extrap_par%i_t_extrap <= max_it_image)
      if(gen_par%debug2) write(0,*)'image_step=',rev_extrap_par%image_step

      rev_extrap_par%snap_wave_second_step= &
        (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_second,rat_t_snap_wave_second)==0).and. &
        (rev_extrap_par%i_t_extrap >= min_it_snap_wave_second).and.&
        (rev_extrap_par%i_t_extrap <= max_it_snap_wave_second)
      if(gen_par%debug2) write(0,*)'snap_wave_second_step=',rev_extrap_par%snap_wave_second_step

      rev_extrap_par%snap_image_step= &
        (mod(rev_extrap_par%i_t_extrap-min_it_snap_image,rat_t_snap_image)==0).and.&
        (rev_extrap_par%i_t_extrap >= min_it_snap_image).and.&
        (rev_extrap_par%i_t_extrap <= max_it_snap_image)
      if(gen_par%debug2) write(0,*)'snap_image_step=',rev_extrap_par%snap_image_step

      rev_extrap_par%snap_data_step= &
        (mod(rev_extrap_par%i_t_extrap-min_it_snap_data,rat_t_snap_data)==0).and. &
        (rev_extrap_par%i_t_extrap >= min_it_snap_data).and.&
        (rev_extrap_par%i_t_extrap <= max_it_snap_data)
      if(gen_par%debug2) write(0,*)'snap_data_step=',rev_extrap_par%snap_data_step

      if(gen_par%check_stability) then
       !{
         norm_wave=sqrt(sum(wave_block**2))
         write(0,*)'norm wavefield at time=',rev_extrap_par%t,norm_wave
       !}
      end if

      if (data_par%file == DATA_INPUT) then
        !{
        if(rev_extrap_par%snap_data_step ) then
          if(gen_par%debug1) write(0,*)'Read receiver Data'
          call SEP_timer_start(14)
          call input_data(data_par%tag,gen_par,data_par,wave_par,& 
                   rev_extrap_par,force_block,wave_block)
          call SEP_timer_stop(14)
        end if
        !}
      end if

      if(rev_extrap_par%adj) then
        if(rev_extrap_par%image_step) then
          if(gen_par%debug1)write(0,*)'Input wavefield_back'
          call input_wave(wave_par%tag_back,gen_par,wave_par,rev_extrap_par,& 
                 wave_back_block)
          if(gen_par%debug1)write(0,*)'Apply Imaging condition'
          stat=rev_scatter_oper(rev_extrap_par,wave_par,slow_par, &
             wave_block,wave_back_block,slow_block,dslow_block) 
        end if
      end if

! extrapolation step

      select case (gen_par%oper_type(1:2))
      case ('co')
        if(gen_par%debug2) write(0,*)'oper_type = Constant'
          stat= rev_extrap_oper( &
            rev_extrap_par=rev_extrap_par, &
            wave_par=wave_par, &
            force_block=force_block, &
            wave_block=wave_block &
            )
      case ('fu')
        if(gen_par%debug2) write(0,*)'oper_type = full'
    call SEP_timer_start(17)
          stat= rev_extrap_oper( &
            rev_extrap_par=rev_extrap_par, &
            wave_par=wave_par, &
            force_block=force_block, &
            wave_block=wave_block, &
            slow_block=slow_block)
    call SEP_timer_stop(17)
      case default
         write(0,*)'gen_par-oper_type=',trim(gen_par%oper_type)
         call seperr('Not supported oper_type')
      end select


      if(.not. rev_extrap_par%adj) then
        if(rev_extrap_par%image_step) then
          if(gen_par%debug1)write(0,*)'Input wavefield_back'
          call input_wave(wave_par%tag_back,gen_par,wave_par,rev_extrap_par,& 
                 wave_back_block)
          if(gen_par%debug1)write(0,*)'Apply scattering term'
          stat=rev_scatter_oper(rev_extrap_par,wave_par,slow_par, &
             wave_block,wave_back_block,slow_block,dslow_block) 
        end if
      end if

      if(wave_par%file == WAVE_OUTPUT) then
      !{
        if(rev_extrap_par%snap_wave_second_step) then
          if(gen_par%debug1) write(0,*)'Output Wavefield_rec'
          call output_wave(wave_par%tag_second,gen_par,wave_par, &
                  rev_extrap_par,wave_block)
        end if
      !}
      end if

      if(data_par%file == DATA_OUTPUT) then
      !{
        if(rev_extrap_par%snap_data_step ) then
          if(gen_par%debug1) write(0,*)'Output Data'
          call output_data(data_par%tag,gen_par,data_par,wave_par, &
                  rev_extrap_par,wave_block)
       end if
      !}
      end if

!      if (image_par%file == IMAGE_OUTPUT) then
!      !{
!        if(rev_extrap_par%snap_image_step ) then
!          if(gen_par%debug1) write(0,*)'Output Image'
!          call output_image(image_par%tag,image_par%tag_hz,gen_par,image_par, &
!                             wave_par,rev_extrap_par, &
!                             max_it_snap_image, &
!                             wave_block,image_mem,image_mem_hz)
!        end if
!      !}
!      end if

!      force_block=0
!      if (data_par%file == DATA_INPUT) then
!        !{
!        if(rev_extrap_par%snap_data_step ) then
!          if(gen_par%debug1) write(0,*)'Read receiver Data'
!          call SEP_timer_start(14)
!          call input_data(data_par%tag,gen_par,data_par,wave_par, &
!                  rev_extrap_par,force_block,wave_block)
!          call SEP_timer_stop(14)
!        end if
!        !}
!      end if


      if (image_par%file == IMAGE_OUTPUT) then
      !{
        if(rev_extrap_par%snap_image_step ) then
        !{
          if(gen_par%debug1) write(0,*)'Output Image'
          call add_output_image_lin(gen_par,image_par,rev_extrap_par,wave_par, &
                    dslow_block,image_mem,image_mem_hz)
        !}
        end if
      !}
      end if

    !}
    end do

  !}
  end do do_s1_par 
  !}
  end do do_s2_par 

  stat=0
  if(any(stat_par /= 0)) stat =1
  return
end function rev_extrap_gen_lin_imag

end module rev_extrap_gen_lin_imag_mod
