!*$*OPTIMIZE(5)
!*$*ROUNDOFF(3)
!*$*SCALAROPTIMIZE(3)

module rev_extrap_gen_mod
  use genrev3d_types
  use data_types
  use image_types
  use slow_types
  use wave_types
  use wavelet_types
  use rev_extrap_types

  use wave_mod
  use wavelet_mod
  use image_mod
  use data_mod
  use slow_mod

#ifdef __CONVEY
  use Rev2D_fd_types
  use Rev2D_fd_mod
#endif

  use timers_mod
#ifdef __CONVEY
  use cny_timers_mod
#endif

!  use down_gen_ser_mod
!  use image_mod
!  use filter_data_mod

  implicit none

!#ifdef __CONVEY
!integer :: cny_i_timer
!!$cny coproc_mem(cny_i_timer)
!#endif

contains
  function rev_extrap_gen( &
    gen_par,data_par,image_par,rev_extrap_par,slow_par,wave_par,wavelet_par, &
    rev_extrap_oper,rev_der2time_oper,rev_scatter_oper, &
    slow_block,&
    force_block,&
    wave_block,&
    wave_back_block,&
    wavelet, &
    image_mem, &
    image_mem_hz) &
  result(stat)
!!$cny dual_target(1)

  interface 
    function rev_extrap_oper( &
                             rev_extrap_par, &
                             wave_par, &
                             force_block, &
                             wave_block, slow_block) &
                             result(stat)
      use rev_extrap_types, only: rev_extrap_param_type
      use data_types, only: data_param_type
      use wave_types, only: wave_param_type
      integer                                   :: stat
      type (rev_extrap_param_type)              :: rev_extrap_par
      type (wave_param_type)                    :: wave_par
      real,  dimension (:,:,:,:)                :: force_block
      real,  dimension (:,:,:,:,:,:)            :: wave_block
      real, optional,  dimension (:,:,:,:,:,:)  :: slow_block
    end function rev_extrap_oper

    function rev_der2time_oper (&
          rev_extrap_par, &
          wave_par, &
          wave_block, &
          wave_back_block) &
        result(stat)
      use rev_extrap_types, only: rev_extrap_param_type
      use wave_types, only: wave_param_type
      integer                                   :: stat
      type (rev_extrap_param_type)              :: rev_extrap_par
      type (wave_param_type)                    :: wave_par
      real,  dimension (:,:,:,:,:,:)            :: wave_block
      real,  dimension (:,:,:,:,:,:)            :: wave_back_block
    end function rev_der2time_oper
  
    function rev_scatter_oper( &
        rev_extrap_par, &
        wave_par, &
        wave_block,wave_back_block, &
        image_mem,image_mem_hz, & 
        slow_block) &
      result(stat)
      use rev_extrap_types, only: rev_extrap_param_type
      use wave_types, only: wave_param_type
      integer                                   :: stat
      type (rev_extrap_param_type)              :: rev_extrap_par
      type (wave_param_type)                    :: wave_par
      real,  dimension (:,:,:,:,:,:)            :: wave_block
      real,  dimension (:,:,:,:,:,:)            :: wave_back_block
      real,  dimension (:,:,:,:)                :: image_mem
      real,  dimension (:,:,:,:)                :: image_mem_hz
      real, optional,  dimension (:,:,:,:,:,:)  :: slow_block
    end function rev_scatter_oper
  
  end interface

    integer stat
    type(genrev3d_param_type) :: gen_par
    type (data_param_type) :: data_par
    type (image_param_type) :: image_par
    type (rev_extrap_param_type) :: rev_extrap_par
    type (slow_param_type) :: slow_par
    type (wave_param_type) :: wave_par
    type (wavelet_param_type) :: wavelet_par

    optional rev_scatter_oper


    integer stat_par(gen_par%n_threads)
    real, dimension    (:,:,:,:,:,:) :: slow_block
    real, dimension (:,:,:,:) :: force_block
    real, dimension (:,:,:,:,:,:) :: wave_block
    real, dimension (:,:,:,:,:,:) :: wave_back_block
    real, dimension (:,:,:,:) :: wavelet
    real, dimension (:,:,:,:) :: image_mem,image_mem_hz 

    integer, external :: omp_get_thread_num

    integer ierr,i_t_snap
    integer rat_t_snap_wave_second,rat_t_snap_image,rat_t_snap_data,rat_t_snap_data_sou
    integer rat_t_snap_wave_first,rat_t_snap_wave_back
    integer min_it_snap_wave_second,min_it_snap_image,min_it_snap_data,min_it_snap_data_sou
    integer min_it_snap_wave_first,min_it_snap_wave_back
    integer max_it_snap_wave_second,max_it_snap_image
    integer max_it_snap_data,max_it_snap_data_sou
    integer max_it_snap_wave_first,max_it_snap_wave_back
    integer n_s1_par,n_s2_par,i_s1_par,i_s2_par,i_t_extrap

    real t_extrap,epsi_image_d_t,epsi_wave_d_t,epsi_data_d_t
    real s1_par_0,s2_par_0,d_s1_par,d_s2_par
    real norm_wave,courant_numb

    stat_par=0 ! in case that we never call the inner routines

    epsi_image_d_t=.01*image_par%d_t
    epsi_wave_d_t=.01*wave_par%d_t
    epsi_data_d_t=.01*data_par%d_t

    ! read appropriate wavelet files
    if(wavelet_par%file == WAVELET_INPUT) then
      if(gen_par%debug1) write(0,*)'Read Wavelet'
      call read_wavelet(wavelet_par%tag,wavelet_par,wavelet)
    end if
    call snap('Wave-nlin.H',wavelet_par%n_t,1,wavelet)

    if (gen_par%source_type(1:6) == "point_") then
      n_s1_par=data_par%n_sx
      n_s2_par=data_par%n_sy
      d_s1_par=data_par%d_sx
      d_s2_par=data_par%d_sy
      s1_par_0=data_par%sx_0
      s2_par_0=data_par%sy_0
    else
      n_s1_par=data_par%n_s1_par
      n_s2_par=data_par%n_s2_par
      d_s1_par=data_par%d_s1_par
      d_s2_par=data_par%d_s2_par
      s1_par_0=data_par%s1_par_0
      s2_par_0=data_par%s2_par_0
    end if

    select case (gen_par%domain_range(1:8))
    case ('image_da')
    !{

      if(gen_par%debug1) write(0,*)'image_par-t_0=',image_par%t_0
      if(gen_par%debug1) write(0,*)'rev_extrap_par-t_0_first=',rev_extrap_par%t_0_first
      if(gen_par%debug1) write(0,*)'t_sou_0=',data_par%t_sou_0

      if_adj_image_data : if (rev_extrap_par%adj) then
      !{

        if(gen_par%image_zero_time) then
!          min_it_snap_wave_back=nint((wave_par%t_back_0-rev_extrap_par%t_0_first)/ &
!                                    rev_extrap_par%d_t_first) Biondo 2/1/2006
!          max_it_snap_wave_back=rev_extrap_par%n_t_first-1 Biondo 2/1/2006
          min_it_snap_wave_back=&
                 nint((wave_par%t_back_0-rev_extrap_par%t_0_first)/ & 
                       rev_extrap_par%d_t_first)-1
          max_it_snap_wave_back=rev_extrap_par%n_t_first-1-1
          rat_t_snap_wave_back=nint(wave_par%d_t_back/rev_extrap_par%d_t_first)
          if(rat_t_snap_wave_back == 0) rat_t_snap_wave_back=-1
        else
          min_it_snap_wave_back=nint((image_par%t_0-rev_extrap_par%t_0_first)/ &
                                    rev_extrap_par%d_t_first)
          max_it_snap_wave_back=rev_extrap_par%n_t_first
          rat_t_snap_wave_back=nint(image_par%d_t/rev_extrap_par%d_t_first)
          if(rat_t_snap_wave_back == 0) rat_t_snap_wave_back=-1
        end if
        if(gen_par%debug1) write(0,*)'min_it_snap_wave_back=',min_it_snap_wave_back
        if(gen_par%debug1) write(0,*)'max_it_snap_wave_back=',max_it_snap_wave_back
        if(gen_par%debug1) write(0,*)'rat_t_snap_wave_back=',rat_t_snap_wave_back

        min_it_snap_wave_second=0
        max_it_snap_wave_second=rev_extrap_par%n_t_second-1
        rat_t_snap_wave_second=nint(wave_par%d_t/rev_extrap_par%d_t_second)
        if(rat_t_snap_wave_second == 0) rat_t_snap_wave_second=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_wave_second=',min_it_snap_wave_second
        if(gen_par%debug1) write(0,*)'max_it_snap_wave_second=',max_it_snap_wave_second
        if(gen_par%debug1) write(0,*)'rat_t_snap_wave_second=',rat_t_snap_wave_second
  
        if (gen_par%source_type(1:6) == "point_") then
          min_it_snap_wave_first=nint((data_par%t_0-rev_extrap_par%t_0_first)/ &
                                  rev_extrap_par%d_t_first)
        else
          min_it_snap_wave_first=nint((data_par%t_sou_0-rev_extrap_par%t_0_first)/ &
                                  rev_extrap_par%d_t_first)
        end if

        max_it_snap_wave_first=rev_extrap_par%n_t_first-1
        rat_t_snap_wave_first=-nint(wave_par%d_t/rev_extrap_par%d_t_first)
        if(rat_t_snap_wave_first == 0) rat_t_snap_wave_first=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_wave_first=',min_it_snap_wave_first
        if(gen_par%debug1) write(0,*)'max_it_snap_wave_first=',max_it_snap_wave_first
        if(gen_par%debug1) write(0,*)'rat_t_snap_wave_first=',rat_t_snap_wave_first
  
  !        min_it_snap_image=nint(image_par%t_0_lag/image_par%d_t_lag) + &
  !                          image_par%n_t_lag-1 &
  !                          +2
  !        max_it_snap_image=rev_extrap_par%n_t_second+1+ &
  !                            nint(image_par%t_0_lag/image_par%d_t_lag)
  
        if(gen_par%image_zero_time) then
!          min_it_snap_image=rev_extrap_par%n_t_second-image_par%n_t + &
!                              nint(image_par%t_0_lag/image_par%d_t_lag) Biondo 2/1/2006
!          max_it_snap_image=rev_extrap_par%n_t_second -1 + &
!                              nint(image_par%t_0_lag/image_par%d_t_lag) Biondo 2/1/2006
          min_it_snap_image=rev_extrap_par%n_t_second-image_par%n_t + &
                              nint(image_par%t_0_lag/image_par%d_t_lag) -1
          max_it_snap_image=rev_extrap_par%n_t_second -1 + &
                              nint(image_par%t_0_lag/image_par%d_t_lag) -1
          rat_t_snap_image=-nint(image_par%d_t/rev_extrap_par%d_t_second)
          if(rat_t_snap_image == 0) rat_t_snap_image=-1
        else
          min_it_snap_image=nint(image_par%t_0_lag/image_par%d_t_lag) + &
                            image_par%n_t_lag-1 
          max_it_snap_image=rev_extrap_par%n_t_second-1+ &
                              nint(image_par%t_0_lag/image_par%d_t_lag)
          rat_t_snap_image=-nint(image_par%d_t/rev_extrap_par%d_t_second)
          if(rat_t_snap_image == 0) rat_t_snap_image=-1
        end if
        if(gen_par%debug1) write(0,*)'min_it_snap_image=',min_it_snap_image
        if(gen_par%debug1) write(0,*)'max_it_snap_image=',max_it_snap_image
        if(gen_par%debug1) write(0,*)'rat_t_snap_image=',rat_t_snap_image
  
        if(gen_par%image_zero_time) then
          min_it_snap_data=0
          max_it_snap_data=data_par%n_t-1
          rat_t_snap_data=abs(nint(data_par%d_t/rev_extrap_par%d_t_second))
          min_it_snap_data_sou=0
          max_it_snap_data_sou=data_par%n_t-1
          rat_t_snap_data_sou=abs(nint(data_par%d_t/rev_extrap_par%d_t_second))
        else
          min_it_snap_data=0
          max_it_snap_data=data_par%n_t-1
          rat_t_snap_data=abs(nint(data_par%d_t/rev_extrap_par%d_t_second))
          min_it_snap_data_sou=min_it_snap_data
          max_it_snap_data_sou=max_it_snap_data
          rat_t_snap_data_sou=rat_t_snap_data
        end if
        if(gen_par%debug1) write(0,*)'min_it_snap_data_sou=',min_it_snap_data_sou
        if(gen_par%debug1) write(0,*)'max_it_snap_data_sou=',max_it_snap_data_sou
        if(gen_par%debug1) write(0,*)'rat_t_snap_data_sou=',rat_t_snap_data_sou
        if(gen_par%debug1) write(0,*)'min_it_snap_data=',min_it_snap_data
        if(gen_par%debug1) write(0,*)'max_it_snap_data=',max_it_snap_data
        if(gen_par%debug1) write(0,*)'rat_t_snap_data=',rat_t_snap_data
  
      !}
      else if_adj_image_data
      !{

        min_it_snap_wave_back=nint((image_par%t_0-rev_extrap_par%t_0_first)/ &
                                  rev_extrap_par%d_t_first)
        max_it_snap_wave_back=rev_extrap_par%n_t_first
        rat_t_snap_wave_back=nint(image_par%d_t/rev_extrap_par%d_t_first)
        if(rat_t_snap_wave_back == 0) rat_t_snap_wave_back=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_wave_back=',min_it_snap_wave_back
        if(gen_par%debug1) write(0,*)'max_it_snap_wave_back=',max_it_snap_wave_back
        if(gen_par%debug1) write(0,*)'rat_t_snap_wave_back=',rat_t_snap_wave_back
  
        min_it_snap_wave_second=nint((data_par%t_0-rev_extrap_par%t_0_second)/ &
                                  rev_extrap_par%d_t_second)
        max_it_snap_wave_second=rev_extrap_par%n_t_second
        rat_t_snap_wave_second=nint(wave_par%d_t/rev_extrap_par%d_t_second)
        if(rat_t_snap_wave_second == 0) rat_t_snap_wave_second=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_wave_second=',min_it_snap_wave_second
        if(gen_par%debug1) write(0,*)'max_it_snap_wave_second=',max_it_snap_wave_second
        if(gen_par%debug1) write(0,*)'rat_t_snap_wave_second=',rat_t_snap_wave_second
  
        min_it_snap_wave_first=nint((data_par%t_0-rev_extrap_par%t_0_first)/ &
                                  rev_extrap_par%d_t_first)
        max_it_snap_wave_first=rev_extrap_par%n_t_first
        rat_t_snap_wave_first=nint(wave_par%d_t/rev_extrap_par%d_t_first)
        if(rat_t_snap_wave_first == 0) rat_t_snap_wave_first=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_wave_first=',min_it_snap_wave_first
        if(gen_par%debug1) write(0,*)'max_it_snap_wave_first=',max_it_snap_wave_first
        if(gen_par%debug1) write(0,*)'rat_t_snap_wave_first=',rat_t_snap_wave_first
  
  !      min_it_snap_wave=nint((wave_par%t_0-rev_extrap_par%t_0_second)/ &
  !                            rev_extrap_par%d_t_second)
  !      max_it_snap_wave=rev_extrap_par%n_t_second
  !      rat_t_snap_wave=nint(wave_par%d_t/rev_extrap_par%d_t_second)
  !      if(rat_t_snap_wave == 0) rat_t_snap_wave=-1
  !      if(gen_par%debug1) write(0,*)'min_it_snap_wave=',min_it_snap_wave
  !      if(gen_par%debug1) write(0,*)'rat_t_snap_wave=',rat_t_snap_wave
  
  !      min_it_snap_image=nint(image_par%t_0_lag/image_par%d_t_lag) + &
  !                        image_par%n_t_lag-1 
  !      max_it_snap_image=rev_extrap_par%n_t_second-1+ &
  !                          nint(image_par%t_0_lag/image_par%d_t_lag)
  !      rat_t_snap_image=nint(image_par%d_t/rev_extrap_par%d_t_second)
  
        min_it_snap_image=nint((image_par%t_0-rev_extrap_par%t_0_first)/ &
                                rev_extrap_par%d_t_first)
        max_it_snap_image=rev_extrap_par%n_t_first-1
        rat_t_snap_image=nint(image_par%d_t/rev_extrap_par%d_t_first)
  
        if(rat_t_snap_image == 0) rat_t_snap_image=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_image=',min_it_snap_image
        if(gen_par%debug1) write(0,*)'max_it_snap_image=',max_it_snap_image
        if(gen_par%debug1) write(0,*)'rat_t_snap_image=',rat_t_snap_image
  
  !        min_it_snap_image=nint((image_par%t_0-rev_extrap_par%t_0_second)/ &
  !                              rev_extrap_par%d_t_second)
  !        max_it_snap_image=rev_extrap_par%n_t_second
  !        rat_t_snap_image=nint(image_par%d_t/rev_extrap_par%d_t_second)
  !        if(rat_t_snap_image == 0) rat_t_snap_image=-1
  !        if(gen_par%debug1) write(0,*)'min_it_snap_image=',min_it_snap_image
  !        if(gen_par%debug1) write(0,*)'rat_t_snap_image=',rat_t_snap_image
    
        min_it_snap_data_sou=rev_extrap_par%n_t_first
        max_it_snap_data_sou=0
        rat_t_snap_data_sou=1

        min_it_snap_data=nint((data_par%t_0-rev_extrap_par%t_0_second)/ &
                              rev_extrap_par%d_t_second)
        max_it_snap_data=rev_extrap_par%n_t_second-1
        rat_t_snap_data=abs(nint(data_par%d_t/rev_extrap_par%d_t_second))
  

        if(gen_par%debug1) write(0,*)'min_it_snap_data_sou=',min_it_snap_data_sou
        if(gen_par%debug1) write(0,*)'max_it_snap_data_sou=',max_it_snap_data_sou
        if(gen_par%debug1) write(0,*)'rat_t_snap_data_sou=',rat_t_snap_data_sou
        if(gen_par%debug1) write(0,*)'min_it_snap_data=',min_it_snap_data
        if(gen_par%debug1) write(0,*)'max_it_snap_data=',max_it_snap_data
        if(gen_par%debug1) write(0,*)'rat_t_snap_data=',rat_t_snap_data

      !}
      end if if_adj_image_data
    !}
    case ('lin_imag')
    !{

      min_it_snap_wave_first=nint((data_par%t_0-rev_extrap_par%t_0_first)/ &
                                rev_extrap_par%d_t_first)
      max_it_snap_wave_first=rev_extrap_par%n_t_first-1
      rat_t_snap_wave_first=nint(wave_par%d_t/rev_extrap_par%d_t_first)
      if(rat_t_snap_wave_first == 0) rat_t_snap_wave_first=-1
      if(gen_par%debug1) &
        write(0,*)'min_it_snap_wave_first=',min_it_snap_wave_first
      if(gen_par%debug1) &
        write(0,*)'max_it_snap_wave_first=',max_it_snap_wave_first
      if(gen_par%debug1) &
        write(0,*)'rat_t_snap_wave_first=',rat_t_snap_wave_first

      min_it_snap_wave_second=nint((wave_par%t_0-rev_extrap_par%t_0_second)/ &
                            rev_extrap_par%d_t_second)
      max_it_snap_wave_second=rev_extrap_par%n_t_second-1
      rat_t_snap_wave_second=nint(wave_par%d_t/rev_extrap_par%d_t_second)
      if(rat_t_snap_wave_second == 0) rat_t_snap_wave_second=-1
      if(gen_par%debug1) &
        write(0,*)'min_it_snap_wave_second=',min_it_snap_wave_second
      if(gen_par%debug1) &
        write(0,*)'max_it_snap_wave_second=',max_it_snap_wave_second
      if(gen_par%debug1) &
        write(0,*)'rat_t_snap_wave_second=',rat_t_snap_wave_second

      min_it_snap_wave_back=&
                 nint((wave_par%t_back_0-rev_extrap_par%t_0_first)/ & 
                       rev_extrap_par%d_t_first)
      max_it_snap_wave_back=rev_extrap_par%n_t_first-1
      rat_t_snap_wave_back=nint(wave_par%d_t_back/rev_extrap_par%d_t_first)
      if(rat_t_snap_wave_back == 0) rat_t_snap_wave_back=-1
      if(gen_par%debug1)write(0,*)'min_it_snap_wave_back=',min_it_snap_wave_back
      if(gen_par%debug1)write(0,*)'max_it_snap_wave_back=',max_it_snap_wave_back
      if(gen_par%debug1)write(0,*)'rat_t_snap_wave_back=',rat_t_snap_wave_back

      min_it_snap_image=-1
      max_it_snap_image=-1
      rat_t_snap_image=-1

      min_it_snap_data=nint((data_par%t_0-rev_extrap_par%t_0_second)/ &
                            rev_extrap_par%d_t_second)
      max_it_snap_data=rev_extrap_par%n_t_second-1
      rat_t_snap_data=nint(data_par%d_t/rev_extrap_par%d_t_second)
      if(rat_t_snap_data == 0) rat_t_snap_data=-1
      if(gen_par%debug1) write(0,*)'min_it_snap_data=',min_it_snap_data
      if(gen_par%debug1) write(0,*)'max_it_snap_data=',max_it_snap_data
      if(gen_par%debug1) write(0,*)'rat_t_snap_data=',rat_t_snap_data

      rat_t_snap_data_sou=-1

    !}
    case ('image_gl')
    !{
      if(gen_par%debug1) write(0,*)'image_par-t_0=',image_par%t_0
      if(gen_par%debug1) write(0,*)'rev_extrap_par-t_0_first=',rev_extrap_par%t_0_first
      min_it_snap_wave_back=nint((image_par%t_0-rev_extrap_par%t_0_first)/ &
                                rev_extrap_par%d_t_first)
      max_it_snap_wave_back=rev_extrap_par%n_t_first
      rat_t_snap_wave_back=nint(image_par%d_t/rev_extrap_par%d_t_first)
      if(rat_t_snap_wave_back == 0) rat_t_snap_wave_back=-1
      if(gen_par%debug1) write(0,*)'min_it_snap_wave_back=',min_it_snap_wave_back
      if(gen_par%debug1) write(0,*)'max_it_snap_wave_back=',max_it_snap_wave_back
      if(gen_par%debug1) write(0,*)'rat_t_snap_wave_back=',rat_t_snap_wave_back
  
      min_it_snap_wave_second=nint((data_par%t_0-rev_extrap_par%t_0_second)/ &
                                rev_extrap_par%d_t_second)
      max_it_snap_wave_second=rev_extrap_par%n_t_second
      rat_t_snap_wave_second=nint(wave_par%d_t/rev_extrap_par%d_t_second)
      if(rat_t_snap_wave_second == 0) rat_t_snap_wave_second=-1
      if(gen_par%debug1) write(0,*)'min_it_snap_wave_second=',min_it_snap_wave_second
      if(gen_par%debug1) write(0,*)'max_it_snap_wave_second=',max_it_snap_wave_second
      if(gen_par%debug1) write(0,*)'rat_t_snap_wave_second=',rat_t_snap_wave_second
  
!      min_it_snap_wave_first=nint((data_par%t_sou_0-rev_extrap_par%t_0_first)/ &
!                                rev_extrap_par%d_t_first)
      min_it_snap_wave_first=nint((data_par%t_0-rev_extrap_par%t_0_first)/ &
                                rev_extrap_par%d_t_first)
!      min_it_snap_wave_first=0
      max_it_snap_wave_first=rev_extrap_par%n_t_first
      rat_t_snap_wave_first=nint(-wave_par%d_t/rev_extrap_par%d_t_first)
      if(rat_t_snap_wave_first == 0) rat_t_snap_wave_first=1
      if(gen_par%debug1) write(0,*)'min_it_snap_wave_first=',min_it_snap_wave_first
      if(gen_par%debug1) write(0,*)'max_it_snap_wave_first=',max_it_snap_wave_first
      if(gen_par%debug1) write(0,*)'rat_t_snap_wave_first=',rat_t_snap_wave_first
  
      min_it_snap_image=0
      max_it_snap_image=wavelet_par%n_t-1
      rat_t_snap_image=1

      if(gen_par%debug1) write(0,*)'min_it_snap_image=',min_it_snap_image
      if(gen_par%debug1) write(0,*)'max_it_snap_image=',max_it_snap_image
      if(gen_par%debug1) write(0,*)'rat_t_snap_image=',rat_t_snap_image
  
      min_it_snap_data=nint((data_par%t_0-rev_extrap_par%t_0_second)/ &
                            rev_extrap_par%d_t_second)
      max_it_snap_data=rev_extrap_par%n_t_second-1
      rat_t_snap_data=abs(nint(data_par%d_t/rev_extrap_par%d_t_second))

      min_it_snap_data_sou=min_it_snap_data
      max_it_snap_data_sou=max_it_snap_data
      rat_t_snap_data_sou=rat_t_snap_data
  
      if(gen_par%debug1) write(0,*)'min_it_snap_data_sou=',min_it_snap_data_sou
      if(gen_par%debug1) write(0,*)'max_it_snap_data_sou=',max_it_snap_data_sou
      if(gen_par%debug1) write(0,*)'rat_t_snap_data_sou=',rat_t_snap_data_sou
      if(gen_par%debug1) write(0,*)'min_it_snap_data=',min_it_snap_data
      if(gen_par%debug1) write(0,*)'max_it_snap_data=',max_it_snap_data
      if(gen_par%debug1) write(0,*)'rat_t_snap_data=',rat_t_snap_data
  
    !}
    case ('vel_data')
    !{

      if_vel_data : if (rev_extrap_par%adj) then
        call seperr('Not supported domain_range=vel_data and adj=yes')
      else if_vel_data
  
        min_it_snap_wave_second=nint((wave_par%t_0-rev_extrap_par%t_0_second)/ &
                              rev_extrap_par%d_t_second)
        max_it_snap_wave_second=rev_extrap_par%n_t_second-1
        rat_t_snap_wave_second=nint(wave_par%d_t/rev_extrap_par%d_t_second)
        if(rat_t_snap_wave_second == 0) rat_t_snap_wave_second=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_wave_second=',min_it_snap_wave_second
        if(gen_par%debug1) write(0,*)'max_it_snap_wave_second=',max_it_snap_wave_second
        if(gen_par%debug1) write(0,*)'rat_t_snap_wave_second=',rat_t_snap_wave_second
  
        min_it_snap_image=-1
        max_it_snap_image=-1
        rat_t_snap_image=-1
  
        min_it_snap_data=nint((data_par%t_0-rev_extrap_par%t_0_second)/ &
                              rev_extrap_par%d_t_second)
        max_it_snap_data=rev_extrap_par%n_t_second-1
        rat_t_snap_data=nint(data_par%d_t/rev_extrap_par%d_t_second)
  
        if(rat_t_snap_data == 0) rat_t_snap_data=-1
        if(gen_par%debug1) write(0,*)'min_it_snap_data=',min_it_snap_data
        if(gen_par%debug1) write(0,*)'max_it_snap_data=',max_it_snap_data
        if(gen_par%debug1) write(0,*)'rat_t_snap_data=',rat_t_snap_data
      end if if_vel_data
  
    !}
    case default
       write(0,*)'gen_par-domain_range=',trim(gen_par%domain_range)
     call seperr('Not supported domain_range')
  end select

  do_s2_par : do i_s2_par=1,n_s2_par
  do_s1_par : do i_s1_par=1,n_s1_par

    rev_extrap_par%s1_par=s1_par_0 + (i_s1_par-1)*d_s1_par
    rev_extrap_par%s2_par=s2_par_0 + (i_s2_par-1)*d_s2_par 

    rev_extrap_par%i_s1_par=i_s1_par
    rev_extrap_par%i_s2_par=i_s2_par

    data_par%i_s1_par=i_s1_par
    data_par%i_s2_par=i_s2_par

    if (gen_par%source_type(1:6) == "point_") then
      if(gen_par%verb) then
        write(0,*)'i_s1_par=',i_s1_par,'at location x=',rev_extrap_par%s1_par
        write(0,*)'i_s2_par=',i_s2_par,'at location y=',rev_extrap_par%s2_par
      end if

      wave_par%mx_0=rev_extrap_par%s1_par + wave_par%off_x_0
      wave_par%my_0=rev_extrap_par%s2_par + wave_par%off_y_0

      if(gen_par%debug1) write(0,*)'wave_par-mx_0=',wave_par%mx_0
      if(gen_par%debug1) write(0,*)'wave_par-my_0=',wave_par%my_0

    else
      if(gen_par%verb) then
        write(0,*)'i_s1_par=',i_s1_par,'with value=',rev_extrap_par%s1_par
        write(0,*)'i_s2_par=',i_s2_par,'with value=',rev_extrap_par%s2_par
      end if
    end if

!$cny begin_coproc
    wave_block=0.
    force_block=0
!$cny end_coproc

    ! read appropriate veocity files
    if(slow_par%file == SLOW_INPUT) then
       if(gen_par%debug1) write(0,*)'Read Slow'
         call SEP_timer_start(11)
         call read_slow(gen_par,slow_par%tag, wave_par,slow_par, slow_block)
         call SEP_timer_stop(11)

    else
      if(gen_par%debug1) write(0,*)'DO NOT read Slow'
      slow_par%slow_min=gen_par%slow_const
    end if

    courant_numb=rev_extrap_par%d_t/ &
              (slow_par%slow_min*min(wave_par%d_mx,wave_par%d_z))
    if(gen_par%verb) write(0,*)'Courant number=',courant_numb, &
        'Maximum allowed is',gen_par%max_courant_numb

    if(gen_par%check_stability) then
      if(courant_numb > gen_par%max_courant_numb) then
        write(0,*)'Courant number=',courant_numb, &
        'larger than maximum allowed of',gen_par%max_courant_numb
        call seperr('Exceeded CFL stability condition')
      end if
    end if


    if((.not. gen_par%skip_source_prop) .and. &
       (((i_s1_par==1).and.(i_s2_par==1)).or. &
         (.not. gen_par%layered_mig_vel)) .and. &
       (gen_par%domain_range(1:8) /= 'vel_data') &
      ) then
    !{

      if(gen_par%verb) write(0,*)'Start first time loop'

      rev_extrap_par%sou_rec_wave=SOU_WAVE

      rev_extrap_par%i_t_extrap=0
      rev_extrap_par%t= rev_extrap_par%t_0_first+ &
                          (rev_extrap_par%i_t_extrap)*rev_extrap_par%d_t_first

      if(gen_par%verb) then 
        write(0,*)'Source wavefield: Doing time step #', &
                      rev_extrap_par%i_t_extrap
        write(0,*)'Source wavefield: time=',rev_extrap_par%t
      end if

! determine wether to write out wavefield
      rev_extrap_par%snap_wave_first_step= &
 (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_first,rat_t_snap_wave_first)==0)&
   .and.&
       (rev_extrap_par%i_t_extrap >= min_it_snap_wave_first).and.&
       (rev_extrap_par%i_t_extrap <= max_it_snap_wave_first)
      if(gen_par%verb) write(0,*)'snap_wave_first_step=',rev_extrap_par%snap_wave_first_step
      if(gen_par%verb) write(0,*)'i_t_extrap=',rev_extrap_par%i_t_extrap


      rev_extrap_par%wave_back_step= &
 (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_back,rat_t_snap_wave_back)==0)&
   .and.&
       (rev_extrap_par%i_t_extrap >= min_it_snap_wave_back).and.&
       (rev_extrap_par%i_t_extrap <= max_it_snap_wave_back)
      if(gen_par%verb) write(0,*)'wave_back_step=',rev_extrap_par%wave_back_step


      rev_extrap_par%data_sou_step= &
 (mod(rev_extrap_par%i_t_extrap-min_it_snap_data_sou,rat_t_snap_data_sou)==0)&
    .and.&
        (rev_extrap_par%i_t_extrap >= min_it_snap_data_sou).and.&
        (rev_extrap_par%i_t_extrap <= max_it_snap_data_sou)
      if(gen_par%verb) write(0,*)'data_sou_step=',rev_extrap_par%data_sou_step


      rev_extrap_par%snap_image_step= &
 (mod(rev_extrap_par%i_t_extrap-min_it_snap_image,rat_t_snap_image)==0).and.&
        (rev_extrap_par%i_t_extrap >= min_it_snap_image).and.&
        (rev_extrap_par%i_t_extrap <= max_it_snap_image)
      if(gen_par%verb) write(0,*)'snap_image_step=',rev_extrap_par%snap_image_step

      if (gen_par%source_type(1:6) == "point_") then
      !{
        if(wavelet_par%file == WAVELET_INPUT) then
          call add_wavelet(gen_par,wave_par,wavelet_par,rev_extrap_par, &
                             force_block,wavelet)
        end if
      !}
      else
      !{
        if (data_par%file == DATA_INPUT) then
          !{
          if(rev_extrap_par%data_sou_step ) then
            if(gen_par%debug1) write(0,*)'Read Data_sou'
            call SEP_timer_start(14)
            call input_data(data_par%tag_sou,gen_par,data_par,wave_par, &
                 rev_extrap_par,force_block,wave_block)
            call SEP_timer_stop(14)
          end if
          !}
         end if
      !}
      end if


      if ((gen_par%domain_range(1:8) == 'image_gl') .or. &
          (gen_par%domain_range(1:8) == 'image_da')) then
      !{
        if (image_par%file == IMAGE_INPUT) then
        !{

          if(rev_extrap_par%snap_image_step ) then
            call SEP_timer_start(13)
            if(gen_par%debug1) write(0,*)'Add Image to Source wavefield'
            call add_image(gen_par,image_par,wave_par,rev_extrap_par, &
                 wavelet_par,wave_block,wavelet,image_mem,image_mem_hz)
            call SEP_timer_stop(13)
          end if
        !}
        end if
      !}
      end if

      if(wave_par%file == WAVE_OUTPUT) then
        if(rev_extrap_par%snap_wave_first_step) then
          if(gen_par%debug1) write(0,*)'Output Wavefield_back'
          call SEP_timer_start(12)
          call output_wave(wave_par%tag_first,gen_par,wave_par, &
                     rev_extrap_par,wave_block)
          call SEP_timer_stop(12)
        end if
      end if

      if(wave_par%file_back == WAVE_OUTPUT) then
        if(rev_extrap_par%wave_back_step) then
          if (gen_par%domain_range(1:8) == 'lin_imag') then
            stat= rev_der2time_oper (&
              rev_extrap_par=rev_extrap_par, &
              wave_par=wave_par, &
              wave_block=wave_block, &
              wave_back_block=wave_back_block)
            if(gen_par%debug1) write(0,*)'Output Der2Time Wavefield_sou'
            call SEP_timer_start(12)
            call output_wave(wave_par%tag_back,gen_par,wave_par, &
                       rev_extrap_par,wave_back_block)
            call SEP_timer_stop(12)
          else
            if(gen_par%debug1) write(0,*)'Output Wavefield_sou'
            call SEP_timer_start(12)
            call output_wave(wave_par%tag_back,gen_par,wave_par, &
                       rev_extrap_par,wave_block)
            call SEP_timer_stop(12)
          end if
        end if
      end if

      if (gen_par%domain_range(1:8) == 'image_gl') then
      !{
        if(data_par%file == DATA_OUTPUT) then
        !{
          if(rev_extrap_par%data_sou_step) then
          !{
            if(gen_par%debug1) write(0,*)'Output Data_sou'
            call SEP_timer_start(14)
            call output_data(data_par%tag_sou,gen_par,data_par, &
                     wave_par,rev_extrap_par,wave_block)
            call SEP_timer_stop(14)
          !}
          end if
        !}
        end if
      !}
      end if

      do i_t_extrap=1,rev_extrap_par%n_t_first
      !{


        rev_extrap_par%i_t_extrap=i_t_extrap

        rev_extrap_par%t= rev_extrap_par%t_0_first+ &
                          (rev_extrap_par%i_t_extrap)*rev_extrap_par%d_t_first

        if(gen_par%verb) then 
          write(0,*)'Source wavefield: Doing time step #', &
                        rev_extrap_par%i_t_extrap
          write(0,*)'Source wavefield: time=',rev_extrap_par%t
        end if

        if(gen_par%check_stability) then
         !{
           norm_wave=sqrt(sum(wave_block**2))
           write(0,*)'norm wavefield at time=',rev_extrap_par%t,norm_wave
         !}
        end if

! determine wether to write out wavefield
        rev_extrap_par%snap_wave_first_step= &
          (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_first, &
           rat_t_snap_wave_first)==0).and.&
          (rev_extrap_par%i_t_extrap >= min_it_snap_wave_first).and.&
          (rev_extrap_par%i_t_extrap <= max_it_snap_wave_first)
        if(gen_par%verb) write(0,*)'snap_wave_first_step=',rev_extrap_par%snap_wave_first_step
!        if(gen_par%verb) write(0,*)'i_t_extrap=',rev_extrap_par%i_t_extrap
!        if(gen_par%verb) write(0,*)'rat_t_snap_wave_first=',rat_t_snap_wave_first
!        if(gen_par%verb) write(0,*)'min_it_snap_wave_first=',min_it_snap_wave_first
!        if(gen_par%verb) write(0,*)'max_it_snap_wave_first=',max_it_snap_wave_first
  
        rev_extrap_par%wave_back_step= &
          (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_back, &
          rat_t_snap_wave_back)==0).and.&
          (rev_extrap_par%i_t_extrap >= min_it_snap_wave_back).and.&
          (rev_extrap_par%i_t_extrap <= max_it_snap_wave_back)
        if(gen_par%verb) write(0,*)'wave_back_step=',rev_extrap_par%wave_back_step
  
        rev_extrap_par%data_sou_step= &
           (mod(rev_extrap_par%i_t_extrap-min_it_snap_data_sou, &
            rat_t_snap_data_sou) == 0).and.&
           (rev_extrap_par%i_t_extrap >= min_it_snap_data_sou).and.&
           (rev_extrap_par%i_t_extrap <= max_it_snap_data_sou)
        if(gen_par%verb) write(0,*)'data_sou_step=',rev_extrap_par%data_sou_step

        rev_extrap_par%snap_image_step= &
          (mod(rev_extrap_par%i_t_extrap-min_it_snap_image,rat_t_snap_image)==0).and.&
          (rev_extrap_par%i_t_extrap >= min_it_snap_image).and.&
          (rev_extrap_par%i_t_extrap <= max_it_snap_image)
        if(gen_par%verb) write(0,*)'snap_image_step=',rev_extrap_par%snap_image_step

  ! extrapolation step
  
        select case (gen_par%oper_type(1:2))
        case ('co')
          if(gen_par%debug1) write(0,*)'oper_type = Constant'
            stat= rev_extrap_oper( &
              rev_extrap_par=rev_extrap_par, &
              wave_par=wave_par, &
              force_block=force_block, &
              wave_block=wave_block &
              )
        case ('fu')
          if(gen_par%debug1) write(0,*)'oper_type = full'
            stat= rev_extrap_oper( &
              rev_extrap_par=rev_extrap_par, &
              wave_par=wave_par, &
              force_block=force_block, &
              wave_block=wave_block, &
              slow_block=slow_block)
        case default
           write(0,*)'gen_par-oper_type=',trim(gen_par%oper_type)
           call seperr('Not supported oper_type')
        end select

        if(wave_par%file == WAVE_OUTPUT) then
        !{
          if(rev_extrap_par%snap_wave_first_step) then
            if(gen_par%debug1) write(0,*)'Output Wavefield_sou'
            call SEP_timer_start(12)
            call output_wave(wave_par%tag_first,gen_par,wave_par, &
                       rev_extrap_par,wave_block)
            call SEP_timer_stop(12)
          end if
        !}
        end if
  
        if(wave_par%file_back == WAVE_OUTPUT) then
        !{
          if(rev_extrap_par%wave_back_step) then
            if (gen_par%domain_range(1:8) == 'lin_imag') then
              stat= rev_der2time_oper (&
                rev_extrap_par=rev_extrap_par, &
                wave_par=wave_par, &
                wave_block=wave_block, &
                wave_back_block=wave_back_block)

              if(gen_par%debug1) write(0,*)'Output Der2Time Wavefield_back'
              call SEP_timer_start(12)
              call output_wave(wave_par%tag_back,gen_par,wave_par, &
                  rev_extrap_par,wave_back_block)
              call SEP_timer_stop(12)
            else
              if(gen_par%debug1) write(0,*)'Output Wavefield_back'
              call SEP_timer_start(12)
              call output_wave(wave_par%tag_back,gen_par,wave_par, &
                  rev_extrap_par,wave_block)
              call SEP_timer_stop(12)
            end if
          end if
        !}
        end if

        if (gen_par%domain_range(1:8) == 'image_gl') then
        !{
          if(data_par%file == DATA_OUTPUT) then
          !{
            if(rev_extrap_par%data_sou_step ) then
              if(gen_par%debug1) write(0,*)'Output Data_sou'
              call SEP_timer_start(14)
              call output_data(data_par%tag_sou,gen_par,data_par, &
                    wave_par,rev_extrap_par,wave_block)
              call SEP_timer_stop(14)
           end if
          !}
          end if
        !}
        end if

        force_block=0

        if (gen_par%source_type(1:6) == "point_") then
        !{
          if(wavelet_par%file == WAVELET_INPUT) then
            call add_wavelet(gen_par,wave_par,wavelet_par,rev_extrap_par, &
                               force_block,wavelet)
          end if
        !}
        else
        !{
          if (data_par%file == DATA_INPUT) then
            !{
            if(rev_extrap_par%data_sou_step ) then
              if(gen_par%debug1) write(0,*)'Read Data_sou'
              call SEP_timer_start(14)
              call input_data(data_par%tag_sou,gen_par,data_par,wave_par, &
                 rev_extrap_par,force_block,wave_block)
              call SEP_timer_stop(14)
            end if
            !}
           end if
        !}
        end if

        if ((gen_par%domain_range(1:8) == 'image_gl') .or. &
            (gen_par%domain_range(1:8) == 'image_da')) then
        !{
          if (image_par%file == IMAGE_INPUT) then
          !{
            if(rev_extrap_par%snap_image_step ) then
              call SEP_timer_start(13)
              if(gen_par%debug1) write(0,*)'Add Image to Source wavefield'
              call add_image(gen_par,image_par,wave_par,rev_extrap_par, &
                         wavelet_par,wave_block,wavelet,image_mem,image_mem_hz)
              call SEP_timer_stop(13)
            end if
          !}
          end if
        !}
        end if

      !}
      end do
    !}
    end if


! now propagate recorded wavefield and image or forwar scattered wavefiled
    if(gen_par%verb) write(0,*)'Start second time loop'

    wave_block=0.
    wave_back_block=0.
    force_block=0

    if (image_par%file == IMAGE_INPUT) then
    !{
      if(gen_par%debug1) write(0,*)'Input Image'
      call input_image(gen_par,image_par,rev_extrap_par,image_mem,image_mem_hz)
    !}
    end if

    rev_extrap_par%i_t_extrap=0
    rev_extrap_par%t= rev_extrap_par%t_0_second

    if (gen_par%domain_range(1:8) == 'vel_data') then
      rev_extrap_par%sou_rec_wave=SOU_WAVE
      if(gen_par%verb) then 
        write(0,*)'Source wavefield: Doing time step #', &
                      rev_extrap_par%i_t_extrap
        write(0,*)'Source wavefield: time=',rev_extrap_par%t
      end if
    else
      rev_extrap_par%sou_rec_wave=REC_WAVE
      if(gen_par%verb) then 
        write(0,*)'Receiver wavefield: Doing time step #', &
                      rev_extrap_par%i_t_extrap
        write(0,*)'Receiver wavefield: time=',rev_extrap_par%t
      end if
    end if

      min_it_snap_wave_back=&
                 nint((wave_par%t_back_0-rev_extrap_par%t_0_second)/ & 
                       rev_extrap_par%d_t_second)
      max_it_snap_wave_back=rev_extrap_par%n_t_second-1
      rat_t_snap_wave_back=nint(wave_par%d_t_back/rev_extrap_par%d_t_second)
      if(rat_t_snap_wave_back == 0) rat_t_snap_wave_back=-1
      if(gen_par%debug1)write(0,*)'min_it_snap_wave_back=',min_it_snap_wave_back
      if(gen_par%debug1)write(0,*)'max_it_snap_wave_back=',max_it_snap_wave_back
      if(gen_par%debug1)write(0,*)'rat_t_snap_wave_back=',rat_t_snap_wave_back

      min_it_snap_image=-1

      rev_extrap_par%wave_back_step= &
 (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_back,rat_t_snap_wave_back)==0)&
   .and.&
       (rev_extrap_par%i_t_extrap >= min_it_snap_wave_back).and.&
       (rev_extrap_par%i_t_extrap <= max_it_snap_wave_back)
      if(gen_par%verb) &
        write(0,*)'wave_back_step=',rev_extrap_par%wave_back_step

    rev_extrap_par%snap_wave_second_step= &
      (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_second,rat_t_snap_wave_second) == 0).and.&
      (rev_extrap_par%i_t_extrap >= min_it_snap_wave_second).and.&
      (rev_extrap_par%i_t_extrap <= max_it_snap_wave_second)
    if(gen_par%verb) &
        write(0,*)'snap_wave_second_step=',rev_extrap_par%snap_wave_second_step

    rev_extrap_par%snap_image_step= &
      (mod(rev_extrap_par%i_t_extrap-min_it_snap_image,rat_t_snap_image)==0).and.&
      (rev_extrap_par%i_t_extrap >= min_it_snap_image).and.&
      (rev_extrap_par%i_t_extrap <= max_it_snap_image)
    if(gen_par%verb) write(0,*)'snap_image_step=',rev_extrap_par%snap_image_step

    rev_extrap_par%snap_data_step= &
      (mod(rev_extrap_par%i_t_extrap-min_it_snap_data,rat_t_snap_data)==0).and.&
      (rev_extrap_par%i_t_extrap >= min_it_snap_data).and.&
      (rev_extrap_par%i_t_extrap <= max_it_snap_data)
    if(gen_par%verb) write(0,*)'snap_data_step=',rev_extrap_par%snap_data_step

    if (gen_par%domain_range(1:8) == 'lin_imag') then
    !{
      if(rev_extrap_par%wave_back_step) then
        call SEP_timer_start(12)
        call input_wave(wave_par%tag_back,gen_par,wave_par,rev_extrap_par,& 
                 wave_back_block)
        call SEP_timer_stop(12)
        stat=rev_scatter_oper(rev_extrap_par,wave_par, &
           wave_block,wave_back_block,image_mem,image_mem_hz,slow_block) 
      end if
    !}
    end if

    if(wave_par%file == WAVE_OUTPUT) then
      if(rev_extrap_par%snap_wave_second_step) then
        if(gen_par%debug1) write(0,*)'Output Wavefield'
        call SEP_timer_start(12)
        call output_wave(wave_par%tag_second,gen_par,wave_par,rev_extrap_par, &
                          wave_block)
        call SEP_timer_stop(12)
      end if
    end if
  
    if(data_par%file == DATA_OUTPUT) then
      if(rev_extrap_par%snap_data_step) then
        if(gen_par%debug1) write(0,*)'Output Data'
        call SEP_timer_start(14)
        call output_data(data_par%tag,gen_par,data_par,wave_par,rev_extrap_par,&
                         wave_block)
        call SEP_timer_stop(14)
      end if
    end if

    if (data_par%file == DATA_INPUT) then
      !{
      if(rev_extrap_par%snap_data_step ) then
        if(gen_par%debug1) write(0,*)'Read receiver Data'
        call SEP_timer_start(14)
        call input_data(data_par%tag,gen_par,data_par,wave_par,rev_extrap_par, &
                         force_block,wave_block)
        call SEP_timer_stop(14)
      end if
      !}
    end if

    if (gen_par%domain_range(1:8) == 'vel_data') then
      if(wavelet_par%file == WAVELET_INPUT) then
        call add_wavelet(gen_par,wave_par,wavelet_par,rev_extrap_par, &
                         force_block,wavelet)
      end if
    end if

    if (image_par%file == IMAGE_INPUT) then
    !{
      if(image_par%two_images) then
        if(gen_par%debug1) write(0,*)'Input Image_rec'
        image_par%tag=trim(image_par%tag_rec)
        call input_image(gen_par,image_par,rev_extrap_par, &
            image_mem,image_mem_hz)
      end if

      if(rev_extrap_par%snap_image_step ) then
        call SEP_timer_start(13)
        if(gen_par%debug1) write(0,*)'Add Image to Receiver wavefield'
          call add_image(gen_par,image_par,wave_par,rev_extrap_par, &
                    wavelet_par,wave_block,wavelet,image_mem,image_mem_hz)

        call SEP_timer_stop(13)
      end if
    !}
    end if



!    !*$*assert do (serial) 
    do i_t_extrap=1,rev_extrap_par%n_t_second
    !{
      rev_extrap_par%i_t_extrap=i_t_extrap
      rev_extrap_par%t= rev_extrap_par%t_0_second+ &
         (rev_extrap_par%i_t_extrap)*rev_extrap_par%d_t_second

      if(gen_par%verb) then 
        write(0,*)'Receiver (source) wavefield: Doing time step #', &
                      rev_extrap_par%i_t_extrap
        write(0,*)'Receiver (source) wavefield: time=',rev_extrap_par%t
      end if



      rev_extrap_par%wave_back_step= &
 (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_back,rat_t_snap_wave_back)==0)&
   .and.&
       (rev_extrap_par%i_t_extrap >= min_it_snap_wave_back).and.&
       (rev_extrap_par%i_t_extrap <= max_it_snap_wave_back)
      if(gen_par%verb) write(0,*)'wave_back_step=',rev_extrap_par%wave_back_step

      rev_extrap_par%snap_wave_second_step= &
        (mod(rev_extrap_par%i_t_extrap-min_it_snap_wave_second,rat_t_snap_wave_second)==0).and. &
        (rev_extrap_par%i_t_extrap >= min_it_snap_wave_second).and.&
        (rev_extrap_par%i_t_extrap <= max_it_snap_wave_second)
      if(gen_par%verb) write(0,*)'snap_wave_second_step=',rev_extrap_par%snap_wave_second_step

      rev_extrap_par%snap_image_step= &
        (mod(rev_extrap_par%i_t_extrap-min_it_snap_image,rat_t_snap_image)==0).and.&
        (rev_extrap_par%i_t_extrap >= min_it_snap_image).and.&
        (rev_extrap_par%i_t_extrap <= max_it_snap_image)
      if(gen_par%verb) write(0,*)'snap_image_step=',rev_extrap_par%snap_image_step

      rev_extrap_par%snap_data_step= &
        (mod(rev_extrap_par%i_t_extrap-min_it_snap_data,rat_t_snap_data)==0).and. &
        (rev_extrap_par%i_t_extrap >= min_it_snap_data).and.&
        (rev_extrap_par%i_t_extrap <= max_it_snap_data)
      if(gen_par%verb) write(0,*)'snap_data_step=',rev_extrap_par%snap_data_step

!      rev_extrap_par%data_step= &
!    (rev_extrap_par%t >= data_par%t_0) .and. &
!   ((abs(amod(rev_extrap_par%t-data_par%t_0,data_par%d_t)) < epsi_data_d_t) &
!     .or. &
!    (data_par%d_t-abs(amod(rev_extrap_par%t-data_par%t_0,data_par%d_t)) <  &
!           epsi_data_d_t) &
!   )
!
!      rev_extrap_par%wave_step= &
!    (rev_extrap_par%t >= data_par%t_0) .and. &
!   ((abs(amod(rev_extrap_par%t-wave_par%t_0,wave_par%d_t)) < epsi_wave_d_t) &
!     .or. &
!    (wave_par%d_t-abs(amod(rev_extrap_par%t-wave_par%t_0,wave_par%d_t)) <  &
!           epsi_wave_d_t) &
!   )
!
!      rev_extrap_par%snap_image_step= &
!    (rev_extrap_par%t >= data_par%t_0) .and. &
!   ((abs(amod(rev_extrap_par%t-image_par%t_0,image_par%d_t)) < epsi_image_d_t) &
!     .or. &
!    (image_par%d_t-abs(amod(rev_extrap_par%t-image_par%t_0,image_par%d_t)) <  &
!           epsi_image_d_t) &
!   )

      if(gen_par%check_stability) then
       !{
         norm_wave=sqrt(sum(wave_block**2))
         write(0,*)'norm wavefield at time=',rev_extrap_par%t,norm_wave
       !}
      end if

! extrapolation step

      select case (gen_par%oper_type(1:2))
      case ('co')
        if(gen_par%debug1) write(0,*)'oper_type = Constant'
          stat= rev_extrap_oper( &
            rev_extrap_par=rev_extrap_par, &
            wave_par=wave_par, &
            force_block=force_block, &
            wave_block=wave_block &
            )
      case ('fu')
        if(gen_par%debug1) write(0,*)'oper_type = full'
    call SEP_timer_start(15)

#ifdef __CONVEY
!$cny begin_coproc
cny_i_timer=2
call CNY_timer_start(cny_i_timer)
#endif

!call cny$mput("#1", 2)

#ifdef __CONVEY

!          stat= rev_extrap_oper( &
          stat= Rev2D_fd_extrap_oper( &
            rev_extrap_par=rev_extrap_par, &
            wave_par=wave_par, &
            force_block=force_block, &
            wave_block=wave_block, &
            slow_block=slow_block)

!call cny$mput("#2", 2)

!$cny end_coproc
cny_i_timer=2
call CNY_timer_stop(cny_i_timer)
#else
          stat= rev_extrap_oper( &
            rev_extrap_par=rev_extrap_par, &
            wave_par=wave_par, &
            force_block=force_block, &
            wave_block=wave_block, &
            slow_block=slow_block)

!call cny$mput("#2", 2)

#endif

    call SEP_timer_stop(15)
      case default
         write(0,*)'gen_par-oper_type=',trim(gen_par%oper_type)
         call seperr('Not supported oper_type')
      end select

      if(wave_par%file == WAVE_OUTPUT) then
      !{
        if(rev_extrap_par%snap_wave_second_step) then
          if(gen_par%debug1) write(0,*)'Output Wavefield'
          call SEP_timer_start(12)
          call output_wave(wave_par%tag_second,gen_par,wave_par, &
                           rev_extrap_par,wave_block)
          call SEP_timer_stop(12)
        end if
      !}
      end if

      if(data_par%file == DATA_OUTPUT) then
      !{
        if(rev_extrap_par%snap_data_step ) then
          if(gen_par%debug1) write(0,*)'Output Data'
          call SEP_timer_start(14)
          call output_data(data_par%tag,gen_par,data_par,wave_par, &
                 rev_extrap_par,wave_block)
          call SEP_timer_stop(14)
       end if
      !}
      end if

      if (image_par%file == IMAGE_OUTPUT) then
      !{
        if(rev_extrap_par%snap_image_step ) then
          if(gen_par%debug1) write(0,*)'Output Image'
          call output_image(image_par%tag,image_par%tag_hz,gen_par,image_par, &
                             wave_par,rev_extrap_par, &
                             max_it_snap_image, &
                             wave_block,image_mem,image_mem_hz)
        end if
      !}
      end if

      force_block=0

      if (data_par%file == DATA_INPUT) then
        !{
        if(rev_extrap_par%snap_data_step ) then
          if(gen_par%debug1) write(0,*)'Read receiver Data'
          call SEP_timer_start(14)
          call input_data(data_par%tag,gen_par,data_par,wave_par, &
                  rev_extrap_par,force_block,wave_block)
          call SEP_timer_stop(14)
        end if
        !}
      end if

      if (gen_par%domain_range(1:8) == 'vel_data')  then
        if(wavelet_par%file == WAVELET_INPUT) then
          call add_wavelet(gen_par,wave_par,wavelet_par,rev_extrap_par, &
                           force_block,wavelet)
        end if
      end if

      if (image_par%file == IMAGE_INPUT) then
      !{
        if(rev_extrap_par%snap_image_step ) then
          call SEP_timer_start(13)
          if(gen_par%debug1) write(0,*)'Add Image to Receiver wavefield'
          call add_image(gen_par,image_par,wave_par,rev_extrap_par, &
                    wavelet_par,wave_block,wavelet,image_mem,image_mem_hz)

          call SEP_timer_stop(13)
        end if
      !}
      end if

      if (gen_par%domain_range(1:8) == 'lin_imag') then
      !{
        if(rev_extrap_par%wave_back_step) then
          call SEP_timer_start(12)
          call input_wave(wave_par%tag_back,gen_par,wave_par,rev_extrap_par,& 
                 wave_back_block)
          call SEP_timer_stop(12)
          stat=rev_scatter_oper(rev_extrap_par,wave_par, &
             wave_block,wave_back_block,image_mem,image_mem_hz,slow_block) 
        end if
      !}
      end if

    !}
    end do


    end do do_s1_par 
    end do do_s2_par 

    stat=0
    if(any(stat_par /= 0)) stat =1
    return
  end function rev_extrap_gen
end module rev_extrap_gen_mod
