module slow_mod
  use genrev3d_types
  use data_types
  use slow_types
  use wave_types

  implicit none
  logical, private :: interp_x_slow,interp_y_slow
  integer, external ::  sreed,srite,sseek_block
  integer, private :: n_bytes,n_ztau
  integer, private :: sou_slow,rec_slow

  integer, allocatable, dimension (:,:,:), private :: ix_slow,iy_slow
  real, allocatable, dimension (:,:,:), private :: wx_slow,wy_slow

contains
  subroutine init_io_slow(gen_par,wave_par,slow_par)
    implicit none
    integer ierr
    type (genrev3d_param_type) gen_par
    type (wave_param_type) wave_par
    type (slow_param_type) slow_par
    integer rat_dzdtau

    n_ztau=wave_par%n_z
    n_bytes=slow_par%n_mx*slow_par%n_my*slow_par%n_z* 4
       write(0,*)'n_bytes=',n_bytes
       write(0,*)'slow_par%n_mx=',slow_par%n_mx
       write(0,*)'slow_par%n_my=',slow_par%n_my
       write(0,*)'n_ztau=',n_ztau

    rat_dzdtau=nint(slow_par%d_z/wave_par%d_z)
    if((rat_dzdtau /= 1) .and. (rat_dzdtau /= -1)) then
       write(0,*)'slow_par-d_z',slow_par%d_z
       write(0,*)'wave_par-d_z',wave_par%d_z
       call seperr('rat_dzdtau /= 1 not implemented yet')
    end if

    call init_interpolate_slow(gen_par,wave_par,slow_par)

  end subroutine init_io_slow

  subroutine init_interpolate_slow(gen_par,wave_par,slow_par)
    implicit none
    type (genrev3d_param_type) gen_par
    type (wave_param_type) wave_par
    type (slow_param_type) slow_par
    integer ierr
    integer i_mx,i_my,i_hx,i_hy
    real mx,my,hx,hy,rx,ry,sx,sy,inv_d_mx,inv_d_my

!    if(slow_par%slow_shear_file /= SLOW_NO_FILE) then
!      if(slow_par%shear_receiver) then
!        sou_slow=P_SLOW
!        rec_slow=S_SLOW
!      else
!        sou_slow=S_SLOW
!        rec_slow=P_SLOW
!      end if
!    else
        sou_slow=P_SLOW
        rec_slow=P_SLOW
!    end if

    interp_x_slow=.false.
    interp_y_slow=.false.

    if ((wave_par%n_mx /= 1).or.(wave_par%n_hx /= 1))  interp_x_slow=.true.
    if ((wave_par%n_my /= 1).or.(wave_par%n_hy /= 1))  interp_y_slow=.true.

!    if(((slow_par%n_mx /= wave_par%n_mx) .or.  &
!        (slow_par%d_mx /= wave_par%d_mx) .or.  &
!        (slow_par%mx_0 /= wave_par%mx_0) .or.  &
!        (wave_par%n_hx /= 1)) .and. &
!        (slow_par%n_mx /= 1) &
!      ) interp_x_slow=.true.
!
!    if(((slow_par%n_my /= wave_par%n_my) .or.  &
!        (slow_par%d_my /= wave_par%d_my) .or.  &
!        (slow_par%my_0 /= wave_par%my_0) .or.  &
!        (wave_par%n_hy /= 1)) .and. &
!        (slow_par%n_my /= 1) &
!      ) interp_y_slow=.true.
!
!    if((.not. interp_y_slow) .and. (.not. slow_par%extrapolate_slow)) then
!      if(wave_par%my_0 < slow_par%my_0) then
!        call seperr('wave_par-my_0 < slow_par-my_0')
!      end if
!      if(wave_par%my_0 + wave_par%d_my*(wave_par%n_my-1) > & 
!         slow_par%my_0 + slow_par%d_my*(slow_par%n_my-1)) then
!        call seperr('Data my larger than max slow my')
!      end if
!    end if
    

    if(gen_par%debug2) write(0,*)'interp_x_slow=',interp_x_slow
    if(gen_par%debug2) write(0,*)'interp_y_slow=',interp_y_slow

    if(interp_x_slow) then
       !{
       allocate (ix_slow(wave_par%n_mx, &
       wave_par%n_hx, &
       slow_par%n_comp), &
       stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of ix_slow")
       if(gen_par%debug2) write(0,*)'Allocate ix_slow with shape=',shape(ix_slow)

       allocate (wx_slow(wave_par%n_mx, &
       wave_par%n_hx, &
       slow_par%n_comp), &
       stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of wx_slow")
       if(gen_par%debug2) write(0,*)'Allocate wx_slow with shape=',shape(wx_slow)
       !}
    end if

    if(interp_y_slow) then
       !{
       allocate (iy_slow(wave_par%n_my, &
       wave_par%n_hy, &
       slow_par%n_comp), &
       stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of iy_slow")
       if(gen_par%debug2) write(0,*)'Allocate iy_slow with shape=',shape(iy_slow)

       allocate (wy_slow(wave_par%n_my, &
       wave_par%n_hy, &
       slow_par%n_comp), &
       stat=ierr)  
       if(ierr .ne. 0) call seperr("Problems with allocation of wy_slow")
       if(gen_par%debug2) write(0,*)'Allocate wy_slow with shape=',shape(wy_slow)
       !}
    end if


  end subroutine init_interpolate_slow


  subroutine init_coeff_interpolate_slow(gen_par,wave_par,slow_par)
    implicit none
    type (genrev3d_param_type) gen_par
    type (wave_param_type) wave_par
    type (slow_param_type) slow_par
    integer ierr
    integer i_mx,i_my,i_hx,i_hy
    real mx,my,hx,hy,rx,ry,sx,sy,inv_d_mx,inv_d_my

    if(interp_x_slow) then
       !{

       inv_d_mx=1./slow_par%d_mx

       do i_hx=1,wave_par%n_hx 
          !{
          hx=wave_par%hx_0 + (i_hx-1)* wave_par%d_hx
          do i_mx=1,wave_par%n_mx 
             !{
             mx=wave_par%mx_0 + (i_mx-1)* wave_par%d_mx
             sx=mx-hx
             sx=(sx-slow_par%mx_0)*inv_d_mx+1
             if((sx < 1) .or. (sx >= slow_par%n_mx)) then
                if(slow_par%extrapolate_slow) then
                   ix_slow(i_mx,i_hx,SOU_COMP)=min(slow_par%n_mx-1,max(1,int(sx)))
                   wx_slow(i_mx,i_hx,SOU_COMP)=.5
                else
                   write(0,*)'sx=',sx
                   call seperr('Try to use velocity/slowness outside range of mx')
                end if
             else
                ix_slow(i_mx,i_hx,SOU_COMP)=int(sx)
                wx_slow(i_mx,i_hx,SOU_COMP)=sx-int(sx)
             end if

             if(slow_par%n_comp > 1) then
                rx=mx+hx
                rx=(rx-slow_par%mx_0)*inv_d_mx+1
                if((rx < 1) .or. (rx >= slow_par%n_mx)) then
                   if(slow_par%extrapolate_slow) then
                      ix_slow(i_mx,i_hx,REC_COMP)=min(slow_par%n_mx-1,max(1,int(rx)))
                      wx_slow(i_mx,i_hx,REC_COMP)=.5
                   else
                      write(0,*)'rx=',rx
                      call seperr('Try to use velocity/slowness outside range of my')
                   end if
                else
                   ix_slow(i_mx,i_hx,REC_COMP)=int(rx)
                   wx_slow(i_mx,i_hx,REC_COMP)=rx-int(rx)
                end if
             end if
             !}
          end do
          !}
       end do
       !}
    end if

    if(interp_y_slow) then
       !{
       inv_d_my=1./slow_par%d_my

       do i_hy=1,wave_par%n_hy 
          !{
          hy=wave_par%hy_0 + (i_hy-1)* wave_par%d_hy
          do i_my=1,wave_par%n_my 
             !{
             my=wave_par%my_0 + (i_my-1)* wave_par%d_my
             sy=my-hy
             sy=(sy-slow_par%my_0)*inv_d_my+1
             if((sy < 1) .or. (sy >= slow_par%n_my)) then
                if(slow_par%extrapolate_slow) then
                   iy_slow(i_my,i_hy,SOU_COMP)=min(slow_par%n_my-1,max(1,int(sy)))
                   wy_slow(i_my,i_hy,SOU_COMP)=.5
                else
                   write(0,*)'sx=',sx
                   call seperr('Try to use velocity/slowness outside range of my')
                end if
             else
                iy_slow(i_my,i_hy,SOU_COMP)=int(sy)
                wy_slow(i_my,i_hy,SOU_COMP)=sy-int(sy)
             end if

             if(slow_par%n_comp > 1) then
                ry=my+hy
                ry=(ry-slow_par%my_0)*inv_d_my+1
                if((ry < 1) .or. (ry >= slow_par%n_my)) then
                   if(slow_par%extrapolate_slow) then
                      iy_slow(i_my,i_hy,REC_COMP)=min(slow_par%n_my-1,max(1,int(ry)))
                      wy_slow(i_my,i_hy,REC_COMP)=.5
                   else
                      write(0,*)'rx=',rx
                      call seperr('Try to use velocity/slowness outside range of my')
                   end if
                else
                   iy_slow(i_my,i_hy,REC_COMP)=int(ry)
                   wy_slow(i_my,i_hy,REC_COMP)=ry-int(ry)
                end if
             end if
             !}
          end do
          !}
       end do

       !}
    end if
    return
  end subroutine init_coeff_interpolate_slow

  subroutine read_slow(gen_par,tag_slow, wave_par,slow_par, slow_block)
    implicit none

    character(len=128) :: tag_slow
    type (genrev3d_param_type) gen_par
    type (wave_param_type) wave_par
    type (slow_param_type) slow_par
    real, dimension (:,:,:,:,:,:) :: slow_block

    real, allocatable, dimension (:,:,:,:) :: slow_cube

    integer :: ierr,iseek,len_tag,n_bytes_io

    if(slow_par%file /= SLOW_NO_FILE) then
       allocate (slow_cube(slow_par%n_z, &
       slow_par%n_mx,slow_par%n_my,1) , &
       stat=ierr)  
!       !##############
!       write(0,*) shape(slow_cube),ierr
!       write(0,*) slow_par%n_z,slow_par%n_mx,slow_par%n_my,slow_par%n_comp
!       !##############
       if(ierr .ne. 0) call seperr("Problems with allocation of slow_cube")
       if(gen_par%debug2) write(0,*)'Allocate slow_cube with shape=',shape(slow_cube)
    end if
    
    len_tag=len_trim(tag_slow)

    iseek=0
    if(sseek_block(trim(tag_slow),iseek,n_bytes,0) .ne. iseek ) then
       write(0,*)'tag_slow=',trim(tag_slow)
       call seperr('Problems with seeking in tag_slow file (first)')
    end if

!    write(0,*)'n_bytes=',n_bytes

    n_bytes_io=sreed(trim(tag_slow),slow_cube(:,:,:,P_SLOW),n_bytes)
    if(n_bytes_io .ne. n_bytes ) then
       write(0,*)'iseek=',iseek
       write(0,*)'n_bytes_io=',n_bytes_io
       write(0,*)'n_bytes=',n_bytes
       write(0,*)'tag_slow=',trim(tag_slow)
       call seperr('Problems with reading tag_slow file (real)')
    end if

    call init_coeff_interpolate_slow(gen_par,wave_par,slow_par)
    call interpolate_slow(.false.,wave_par,slow_par, slow_block,slow_cube)

    slow_par%slow_min=minval(slow_block)
!    write(0,*)"slow_min=",slow_par%slow_min

    if(allocated(slow_cube)) then
       deallocate (slow_cube,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating slow_cube')
    end if

#ifdef __CONVEY
!$cny migrate_coproc(slow_block, n_bytes)
#endif

    return
  end subroutine read_slow

  subroutine write_slow(tag_slow, wave_par,slow_par, slow_block)
    implicit none

    character(len=128) :: tag_slow
    type (wave_param_type) wave_par
    type (slow_param_type) slow_par
    real, dimension (:,:,:,:,:,:) :: slow_block

    logical add_to_dslow
    integer :: ierr,iseek,iseek_slice,len_tag
    integer :: n_bytes_block,n_bytes_slice

    call seperr('write slow not implemented yet')

  end subroutine write_slow

  subroutine interpolate_slow(adj,wave_par,slow_par, slow_block,slow_cube)
    implicit none

    logical adj
    type (wave_param_type) wave_par
    type (slow_param_type) slow_par
    real, dimension (:,:,:,:,:,:) :: slow_block
    real, dimension (:,:,:,:) :: slow_cube

    character(len=1) :: type_slow
    integer :: ierr,i_ztau,len_tag
    integer i_mx,i_my,i_hx,i_hy,i_mx_slow
    real w00_sou,w10_sou,w01_sou,w11_sou
    real w00_rec,w10_rec,w01_rec,w11_rec

    if(adj) then
       !{
       if(interp_x_slow .and. interp_y_slow) then
          call seperr('interpolate_slow: adj=.true. and 3-D interp')
       else
          !{
          if(interp_x_slow) then
             !{
           do i_hy=1,wave_par%n_hy 
              !{
              do i_hx=1,wave_par%n_hx 
                 !{
                 do i_my=1,wave_par%n_my 
                    !{
                    do i_mx=1,wave_par%n_mx 
                       !{
                       w00_sou=(1.-wx_slow(i_mx,i_hx,SOU_COMP))
                       w10_sou=(   wx_slow(i_mx,i_hx,SOU_COMP))

                        i_mx_slow=ix_slow(i_mx,i_hx,SOU_COMP)
                        !write(0,*)'i_mx_slow',i_mx_slow
                        do i_ztau=1,n_ztau
                           !{
                          slow_cube(i_ztau, i_mx_slow ,i_my,1) =             & 
                           slow_cube(i_ztau, i_mx_slow  ,i_my,1) +             & 
                           slow_block(i_ztau, i_mx,i_my,i_hx,i_hy,SOU_COMP)* &
                           w00_sou 

                          slow_cube(i_ztau, i_mx_slow+1, i_my,1) =             &
                           slow_cube(i_ztau, i_mx_slow+1, i_my,1) +             &
                           slow_block(i_ztau, i_mx,i_my,i_hx,i_hy,SOU_COMP)* &
                           w10_sou 
                           !}
                        end do
                       !}
                    end do
                    !}
                 end do
                 !}
              end do
              !}
           end do

           if(slow_par%n_comp > 1) then
              !{
              do i_hy=1,wave_par%n_hy 
                 !{
                 do i_hx=1,wave_par%n_hx 
                    !{
                    do i_my=1,wave_par%n_my 
                       !{
                       do i_mx=1,wave_par%n_mx 
                          !{
                          w00_rec=(1.-wx_slow(i_mx,i_hx,REC_COMP))
                          w10_rec=(   wx_slow(i_mx,i_hx,REC_COMP))

                          i_mx_slow=ix_slow(i_mx,i_hx,REC_COMP)
                          do i_ztau=1,n_ztau
                             !{
                             slow_cube(i_ztau, i_mx_slow , i_my,1) =             & 
                             slow_cube(i_ztau, i_mx_slow , i_my,1) +             & 
                             slow_block(i_ztau, i_mx,i_my,i_hx,i_hy,REC_COMP)* &
                             w00_rec

                             slow_cube(i_ztau, i_mx_slow+1, i_my,1) =             &
                             slow_cube(i_ztau, i_mx_slow+1, i_my,1) +             &
                             slow_block(i_ztau, i_mx,i_my,i_hx,i_hy,REC_COMP)* &
                              w10_rec 
                              !}
                           end do
                          !}
                       end do
                       !}
                    end do
                    !}
                 end do
                 !}
              end do

              !}
           end if
           !}
          else
             !{
             if(interp_y_slow) then
                call seperr ('Interpolation in y not implemented yet')
             else
                !{
                 if(slow_par%n_comp > 1) then
                    slow_cube(:,:,:,1) = &
                    .5*(slow_block(:,:,:,1,1,SOU_COMP) + &
                        slow_block(:,:,:,1,1,REC_COMP))
                 else
                    slow_cube(:,:,:,1)=slow_block(:,:,:,1,1,1)
                 end if
                !}
             end if
             !}
          end if
          !}
       end if

       !}
    else
       !{
       if(interp_x_slow .and. interp_y_slow) then
          !{
          do i_hy=1,wave_par%n_hy 
             !{
             do i_hx=1,wave_par%n_hx 
                !{
                do i_my=1,wave_par%n_my 
                   !{
                   do i_mx=1,wave_par%n_mx 
                      !{
                      w00_sou=(1.-wx_slow(i_mx,i_hx,SOU_COMP))* &
                      (1.-wy_slow(i_my,i_hy,SOU_COMP))
                      w10_sou=(   wx_slow(i_mx,i_hx,SOU_COMP))* &
                      (1.-wy_slow(i_my,i_hy,SOU_COMP))
                      w01_sou=(1.-wx_slow(i_mx,i_hx,SOU_COMP))* &
                      (   wy_slow(i_my,i_hy,SOU_COMP))
                      w11_sou=(   wx_slow(i_mx,i_hx,SOU_COMP))* &
                      (   wy_slow(i_my,i_hy,SOU_COMP))
                       do i_ztau=1,n_ztau
                          !{
                        slow_block(i_ztau,i_mx,i_my,i_hx,i_hy,SOU_COMP) = &
                          slow_cube(i_ztau,ix_slow(i_mx,i_hx,SOU_COMP)  , &
                            iy_slow(i_my,i_hy,SOU_COMP)  , &
                            sou_slow) * w00_sou +         &
                          slow_cube(i_ztau,ix_slow(i_mx,i_hx,SOU_COMP)+1, &
                            iy_slow(i_my,i_hy,SOU_COMP)  , &
                            sou_slow) * w10_sou +         &
                          slow_cube(i_ztau,ix_slow(i_mx,i_hx,SOU_COMP)  , &
                            iy_slow(i_my,i_hy,SOU_COMP)+1, &
                            sou_slow) * w01_sou +         &
                          slow_cube(i_ztau,ix_slow(i_mx,i_hx,SOU_COMP)+1, &
                            iy_slow(i_my,i_hy,SOU_COMP)+1, &
                            sou_slow) * w11_sou 
                          !}
                       end do
                      !}
                   end do
                   !}
                end do
                !}
             end do
             !}
          end do

          if(slow_par%n_comp > 1) then
             !{
             do i_hy=1,wave_par%n_hy 
                !{
                do i_hx=1,wave_par%n_hx 
                   !{
                   do i_my=1,wave_par%n_my 
                      !{
                      do i_mx=1,wave_par%n_mx 
                         !{
                         w00_rec=(1.-wx_slow(i_mx,i_hx,REC_COMP))* &
                         (1.-wy_slow(i_my,i_hy,REC_COMP))
                         w10_rec=(   wx_slow(i_mx,i_hx,REC_COMP))* &
                         (1.-wy_slow(i_my,i_hy,REC_COMP))
                         w01_rec=(1.-wx_slow(i_mx,i_hx,REC_COMP))* &
                         (   wy_slow(i_my,i_hy,REC_COMP))
                         w11_rec=(   wx_slow(i_mx,i_hx,REC_COMP))* &
                         (   wy_slow(i_my,i_hy,REC_COMP))
                          do i_ztau=1,n_ztau
                             !{
                           slow_block(i_ztau,i_mx,i_my,i_hx,i_hy,REC_COMP) = &
                             slow_cube(i_ztau,ix_slow(i_mx,i_hx,REC_COMP)  , &
                               iy_slow(i_my,i_hy,REC_COMP)  , &
                               rec_slow) * w00_rec +         &
                             slow_cube(i_ztau,ix_slow(i_mx,i_hx,REC_COMP)+1, &
                               iy_slow(i_my,i_hy,REC_COMP)  , &
                               rec_slow) * w10_rec +         &
                             slow_cube(i_ztau,ix_slow(i_mx,i_hx,REC_COMP)  , &
                               iy_slow(i_my,i_hy,REC_COMP)+1, &
                               rec_slow) * w01_rec +         &
                             slow_cube(i_ztau,ix_slow(i_mx,i_hx,REC_COMP)+1, &
                               iy_slow(i_my,i_hy,REC_COMP)+1, &
                               rec_slow) * w11_rec 
                             !}
                          end do
                         !}
                      end do
                      !}
                   end do
                   !}
                end do
                !}
             end do

             !}
          end if
          !}
       else
          !{
          if(interp_x_slow) then
             !{

             do i_hy=1,wave_par%n_hy 
                !{
                do i_hx=1,wave_par%n_hx 
                   !{
                   do i_my=1,wave_par%n_my 
                      !{
                      do i_mx=1,wave_par%n_mx 
                         !{
                         w00_sou=(1.-wx_slow(i_mx,i_hx,SOU_COMP))
                         w10_sou=(   wx_slow(i_mx,i_hx,SOU_COMP))

                          do i_ztau=1,n_ztau
                             !{
                             slow_block(i_ztau,i_mx,i_my,i_hx,i_hy,SOU_COMP) = &
                               slow_cube(i_ztau,ix_slow(i_mx,i_hx,SOU_COMP)  , &
                                 i_my,                         &
                                 sou_slow) * w00_sou +        &
                               slow_cube(i_ztau,ix_slow(i_mx,i_hx,SOU_COMP)+1, &
                                 i_my,                         &
                                 sou_slow) * w10_sou 

                             !}
                          end do
                         !}
                      end do
                      !}
                   end do
                   !}
                end do
                !}
             end do

!write(0,*)'slow_par%n_comp=',slow_par%n_comp
             if(slow_par%n_comp > 1) then
                !{
                do i_hy=1,wave_par%n_hy 
                   !{
                   do i_hx=1,wave_par%n_hx 
                      !{
                      do i_my=1,wave_par%n_my 
                         !{
                         do i_mx=1,wave_par%n_mx 
                            !{
                            w00_rec=(1.-wx_slow(i_mx,i_hx,REC_COMP))
                            w10_rec=(   wx_slow(i_mx,i_hx,REC_COMP))
                             do i_ztau=1,n_ztau
                                !{
                                slow_block(i_ztau,i_mx,i_my,i_hx,i_hy,REC_COMP) = &
                                  slow_cube(i_ztau,ix_slow(i_mx,i_hx,REC_COMP)  , &
                                    i_my,                          &
                                    rec_slow) * w00_rec +         &
                                  slow_cube(i_ztau,ix_slow(i_mx,i_hx,REC_COMP)+1, &
                                    i_my,                          &
                                    rec_slow) * w10_rec 
                                !}
                             end do
                            !}
                         end do
                         !}
                      end do
                      !}
                   end do
                   !}
                end do

                !}
             end if
             !}
          else
             !{
             if(interp_y_slow) then
                !{
                call seperr ('Interpolation in y not implemented yet')
                !}
             else
                !{
                 if(slow_par%n_comp > 1) then
                    slow_block(:,:,:,1,1,SOU_COMP)= &
                      slow_cube(:,:,:,1)
                    slow_block(:,:,:,1,1,REC_COMP)= &
                      slow_cube(:,:,:,1)
                 else
                    slow_block(:,:,:,1,1,1)=slow_cube(:,:,:,sou_slow)
                 end if
                !}
             end if
             !}
          end if
          !}
       end if

       if(.not. slow_par%slowness) then
          slow_block=1./slow_block
       end if
       if(slow_par%scale /= 1.) then
          write(0,*)'Scale slowness by scale=',slow_par%scale
          slow_block=slow_par%scale*slow_block
       end if
       !}
    end if

    return
  end subroutine interpolate_slow


  subroutine clean_io_slow()
    implicit none
    integer ierr

    if(allocated(ix_slow)) then
       deallocate (ix_slow,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating ix_slow')
    end if
    if(allocated(iy_slow)) then
       deallocate (iy_slow,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating iy_slow')
    end if
    if(allocated(wx_slow)) then
       deallocate (wx_slow,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating wx_slow')
    end if
    if(allocated(wy_slow)) then
       deallocate (wy_slow,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating wy_slow')
    end if
    return
  end subroutine clean_io_slow
end module slow_mod
