module wave_mod

  use genrev3d_types
  use wave_types
  use slow_types
  use rev_extrap_types


  implicit none

  logical, private :: adj,add

  integer, external ::  sreed,srite,sseek_block
  integer, private :: n_bytes_io,n_ztau

  real, allocatable, dimension (:,:,:), private :: wave
  
contains

  !-----------------------------------------------------------------------------
  subroutine init_io_wave(gen_par,wave_par)

    type (genrev3d_param_type) gen_par
    type (wave_param_type) wave_par
    integer ierr

    !+++
    n_ztau=wave_par%n_z
    n_bytes_io=n_ztau*wave_par%n_mx* &
               wave_par%n_my*wave_par%esize

    allocate (wave(n_ztau,wave_par%n_mx,wave_par%n_my),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of wave in init_io_wave")

!#ifdef __CONVEY
!!$cny migrate_coproc(wave, n_bytes_io)
!#endif

  end subroutine init_io_wave


  !-----------------------------------------------------------------------------
  subroutine output_wave(tag_wave,gen_par,wave_par,rev_extrap_par,wave_block)

   character(len=128) :: tag_wave
   type (genrev3d_param_type) gen_par
   type (wave_param_type) wave_par
   type (rev_extrap_param_type) :: rev_extrap_par
   real,  dimension (:,:,:,:,:,:)   :: wave_block

   logical :: rewind
   integer :: len_tag,i_hx,i_hy,ret_sseek,min_it_snap_wave_back_first
   integer :: i_t_step_wave

   min_it_snap_wave_back_first=&
               nint((wave_par%t_back_0-rev_extrap_par%t_0_first)/ & 
                     rev_extrap_par%d_t_first)+1

   len_tag=min(len_trim(tag_wave),len_trim(wave_par%tag_back))

   if (tag_wave(1:len_tag) == wave_par%tag_back(1:len_tag)) then
     i_t_step_wave=wave_par%i_t_step_back(1)
   else
     i_t_step_wave=wave_par%i_t_step(1)
   end if

   rewind = ((tag_wave(1:len_tag) == wave_par%tag_back(1:len_tag)) .and. &
             (rev_extrap_par%i_t_extrap == min_it_snap_wave_back_first) .and.  &
             ((rev_extrap_par%i_s1_par /= 1) .or.  &
              (rev_extrap_par%i_s2_par /= 1)) &
            )


!  write(0,*)'tag_wave=',tag_wave(1:len_tag)
!  write(0,*)'tag_back=',wave_par%tag_back(1:len_trim(wave_par%tag_back))
!  write(0,*)'i_t_extrap=',rev_extrap_par%i_t_extrap
!  write(0,*)'i_s1_par=',rev_extrap_par%i_s1_par
!  write(0,*)'i_s2_par=',rev_extrap_par%i_s2_par
!  write(0,*)'rewind=',rewind
!
!write(0,*)'n_ztau=',n_ztau
!write(0,*)'n_my=',wave_par%n_my
!write(0,*)'n_mx=',wave_par%n_mx
!write(0,*)'n_hy=',wave_par%n_hy
!write(0,*)'n_hx=',wave_par%n_hx
!
!   if(gen_par%debug2) write(0,*)'len_tag',len_tag
!
!   if(gen_par%debug2) write(0,*)'Writing to ',tag_wave(1:len_tag)
!
!   if(gen_par%debug2) write(0,*)'shape(wave)',shape(wave)
!   if(gen_par%debug2) write(0,*)'shape(wave_block)',shape(wave_block)
!
!
!   if(gen_par%debug2) write(0,*)'Before seek'

   if(rewind) then
   ret_sseek=sseek_block(trim(tag_wave),0,n_bytes_io,0) 
     if (ret_sseek .ne. 0) then
     !{
       write(0,*)'ret_sseek=',ret_sseek
       write(0,*)'tag_wave=',trim(tag_wave)
       call seperr("Error block seeking wave in output")
     !}
     end if
   end if

!   if(gen_par%debug2) write(0,*)'Before srite'
!   if(gen_par%debug2) write(0,*)'i_t_step_wave=',i_t_step_wave

   do i_hy=1,wave_par%n_hy
     do i_hx=1,wave_par%n_hx
       wave=wave_block(:,:,:,i_hx,i_hy,i_t_step_wave)
!!#ifdef __CONVEY
!!!$cny migrate_host(wave, n_bytes_io)
!!#endif
       if(srite(trim(tag_wave),wave,n_bytes_io) .ne. n_bytes_io ) then
         write(0,*)'tag_wave=',trim(tag_wave)
         write(0,*)'i_hx=',i_hx
         write(0,*)'i_hy=',i_hy
         call seperr('Problems with writing tag_wave file')
       end if
!!#ifdef __CONVEY
!!!$cny migrate_coproc(wave, n_bytes_io)
!!#endif
     end do
   end do

!   if(gen_par%debug2) write(0,*)'AFter srite'


  return	
  end subroutine output_wave


  !-----------------------------------------------------------------------------
  subroutine input_wave(tag_wave,gen_par,wave_par,rev_extrap_par,wave_block)

   character(len=128) :: tag_wave
   type (genrev3d_param_type) gen_par
   type (wave_param_type) wave_par
   type (rev_extrap_param_type) rev_extrap_par
   real, dimension (:,:,:,:,:,:)   :: wave_block

   integer :: i_hx,i_hy
   integer :: i_t_wave,block_wave,nsseek_blocks,ret_sseek
   integer :: len_tag,i_t_step_wave
   real   :: wave(n_ztau,wave_par%n_mx,wave_par%n_my)

   i_t_wave=nint((rev_extrap_par%t - wave_par%t_back_0)/wave_par%d_t_back)+1
   if(gen_par%debug2) write(0,*)'i_t_wave=',i_t_wave

   len_tag=min(len_trim(tag_wave),len_trim(wave_par%tag_back))

   if (tag_wave(1:len_tag) == wave_par%tag_back(1:len_tag)) then
     i_t_step_wave=wave_par%i_t_step_back(1)
   else
     i_t_step_wave=wave_par%i_t_step(1)
   end if

   block_wave= ((i_t_wave-1))*wave_par%n_my ! Biondo 2/1/2006
   nsseek_blocks= block_wave 
!     write(0,*)'block_wave=',block_wave
!         write(0,*)'nsseek_blocks (wave)=',nsseek_blocks

   if(i_t_wave > wave_par%n_t_back) then
     write(0,*)'i_t_wave=',i_t_wave
     write(0,*)'block_wave=',block_wave
     write(0,*)'n_t_back=',wave_par%n_t_back
     call seperr("Tryng to read beyond saved time step in Wavefield_back")
   end if

   ret_sseek=sseek_block(trim(tag_wave),nsseek_blocks,n_bytes_io,0) 
   if (ret_sseek .ne. nsseek_blocks) then
   !{
     write(0,*)'nsseek_blocks=',nsseek_blocks
     write(0,*)'ret_sseek=',ret_sseek
     write(0,*)'tag_wave=',trim(tag_wave)
     call seperr("Error block seeking wave")
   !}
   end if

   do i_hy=1,wave_par%n_hy
     do i_hx=1,wave_par%n_hx
!#ifdef __CONVEY
!!$cny migrate_host(wave, n_bytes_io)
!#endif
       if(sreed(trim(tag_wave),wave,n_bytes_io) .ne. n_bytes_io ) then
         write(0,*)'tag_wave=',trim(tag_wave)
         write(0,*)'i_t_wave=',i_t_wave
         write(0,*)'block_wave=',block_wave
         write(0,*)'i_hx=',i_hx
         write(0,*)'i_hy=',i_hy
           call seperr('Problems with reading tag_wave file')
       end if
!#ifdef __CONVEY
!!$cny migrate_coproc(wave, n_bytes_io)
!#endif
       wave_block(:,:,:,i_hx,i_hy,i_t_step_wave)=wave
     end do
   end do


   return	
   end subroutine input_wave

  !-----------------------------------------------------------------------------


  !-----------------------------------------------------------------------------
  subroutine clean_io_wave()
    integer ierr

    if(allocated(wave)) deallocate (wave,stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with deallocation of wave in clean_io_wave")

    return
  end subroutine clean_io_wave


end module wave_mod
