module wavelet_mod

  use genrev3d_types
  use wave_types
  use wavelet_types
  use rev_extrap_types


  implicit none

  integer, external ::  sreed,srite,sseek_block
  integer, private :: n_bytes_io
  
contains

  !-----------------------------------------------------------------------------
  subroutine init_io_wavelet(gen_par,wavelet_par)

    type (genrev3d_param_type) gen_par
    type (wavelet_param_type) wavelet_par

    !+++
      n_bytes_io=wavelet_par%n_t*wavelet_par%n_z* &
                 wavelet_par%n_off_x*wavelet_par%n_off_y* &
                 wavelet_par%esize

    return
  end subroutine init_io_wavelet


  !-----------------------------------------------------------------------------
  subroutine read_wavelet(tag_wavelet,wavelet_par,wavelet)

   character(len=128) :: tag_wavelet
   type (wavelet_param_type) wavelet_par
   real,  dimension (:,:,:,:)   :: wavelet

   if(sreed(trim(tag_wavelet),wavelet,n_bytes_io) .ne. n_bytes_io ) then
     write(0,*)'tag_wavelet=',trim(tag_wavelet)
     call seperr('Problems with reading tag_wavelet file')
   end if

#ifdef __CONVEY
!$cny migrate_coproc(wavelet, n_bytes_io)
#endif

!write(0,*)'wavelet=',wavelet
!
   return	

   end subroutine read_wavelet

  !-----------------------------------------------------------------------------

  subroutine add_wavelet(gen_par,wave_par,wavelet_par,rev_extrap_par,force_block,wavelet)

   character(len=128) :: tag_wavelet
   type (genrev3d_param_type) gen_par
   type (wave_param_type) wave_par
   type (wavelet_param_type) wavelet_par
   type (rev_extrap_param_type) rev_extrap_par
   real,  dimension (:,:,:,:)   :: force_block
   real,  dimension (:,:,:,:)   :: wavelet

   integer i_t_source,i_mx_source,i_my_source,i_off_x,i_off_y,i_z

!   i_t_source =  &
!     nint( &
!      (rev_extrap_par%t-wavelet_par%t_0-abs(rev_extrap_par%d_t_second)+ &
!       wavelet_par%t_delay)/ &
!       wavelet_par%d_t &
!         ) + 1

   i_t_source =  &
     nint( &
      (rev_extrap_par%t-wavelet_par%t_0+ &
       wavelet_par%t_delay)/ &
       wavelet_par%d_t &
         ) + 1

     if(gen_par%debug2) write(0,*)'rev_extrap_par-t=',rev_extrap_par%t
     if(gen_par%debug2) write(0,*)'wavelet_par-t_0=',wavelet_par%t_0
     if(gen_par%debug2) write(0,*)'wavelet_par-t_delay=',wavelet_par%t_delay
     if(gen_par%debug2) write(0,*)'wavelet_par-d_t=',wavelet_par%d_t
     if(gen_par%debug2) write(0,*)'i_t_source=',i_t_source

   if(i_t_source < 1) then 
     write(0,*)'i_t_source=',i_t_source
     call seperr ("i_t_source < 1")
   end if

#ifdef __CONVEY
!$cny begin_coproc
#endif

   if(i_t_source < wavelet_par%n_t) then
     i_mx_source=& 
   ((rev_extrap_par%s1_par+wavelet_par%off_x_0-wave_par%mx_0)/wave_par%d_mx)+1
     i_my_source=& 
   ((rev_extrap_par%s2_par+wavelet_par%off_y_0-wave_par%my_0)/wave_par%d_my)+1
   !{
  
#ifndef __CONVEY
     if(gen_par%debug2) write(0,*)'i_t_source=',i_t_source
     if(gen_par%debug2) write(0,*)'i_mx_source=',i_mx_source
     if(gen_par%debug2) write(0,*)'i_my_source=',i_my_source
  
     if(i_mx_source < 1) call seperr ("i_mx_source < 1")
     if(i_my_source < 1) call seperr ("i_my_source < 1")
  
     if(i_mx_source+wavelet_par%n_off_x > wave_par%n_mx) &
        call seperr ("i_mx_source + wavelet-n_off_x > wave_par-n_mx")
  
#endif
     force_block(:,:,:,1)= 0.

     do i_off_y=1,wavelet_par%n_off_y
       do i_off_x=1,wavelet_par%n_off_x
         do i_z=1,wavelet_par%n_z
  
         force_block(i_z,i_mx_source+i_off_x-1,i_my_source+i_off_y-1,1)= &
           wavelet(i_t_source,i_z,i_off_x,i_off_y)
         end do
       end do
     end do
!write(0,*)'Inside minval(force_block)=',minval(force_block)
!write(0,*)'Inside maxval(force_block)=',maxval(force_block)
   !}
   else
   !{
     force_block(:,:,:,1)= 0.
   !}
   end if

#ifdef __CONVEY
!$cny end_coproc
#endif



   return	

   end subroutine add_wavelet

  !-----------------------------------------------------------------------------
  subroutine clean_io_wavelet()
    integer ierr

    return
  end subroutine clean_io_wavelet


end module wavelet_mod
