module Rev2D_fd_mod

  use Rev2D_fd_types

  use genrev3d_types
  use rev_extrap_types
  use data_types
  use slow_types
  use wave_types
  use wavelet_types
  use bound_types

  use bound_mod
  use timers_mod
#ifdef __CONVEY
  use cny_timers_mod
#endif

  implicit none

  logical inject_data
  logical, private :: apply_any_bound
  integer, private :: i_z_sou,i_z_rec,half_n_z_sou,half_n_z_rec,nop
!  integer*8, external :: cit_read
  integer*8 :: beg_timer,end_timer
  real, private :: c0,c1,c2
  type (Rev2D_fd_param_type), private :: Rev2D_fd_par_loc

#ifdef __CONVEY
integer :: cny_i_timer
!$cny coproc_mem(cny_i_timer)
#endif

contains

  subroutine init_Rev2D_fd_par(Rev2D_fd_par)
!$cny dual_target(0)
    type (Rev2D_fd_param_type) Rev2D_fd_par

    Rev2D_fd_par_loc=Rev2D_fd_par

  end subroutine init_Rev2D_fd_par

  function Rev2D_fd_init(gen_par,data_par,slow_par, &
		wave_par,wavelet_par,bound_par) result(info)
!$cny dual_target(0)

  type (genrev3d_info_type) info
  type (genrev3d_param_type) gen_par
  type (data_param_type) data_par
  type (slow_param_type) slow_par
  type (wave_param_type) wave_par
  type (wavelet_param_type) wavelet_par
  type (bound_param_type) bound_par
  
  integer ierr,i_z,n_z,i_thread


  info=info_default

! ATTENTION: Wrong need to be changed
  gen_par%max_courant_numb=sqrt(2.)/acos(-1.)

  apply_any_bound=gen_par%apply_bound

  inject_data=gen_par%inject_data
!    write(0,*)'inject_data=',inject_data

  wave_par%n_t_step=3

  c0 = -30./12.
  c1 = +16./12.
  c2 = -1./12.
  nop=2

  i_z_sou=wave_par%i_z_sou
  i_z_rec=wave_par%i_z_rec
  half_n_z_sou=(wavelet_par%n_z-1)/2
  half_n_z_rec=0 ! ATTENTION if this changes data_mod.f90 must change too!

!  if(gen_par%debug1) write(0,*)'SOU_WAVE: i_z_sou=',i_z_sou
!  if(gen_par%debug1) write(0,*)'SOU_WAVE: half_n_z_sou=',half_n_z_sou
!  if(gen_par%debug1) write(0,*)'REC_WAVE: i_z_rec=',i_z_rec
!  if(gen_par%debug1) write(0,*)'REC_WAVE: half_n_z_rec=',half_n_z_rec
!  if(gen_par%debug1) write(0,*)'bound nz=',bound_par%n_z_bound

  if(i_z_sou-half_n_z_sou < nop+1) then
    call seperr('STOP: you are placing source in top absorbing boundary')
  end if
  if (i_z_sou+half_n_z_sou > wave_par%n_z-nop) then
    call seperr('STOP: you are placing source in bottom absorbing boundary')
  end if

  if(i_z_rec-half_n_z_rec < nop+1) then
    call seperr('STOP: you are placing receivers in top absorbing boundary')
  end if
  if (i_z_rec+half_n_z_rec > wave_par%n_z-nop) then
   call seperr('STOP: you are placing receivers in bottom absorbing boundary')
  end if

  if(i_z_sou-half_n_z_sou < bound_par%n_z_bound+1) then
    call seperr('STOP: you are placing source in top damping boundary')
  end if
  if (i_z_sou+half_n_z_sou > wave_par%n_z-bound_par%n_z_bound) then
    call seperr('STOP: you are placing source in bottom damping boundary')
  end if

  if(i_z_rec-half_n_z_rec < bound_par%n_z_bound+1) then
    call seperr('STOP: you are placing receivers in top damping boundary')
  end if
  if (i_z_rec+half_n_z_rec > wave_par%n_z-bound_par%n_z_bound) then
   call seperr('STOP: you are placing receivers in bottom damping boundary')
  end if

  end function Rev2D_fd_init 



  function Rev2D_fd_extrap_oper(rev_extrap_par,wave_par,&
                        force_block,wave_block,slow_block) result(stat)
!$cny dual_target(1)
  integer stat
  type (rev_extrap_param_type) rev_extrap_par
  type (wave_param_type) wave_par

  integer omp_get_thread_num
  integer n_threads,i_thread
  
  logical add,adj,inject_data_loc
  integer n_z,n_mx,i_z,i_mx,ierr
  integer i_t0,i_tm1,i_tm2,i_z_force,half_n_z_force
  real vel_sq
  real d_t_sq

  real d_x_sq_reci,d_z_sq_reci
  real lap_tm1_x,lap_tm1_z,lap_tm1
  real bndz0,delp,vel,d_tz,d_tx,bndx0,bndx1
  integer iop

  optional slow_block
  real,  dimension (:,:,:,:)      :: force_block
  real,  dimension (:,:,:,:,:,:)  :: wave_block
  real,  dimension (:,:,:,:,:,:)  :: slow_block

!!$cny begin_coproc

!call cny$mput("#3", 2)

#ifdef __CONVEY
cny_i_timer=1
call CNY_timer_start(cny_i_timer)
#endif

!call cny$mput("#4", 2)

  adj=rev_extrap_par%adj
  add=rev_extrap_par%add

  n_threads=rev_extrap_par%n_threads

  d_t_sq=(rev_extrap_par%d_t)**2
  d_x_sq_reci=1.0/((wave_par%d_mx)**2)
  d_z_sq_reci=1.0/((wave_par%d_z)**2)
  d_tz=rev_extrap_par%d_t/wave_par%d_z
  d_tx=rev_extrap_par%d_t/wave_par%d_mx

  i_t0=wave_par%i_t_step(1)
  i_tm1=wave_par%i_t_step(2)
  i_tm2=wave_par%i_t_step(3)

  n_z=wave_par%n_z
  n_mx=wave_par%n_mx

  select case (rev_extrap_par%sou_rec_wave)
   case (SOU_WAVE)
     i_z_force=i_z_sou
     half_n_z_force=half_n_z_sou
     inject_data_loc=.false.
   case (REC_WAVE)
     i_z_force=i_z_rec
     half_n_z_force=half_n_z_rec
     inject_data_loc=inject_data
   case default
!     write(0,*)'sou_rec_wave=',rev_extrap_par%sou_rec_wave
#ifndef __CONVEY
     call seperr('Rev2D_fd_mod: Not supported sou_rec_wave')
#endif
  end select
  

!  !$omp parallel default (shared), &
!  !$omp& private(i_mx,i_z)
!  !$omp do
#ifdef __CONVEY
cny_i_timer=3
! loop #1
call CNY_timer_start(cny_i_timer)
#endif
  do i_mx=1,n_mx
    do i_z=1,n_z
       wave_block(i_z,i_mx,1,1,1,i_tm2)= wave_block(i_z,i_mx,1,1,1,i_tm1)
       wave_block(i_z,i_mx,1,1,1,i_tm1)= wave_block(i_z,i_mx,1,1,1,i_t0)
    end do
  end do
#ifdef __CONVEY
cny_i_timer=3
call CNY_timer_stop(cny_i_timer)
#endif
!!$omp end parallel

!  !$omp parallel default (shared), &
!  !$omp& private(i_mx,i_z,vel_sq,lap_tm1_x,lap_tm1_z,lap_tm1)
!  !$omp do
  do i_mx=1+nop,n_mx-nop
  !{
#ifdef __CONVEY
cny_i_timer=4
! loop #2
call CNY_timer_start(cny_i_timer)
#endif
    do i_z=1+nop,i_z_force-half_n_z_force-1
    !{
      vel_sq=1.0/(slow_block(i_z,i_mx,1,1,1,1)**2)

      lap_tm1_x=c0*wave_block(i_z,i_mx,1,1,1,i_tm1)+ &
        c1*(wave_block(i_z,i_mx-1,1,1,1,i_tm1) &
                              +wave_block(i_z,i_mx+1,1,1,1,i_tm1))&
       +c2*(wave_block(i_z,i_mx-2,1,1,1,i_tm1) &
                              +wave_block(i_z,i_mx+2,1,1,1,i_tm1))
      lap_tm1_x=lap_tm1_x*d_x_sq_reci
      lap_tm1_z=c0*wave_block(i_z,i_mx,1,1,1,i_tm1)+ &
                c1*(wave_block(i_z-1,i_mx,1,1,1,i_tm1) &
                            +wave_block(i_z+1,i_mx,1,1,1,i_tm1) )+ &
                c2*(wave_block(i_z-2,i_mx,1,1,1,i_tm1) &
                              +wave_block(i_z+2,i_mx,1,1,1,i_tm1))
      lap_tm1_z=lap_tm1_z*d_z_sq_reci
      lap_tm1=lap_tm1_x+lap_tm1_z

      wave_block(i_z,i_mx,1,1,1,i_t0)= 2*wave_block(i_z,i_mx,1,1,1,i_tm1)&
      - wave_block(i_z,i_mx,1,1,1,i_tm2)&
      + d_t_sq*vel_sq*lap_tm1     
    !}
    end do

#ifdef __CONVEY
cny_i_timer=4
call CNY_timer_stop(cny_i_timer)
#endif

#ifdef __CONVEY
cny_i_timer=5
! loop #3
call CNY_timer_start(cny_i_timer)
#endif

! Add forcing function to laplacian
    if(inject_data_loc) then
    !{
      do i_z=i_z_force-half_n_z_force,i_z_force+half_n_z_force
      !{
        wave_block(i_z,i_mx,1,1,1,i_t0)= &
                    force_block(i_z-i_z_force+half_n_z_force+1,i_mx,1,1)
      !}
      end do
    !}
    else 
    !{
      do i_z=i_z_force-half_n_z_force,i_z_force+half_n_z_force
      !{
        vel_sq=1./(slow_block(i_z,i_mx,1,1,1,1)**2)

        lap_tm1_x=c0*wave_block(i_z,i_mx,1,1,1,i_tm1)+ &
          c1*(wave_block(i_z,i_mx-1,1,1,1,i_tm1) &
                                +wave_block(i_z,i_mx+1,1,1,1,i_tm1))&
         +c2*(wave_block(i_z,i_mx-2,1,1,1,i_tm1) &
                                +wave_block(i_z,i_mx+2,1,1,1,i_tm1))
        lap_tm1_x=lap_tm1_x*d_x_sq_reci
        lap_tm1_z=c0*wave_block(i_z,i_mx,1,1,1,i_tm1)+ &
                  c1*(wave_block(i_z-1,i_mx,1,1,1,i_tm1) &
                              +wave_block(i_z+1,i_mx,1,1,1,i_tm1) )+ &
                  c2*(wave_block(i_z-2,i_mx,1,1,1,i_tm1) &
                                +wave_block(i_z+2,i_mx,1,1,1,i_tm1))
        lap_tm1_z=lap_tm1_z*d_z_sq_reci
        lap_tm1=lap_tm1_x+lap_tm1_z

        wave_block(i_z,i_mx,1,1,1,i_t0)= 2*wave_block(i_z,i_mx,1,1,1,i_tm1)&
        - wave_block(i_z,i_mx,1,1,1,i_tm2)&
        + d_t_sq*(vel_sq*lap_tm1 + &     
               force_block(i_z-i_z_force+half_n_z_force+1,i_mx,1,1))
      !}
      end do
  
    !}
    end if

! Finished adding forcing function

#ifdef __CONVEY
cny_i_timer=5
call CNY_timer_stop(cny_i_timer)
#endif

!$cny begin_coproc
#ifdef __CONVEY
cny_i_timer=6
! loop #4
call CNY_timer_start(cny_i_timer)
#endif
! after inject layer
!    do i_z=i_z_force+half_n_z_force+1,n_z-nop
!    !{
!      vel_sq=1.0/(slow_block(i_z,i_mx,1,1,1,1)**2)
!
!      lap_tm1_x=c0*wave_block(i_tm1,i_z,i_mx,1,1,1)+ &
!        c1*(wave_block(i_tm1,i_z,i_mx-1,1,1,1) &
!                              +wave_block(i_tm1,i_z,i_mx+1,1,1,1))&
!       +c2*(wave_block(i_tm1,i_z,i_mx-2,1,1,1) &
!                              +wave_block(i_tm1,i_z,i_mx+2,1,1,1))
!      lap_tm1_x=lap_tm1_x*d_x_sq_reci
!      lap_tm1_z=c0*wave_block(i_tm1,i_z,i_mx,1,1,1)+ &
!                c1*(wave_block(i_tm1,i_z-1,i_mx,1,1,1) &
!                            +wave_block(i_tm1,i_z+1,i_mx,1,1,1) )+ &
!                c2*(wave_block(i_tm1,i_z-2,i_mx,1,1,1) &
!                              +wave_block(i_tm1,i_z+2,i_mx,1,1,1))
!      lap_tm1_z=lap_tm1_z*d_z_sq_reci
!      lap_tm1=lap_tm1_x+lap_tm1_z
!
!      wave_block(i_t0,i_z,i_mx,1,1,1)= 2*wave_block(i_tm1,i_z,i_mx,1,1,1)&
!      - wave_block(i_tm2,i_z,i_mx,1,1,1)&
!      + d_t_sq*vel_sq*lap_tm1     
!    !}
!    end do

    do i_z=i_z_force+half_n_z_force+1,n_z-nop
    !{
      vel_sq=1.0/(slow_block(i_z,i_mx,1,1,1,1)**2)

      lap_tm1_x=c0*wave_block(i_z,i_mx,1,1,1,2)+ &
        c1*(wave_block(i_z,i_mx-1,1,1,1,2) &
                              +wave_block(i_z,i_mx+1,1,1,1,2))&
       +c2*(wave_block(i_z,i_mx-2,1,1,1,2) &
                              +wave_block(i_z,i_mx+2,1,1,1,2))
      lap_tm1_x=lap_tm1_x*d_x_sq_reci
      lap_tm1_z=c0*wave_block(i_z,i_mx,1,1,1,2)+ &
                c1*(wave_block(i_z-1,i_mx,1,1,1,2) &
                            +wave_block(i_z+1,i_mx,1,1,1,2) )+ &
                c2*(wave_block(i_z-2,i_mx,1,1,1,2) &
                              +wave_block(i_z+2,i_mx,1,1,1,2))
      lap_tm1_z=lap_tm1_z*d_z_sq_reci
      lap_tm1=lap_tm1_x+lap_tm1_z

      wave_block(i_z,i_mx,1,1,1,1)= 2*wave_block(i_z,i_mx,1,1,1,2)&
      - wave_block(i_z,i_mx,1,1,1,3)&
      + d_t_sq*vel_sq*lap_tm1     
    !}
    end do

#ifdef __CONVEY
cny_i_timer=6
call CNY_timer_stop(cny_i_timer)
#endif
!$cny end_coproc
  !}
  end do
!!$omp end parallel

#ifdef __CONVEY
cny_i_timer=7
! loop #5
call CNY_timer_start(cny_i_timer)
#endif
!! Not easy to parallelize
!!!  !$omp parallel default (shared), &
!!!  !$omp& private(iop,i_mx,i_z,vel,delp,bndz0,bndx0,bndx1)
!!!  !$omp do
  do iop=nop,1,-1
  !{
    do i_mx=1,n_mx
    !{
!        vel=1.0/slow_block(iop,i_mx,1,1,1,1)
!        delp=vel*d_tz
        delp=d_tz/slow_block(iop,i_mx,1,1,1,1)
        bndz0= (1.-delp)/(1.+delp) 
        wave_block(iop,i_mx,1,1,1,i_t0)= &
                    wave_block(iop+1,i_mx,1,1,1,i_tm1)+ &
             bndz0*(wave_block(iop,i_mx,1,1,1,i_tm1)- &
              wave_block(iop+1,i_mx,1,1,1,i_t0) )  
    !}
    end do
    do i_mx=1,n_mx
    !{
!        vel=1.0/slow_block(n_z-iop+1,i_mx,1,1,1,1)
!        delp=vel*d_tz
        delp=d_tz/slow_block(n_z-iop+1,i_mx,1,1,1,1)
        bndz0= (1.-delp)/(1.+delp) 
        wave_block(n_z-iop+1,i_mx,1,1,1,i_t0)= &
                    wave_block(n_z-iop,i_mx,1,1,1,i_tm1)+ &
             bndz0*(wave_block(n_z-iop+1,i_mx,1,1,1,i_tm1)- &
              wave_block(n_z-iop,i_mx,1,1,1,i_t0) )  
    !}
    end do
     
    do i_z=1,n_z
    !{
!        vel=1.0/slow_block(i_z,iop,1,1,1,1)
!        delp=vel*d_tx
        delp=d_tx/slow_block(i_z,iop,1,1,1,1)
        bndx0= (1.-delp)/(1.+delp)
        wave_block(i_z,iop,1,1,1,i_t0)=&
           wave_block(i_z,iop+1,1,1,1,i_tm1)+ &
           bndx0*(wave_block(i_z,iop,1,1,1,i_tm1)- &     
                wave_block(i_z,iop+1,1,1,1,i_t0) )     
    !}
    end do

    do i_z=1,n_z
    !{
!        vel=1.0/slow_block(i_z,n_mx-iop+1,1,1,1,1)
!        delp = vel*d_tx
        delp=d_tx/slow_block(i_z,n_mx-iop+1,1,1,1,1)
        bndx1= (1.-delp)/(1.+delp) 
        wave_block(i_z,n_mx-iop+1,1,1,1,i_t0)=&
            wave_block(i_z,n_mx-iop,1,1,1,i_tm1)+ &   
            bndx1*(wave_block(i_z,n_mx-iop+1,1,1,1,i_tm1)- &
                  wave_block(i_z,n_mx-iop,1,1,1,i_t0) )  
    !}
    end do

  !}
  end do
!!!$omp end parallel
#ifdef __CONVEY
cny_i_timer=7
call CNY_timer_stop(cny_i_timer)
#endif


!!####################### end absorbing boundary condition ####
  
!@@@@@@@@@@@@@@@@@@@@@@ source force @@@@@@@@@@@@@@@@@@@@@@@@@
!  if(inject_data_loc) then
!!  !$omp parallel default (shared), &
!!  !$omp& private(i_mx)
!!  !$omp do
!    do i_mx=1,n_mx
!      wave_block(i_t0,i_z_force,i_mx,1,1,1)= &
!                  force_block(i_mx,1,1,1)
!    end do
!!  !$omp end parallel
!  else 
!!  !$omp parallel default (shared), &
!!  !$omp& private(i_mx)
!!  !$omp do
!    do i_mx=1,n_mx
!       wave_block(i_t0,i_z_force,i_mx,1,1,1)=&
!           wave_block(i_t0,i_z_force,i_mx,1,1,1) &
!           +d_t_sq*force_block(i_mx,1,1,1)
!    end do
!!  !$omp end parallel
!  end if
!
!   if(apply_any_bound) then
!!write(0,*)'Apply damping boundaries'
!!     call apply_bound(wave_par,wave_block(i_t0,:,:,:,:,:))
!   end if
!
!
!!@@@@@@@@@@@@@@@@@@@@@@ end source force@@@@@@@@@@@@@@@@@@@@@@

#ifdef __CONVEY
cny_i_timer=1
call CNY_timer_stop(cny_i_timer)
#endif

!!$cny end_coproc

!!###################### absorbing boundary condition #######
  stat=0
  return
  end function Rev2D_fd_extrap_oper

  function Rev2D_fd_scatter_oper( &
    rev_extrap_par, &
    wave_par, &
    wave_block,wave_back_block, &
    slow_block,dslow_block) &
    result(stat)
!$cny dual_target(0)

    integer stat
    type (rev_extrap_param_type) rev_extrap_par
    type (wave_param_type) wave_par

    integer omp_get_thread_num

    logical add,adj
    integer n_z,n_mx,i_z,i_mx,i_kmx
    integer i_t0
    integer n_threads,i_thread
    real d_t_sq,s_back,delta_s

    optional slow_block

    real,  dimension (:,:,:,:,:,:)  :: wave_block
    real,  dimension (:,:,:,:,:,:)  :: wave_back_block
    real,  dimension (:,:,:,:,:,:)  :: slow_block,dslow_block


    adj=rev_extrap_par%adj
    add=rev_extrap_par%add

    n_threads=rev_extrap_par%n_threads

    i_t0=wave_par%i_t_step(1)
    n_z=wave_par%n_z
    n_mx=wave_par%n_mx

    d_t_sq=(rev_extrap_par%d_t)**2

    if(.not. adj ) then
    !{
!!$omp parallel default (private) ,shared (wave_block,wave_back_block,slow_block,dslow_block),&
!!$omp& shared(n_z,n_mx,n_threads,i_t0)
!!$omp do
      do i_thread=1,n_threads
      !{
        do i_mx=i_thread,n_mx,n_threads
        !{
          do i_z=1,n_z
          !{
            s_back=slow_block(i_z,i_mx,1,1,1,1)
            delta_s=dslow_block(i_z,i_mx,1,1,1,1)
            wave_block(i_z,i_mx,1,1,1,i_t0) = &
              wave_block(i_z,i_mx,1,1,1,i_t0) - &
              2.*s_back*delta_s*wave_back_block(1,i_z,i_mx,1,1,1)
          !}
          end do
        !}
        end do 
      !}
      end do 
!!$omp end parallel
    !}
    else
    !{

!!$omp parallel default (private) ,shared (wave_block,wave_back_block,slow_block,dslow_block),&
!!$omp& shared(n_z,n_mx,n_threads,i_t0)
!!$omp do
      do i_thread=1,n_threads
      !{
        do i_mx=i_thread,n_mx,n_threads
        !{
          do i_z=1,n_z
          !{
            s_back=slow_block(i_z,i_mx,1,1,1,1)
            dslow_block(i_z,i_mx,1,1,1,1)= &
              dslow_block(i_z,i_mx,1,1,1,1) - &
                2.*s_back*wave_back_block(1,i_z,i_mx,1,1,1)* &
                wave_block(i_z,i_mx,1,1,1,i_t0)
          !}
          end do
        !}
        end do 
      !}
      end do 
!!$omp end parallel
    !}
    end if

    stat=0
    return	
  end function Rev2D_fd_scatter_oper

  function Rev2D_fd_der2time_oper (&
    rev_extrap_par, &
    wave_par, &
    wave_block, &
    wave_back_block) &
    result(stat)
!$cny dual_target(0)

    integer stat
    type (rev_extrap_param_type) rev_extrap_par
    type (wave_param_type) wave_par

    integer omp_get_thread_num

    logical add,adj
    integer n_z,n_mx,i_z,i_mx
    integer i_t0,i_tm1,i_tm2
    integer n_threads,i_thread

    real,  dimension (:,:,:,:,:,:)  :: wave_block
    real,  dimension (:,:,:,:,:,:)  :: wave_back_block

    adj=rev_extrap_par%adj
    add=rev_extrap_par%add

    n_threads=rev_extrap_par%n_threads

    i_t0=wave_par%i_t_step(1)
    i_tm1=wave_par%i_t_step(2)
    i_tm2=wave_par%i_t_step(3)

    n_z=wave_par%n_z
    n_mx=wave_par%n_mx

! second derivative in time 

!!$omp parallel default (private) ,shared (wave_block,wave_back_block),&
!!$omp& shared(n_z,n_mx,n_threads,i_t0,i_tm1,i_tm2)
!!$omp do
    do i_thread=1,n_threads
    !{
      do i_mx=i_thread,n_mx,n_threads
      !{
        do i_z=1,n_z
        !{
          wave_back_block(1,i_z,i_mx,1,1,1)= &
              wave_block(i_z,i_mx,1,1,1,i_t0) - &
            2*wave_block(i_z,i_mx,1,1,1,i_tm1) + &
              wave_block(i_z,i_mx,1,1,1,i_tm2) 
        !}
        end do
      !}
      end do 
    !}
    end do 
!!$omp end parallel

    stat=0
    return	
  end function Rev2D_fd_der2time_oper


  function Rev2D_fd_clean()  result(stat)
    integer stat

    integer ierr

    stat=0

!    write(0,*)' Rev2D_fd_clean: Done with cleaning'

  end function Rev2D_fd_clean


end module Rev2D_fd_mod
