module Rev2D_ho_fd_mod

  use Rev2D_ho_fd_types

  use genrev3d_types
  use rev_extrap_types
  use data_types
  use slow_types
  use wave_types
  use wavelet_types
  use bound_types

  use bound_mod
  use timers_mod
#ifdef __CONVEY
  use cny_timers_mod
#endif

  implicit none

abstract interface
       subroutine spat_derivatives(n_z,n_mx,nop,wave_slice,coeff,lap,fourth_der) 
!                  result(spat_der)
!         use Rev2D_ho_fd_types, only: Rev2D_spat_der_type
!         type (Rev2D_spat_der_type) :: spat_der
!         real :: spat_der

         integer n_z,n_mx,nop
         real,  dimension (0:nop)  :: coeff
         real,  dimension (:,:,:,:,:)  :: wave_slice
         real,  dimension (:,:,:,:,:)  :: lap,fourth_der

       end  subroutine spat_derivatives
end interface


  procedure (spat_derivatives), pointer :: spat_der_oper

  logical inject_data,fourth_time
  logical, private :: apply_any_bound
  integer, private :: i_z_sou,i_z_rec,half_n_z_sou,half_n_z_rec,nop
  integer, private :: i_hx_0_dslow
  real, private :: d_x_sq_recip,d_z_sq_recip,d_t_sq,d_t_fourth_12,d_tz,d_tx
  real, private :: inv_d_t_sq
  type (Rev2D_ho_fd_param_type), private :: Rev2D_ho_fd_par_loc

  real, private, allocatable,  dimension (:)      :: coeff


#ifdef __CONVEY
integer :: cny_i_timer
!$cny coproc_mem(cny_i_timer)
#endif

contains

  subroutine init_Rev2D_ho_fd_par(Rev2D_ho_fd_par)
!$cny dual_target(0)
    type (Rev2D_ho_fd_param_type) Rev2D_ho_fd_par

    Rev2D_ho_fd_par_loc=Rev2D_ho_fd_par

  end subroutine init_Rev2D_ho_fd_par

  function Rev2D_ho_fd_init(gen_par,data_par,slow_par, &
                            rev_extrap_par,wave_par,wavelet_par,bound_par) &
                            result(info)
!$cny dual_target(0)

  type (genrev3d_info_type) info
  type (genrev3d_param_type) gen_par
  type (data_param_type) data_par
  type (slow_param_type) slow_par
  type (rev_extrap_param_type) rev_extrap_par
  type (wave_param_type) wave_par
  type (wavelet_param_type) wavelet_par
  type (bound_param_type) bound_par
  
  integer ierr,i_z,n_z,i_thread


  info=info_default

! From Daublain page 58

  apply_any_bound=gen_par%apply_bound

  inject_data=gen_par%inject_data
    write(0,*)'inject_data=',inject_data

  wave_par%n_t_step=3

  d_t_sq=(rev_extrap_par%d_t)**2
  inv_d_t_sq=1./d_t_sq

  d_t_fourth_12=((rev_extrap_par%d_t)**4)/12
  d_x_sq_recip=1.0/((wave_par%d_mx)**2)
  d_z_sq_recip=1.0/((wave_par%d_z)**2)
  d_tz=rev_extrap_par%d_t/wave_par%d_z
  d_tx=rev_extrap_par%d_t/wave_par%d_mx

!write(0,*)'d_t_sq=',d_t_sq
!write(0,*)'d_x_sq_recip=',d_x_sq_recip
!write(0,*)'d_z_sq_recip=',d_z_sq_recip
!write(0,*)'d_tz=',d_tz
!write(0,*)'d_tx=',d_tx

  select case (Rev2D_ho_fd_par_loc%fd_time_order)
    case (2)
      fourth_time=.false.
    case (4)
      fourth_time=.true.
    case default
#ifndef __CONVEY
     call seperr('Rev2D_ho_fd_mod: Not supported fd_time_order')
#endif
  end select

  select case (Rev2D_ho_fd_par_loc%fd_space_order)
    case (2)

    if(fourth_time) then
      call seperr("Fourth in time and second in space is not implemented")
    end if

    spat_der_oper => laplace_second
    gen_par%max_courant_numb=sqrt(3./4.)*sqrt(2.)/2

    nop=1
    allocate (coeff(0:nop), stat=ierr)
    if(gen_par%debug1) write(0,*) 'coeff shape=',shape(coeff),ierr
    if(ierr .ne. 0) call seperr("Problems with allocation of coeff")

    coeff(0)=-2.
    coeff(1)=1
!    c0 = -30./12.
!    c1 = +16./12.

!#ifndef __CONVEY
!     call seperr('Rev2D_ho_fd_mod: Not supported yet fd_space_order=2')
!#endif

   case (4)

    if(fourth_time) then
      call seperr("Fourth in time and space is not implemented")
    end if

    spat_der_oper => laplace_fourth

    gen_par%max_courant_numb=sqrt(3./4.)*sqrt(2.)/2

    nop=2
    allocate (coeff(0:nop), stat=ierr)
    if(gen_par%debug1) write(0,*) 'coeff shape=',shape(coeff),ierr
    if(ierr .ne. 0) call seperr("Problems with allocation of coeff")

    coeff(0)=-30./12.
    coeff(1)=+16./12.
    coeff(2)=-1./12.
!    c0 = -30./12.
!    c1 = +16./12.
!    c2 = -1./12.


   case (10)

    gen_par%max_courant_numb=sqrt(3./4.)*sqrt(2.)/2

    if(fourth_time) then
      spat_der_oper => forth_order_time
! See Dbalain page 65
      gen_par%max_courant_numb=.5
    else
      spat_der_oper => laplace_tenth
    end if

    nop=5

    allocate (coeff(0:nop), stat=ierr)
    if(gen_par%debug1) write(0,*) 'coeff shape=',shape(coeff),ierr
    if(ierr .ne. 0) call seperr("Problems with allocation of coeff")

    coeff(0)=-5.8544444444/2.
    coeff(1)=+3.3333333333/2.
    coeff(2)=-0.4761904762/2.
    coeff(3)=+0.0793650794/2.
    coeff(4)=-0.0099206349/2.
    coeff(5)=+0.0006349206/2.

   case default
#ifndef __CONVEY
     call seperr('Rev2D_ho_fd_mod: Not supported fd_space_order')
#endif
  end select

  i_z_sou=wave_par%i_z_sou
  i_z_rec=wave_par%i_z_rec
  half_n_z_sou=(wavelet_par%n_z-1)/2
  half_n_z_rec=0 ! ATTENTION if this changes data_mod.f90 must change too!

!  if(gen_par%debug1) write(0,*)'SOU_WAVE: i_z_sou=',i_z_sou
!  if(gen_par%debug1) write(0,*)'SOU_WAVE: half_n_z_sou=',half_n_z_sou
!  if(gen_par%debug1) write(0,*)'REC_WAVE: i_z_rec=',i_z_rec
!  if(gen_par%debug1) write(0,*)'REC_WAVE: half_n_z_rec=',half_n_z_rec
!  if(gen_par%debug1) write(0,*)'bound nz=',bound_par%n_z_bound

!  if(gen_par%image_zero_offset) then
!  !{
!    i_hx_0_dslow=0
!  !}
!  else
!  !{
    i_hx_0_dslow=nint(-slow_par%hx_0/slow_par%d_hx)+1
!  !}
!  end if
  if(gen_par%debug2) write(0,*)'(Scatter) i_hx_0_dslow=',i_hx_0_dslow



  if(i_z_sou-half_n_z_sou < nop+1) then
    call seperr('STOP: you are placing source in top absorbing boundary')
  end if
  if (i_z_sou+half_n_z_sou > wave_par%n_z-nop) then
    call seperr('STOP: you are placing source in bottom absorbing boundary')
  end if

  if(i_z_rec-half_n_z_rec < nop+1) then
    call seperr('STOP: you are placing receivers in top absorbing boundary')
  end if
  if (i_z_rec+half_n_z_rec > wave_par%n_z-nop) then
   call seperr('STOP: you are placing receivers in bottom absorbing boundary')
  end if

  if(i_z_sou-half_n_z_sou < bound_par%n_z_bound+1) then
    call seperr('STOP: you are placing source in top damping boundary')
  end if
  if (i_z_sou+half_n_z_sou > wave_par%n_z-bound_par%n_z_bound) then
    call seperr('STOP: you are placing source in bottom damping boundary')
  end if

  if(i_z_rec-half_n_z_rec < bound_par%n_z_bound+1) then
    call seperr('STOP: you are placing receivers in top damping boundary')
  end if
  if (i_z_rec+half_n_z_rec > wave_par%n_z-bound_par%n_z_bound) then
   call seperr('STOP: you are placing receivers in bottom damping boundary')
  end if

  end function Rev2D_ho_fd_init 



  function Rev2D_ho_fd_extrap_oper(rev_extrap_par,wave_par,&
                        force_block,wave_block,slow_block) result(stat)
!$cny dual_target(1)
  integer stat
  type (rev_extrap_param_type) rev_extrap_par
  type (wave_param_type) wave_par

!  integer omp_get_thread_num
  integer n_threads,i_thread
  
  logical add,adj,inject_data_loc
  integer n_z,n_mx,i_z,i_mx,ierr
  integer i_t0,i_tm1,i_tm2,i_z_force,half_n_z_force
  real vel_sq

  integer iop
!  real lap_tm1_x,lap_tm1_z,lap_tm1
  real bndz0,delp,vel,bndx0,bndx1
!  type (Rev2D_spat_der_type) :: spat_der
!  real spat_der

  optional slow_block
  real,  dimension (:,:,:,:)      :: force_block
  real, target,  dimension (:,:,:,:,:,:)  :: wave_block
  real,  dimension (:,:,:,:,:,:)  :: slow_block

  real, allocatable,  dimension (:,:,:,:,:)  :: lap,fourth_der

  real, pointer, dimension (:,:,:,:,:)      :: wave_t0,wave_tm1,wave_tm2

  allocate (lap(wave_par%n_z,wave_par%n_mx,1,1,1), stat=ierr)
!  if(gen_par%debug1)  &
!      write(0,*) 'shape(lap)=',shape(lap),ierr
  if(ierr .ne. 0) call seperr("Problems with allocation of lap")

  allocate (fourth_der(wave_par%n_z,wave_par%n_mx,1,1,1), stat=ierr)
!  if(gen_par%debug1)  &
!      write(0,*) 'shape(fourth_der)=',shape(fourth_der),ierr
  if(ierr .ne. 0) call seperr("Problems with allocation of fourth_der")

!!$cny begin_coproc

!call cny$mput("#3", 2)

#ifdef __CONVEY
cny_i_timer=1
call CNY_timer_start(cny_i_timer)
#endif

!call cny$mput("#4", 2)

  adj=rev_extrap_par%adj
  add=rev_extrap_par%add

  n_threads=rev_extrap_par%n_threads

!write(0,*)'d_t_sq=',d_t_sq
!write(0,*)'d_x_sq_recip=',d_x_sq_recip
!write(0,*)'d_z_sq_recip=',d_z_sq_recip
!write(0,*)'d_tz=',d_tz
!write(0,*)'d_tx=',d_tx

!  d_t_sq=(rev_extrap_par%d_t)**2
!  d_x_sq_recip=1.0/((wave_par%d_mx)**2)
!  d_z_sq_recip=1.0/((wave_par%d_z)**2)
!  d_tz=rev_extrap_par%d_t/wave_par%d_z
!  d_tx=rev_extrap_par%d_t/wave_par%d_mx

!write(0,*)'before i_t_step=',wave_par%i_t_step

  i_tm2=wave_par%i_t_step(3)
  wave_par%i_t_step(3)=wave_par%i_t_step(2)
  wave_par%i_t_step(2)=wave_par%i_t_step(1)
  wave_par%i_t_step(1)=i_tm2

  i_t0=wave_par%i_t_step(1)
  i_tm1=wave_par%i_t_step(2)
  i_tm2=wave_par%i_t_step(3)

  wave_t0 => wave_block(:,:,:,:,:,i_t0)
  wave_tm1 => wave_block(:,:,:,:,:,i_tm1)
  wave_tm2 => wave_block(:,:,:,:,:,i_tm2)
!
!       wave_block(i_z,i_mx,1,1,1,i_tm2)= wave_block(i_z,i_mx,1,1,1,i_tm1)
!       wave_block(i_z,i_mx,1,1,1,i_tm1)= wave_block(i_z,i_mx,1,1,1,i_t0)

  n_z=wave_par%n_z
  n_mx=wave_par%n_mx

  select case (rev_extrap_par%sou_rec_wave)
   case (SOU_WAVE)
     i_z_force=i_z_sou
     half_n_z_force=half_n_z_sou
     inject_data_loc=.false.
   case (REC_WAVE)
     i_z_force=i_z_rec
     half_n_z_force=half_n_z_rec
     inject_data_loc=inject_data
   case default
!     write(0,*)'sou_rec_wave=',rev_extrap_par%sou_rec_wave
#ifndef __CONVEY
     call seperr('Rev2D_ho_fd_mod: Not supported sou_rec_wave')
#endif
  end select
  

!#ifdef __CONVEY
!cny_i_timer=3
!! loop #1
!call CNY_timer_start(cny_i_timer)
!#else
!  !$omp parallel default (shared), &
!  !$omp& private(i_mx,i_z)
!  !$omp do
!#endif
!
!!  do i_mx=1,n_mx
!!    do i_z=1,n_z
!!       wave_block(i_z,i_mx,1,1,1,i_tm2)= wave_block(i_z,i_mx,1,1,1,i_tm1)
!!       wave_block(i_z,i_mx,1,1,1,i_tm1)= wave_block(i_z,i_mx,1,1,1,i_t0)
!!    end do
!!  end do
!
!
!#ifdef __CONVEY
!cny_i_timer=3
!call CNY_timer_stop(cny_i_timer)
!#endif
!
!!$omp end parallel

!write(0,*)'#31'
!pause 

  call spat_der_oper(n_z,n_mx,nop,wave_tm1,coeff,lap,fourth_der) 

!write(0,*)'#32'
!pause 

#ifndef __CONVEY
!$omp parallel default (shared), &
!$omp& private(i_mx,i_z,vel_sq)
!$omp do
#endif
  do i_mx=1+nop,n_mx-nop
  !{
    do i_z=1+nop,i_z_force-half_n_z_force-1
    !{
      vel_sq=1.0/(slow_block(i_z,i_mx,1,1,1,1)**2)
      wave_t0(i_z,i_mx,1,1,1)= 2*wave_tm1(i_z,i_mx,1,1,1)&
      - wave_tm2(i_z,i_mx,1,1,1)&
      + d_t_sq*vel_sq*lap(i_z,i_mx,1,1,1) &     
      + d_t_fourth_12*(vel_sq*vel_sq)*fourth_der(i_z,i_mx,1,1,1)     

    !}
    end do

!    write(0,*)'inject_data_loc=',inject_data_loc
! Add forcing function to laplacian
    if(inject_data_loc) then
    !{
      do i_z=i_z_force-half_n_z_force,i_z_force+half_n_z_force
      !{
        wave_block(i_z,i_mx,1,1,1,i_t0)= &
                    force_block(i_z-i_z_force+half_n_z_force+1,i_mx,1,1)
      !}
      end do
    !}
    else 
    !{
      do i_z=i_z_force-half_n_z_force,i_z_force+half_n_z_force
      !{

      vel_sq=1./(slow_block(i_z,i_mx,1,1,1,1)**2)

      wave_t0(i_z,i_mx,1,1,1)= 2*wave_tm1(i_z,i_mx,1,1,1)&
      - wave_tm2(i_z,i_mx,1,1,1)&
      + d_t_sq*(vel_sq*lap(i_z,i_mx,1,1,1) + &     
               force_block(i_z-i_z_force+half_n_z_force+1,i_mx,1,1)) &
      + d_t_fourth_12*(vel_sq*vel_sq)*fourth_der(i_z,i_mx,1,1,1)     
      !}
      end do
  
    !}
    end if

! Finished adding forcing function

! after inject layer
    do i_z=i_z_force+half_n_z_force+1,n_z-nop
    !{
      vel_sq=1.0/(slow_block(i_z,i_mx,1,1,1,1)**2)

      wave_t0(i_z,i_mx,1,1,1)= 2*wave_tm1(i_z,i_mx,1,1,1)&
      - wave_tm2(i_z,i_mx,1,1,1)&
      + d_t_sq*vel_sq*lap(i_z,i_mx,1,1,1) &
      + d_t_fourth_12*(vel_sq*vel_sq)*fourth_der(i_z,i_mx,1,1,1)     

    !}
    end do

#ifdef __CONVEY
cny_i_timer=6
call CNY_timer_stop(cny_i_timer)
#endif
!$cny end_coproc
  !}
  end do
#ifndef __CONVEY
!$omp end parallel
#endif

!write(0,*)'#33'
!pause 

#ifdef __CONVEY
cny_i_timer=7
! loop #5
call CNY_timer_start(cny_i_timer)
#endif
!! Not easy to parallelize
!!!  !$omp parallel default (shared), &
!!!  !$omp& private(iop,i_mx,i_z,vel,delp,bndz0,bndx0,bndx1)
!!!  !$omp do
!      write(0,*) '#4'
  do iop=nop,1,-1
  !{
    do i_mx=1,n_mx
    !{
!        vel=1.0/slow_block(iop,i_mx,1,1,1,1)
!        delp=vel*d_tz
        delp=d_tz/slow_block(iop,i_mx,1,1,1,1)
        bndz0= (1.-delp)/(1.+delp) 
        wave_block(iop,i_mx,1,1,1,i_t0)= &
                    wave_block(iop+1,i_mx,1,1,1,i_tm1)+ &
             bndz0*(wave_block(iop,i_mx,1,1,1,i_tm1)- &
              wave_block(iop+1,i_mx,1,1,1,i_t0) )  
    !}
    end do
    do i_mx=1,n_mx
    !{
!        vel=1.0/slow_block(n_z-iop+1,i_mx,1,1,1,1)
!        delp=vel*d_tz
        delp=d_tz/slow_block(n_z-iop+1,i_mx,1,1,1,1)
        bndz0= (1.-delp)/(1.+delp) 
        wave_block(n_z-iop+1,i_mx,1,1,1,i_t0)= &
                    wave_block(n_z-iop,i_mx,1,1,1,i_tm1)+ &
             bndz0*(wave_block(n_z-iop+1,i_mx,1,1,1,i_tm1)- &
              wave_block(n_z-iop,i_mx,1,1,1,i_t0) )  
    !}
    end do
     
    do i_z=1,n_z
    !{
!        vel=1.0/slow_block(i_z,iop,1,1,1,1)
!        delp=vel*d_tx
        delp=d_tx/slow_block(i_z,iop,1,1,1,1)
        bndx0= (1.-delp)/(1.+delp)
        wave_block(i_z,iop,1,1,1,i_t0)=&
           wave_block(i_z,iop+1,1,1,1,i_tm1)+ &
           bndx0*(wave_block(i_z,iop,1,1,1,i_tm1)- &     
                wave_block(i_z,iop+1,1,1,1,i_t0) )     
    !}
    end do

    do i_z=1,n_z
    !{
!        vel=1.0/slow_block(i_z,n_mx-iop+1,1,1,1,1)
!        delp = vel*d_tx
        delp=d_tx/slow_block(i_z,n_mx-iop+1,1,1,1,1)
        bndx1= (1.-delp)/(1.+delp) 
        wave_block(i_z,n_mx-iop+1,1,1,1,i_t0)=&
            wave_block(i_z,n_mx-iop,1,1,1,i_tm1)+ &   
            bndx1*(wave_block(i_z,n_mx-iop+1,1,1,1,i_tm1)- &
                  wave_block(i_z,n_mx-iop,1,1,1,i_t0) )  
    !}
    end do

  !}
  end do
!!!$omp end parallel
#ifdef __CONVEY
cny_i_timer=7
call CNY_timer_stop(cny_i_timer)
#endif

!write(0,*)'#34'
!pause 

!!####################### end absorbing boundary condition ####
  
!@@@@@@@@@@@@@@@@@@@@@@ source force @@@@@@@@@@@@@@@@@@@@@@@@@
!  if(inject_data_loc) then
!!  !$omp parallel default (shared), &
!!  !$omp& private(i_mx)
!!  !$omp do
!    do i_mx=1,n_mx
!      wave_block(i_t0,i_z_force,i_mx,1,1,1)= &
!                  force_block(i_mx,1,1,1)
!    end do
!!  !$omp end parallel
!  else 
!!  !$omp parallel default (shared), &
!!  !$omp& private(i_mx)
!!  !$omp do
!    do i_mx=1,n_mx
!       wave_block(i_t0,i_z_force,i_mx,1,1,1)=&
!           wave_block(i_t0,i_z_force,i_mx,1,1,1) &
!           +d_t_sq*force_block(i_mx,1,1,1)
!    end do
!!  !$omp end parallel
!  end if
!
   if(apply_any_bound) then
!write(0,*)'Apply damping boundaries'
     call apply_bound(wave_par,wave_block(:,:,:,:,:,i_t0))
   end if
!
!
!!@@@@@@@@@@@@@@@@@@@@@@ end source force@@@@@@@@@@@@@@@@@@@@@@

!write(0,*)'#35'
!pause 

#ifdef __CONVEY
cny_i_timer=1
call CNY_timer_stop(cny_i_timer)
#endif

    if(allocated(lap)) then
       deallocate (lap,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating lap')
    end if
    if(allocated(fourth_der)) then
       deallocate (fourth_der,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating fourth_der')
    end if

!write(0,*)'#36'
!pause 

!!$cny end_coproc

!!###################### absorbing boundary condition #######
  stat=0
  return
  end function Rev2D_ho_fd_extrap_oper

  function Rev2D_ho_fd_scatter_oper( &
    rev_extrap_par, &
    wave_par, &
    slow_par, &
    wave_block,wave_back_block, &
    slow_block,dslow_block) &
    result(stat)
!$cny dual_target(0)

    integer stat
    type (rev_extrap_param_type) rev_extrap_par
    type (wave_param_type) wave_par
    type (slow_param_type) slow_par

!    integer omp_get_thread_num

    logical add,adj
    integer n_z,n_mx,i_z,i_mx,i_kmx,i_hx
    integer i_t0
!    integer :: i_mx_image,i_mx_wave
!    integer :: i_mx_0_image,i_my_0_image
    integer :: i_mx_wave_beg,i_mx_wave_end
    integer :: i_hx_shift_sou,i_hx_shift_rec
    integer n_threads,i_thread
    real s_back,delta_s

    optional slow_block

    real,  dimension (:,:,:,:,:,:)  :: wave_block
    real,  dimension (:,:,:,:,:,:)  :: wave_back_block
    real,  dimension (:,:,:,:,:,:)  :: slow_block,dslow_block


    adj=rev_extrap_par%adj
    add=rev_extrap_par%add

    n_threads=rev_extrap_par%n_threads

    i_t0=wave_par%i_t_step(1)
    n_z=wave_par%n_z
    n_mx=wave_par%n_mx

!    i_mx_0_image=nint((wave_par%mx_0-image_par%mx_0)/image_par%d_mx)+1
!    i_my_0_image=nint((wave_par%my_0-image_par%my_0)/image_par%d_my)+1
!!write(0,*)'i_mx_0_image=',i_mx_0_image
!!write(0,*)'i_my_0_image=',i_my_0_image
!
!

    if(.not. adj ) then
    !{
!$omp parallel default (private) ,shared (wave_block,wave_back_block,slow_block,dslow_block),&
!$omp& shared(n_z,n_mx,n_threads,i_t0)
!$omp do
      do i_thread=1,n_threads
      !{
        do i_mx=i_thread,n_mx,n_threads
        !{
          do i_z=1,n_z
          !{
            s_back=slow_block(i_z,i_mx,1,1,1,1)
            delta_s=dslow_block(i_z,i_mx,1,1,1,1)
            wave_block(i_z,i_mx,1,1,1,i_t0) = &
              wave_block(i_z,i_mx,1,1,1,i_t0) - &
              2.*s_back*delta_s*wave_back_block(i_z,i_mx,1,1,1,1)
          !}
          end do
        !}
        end do 
      !}
      end do 
!$omp end parallel
    !}
    else
    !{
      do i_hx=1,slow_par%n_hx
      !{
        i_hx_shift_sou=i_hx_0_dslow-i_hx
        i_hx_shift_rec=-(i_hx_0_dslow-i_hx)
        i_mx_wave_beg=1-min(i_hx_shift_sou,i_hx_shift_rec)
        i_mx_wave_end=n_mx-max(i_hx_shift_sou,i_hx_shift_rec)
!write(0,*)'i_hx=',i_hx
!write(0,*)'i_mx_wave_beg=',i_mx_wave_beg
!write(0,*)'i_mx_wave_end=',i_mx_wave_end

!!$omp parallel default (private) ,shared (wave_block,wave_back_block,slow_block,dslow_block),&
!!$omp& shared(n_z,n_mx,n_threads,i_t0,i_hx_shift_sou,i_hx_shift_rec),&
!!$omp& shared(i_mx_wave_beg,i_mx_wave_end,i_mx_0_image)
!!$omp do
!        do i_thread=1,n_threads
!        !{
!          do i_mx=i_thread,n_mx,n_threads
!          !{

!$omp parallel default (shared), &
!$omp& private(i_mx,i_z)
!$omp do
            do i_mx=i_mx_wave_beg,i_mx_wave_end
            !{
              do i_z=1,n_z
              !{
                s_back=slow_block(i_z,i_mx,1,1,1,1)
                dslow_block(i_z,i_mx,i_hx,1,1,1)= &
                  dslow_block(i_z,i_mx,i_hx,1,1,1) - &
                    2.*s_back* &
                    wave_back_block(i_z,i_mx+i_hx_shift_sou,1,1,1,1)* &
                    wave_block(i_z,i_mx+i_hx_shift_rec,1,1,1,i_t0)
              !}
              end do
            !}
            end do 
!$omp end parallel
!          !}
!          end do 
!        !}
!        end do 
      !}
      end do 
    !}
    end if

    stat=0
    return	
  end function Rev2D_ho_fd_scatter_oper

  function Rev2D_ho_fd_der2time_oper (&
    rev_extrap_par, &
    wave_par, &
    wave_block, &
    wave_back_block, &
    slow_block) &
    result(stat)
!$cny dual_target(0)

    integer stat
    type (rev_extrap_param_type) rev_extrap_par
    type (wave_param_type) wave_par

!    integer omp_get_thread_num

    logical add,adj
    integer n_z,n_mx,i_z,i_mx
    integer i_t0,i_tm1,i_tm2
    integer n_threads,i_thread,ierr
    real vel_sq

    real, target,  dimension (:,:,:,:,:,:)  :: wave_block
    real,  dimension (:,:,:,:,:,:)  :: wave_back_block
    real,  dimension (:,:,:,:,:,:)  :: slow_block

    real, allocatable,  dimension (:,:,:,:,:)  :: lap,fourth_der

    real, pointer, dimension (:,:,:,:,:)      :: wave_t0,wave_tm1,wave_tm2


    adj=rev_extrap_par%adj
    add=rev_extrap_par%add

    n_threads=rev_extrap_par%n_threads

    i_t0=wave_par%i_t_step(1)
    i_tm1=wave_par%i_t_step(2)
    i_tm2=wave_par%i_t_step(3)

!write(0,*)'i_t0,i_tm1,i_tm2=',i_t0,i_tm1,i_tm2

    wave_t0 => wave_block(:,:,:,:,:,i_t0)
    wave_tm1 => wave_block(:,:,:,:,:,i_tm1)
    wave_tm2 => wave_block(:,:,:,:,:,i_tm2)

!write(0,*)'###1'
!

    n_z=wave_par%n_z
    n_mx=wave_par%n_mx

! second derivative in time 

!    if(fourth_time) then
!    !{
!
!call seperr ("I have implemented it but the cost is not worth the improved accuracy")
!      allocate (lap(wave_par%n_z,wave_par%n_mx,1,1,1), stat=ierr)
!      if(ierr .ne. 0) call seperr("Problems with allocation of lap")
!  
!      allocate (fourth_der(wave_par%n_z,wave_par%n_mx,1,1,1), stat=ierr)
!      if(ierr .ne. 0) call seperr("Problems with allocation of fourth_der")
!
!      call spat_der_oper(n_z,n_mx,nop,wave_tm1,coeff,lap,fourth_der) 
!
!!$omp parallel default (shared), &
!!$omp& private(i_mx,i_z,vel_sq)
!!$omp do
!      do i_mx=1,n_mx
!      !{
!        do i_z=1,n_z
!        !{
!          vel_sq=1.0/(slow_block(i_z,i_mx,1,1,1,1)**2)
!          wave_back_block(i_z,i_mx,1,1,1,1)= &
!              inv_d_t_sq*(  wave_t0(i_z,i_mx,1,1,1) - &
!                         2*wave_tm1(i_z,i_mx,1,1,1) + &
!                           wave_tm2(i_z,i_mx,1,1,1) + &
!                         d_t_fourth_12*(vel_sq*vel_sq) * &
!                           fourth_der(i_z,i_mx,1,1,1) &
!!                           wave_tm2(i_z,i_mx,1,1,1)  &
!                          )     
!        !}
!        end do
!      !}
!      end do 
!!$omp end parallel
!
!      if(allocated(lap)) then
!         deallocate (lap,stat=ierr)  
!         if(ierr /= 0) call seperr('Problems with deallocating lap')
!      end if
!      if(allocated(fourth_der)) then
!         deallocate (fourth_der,stat=ierr)  
!      end if
!
!    !}
!    else
!    !{

!$omp parallel default (shared), &
!$omp& private(i_mx,i_z)
!$omp do
      do i_mx=1,n_mx
      !{
        do i_z=1,n_z
        !{
          wave_back_block(i_z,i_mx,1,1,1,1)= &
              inv_d_t_sq*(  wave_t0(i_z,i_mx,1,1,1) - &
                         2*wave_tm1(i_z,i_mx,1,1,1) + &
                           wave_tm2(i_z,i_mx,1,1,1))

!            wave_back_block(i_z,i_mx,1,1,1,1)= &
!                wave_block(i_z,i_mx,1,1,1,i_t0) - &
!              2*wave_block(i_z,i_mx,1,1,1,i_tm1) + &
!                wave_block(i_z,i_mx,1,1,1,i_tm2) 

        !}
        end do
      !}
      end do 
!$omp end parallel

!    !}
!    end if

    stat=0
    return	
  end function Rev2D_ho_fd_der2time_oper

  subroutine laplace_second(n_z,n_mx,nop,wave_slice,coeff,lap,fourth_der) 
!                            result(spat_der)
!  type (Rev2D_spat_der_type) :: spat_der
!  real  :: spat_der
  integer n_z,n_mx,nop

  integer i_z,i_mx
  real lap_x,lap_z

  real,  dimension (0:nop)  :: coeff
  real,  dimension (:,:,:,:,:)  :: wave_slice
  real,  dimension (:,:,:,:,:)  :: lap,fourth_der

!$omp parallel default (shared), &
!$omp& private(i_mx,i_z,lap_x,lap_z)
!$omp do
  do i_mx=1+nop,n_mx-nop
  !{
    do i_z=1+nop,n_z-nop
    !{

      lap_x=coeff(0)*wave_slice(i_z,i_mx,1,1,1)+ &
            coeff(1)*(wave_slice(i_z,i_mx-1,1,1,1) &
                          +wave_slice(i_z,i_mx+1,1,1,1))
      lap_x=lap_x*d_x_sq_recip
      lap_z=coeff(0)*wave_slice(i_z,i_mx,1,1,1)+ &
            coeff(1)*(wave_slice(i_z-1,i_mx,1,1,1) &
                        +wave_slice(i_z+1,i_mx,1,1,1) )
      lap_z=lap_z*d_z_sq_recip

      lap(i_z,i_mx,1,1,1)=lap_x+lap_z
      fourth_der(i_z,i_mx,1,1,1)=0.
    !}
    end do
  !}
  end do
!$omp end parallel

  end subroutine laplace_second

  subroutine laplace_fourth(n_z,n_mx,nop,wave_slice,coeff,lap,fourth_der) 
!                            result(spat_der)
!  type (Rev2D_spat_der_type) :: spat_der
!  real  :: spat_der
  integer n_z,n_mx,nop

  integer i_z,i_mx
  real lap_x,lap_z

  real,  dimension (0:nop)  :: coeff
  real,  dimension (:,:,:,:,:)  :: wave_slice
  real,  dimension (:,:,:,:,:)  :: lap,fourth_der

!$omp parallel default (shared), &
!$omp& private(i_mx,i_z,lap_x,lap_z)
!$omp do
  do i_mx=1+nop,n_mx-nop
  !{
    do i_z=1+nop,n_z-nop
    !{

      lap_x=coeff(0)*wave_slice(i_z,i_mx,1,1,1)+ &
            coeff(1)*(wave_slice(i_z,i_mx-1,1,1,1) &
                          +wave_slice(i_z,i_mx+1,1,1,1))+&
            coeff(2)*(wave_slice(i_z,i_mx-2,1,1,1) &
                          +wave_slice(i_z,i_mx+2,1,1,1))
      lap_x=lap_x*d_x_sq_recip
      lap_z=coeff(0)*wave_slice(i_z,i_mx,1,1,1)+ &
            coeff(1)*(wave_slice(i_z-1,i_mx,1,1,1) &
                        +wave_slice(i_z+1,i_mx,1,1,1) )+ &
            coeff(2)*(wave_slice(i_z-2,i_mx,1,1,1) &
                          +wave_slice(i_z+2,i_mx,1,1,1))
      lap_z=lap_z*d_z_sq_recip

      lap(i_z,i_mx,1,1,1)=lap_x+lap_z
      fourth_der(i_z,i_mx,1,1,1)=0.
    !}
    end do
  !}
  end do
!$omp end parallel


  end subroutine laplace_fourth

  subroutine laplace_tenth(n_z,n_mx,nop,wave_slice,coeff,lap,fourth_der) 
!                            result(spat_der)

  integer n_z,n_mx,nop

  integer i_z,i_mx
  real lap_x,lap_z

  real,  dimension (0:nop)  :: coeff
  real,  dimension (:,:,:,:,:)  :: wave_slice
  real,  dimension (:,:,:,:,:)  :: lap,fourth_der

!$omp parallel default (shared), &
!$omp& private(i_mx,i_z,lap_x,lap_z)
!$omp do
  do i_mx=1+nop,n_mx-nop
  !{
    do i_z=1+nop,n_z-nop
    !{
      lap_x=coeff(0)*wave_slice(i_z,i_mx,1,1,1)+ &
            coeff(1)*(wave_slice(i_z,i_mx-1,1,1,1) &
                          +wave_slice(i_z,i_mx+1,1,1,1))+&
            coeff(2)*(wave_slice(i_z,i_mx-2,1,1,1) &
                          +wave_slice(i_z,i_mx+2,1,1,1))+&
            coeff(3)*(wave_slice(i_z,i_mx-3,1,1,1) &
                          +wave_slice(i_z,i_mx+3,1,1,1))+&
            coeff(4)*(wave_slice(i_z,i_mx-4,1,1,1) &
                          +wave_slice(i_z,i_mx+4,1,1,1))+&
            coeff(5)*(wave_slice(i_z,i_mx-5,1,1,1) &
                          +wave_slice(i_z,i_mx+5,1,1,1))
      lap_x=lap_x*d_x_sq_recip
      lap_z=coeff(0)*wave_slice(i_z,i_mx,1,1,1)+ &
            coeff(1)*(wave_slice(i_z-1,i_mx,1,1,1) &
                        +wave_slice(i_z+1,i_mx,1,1,1) )+ &
            coeff(2)*(wave_slice(i_z-2,i_mx,1,1,1) &
                          +wave_slice(i_z+2,i_mx,1,1,1))+ &
            coeff(3)*(wave_slice(i_z-3,i_mx,1,1,1) &
                          +wave_slice(i_z+3,i_mx,1,1,1))+ &
            coeff(4)*(wave_slice(i_z-4,i_mx,1,1,1) &
                          +wave_slice(i_z+4,i_mx,1,1,1))+ &
            coeff(5)*(wave_slice(i_z-5,i_mx,1,1,1) &
                          +wave_slice(i_z+5,i_mx,1,1,1))
      lap_z=lap_z*d_z_sq_recip

      lap(i_z,i_mx,1,1,1)=lap_x+lap_z
      fourth_der(i_z,i_mx,1,1,1)=0.

    !}
    end do
  !}
  end do
!$omp end parallel

  end subroutine laplace_tenth

  subroutine forth_order_time(n_z,n_mx,nop,wave_slice,coeff,lap,fourth_der) 

  integer n_z,n_mx,nop

  integer i_z,i_mx
  real lap_x,lap_z

  real,  dimension (0:nop)  :: coeff
  real,  dimension (:,:,:,:,:)  :: wave_slice
  real,  dimension (:,:,:,:,:)  :: lap,fourth_der

  real,  dimension (n_z,n_mx,1,1,1)  :: temp

!write(0,*)'#41'
!pause 
!$omp parallel default (shared), &
!$omp& private(i_mx,i_z,lap_x,lap_z)


!$omp do
  do i_mx=1,n_mx
    do i_z=1,n_z
       temp(i_z,i_mx,1,1,1)=0.
    end do
  end do
!write(0,*)'#42'
!pause 
!$omp do
  do i_mx=1+nop,n_mx-nop
  !{
    do i_z=1+nop,n_z-nop
    !{
      lap_x=coeff(0)*wave_slice(i_z,i_mx,1,1,1)+ &
            coeff(1)*(wave_slice(i_z,i_mx-1,1,1,1) &
                          +wave_slice(i_z,i_mx+1,1,1,1))+&
            coeff(2)*(wave_slice(i_z,i_mx-2,1,1,1) &
                          +wave_slice(i_z,i_mx+2,1,1,1))+&
            coeff(3)*(wave_slice(i_z,i_mx-3,1,1,1) &
                          +wave_slice(i_z,i_mx+3,1,1,1))+&
            coeff(4)*(wave_slice(i_z,i_mx-4,1,1,1) &
                          +wave_slice(i_z,i_mx+4,1,1,1))+&
            coeff(5)*(wave_slice(i_z,i_mx-5,1,1,1) &
                          +wave_slice(i_z,i_mx+5,1,1,1))
      lap_x=lap_x*d_x_sq_recip

      lap(i_z,i_mx,1,1,1)=lap_x ! lap=der2/dx2
      temp(i_z,i_mx,1,1,1)=lap_x ! temp=der2/dx2

    !}
    end do
  !}
  end do

!$omp do
  do i_mx=1+nop,n_mx-nop
  !{
    do i_z=1+nop,n_z-nop
    !{

      lap_z=coeff(0)*temp(i_z,i_mx,1,1,1)+ &
            coeff(1)*(temp(i_z-1,i_mx,1,1,1) &
                        +temp(i_z+1,i_mx,1,1,1) )+ &
            coeff(2)*(temp(i_z-2,i_mx,1,1,1) &
                          +temp(i_z+2,i_mx,1,1,1))+ &
            coeff(3)*(temp(i_z-3,i_mx,1,1,1) &
                          +temp(i_z+3,i_mx,1,1,1))+ &
            coeff(4)*(temp(i_z-4,i_mx,1,1,1) &
                          +temp(i_z+4,i_mx,1,1,1))+ &
            coeff(5)*(temp(i_z-5,i_mx,1,1,1) &
                          +temp(i_z+5,i_mx,1,1,1))
      lap_z=lap_z*d_z_sq_recip

      fourth_der(i_z,i_mx,1,1,1)=2*lap_z ! fourth_der=2*der4/dx2dz2
    !}
    end do
  !}
  end do

!$omp do
  do i_mx=1+nop,n_mx-nop
  !{
    do i_z=1+nop,n_z-nop
    !{

      lap_x=coeff(0)*temp(i_z,i_mx,1,1,1)+ &
            coeff(1)*(temp(i_z,i_mx-1,1,1,1) &
                          +temp(i_z,i_mx+1,1,1,1))+&
            coeff(2)*(temp(i_z,i_mx-2,1,1,1) &
                          +temp(i_z,i_mx+2,1,1,1))+&
            coeff(3)*(temp(i_z,i_mx-3,1,1,1) &
                          +temp(i_z,i_mx+3,1,1,1))+&
            coeff(4)*(temp(i_z,i_mx-4,1,1,1) &
                          +temp(i_z,i_mx+4,1,1,1))+&
            coeff(5)*(temp(i_z,i_mx-5,1,1,1) &
                          +temp(i_z,i_mx+5,1,1,1))
      lap_x=lap_x*d_x_sq_recip

      fourth_der(i_z,i_mx,1,1,1)=fourth_der(i_z,i_mx,1,1,1)+lap_x 
             ! fourth_der=2*der4/dx2dz2 + der4/dx4
    !}
    end do
  !}
  end do

!$omp do
  do i_mx=1+nop,n_mx-nop
  !{
    do i_z=1+nop,n_z-nop
    !{
      lap_z=coeff(0)*wave_slice(i_z,i_mx,1,1,1)+ &
            coeff(1)*(wave_slice(i_z-1,i_mx,1,1,1) &
                        +wave_slice(i_z+1,i_mx,1,1,1) )+ &
            coeff(2)*(wave_slice(i_z-2,i_mx,1,1,1) &
                          +wave_slice(i_z+2,i_mx,1,1,1))+ &
            coeff(3)*(wave_slice(i_z-3,i_mx,1,1,1) &
                          +wave_slice(i_z+3,i_mx,1,1,1))+ &
            coeff(4)*(wave_slice(i_z-4,i_mx,1,1,1) &
                          +wave_slice(i_z+4,i_mx,1,1,1))+ &
            coeff(5)*(wave_slice(i_z-5,i_mx,1,1,1) &
                          +wave_slice(i_z+5,i_mx,1,1,1))
      lap_z=lap_z*d_z_sq_recip

      lap(i_z,i_mx,1,1,1)=lap(i_z,i_mx,1,1,1)+lap_z ! lap=der2/dx2 + der2/dz2
      temp(i_z,i_mx,1,1,1)=lap_z ! temp=der2/dz2

    !}
    end do
  !}
  end do

!$omp do
  do i_mx=1+nop,n_mx-nop
  !{
    do i_z=1+nop,n_z-nop
    !{
      lap_z=coeff(0)*temp(i_z,i_mx,1,1,1)+ &
            coeff(1)*(temp(i_z-1,i_mx,1,1,1) &
                        +temp(i_z+1,i_mx,1,1,1) )+ &
            coeff(2)*(temp(i_z-2,i_mx,1,1,1) &
                          +temp(i_z+2,i_mx,1,1,1))+ &
            coeff(3)*(temp(i_z-3,i_mx,1,1,1) &
                          +temp(i_z+3,i_mx,1,1,1))+ &
            coeff(4)*(temp(i_z-4,i_mx,1,1,1) &
                          +temp(i_z+4,i_mx,1,1,1))+ &
            coeff(5)*(temp(i_z-5,i_mx,1,1,1) &
                          +temp(i_z+5,i_mx,1,1,1))
      lap_z=lap_z*d_z_sq_recip

      fourth_der(i_z,i_mx,1,1,1)=fourth_der(i_z,i_mx,1,1,1)+lap_z
             ! fourth_der=2*der4/dx2dz2 + der4/dx4 + der4/dz4

!      fourth_der(i_z,i_mx,1,1,1)=0.

    !}
    end do
  !}
  end do
!$omp end parallel

!write(0,*)'#43'
!pause 

  end subroutine forth_order_time


  function Rev2D_ho_fd_clean()  result(stat)
    integer stat

    integer ierr

    if(allocated(coeff)) deallocate (coeff,stat=ierr)

!    write(0,*)' Rev2D_ho_fd_clean: Done with cleaning'

    stat=0
  end function Rev2D_ho_fd_clean


end module Rev2D_ho_fd_mod
