module FocusMes_mod

  use focusmes_types
  use fft_types

  use fft_data_mod

  implicit none

  integer, external ::  sreed,srite,sseek_block

  integer, private :: n_z,n_mx,n_z_avg,n_mx_avg
  integer, private :: n_hx,n_gamma,n3_inp,n4_inp,n_rho,n_dip,n_rad_curv
  integer, private :: i_dip_flip,i_dip_split,i_dip_split_1,i_z_dip_split_1
  integer, private :: n_bytes_cube_inp,n_bytes_cube_out
  integer, private :: n_bytes_apert
  integer, private :: n_bytes_sum_sq,n_bytes_apert_dip,n_bytes_apert_apert
  integer, private :: n_bytes_apert_apert_dip
  integer, private :: n_threads
  real, private :: d_z,d_mx,d_hx,d_gamma,d_rho,d_dip,d_rad_curv
  real, private :: z_0,mx_0,hx_0,hz_0,gamma_0,rho_0,rad_curv_0,dip_0,dip_flip
  real, private :: dip_split,dip_split_1,z_dip_split_1
  real, private :: global_threshold,local_threshold

  real, allocatable, dimension (:,:,:,:), private :: apert
  real, allocatable, dimension (:,:,:), private :: apert_apert,apert_dip,sum_sq
  real, allocatable, dimension (:,:,:,:), private :: apert_apert_dip
  real, allocatable, dimension (:,:,:,:), private :: cube_out
  real, allocatable, dimension (:,:,:,:), private :: cube_inp

contains

  function FocusMes_init(focusmes_par)  result(status)
    type (focusmes_param_type) focusmes_par
    integer status

    integer ierr,i_thread

    status=0

    local_threshold=focusmes_par%local_threshold
    global_threshold=focusmes_par%global_threshold

    n_z=focusmes_par%n_z
    z_0=focusmes_par%z_0
    d_z=focusmes_par%d_z

    n_mx=focusmes_par%n_mx
    mx_0=focusmes_par%mx_0
    d_mx=focusmes_par%d_mx

    n_rho=focusmes_par%n_rho
    rho_0=focusmes_par%rho_0
    d_rho=focusmes_par%d_rho

    n_rad_curv=focusmes_par%n_rad_curv
    rad_curv_0=focusmes_par%rad_curv_0
    d_rad_curv=focusmes_par%d_rad_curv

    n_z_avg=focusmes_par%n_z_avg
    n_mx_avg=focusmes_par%n_mx_avg

    prestack_image_1: if(focusmes_par%prestack_image) then
    !{
      select case (focusmes_par%aperture_type(1:5))
      case('apert ')
        n_gamma=focusmes_par%n_gamma
        gamma_0=focusmes_par%gamma_0
        d_gamma=focusmes_par%d_gamma
        n3_inp=n_gamma
        n4_inp=1
      case('dip  ')
        n_dip=focusmes_par%n_dip
        dip_0=focusmes_par%dip_0
        d_dip=focusmes_par%d_dip
        n3_inp=n_dip
        n4_inp=1
  
      case('both ')
        n_gamma=focusmes_par%n_gamma
        gamma_0=focusmes_par%gamma_0
        d_gamma=focusmes_par%d_gamma
        n_dip=focusmes_par%n_dip
        dip_0=focusmes_par%dip_0
        d_dip=focusmes_par%d_dip
        n3_inp=n_gamma
        n4_inp=n_dip
  
      case default
        write(0,*)'aperture_type=',focusmes_par%aperture_type(1:5)
        call seperr ('Do not know what to do with this aperture_type')
      end select

      select case (focusmes_par%aperture_type(1:5))
      case('both ','dip  ')

        if(focusmes_par%flip_sign_dip < huge(focusmes_par%flip_sign_dip)) then
          dip_flip=focusmes_par%flip_sign_dip 
          if(d_dip <= 0.) call seperr ("d_dip must be >=0 when flip_sign_dip is set")
          if(abs(mod(sign((dip_flip-dip_0),d_dip),d_dip)) > .01*d_dip) then
                write(0,*)'mod(sign(dip_flip-dip_0,d_dip),d_dip)=', &
                        mod(sign(dip_flip-dip_0,d_dip),d_dip)
                call seperr ("dip_flip-dip_0 must be multiple of d_dip")
          end if
          i_dip_flip=(dip_flip-dip_0)/d_dip+1
          if((i_dip_flip <=1) .or. (i_dip_flip > n_dip)) then
            call seperr("Dip range must contain neg. dips when flip_sign_dip is set")
          end if
        else
          i_dip_flip=-1
        end if

  
        if(focusmes_par%split_dip < huge(focusmes_par%split_dip)) then
        !{

          dip_split=focusmes_par%split_dip 
          if(d_dip <= 0.) call seperr ("d_dip must be >=0 when split_dip is set")
          if(abs(mod(sign((dip_split-dip_0),d_dip),d_dip)) > .01*d_dip) then
                write(0,*)'mod(sign(dip_split-dip_0,d_dip),d_dip)=', &
                        mod(sign(dip_split-dip_0,d_dip),d_dip)
                call seperr ("dip_split-dip_0 must be multiple of d_dip")
          end if
          i_dip_split=(dip_split-dip_0)/d_dip+1
          if((i_dip_split <=1) .or. (i_dip_split > n_dip)) then
            call seperr("Dip range must contain neg. dips when split_dip is set")
          end if
          dip_split_1=focusmes_par%split_dip_1 
          if(dip_split_1 < huge(dip_split_1)) then
          !{
  
            if(abs(mod(sign((dip_split_1-dip_0),d_dip),d_dip))>.01*d_dip) then
                  write(0,*)'mod(sign(dip_split_1-dip_0,d_dip),d_dip)=', &
                          mod(sign(dip_split_1-dip_0,d_dip),d_dip)
                  call seperr ("dip_split_1-dip_0 must be multiple of d_dip")
            end if
  
            i_dip_split_1=(dip_split_1-dip_0)/d_dip+1
            if((i_dip_split_1 <=1) .or. (i_dip_split_1 > n_dip)) then
              call seperr("Dip range must contain neg. dips when split_dip_1 is set")
            end if
  
            z_dip_split_1=focusmes_par%z_split_dip_1 
            if(abs(mod(sign((z_dip_split_1-z_0),d_z),d_z))>.01*d_z) then
                  write(0,*)'mod(sign(z_dip_split_1-z_0,d_z),d_z)=', &
                          mod(sign(z_dip_split_1-z_0,d_z),d_z)
                  call seperr ("z_dip_split_1-z_0 must be multiple of d_z")
            end if

            i_z_dip_split_1=(z_dip_split_1-z_0)/d_z+1
            if((i_z_dip_split_1 <1) .or. (i_z_dip_split_1 >= n_z)) then
              call seperr("Depth range wrong plit_dip_1 is set")
            end if
  
          !}
          else
          !{
            i_dip_split_1=-1
            i_z_dip_split_1=n_z+1
          !}
          end if
  
  
        !}
        else
        !{
          i_dip_split=-1
          i_dip_split_1=-1
          i_z_dip_split_1=n_z+1
        !}
        end if
  
        if(focusmes_par%verb) write(0,*)'i_dip_flip=',i_dip_flip
        if(focusmes_par%verb) write(0,*)'i_dip_split=',i_dip_split
        if(focusmes_par%verb) write(0,*)'i_dip_split_1=',i_dip_split_1
        if(focusmes_par%verb) write(0,*)'i_z_dip_split_1=',i_z_dip_split_1
        if((i_dip_split >0).and.(i_dip_flip >0)) &
              call seperr('Cannot flip sign and split dips at the same time')

      case default
        write(0,*)'No dip to flip'
      end select
    !}
    else prestack_image_1
    !{
      n3_inp=1
      n4_inp=1
    !}
    end if prestack_image_1

        if(focusmes_par%verb) write(0,*)'# z_dip_split_1=',z_dip_split_1

    n_bytes_cube_inp=n_z*n_mx*n3_inp*n4_inp*4
    allocate (cube_inp(n_z,n_mx,n3_inp,n4_inp),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of cube_inp ")
    cube_inp=0.


    select case (focusmes_par%oper_name(1:9))
    !{
    case('semblance')
    !{
      select case (focusmes_par%aperture_type(1:5))
      case('apert ')
        n_bytes_apert_apert=n_z*n_mx*n_gamma*4
        allocate (apert_apert(n_z,n_mx,n_gamma),stat=ierr)  
        if(ierr /= 0) call seperr("Problems with allocation of apert_apert")

      case('dip  ')
        n_bytes_apert_dip=n_z*n_mx*n_dip*4
        allocate (apert_dip(n_z,n_mx,n_dip),stat=ierr)  
        if(ierr /= 0) call seperr("Problems with allocation of apert_dip ")

      case('both ')
        n_bytes_apert_apert_dip=n_z*n_mx*n_gamma*n_dip*4
        allocate (apert_apert_dip(n_z,n_mx,n_gamma,n_dip),stat=ierr)  
        if(ierr/=0) call seperr("Problems with allocation of apert_apert_dip ")

      case default
        write(0,*)'aperture_type=',focusmes_par%aperture_type(1:5)
        call seperr ('Do not know what to do with this opern_name')
      end select

      n_bytes_cube_out=n_z*n_mx*4
      allocate (cube_out(n_z,n_mx,1,1),stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of cube_out ")
      cube_out=0.

      if(focusmes_par%sum_squares_input) then
      !{
        n_bytes_sum_sq=n_z*n_mx*4
        allocate (sum_sq(n_z,n_mx,1),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of sum_sq ")
      !}
      end if

    !}
    case('kurtosis')
    !{
      if(focusmes_par%write_prestack) then
        n_bytes_cube_out=n3_inp*n_rho*n_rad_curv*4
        allocate (cube_out(n3_inp,n_rho,n_rad_curv,1),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of cube_out ")
      else
        n_bytes_cube_out=n_rho*n_rad_curv*4
        allocate (cube_out(1,n_rho,n_rad_curv,1),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of cube_out ")
      end if
      cube_out=0.
    !}
    case('aperture ')
    !{
      if(focusmes_par%write_prestack) then
        n_bytes_cube_out=n_z*n_mx*n3_inp*n4_inp*4
        allocate (cube_out(n_z,n_mx,n3_inp,n4_inp),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of cube_out ")

        n_bytes_apert=n_z*n_mx*n3_inp*n4_inp*4
        allocate (apert(n_z,n_mx,n3_inp,n4_inp),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of apert ")

      end if
      cube_out=0.
    !}
    case default
      write(0,*)'oper_name=',focusmes_par%oper_name(1:9)
      call seperr ('Do not know what to do with this opern_name')
    !}
    end select

    n_threads=focusmes_par%n_threads

     if(focusmes_par%debug1) then
write(0,*)'n3_inp=',n3_inp
write(0,*)'n4_inp=',n4_inp
write(0,*)'n_gamma=',n_gamma
write(0,*)'gamma_0=',gamma_0
write(0,*)'d_gamma=',d_gamma
write(0,*)'n_dip=',n_dip
write(0,*)'dip_0=',dip_0
write(0,*)'d_dip=',d_dip
write(0,*)'n_hx=',n_hx
write(0,*)'hx_0=',hx_0
write(0,*)'d_hx=',d_hx
write(0,*)'n_z=',n_z
write(0,*)'n_mx=',n_mx
write(0,*)'d_z=',d_z
write(0,*)'d_mx=',d_mx
write(0,*)'n_z_avg=',n_z_avg
write(0,*)'n_mx_avg=',n_mx_avg
write(0,*)'n_bytes_cube_inp=',n_bytes_cube_inp
write(0,*)'n_bytes_cube_out=',n_bytes_cube_out
write(0,*)'n_bytes_apert_apert=',n_bytes_apert_apert
write(0,*)'n_bytes_apert_dip=',n_bytes_apert_dip
write(0,*)'shape(cube_inp)=',shape(cube_inp)
write(0,*)'shape(cube_out)=',shape(cube_out)
write(0,*)'shape(apert_apert)=',shape(apert_apert)
write(0,*)'shape(apert_dip)=',shape(apert_dip)
write(0,*)'shape(apert)=',shape(apert)
write(0,*)'local_threshold=',local_threshold
write(0,*)'global_threshold=',global_threshold
     end if


  end function FocusMes_init

  function FocusMes_oper(focusmes_par) result(status)

    type (focusmes_param_type) focusmes_par
    integer status

    logical in_aperture,in_aperture_global,split_pos
    integer ret_sseek
    integer i_hx,i_mx,i_z,i_z_avg,i_mx_avg,i_thread,i_3,i_4,i_rho,i_rad_curv
    integer n_samples
    real sample_inp,sample_sq_inp,apert_apert_local,apert_dip_local
    double precision num,num_avg_neg,num_avg_pos,den,apert_width
    double precision global_max_image_rho,local_max_image_rho



    oper_name_1: select case (focusmes_par%oper_name(1:9))
    !{
    case('semblance') oper_name_1
    !{

    do_i_rho_semblance: do i_rho=1,n_rho
    !{
      if(focusmes_par%verb) write(0,*)'i_rho=',i_rho
    do_i_rad_curv_semblance: do i_rad_curv=1,n_rad_curv
    !{
      if(focusmes_par%verb) write(0,*)'i_rad_curv=',i_rad_curv

!        write(0,*)'read tag_cig'
      if(sreed(trim(focusmes_par%tag_cig),cube_inp,n_bytes_cube_inp)/= &
          n_bytes_cube_inp) then
        write(0,*)'tag_cig=',trim(focusmes_par%tag_cig)
        call seperr('Problems with reading tag_cig file')
      end if

      select case (focusmes_par%aperture_type(1:5))
      case('dip  ')
        if(i_dip_flip > 0) then
write(0,*)'flipping signs '
          cube_inp(:,:,1:i_dip_flip-1,:)=-cube_inp(:,:,1:i_dip_flip-1,:)
          cube_inp(:,:,i_dip_flip,:)=0.
        end if
        if((i_dip_split > 0) .and. focusmes_par%zero_split_dip) then
write(0,*)'zeroing_split dip '
          if(i_z_dip_split_1 < n_z) then
!              cube_inp(1:i_z_dip_split_1-1,:,i_dip_split,:)=0.
            cube_inp(i_z_dip_split_1:n_z,:,i_dip_split_1,:)=0.
          else
            cube_inp(:,:,i_dip_split_1,:)=0.
          end if
        end if
      case('both ')
        if(i_dip_flip > 0) then
write(0,*)'flipping signs dip '
          cube_inp(:,:,:,1:i_dip_flip-1)=-cube_inp(:,:,:,1:i_dip_flip-1)
          cube_inp(:,:,:,i_dip_flip)=0.
        end if
        if((i_dip_split > 0) .and. focusmes_par%zero_split_dip) then
          if(i_z_dip_split_1 < n_z) then
write(0,*)'zeroing_split dip '
!              cube_inp(1:i_z_dip_split_1-1,:,:,i_dip_split)=0.
            cube_inp(i_z_dip_split_1:n_z,:,:,i_dip_split_1)=0.
          else
            cube_inp(:,:,:,i_dip_split)=0.
          end if
        end if
      case default
      end select

      select case (focusmes_par%aperture_type(1:5))
      case('apert ')
        if(sreed(trim(focusmes_par%tag_apert_apert),apert_apert, &
            n_bytes_apert_apert)/= n_bytes_apert_apert) then
          write(0,*)'tag_apert_apert=',trim(focusmes_par%tag_apert_apert)
           call seperr('Problems with reading tag_apert_apert file')
        end if
      case('dip  ')
        if(sreed(trim(focusmes_par%tag_apert_dip),apert_dip, &
            n_bytes_apert_dip)/= n_bytes_apert_dip) then
          write(0,*)'tag_apert_dip=',trim(focusmes_par%tag_apert_dip)
            call seperr('Problems with reading tag_apert_dip file')
        end if
      case('both ')
        if(sreed(trim(focusmes_par%tag_apert_apert_dip),apert_apert_dip, &
            n_bytes_apert_apert_dip)/= n_bytes_apert_apert_dip) then
        write(0,*)'tag_apert_apert_dip=',trim(focusmes_par%tag_apert_apert_dip)
          call seperr('Problems with reading tag_apert_apert_dip file')
        end if
      case default
        write(0,*)'aperture_type=',focusmes_par%aperture_type(1:5)
        call seperr ('Do not know what to do with this opern_name')
      end select

      if(focusmes_par%sum_squares_input) then
        if(sreed(trim(focusmes_par%tag_sum_squares),sum_sq,n_bytes_sum_sq)/= &
              n_bytes_sum_sq) then
          write(0,*)'tag_sum_squares=',trim(focusmes_par%tag_sum_squares)
          call seperr('Problems with reading tag_sum_squares file')
        end if
      end if

!      if(global_threshold /= 0.) then
!!        global_max_image_rho=(global_max_image_rho**2)*(n_z_avg*2+1)
!!        global_max_image_rho=maxval(abs(cube_inp))
!        global_max_image_rho=maxval(sum(abs(cube_inp),dim=3))
!        global_max_image_rho=(global_max_image_rho**2)*(n_z_avg*2+n_mx_avg*2+1) !        write(0,*)'global_max_image_rho=',global_max_image_rho
!      else
!        global_max_image_rho=0.
!      end if
!!
!!      if(local_threshold /= 0.) then
!!        global_max_image_rho=maxval(sum(cube_inp,dim=3))
!!      else
!!        global_max_image_rho=0.
!!      end if

      cube_out=0.
      do i_mx=n_mx_avg+1,n_mx-n_mx_avg
      !{
        do i_z=n_z_avg+1,n_z-n_z_avg
        !{
          select case (focusmes_par%aperture_type(1:5))
          case('apert')
            apert_apert_local=maxval(apert_apert(i_z,i_mx,:))
            apert_dip_local=1.
            in_aperture_global=(apert_apert_local /= 0.)
          case('dip  ')
            apert_apert_local=1.
            apert_dip_local=maxval(apert_dip(i_z,i_mx,:))
            in_aperture_global=(apert_dip_local /= 0.)
          case('both ')
            apert_apert_local=maxval(apert_apert_dip(i_z,i_mx,:,:))
            apert_dip_local=1.
            in_aperture_global=(apert_apert_local /= 0.)
          case('none ')
            in_aperture_global=.true.
          case default
            write(0,*)'aperture_type=',focusmes_par%aperture_type(1:5)
            call seperr ('Do not know what to do with this aperture_type')
          end select
          if(in_aperture_global) then
          !{
            num=0
            den=0
            do i_mx_avg=-n_mx_avg,n_mx_avg
            !{
              do i_z_avg=-n_z_avg,n_z_avg
              !{
                num_avg_neg=0
                num_avg_pos=0
                do i_4=1,n4_inp
                !{
                  do i_3=1,n3_inp
                  !{
            select case (focusmes_par%aperture_type(1:5))
            case('apert')
              in_aperture=(apert_apert(i_z+i_z_avg,i_mx+i_mx_avg,i_3) /=0 )
              sample_inp=cube_inp(i_z+i_z_avg,i_mx+i_mx_avg,i_3,1)
              split_pos=.true.
            case('dip  ')
              in_aperture=(apert_dip(i_z+i_z_avg,i_mx+i_mx_avg,i_3) /=0 )
              sample_inp=cube_inp(i_z+i_z_avg,i_mx+i_mx_avg,i_3,1)
              if (i_z+i_z_avg <= i_z_dip_split_1) then
                split_pos=(i_3 >= i_dip_split)
              else
                split_pos=(i_3 >= i_dip_split_1)
              end if
            case('both ')
              in_aperture=apert_apert_dip(i_z+i_z_avg,i_mx+i_mx_avg,i_3,i_4)/=0
              sample_inp=cube_inp(i_z+i_z_avg,i_mx+i_mx_avg,i_3,i_4)
              if (i_z+i_z_avg <= i_z_dip_split_1) then
                split_pos=(i_4 >= i_dip_split)
              else
                split_pos=(i_4 >= i_dip_split_1)
              end if
            case('none ')
              in_aperture= .true.
              sample_inp=cube_inp(i_z+i_z_avg,i_mx+i_mx_avg,i_3,1)
            case default
              write(0,*)'aperture_type=',focusmes_par%aperture_type(1:5)
              call seperr ('Do not know what to do with this aperture_type')
            end select

            if(in_aperture) then
              if(split_pos) then
                num_avg_pos=num_avg_pos+sample_inp
              else
                num_avg_neg=num_avg_neg+sample_inp
              end if
              den=den+sample_inp**2
            end if
                  !}
                  end do
                !}
                end do
                num=num + num_avg_neg**2 + num_avg_pos**2
              !}
              end do
            !}
            end do
!            cube_out(i_z,i_mx,1)=max(apert_apert_local,apert_dip_local)*num/den
            if(focusmes_par%norm_semblance) then
              cube_out(i_z,i_mx,1,1)=apert_apert_local*apert_dip_local*num/den
            else
              cube_out(i_z,i_mx,1,1)=apert_apert_local*apert_dip_local*num
            end if
          !}
          end if
        !}
        end do
      !}
      end do

      if(srite(trim(focusmes_par%tag_semblance),cube_out,n_bytes_cube_out)/= &
          n_bytes_cube_out) then
        write(0,*)'tag_cig=',trim(focusmes_par%tag_semblance)
        call seperr('Problems with writing tag_semblance file')
      end if

    !}
    end do do_i_rad_curv_semblance
    !}
    end do do_i_rho_semblance

    !}
    case('kurtosis') oper_name_1
    !{

    do i_rad_curv=1,n_rad_curv
    !{
    do i_rho=1,n_rho
    !{

      if(sreed(trim(focusmes_par%tag_cig),cube_inp,n_bytes_cube_inp)/= &
          n_bytes_cube_inp) then
        write(0,*)'tag_cig=',trim(focusmes_par%tag_cig)
        call seperr('Problems with reading tag_cig file')
      end if

      do i_3=1,n3_inp
      !{
        num=0
        den=0
        do i_mx=1,n_mx
        !{
          do i_z=1,n_z
          !{
            num=num+cube_inp(i_z,i_mx,i_3,1)**4
            den=den+cube_inp(i_z,i_mx,i_3,1)**2
          !}
          end do
        !}
        end do
        if(den > tiny(den)*100000000000) then
          cube_out(i_3,i_rho,i_rad_curv,1)=((n_mx*n_z)*num)/(den**2)
        else
          cube_out(i_3,i_rho,i_rad_curv,1)=0.
        end if
              
      !}
      end do
    !}
    end do
    !}
    end do

        write(0,*)'rite'
        write(0,*)'tag_kurtosis',trim(focusmes_par%tag_kurtosis)
        write(0,*)' 2 rite'
        write(0,*)' cube_out= ',cube_out
      if(srite(trim(focusmes_par%tag_kurtosis),cube_out,n_bytes_cube_out)/= &
          n_bytes_cube_out) then
        write(0,*)'tag_kurtosis=',trim(focusmes_par%tag_kurtosis)
        call seperr('Problems with reading tag_kurtosis file')
      end if
        write(0,*)'after rite'


    !}
    case('aperture ') oper_name_1
    !{

    do_i_rho_aperture: do i_rho=1,n_rho
    !{
      if(focusmes_par%verb) write(0,*)'i_rho=',i_rho
    do_i_rad_curv_aperture: do i_rad_curv=1,n_rad_curv
    !{
      if(focusmes_par%verb) write(0,*)'i_rad_curv=',i_rad_curv

      if(sreed(trim(focusmes_par%tag_cig),cube_inp,n_bytes_cube_inp)/= &
          n_bytes_cube_inp) then
        write(0,*)'tag_cig=',trim(focusmes_par%tag_cig)
        call seperr('Problems with reading tag_cig file')
      end if

      if(focusmes_par%dip_ang_image) then
        select case (focusmes_par%aperture_type(1:5))
        case('dip  ')
          if(i_dip_flip > 0) then
  write(0,*)'flipping signs '
            cube_inp(:,:,1:i_dip_flip-1,:)=-cube_inp(:,:,1:i_dip_flip-1,:)
            cube_inp(:,:,i_dip_flip,:)=0.
          end if
          if((i_dip_split > 0) .and. focusmes_par%zero_split_dip) then
  write(0,*)'zeroing_split dip '
            if(i_z_dip_split_1 < n_z) then
!              cube_inp(1:i_z_dip_split_1-1,:,i_dip_split,:)=0.
              cube_inp(i_z_dip_split_1:n_z,:,i_dip_split_1,:)=0.
            else
              cube_inp(:,:,i_dip_split_1,:)=0.
            end if
          end if
        case('both ')
          if(i_dip_flip > 0) then
  write(0,*)'flipping signs dip '
            cube_inp(:,:,:,1:i_dip_flip-1)=-cube_inp(:,:,:,1:i_dip_flip-1)
            cube_inp(:,:,:,i_dip_flip)=0.
          end if
          if((i_dip_split > 0) .and. focusmes_par%zero_split_dip) then
  write(0,*)'zeroing_split dip '
            if(i_z_dip_split_1 < n_z) then
!              cube_inp(1:i_z_dip_split_1-1,:,:,i_dip_split)=0.
              cube_inp(i_z_dip_split_1:n_z,:,:,i_dip_split_1)=0.
            else
              cube_inp(:,:,:,i_dip_split)=0.
            end if
          end if
        case default
        end select
      end if

      if(global_threshold /= 0.) then
        global_max_image_rho=maxval(abs(sum( sum(cube_inp,dim=4) ,dim=3)))
        global_max_image_rho=(global_max_image_rho**2)*(n_z_avg*2+n_mx_avg*2+1)
        write(0,*)'global_max_image_rho=',global_max_image_rho
      else
        global_max_image_rho=0.
      end if


      cube_out=0
!      do i_mx=n_mx_avg+1,n_mx-n_mx_avg
      do i_mx=1,n_mx
      !{
!        do i_z=n_z_avg+1,n_z-n_z_avg
        do i_z=1,n_z
        !{
          local_max_image_rho=maxval(abs(cube_inp(i_z,i_mx,:,:)))

           do i_4=1,n4_inp
           !{
             do i_3=1,n3_inp
             !{
               sample_inp=cube_inp(i_z,i_mx,i_3,i_4)
               if(abs(sample_inp) > local_max_image_rho*local_threshold) then
                 apert(i_z,i_mx,i_3,i_4)=1.
               else
                 apert(i_z,i_mx,i_3,i_4)=0.
               end if
            !}
            end do
          !}
          end do
        !}
        end do
      !}
      end do
!        write(0,*)'after apert'

      do i_mx=n_mx_avg+1,n_mx-n_mx_avg
      !{
        do i_z=n_z_avg+1,n_z-n_z_avg
        !{
          num=0
          den=0
          do i_mx_avg=-n_mx_avg,n_mx_avg
          !{
            do i_z_avg=-n_z_avg,n_z_avg
            !{
              num_avg_neg=0.
              num_avg_pos=0.
              do i_4=1,n4_inp
              !{
                do i_3=1,n3_inp
                !{
                  select case (focusmes_par%aperture_type(1:5))
                  case('apert')
                    split_pos=.true.
                  case('dip  ')
                    if (i_z+i_z_avg <= i_z_dip_split_1) then
                      split_pos=(i_3 >= i_dip_split)
                    else
                      split_pos=(i_3 >= i_dip_split_1)
                    end if
                  case('both ')
                    if (i_z+i_z_avg <= i_z_dip_split_1) then
                      split_pos=(i_4 >= i_dip_split)
                    else
                      split_pos=(i_4 >= i_dip_split_1)
                    end if
                  case('none ')
                    in_aperture= .true.
                  case default
                  end select

                  if(apert(i_z+i_z_avg,i_mx+i_mx_avg,i_3,i_4) == 1) then
                    sample_inp=cube_inp(i_z+i_z_avg,i_mx+i_mx_avg,i_3,i_4)
                    if(split_pos) then
                      num_avg_pos=num_avg_pos+sample_inp
                    else
                      num_avg_neg=num_avg_neg+sample_inp
                    end if
                    den=den + sample_inp**2
                  end if
                !}
                end do
              !}
              end do
              num=num + num_avg_neg**2 + num_avg_pos**2
            !}
            end do
          !}
          end do

          if((den > tiny(den)*100000000000000000.) .and. &
             (num > global_max_image_rho*global_threshold)) then
!            apert_width= sum(apert(i_z,i_mx,:))
            apert_width=0.
            do i_4=1,n4_inp
              do i_3=1,n3_inp
                do i_mx_avg=-n_mx_avg,n_mx_avg
                  do i_z_avg=-n_z_avg,n_z_avg
           apert_width=apert_width+apert(i_z+i_z_avg,i_mx+i_mx_avg,i_3,i_4)
                  end do
                end do
              end do
            end do
            if(apert_width > 0.) then
              cube_out(i_z,i_mx,:,:)= apert(i_z,i_mx,:,:)/apert_width
            else
              cube_out(i_z,i_mx,:,:)=0.
            end if
          else
              cube_out(i_z,i_mx,:,:)=0.
          end if

        !}
        end do
      !}
      end do

      if(srite(trim(focusmes_par%tag_aperture),cube_out,n_bytes_apert)/= &
          n_bytes_apert) then
        write(0,*)'tag_cig=',trim(focusmes_par%tag_aperture)
        call seperr('Problems with reading tag_aperture file')
      end if

    !}
    end do do_i_rad_curv_aperture
    !}
    end do do_i_rho_aperture

    !}
    case default oper_name_1
      write(0,*)'oper_name=',focusmes_par%oper_name(1:9)
      call seperr ('Do not know what to do with this opern_name')
    !}
    end select oper_name_1


    status=0
    return	

  end function FocusMes_oper

end module FocusMes_mod
