module FocusMes_transp_mod

  use focusmes_types
  use fft_types

  use fft_data_mod

  implicit none

  integer, external ::  sreed,srite,sseek_block

  integer, private :: n_z,n_mx,n_z_avg
  integer, private :: n_hx,n_gamma,n3_inp,n4_inp,n4_avg,n4_sembl,n_rho,n_dip
  integer, private :: n_scan_dip_0,n_rad_curv
  integer, private :: i_dip_flip,i_dip_split,i_dip_split_1,i_z_dip_split_1
  integer, private :: n_bytes_cube_inp,n_bytes_cube_out
  integer, private :: n_bytes_apert
  integer, private :: n_bytes_sum_sq,n_bytes_apert_dip,n_bytes_apert_apert
  integer, private :: n_bytes_apert_apert_dip,n_bytes_dips_0
  integer, private :: n_threads
  real, private :: d_z,d_mx,d_hx
  real, private :: d_gamma,d_gamma_rad,d_rho,d_dip,d_dip_rad,d_rad_curv
  real, private :: z_0,mx_0,hx_0,hz_0
  real, private :: gamma_0,gamma_0_rad,rho_0,dip_0,dip_0_rad,rad_curv_0
  real, private :: dip_0_curv,scan_dip_0_0,d_scan_dip_0
  real, private :: dip_flip,rad_curv_scale,inv_z_ref
  real, private :: dip_split,dip_split_1,z_dip_split_1
  real, private :: global_threshold,local_threshold,pi,gamma
  real, private :: epsi_den


  real, allocatable, dimension (:,:,:,:), private :: apert
  real, allocatable, dimension (:,:,:), private :: apert_apert,apert_dip,sum_sq
  real, allocatable, dimension (:,:,:,:), private :: apert_apert_dip
  real, allocatable, dimension (:,:,:,:), private :: cube_out
  real, allocatable, dimension (:,:,:,:), private :: cube_inp
  real, allocatable, dimension (:,:,:), private :: dips_0

contains

  function FocusMes_transp_init(focusmes_par)  result(status)
    type (focusmes_param_type) focusmes_par
    integer status

    integer ierr,i_thread

    status=0

    pi=acos(-1.)
    epsi_den=.0000000001
    local_threshold=focusmes_par%local_threshold
    global_threshold=focusmes_par%global_threshold

    inv_z_ref=1./focusmes_par%z_ref

    n_z=focusmes_par%n_z
    z_0=focusmes_par%z_0
    d_z=focusmes_par%d_z

    n_mx=focusmes_par%n_mx
    mx_0=focusmes_par%mx_0
    d_mx=focusmes_par%d_mx

    n_rho=focusmes_par%n_rho
    rho_0=focusmes_par%rho_0
    d_rho=focusmes_par%d_rho

    n_rad_curv=focusmes_par%n_rad_curv
    rad_curv_0=focusmes_par%rad_curv_0
    d_rad_curv=focusmes_par%d_rad_curv

    n_z_avg=focusmes_par%n_z_avg

    curvature_semblance_1 : if(focusmes_par%curvature_semblance) then
      dip_0_curv=focusmes_par%dip_0_curv*(pi/180.)

      if (focusmes_par%linear_curvature_scale) then
        n_rad_curv=focusmes_par%n_rad_curv
        rad_curv_0=focusmes_par%rad_curv_0
        d_rad_curv=focusmes_par%d_rad_curv
        rad_curv_scale=focusmes_par%rad_curv_scale
      else
        n_rad_curv=focusmes_par%n_rad_curv
        rad_curv_0=pi*focusmes_par%rad_curv_0
        d_rad_curv=pi*focusmes_par%d_rad_curv
        rad_curv_scale=focusmes_par%rad_curv_scale
      end if
    else curvature_semblance_1 
      n_rad_curv=focusmes_par%n_rad_curv
      rad_curv_0=pi*focusmes_par%rad_curv_0
      d_rad_curv=pi*focusmes_par%d_rad_curv
      rad_curv_scale=focusmes_par%rad_curv_scale
      dip_0_curv=focusmes_par%dip_0_curv*(pi/180.)
    end if curvature_semblance_1 

    if(focusmes_par%scan_dips_0) then
      n_scan_dip_0=focusmes_par%n_scan_dip_0
      scan_dip_0_0=focusmes_par%scan_dip_0_0
      d_scan_dip_0=focusmes_par%d_scan_dip_0
    else
      n_scan_dip_0=focusmes_par%n_scan_dip_0
      scan_dip_0_0=focusmes_par%scan_dip_0_0
      d_scan_dip_0=focusmes_par%d_scan_dip_0
    end if


    prestack_image_1: if(focusmes_par%prestack_image) then
    !{
      select case (focusmes_par%aperture_type(1:5))
      case('apert ')
        n_gamma=focusmes_par%n_gamma
        gamma_0=focusmes_par%gamma_0
        d_gamma=focusmes_par%d_gamma
        gamma_0_rad=gamma_0*(pi/180.)
        d_gamma_rad=d_gamma*(pi/180.)
        n3_inp=n_gamma
        n4_inp=1
        n4_avg=1
        n4_sembl=1
      case('dip  ')
        n_dip=focusmes_par%n_dip
        dip_0=focusmes_par%dip_0
        d_dip=focusmes_par%d_dip
        dip_0_rad=dip_0*(pi/180.)
        d_dip_rad=d_dip*(pi/180.)
        n3_inp=n_dip
        n4_inp=1
        n4_avg=1
        n4_sembl=1
  
      case('both ')
        n_gamma=focusmes_par%n_gamma
        gamma_0=focusmes_par%gamma_0
        d_gamma=focusmes_par%d_gamma
        gamma_0_rad=gamma_0*(pi/180.)
        d_gamma_rad=d_gamma*(pi/180.)
        n_dip=focusmes_par%n_dip
        dip_0=focusmes_par%dip_0
        d_dip=focusmes_par%d_dip
        dip_0_rad=dip_0*(pi/180.)
        d_dip_rad=d_dip*(pi/180.)
        n3_inp=n_gamma
        if(focusmes_par%only_apert_ang_sembl) then
          n4_inp=n_dip
          n4_avg=n_dip
          n4_sembl=1
        else
          n4_inp=n_dip
          n4_avg=1
          n4_sembl=n_dip
        end if
  
      case default
        write(0,*)'aperture_type=',focusmes_par%aperture_type(1:5)
        call seperr ('Do not know what to do with this opern_name')
      end select

      select case (focusmes_par%aperture_type(1:5))
      case('both ','dip  ')

        if(focusmes_par%flip_sign_dip < huge(focusmes_par%flip_sign_dip)) then
          dip_flip=focusmes_par%flip_sign_dip 
          if(d_dip <= 0.) call seperr ("d_dip must be >=0 when flip_sign_dip is set")
          if(abs(mod(sign((dip_flip-dip_0),d_dip),d_dip)) > .01*d_dip) then
                write(0,*)'mod(sign(dip_flip-dip_0,d_dip),d_dip)=', &
                        mod(sign(dip_flip-dip_0,d_dip),d_dip)
                call seperr ("dip_flip-dip_0 must be multiple of d_dip")
          end if
          i_dip_flip=(dip_flip-dip_0)/d_dip+1
          if((i_dip_flip <=1) .or. (i_dip_flip > n_dip)) then
            call seperr("Dip range must contain neg. dips when flip_sign_dip is set")
          end if
        else
          i_dip_flip=-1
        end if

  
        if(focusmes_par%split_dip < huge(focusmes_par%split_dip)) then
        !{

          dip_split=focusmes_par%split_dip 
          if(d_dip <= 0.) call seperr ("d_dip must be >=0 when split_dip is set")
          if(abs(mod(sign((dip_split-dip_0),d_dip),d_dip)) > .01*d_dip) then
                write(0,*)'mod(sign(dip_split-dip_0,d_dip),d_dip)=', &
                        mod(sign(dip_split-dip_0,d_dip),d_dip)
                call seperr ("dip_split-dip_0 must be multiple of d_dip")
          end if
          i_dip_split=(dip_split-dip_0)/d_dip+1
          if((i_dip_split <=1) .or. (i_dip_split > n_dip)) then
            call seperr("Dip range must contain neg. dips when split_dip is set")
          end if
          dip_split_1=focusmes_par%split_dip_1 
          if(dip_split_1 < huge(dip_split_1)) then
          !{
  
            if(abs(mod(sign((dip_split_1-dip_0),d_dip),d_dip))>.01*d_dip) then
                  write(0,*)'mod(sign(dip_split_1-dip_0,d_dip),d_dip)=', &
                          mod(sign(dip_split_1-dip_0,d_dip),d_dip)
                  call seperr ("dip_split_1-dip_0 must be multiple of d_dip")
            end if
  
            i_dip_split_1=(dip_split_1-dip_0)/d_dip+1
            if((i_dip_split_1 <=1) .or. (i_dip_split_1 > n_dip)) then
              call seperr("Dip range must contain neg. dips when split_dip_1 is set")
            end if
  
            z_dip_split_1=focusmes_par%z_split_dip_1 
            if(abs(mod(sign((z_dip_split_1-z_0),d_z),d_z))>.01*d_z) then
                  write(0,*)'mod(sign(z_dip_split_1-z_0,d_z),d_z)=', &
                          mod(sign(z_dip_split_1-z_0,d_z),d_z)
                  call seperr ("z_dip_split_1-z_0 must be multiple of d_z")
            end if

            i_z_dip_split_1=(z_dip_split_1-z_0)/d_z+1
            if((i_z_dip_split_1 <1) .or. (i_z_dip_split_1 >= n_z)) then
              call seperr("Depth range wrong split_dip_1 is set")
            end if
  
          !}
          else
          !{
            i_dip_split_1=-1
            i_z_dip_split_1=n_z+1
          !}
          end if
  
  
        !}
        else
        !{
          i_dip_split=-1
          i_dip_split_1=-1
          i_z_dip_split_1=n_z+1
        !}
        end if
  
        if(focusmes_par%verb) write(0,*)'i_dip_flip=',i_dip_flip
        if(focusmes_par%verb) write(0,*)'i_dip_split=',i_dip_split
        if(focusmes_par%verb) write(0,*)'i_dip_split_1=',i_dip_split_1
        if(focusmes_par%verb) write(0,*)'i_z_dip_split_1=',i_z_dip_split_1
        if((i_dip_split >0).and.(i_dip_flip >0)) &
              call seperr('Cannot flip sign and split dips at the same time')

      case default
        write(0,*)'No dip to flip'
      end select
    !}
    else prestack_image_1
    !{
      n3_inp=1
      n4_inp=1
      n4_avg=1
      n4_sembl=1
    !}
    end if prestack_image_1

    if(focusmes_par%verb) write(0,*)'# z_dip_split_1=',z_dip_split_1

    n_bytes_cube_inp=n_z*n_rho*n3_inp*n4_inp*4
    allocate (cube_inp(n_z,n_rho,n3_inp,n4_inp),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of cube_inp ")
    cube_inp=0.


    select case (focusmes_par%oper_name(1:9))
    !{
    case('semblance')
    !{
      select case (focusmes_par%aperture_type(1:5))
      case('apert ')
        n_bytes_apert_apert=n_z*n_rho*n_gamma*4
        allocate (apert_apert(n_z,n_rho,n_gamma),stat=ierr)  
        if(ierr /= 0) call seperr("Problems with allocation of apert_apert")

      case('dip  ')
        n_bytes_apert_dip=n_z*n_rho*n_dip*4
        allocate (apert_dip(n_z,n_rho,n_dip),stat=ierr)  
        if(ierr /= 0) call seperr("Problems with allocation of apert_dip ")

        read_dips_0_1: if(focusmes_par%read_dips_0) then
          n_bytes_dips_0=n_z*n_rho*4
          allocate (dips_0(n_z,n_rho,1),stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of dips_0 ")
        end if read_dips_0_1


      case('both ')
        n_bytes_apert_apert_dip=n_z*n_rho*n_gamma*n_dip*4
        allocate (apert_apert_dip(n_z,n_rho,n_gamma,n_dip),stat=ierr)  
        if(ierr/=0) call seperr("Problems with allocation of apert_apert_dip ")

        read_dips_0_2: if(focusmes_par%read_dips_0) then
          n_bytes_dips_0=n_z*n_rho*4
          allocate (dips_0(n_z,n_rho,1),stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of dips_0 ")
        end if read_dips_0_2

      case default
        write(0,*)'aperture_type=',focusmes_par%aperture_type(1:5)
        call seperr ('Do not know what to do with this opern_name')
      end select

      n_bytes_cube_out=n_z*n_rho*4
      allocate (cube_out(n_z,n_rho,1,1),stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of cube_out ")
      cube_out=0.

      if(focusmes_par%sum_squares_input) then
      !{
        n_bytes_sum_sq=n_z*n_rho*4
        allocate (sum_sq(n_z,n_rho,1),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of sum_sq ")
      !}
      end if

    !}
    case('aperture ')
    !{
      select case (focusmes_par%aperture_type(1:5))
      case('dip  ','both ')
        read_dips_0_3: if(focusmes_par%read_dips_0) then
          n_bytes_dips_0=n_z*n_rho*4
          allocate (dips_0(n_z,n_rho,1),stat=ierr)  
          if(ierr .ne. 0) call seperr("Problems with allocation of dips_0 ")
        end if read_dips_0_3
      case default
      end select

      if(focusmes_par%write_prestack) then
        n_bytes_cube_out=n_z*n_rho*n3_inp*n4_inp*4
        allocate (cube_out(n_z,n_rho,n3_inp,n4_inp),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of cube_out ")

        n_bytes_apert=n_z*n_rho*n3_inp*n4_inp*4
        allocate (apert(n_z,n_rho,n3_inp,n4_inp),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of apert ")

      end if
      cube_out=0.
    !}
    case default
      write(0,*)'oper_name=',focusmes_par%oper_name(1:9)
      call seperr ('Do not know what to do with this opern_name')
    !}
    end select

    n_threads=focusmes_par%n_threads

     if(focusmes_par%debug1) then
write(0,*)'inv_z_ref=',inv_z_ref
write(0,*)'n3_inp=',n3_inp
write(0,*)'n4_inp=',n4_inp
write(0,*)'n4_avg=',n4_avg
write(0,*)'n4_sembl=',n4_sembl
write(0,*)'n_rho=',n_rho
write(0,*)'rho_0=',rho_0
write(0,*)'d_rho=',d_rho
write(0,*)'n_rad_curv=',n_rad_curv
write(0,*)'rad_curv_0=',rad_curv_0
write(0,*)'d_rad_curv=',d_rad_curv
write(0,*)'n_gamma=',n_gamma
write(0,*)'gamma_0_rad=',gamma_0*(180./pi)
write(0,*)'d_gamma_rad=',d_gamma*(180./pi)
write(0,*)'n_scan_dip_0=',n_scan_dip_0
write(0,*)'d_scan_dip_0=',d_scan_dip_0
write(0,*)'scan_dip_0_0=',scan_dip_0_0
write(0,*)'n_dip=',n_dip
write(0,*)'dip_0=',dip_0_rad*(180./pi)
write(0,*)'d_dip=',d_dip_rad*(180./pi)
write(0,*)'n_hx=',n_hx
write(0,*)'hx_0=',hx_0
write(0,*)'d_hx=',d_hx
write(0,*)'n_z=',n_z
write(0,*)'n_mx=',n_mx
write(0,*)'d_z=',d_z
write(0,*)'d_mx=',d_mx
write(0,*)'n_z_avg=',n_z_avg
write(0,*)'n_bytes_cube_inp=',n_bytes_cube_inp
write(0,*)'n_bytes_cube_out=',n_bytes_cube_out
write(0,*)'n_bytes_apert_apert=',n_bytes_apert_apert
write(0,*)'n_bytes_apert_dip=',n_bytes_apert_dip
write(0,*)'n_bytes_dips_0=',n_bytes_dips_0
write(0,*)'shape(cube_inp)=',shape(cube_inp)
write(0,*)'shape(cube_out)=',shape(cube_out)
write(0,*)'shape(apert_apert)=',shape(apert_apert)
write(0,*)'shape(apert_dip)=',shape(apert_dip)
write(0,*)'shape(apert)=',shape(apert)
write(0,*)'shape(dips_0)=',shape(dips_0)
write(0,*)'local_threshold=',local_threshold
write(0,*)'global_threshold=',global_threshold
     end if


  end function FocusMes_transp_init

  function FocusMes_transp_oper(focusmes_par) result(status)

    type (focusmes_param_type) focusmes_par
    integer status

    logical in_aperture,in_aperture_global,split_pos,in_ang_range
    integer ret_sseek
    integer i_hx,i_mx,i_z,i_z_avg,i_mx_avg,i_thread,i_3,i_4,i_4_avg,i_4_sembl
    integer i_rho,i_rad_curv
    integer n_samples,i_rho_corr,i_scan_dip_0
    integer i_rho_1
    real sample_inp,sample_sq_inp,apert_apert_local,apert_dip_local
    real delta_rho,dip_rad,rad_curv,rho,rho_corr,w_rho_corr,gamma_rad,gamma,dip
!    real cos_dip_0_curv_sh,sin_dip_0_curv_sh
    double precision num,num_avg_neg,num_avg_pos,den,apert_width
    double precision global_max_image_mx,local_max_image_mx

    i_rho_1=nint((1.-rho_0)/d_rho)+1
    write(0,*)'i_rho_1=',i_rho_1

!    cos_dip_0_curv_sh=cos(dip_0_curv)
!    sin_dip_0_curv_sh=sin(dip_0_curv)

    oper_name_1: select case (focusmes_par%oper_name(1:9))
    !{
    case('semblance') oper_name_1
    !{

    do_i_mx_semblance: do i_mx=1,n_mx
    !{
      if(focusmes_par%verb) write(0,*)'i_mx=',i_mx

    do_i_rad_curv_semblance: do i_rad_curv=1,n_rad_curv
    !{

      if(focusmes_par%verb) write(0,*)'i_rad_curv=',i_rad_curv
      cube_out=0

      if(focusmes_par%curvature_image .or. (i_rad_curv ==1))  then
      !{

      if(focusmes_par%verb) write(0,*)'Read input cube'
      if(sreed(trim(focusmes_par%tag_cig),cube_inp,n_bytes_cube_inp)/= &
          n_bytes_cube_inp) then
        write(0,*)'tag_cig=',trim(focusmes_par%tag_cig)
        call seperr('Problems with reading tag_cig file')
      end if

      select case (focusmes_par%aperture_type(1:5))
      case('dip  ')
        if(i_dip_flip > 0) then
          cube_inp(:,:,1:i_dip_flip-1,:)=-cube_inp(:,:,1:i_dip_flip-1,:)
          cube_inp(:,:,i_dip_flip,:)=0.
        end if
        if((i_dip_split > 0) .and. focusmes_par%zero_split_dip) then
          if(i_z_dip_split_1 < n_z) then
            cube_inp(i_z_dip_split_1:n_z,:,i_dip_split_1,:)=0.
          else
            cube_inp(:,:,i_dip_split_1,:)=0.
          end if
        end if
      case('both ')
        if(i_dip_flip > 0) then
          cube_inp(:,:,:,1:i_dip_flip-1)=-cube_inp(:,:,:,1:i_dip_flip-1)
          cube_inp(:,:,:,i_dip_flip)=0.
        end if
        if((i_dip_split > 0) .and. focusmes_par%zero_split_dip) then
          if(i_z_dip_split_1 < n_z) then
            cube_inp(i_z_dip_split_1:n_z,:,:,i_dip_split_1)=0.
          else
            cube_inp(:,:,:,i_dip_split)=0.
          end if
        end if
      case default
      end select

      select case (focusmes_par%aperture_type(1:5))
      case('apert ')
        if(sreed(trim(focusmes_par%tag_apert_apert),apert_apert, &
            n_bytes_apert_apert)/= n_bytes_apert_apert) then
          write(0,*)'tag_apert_apert=',trim(focusmes_par%tag_apert_apert)
           call seperr('Problems with reading tag_apert_apert file')
        end if
      case('dip  ')
        if(sreed(trim(focusmes_par%tag_apert_dip),apert_dip, &
            n_bytes_apert_dip)/= n_bytes_apert_dip) then
          write(0,*)'tag_apert_dip=',trim(focusmes_par%tag_apert_dip)
            call seperr('Problems with reading tag_apert_dip file')
        end if
        read_dips_0_4: if(focusmes_par%read_dips_0) then
          if(sreed(trim(focusmes_par%tag_dips_0),dips_0, &
              n_bytes_dips_0)/= n_bytes_dips_0) then
            write(0,*)'tag_dips_0=',trim(focusmes_par%tag_dips_0)
              call seperr('Problems with reading tag_dips_0 file')
          end if
        end if read_dips_0_4
      case('both ')
        if(sreed(trim(focusmes_par%tag_apert_apert_dip),apert_apert_dip, &
            n_bytes_apert_apert_dip)/= n_bytes_apert_apert_dip) then
        write(0,*)'tag_apert_apert_dip=',trim(focusmes_par%tag_apert_apert_dip)
          call seperr('Problems with reading tag_apert_apert_dip file')
        end if
        read_dips_0_5: if(focusmes_par%read_dips_0) then
          if(sreed(trim(focusmes_par%tag_dips_0),dips_0, &
              n_bytes_dips_0)/= n_bytes_dips_0) then
            write(0,*)'tag_dips_0=',trim(focusmes_par%tag_dips_0)
              call seperr('Problems with reading tag_dips_0 file')
          end if
        end if read_dips_0_5
      case default
        write(0,*)'aperture_type=',focusmes_par%aperture_type(1:5)
        call seperr ('Do not know what to do with this opern_name')
      end select

      if(focusmes_par%sum_squares_input) then
        if(sreed(trim(focusmes_par%tag_sum_squares),sum_sq,n_bytes_sum_sq)/= &
              n_bytes_sum_sq) then
          write(0,*)'tag_sum_squares=',trim(focusmes_par%tag_sum_squares)
          call seperr('Problems with reading tag_sum_squares file')
        end if
      end if

      !}
      end if

      if (focusmes_par%linear_curvature_scale) then
        rad_curv=rad_curv_0+(i_rad_curv-1)*d_rad_curv
      else
        rad_curv=rad_curv_scale*tan(rad_curv_0+(i_rad_curv-1)*d_rad_curv)
      end if
      if(focusmes_par%verb) write(0,*)'i_rad_curv=',i_rad_curv
      if(focusmes_par%verb) write(0,*)'rad_curv=',rad_curv

      cube_out=0.
      do_i_rho: do i_rho=1,n_rho
      !{
        rho=rho_0+(i_rho-1)*d_rho
!      if(focusmes_par%verb) write(0,*)'i_rho=',i_rho
!      if(focusmes_par%verb) write(0,*)'rho=',rho
        do_i_z: do i_z=n_z_avg+1,n_z-n_z_avg
        !{
          select case (focusmes_par%aperture_type(1:5))
          case('apert')
            apert_apert_local=maxval(apert_apert(i_z,i_rho,:))
            apert_dip_local=1.
            in_aperture_global=(apert_apert_local /= 0.)
          case('dip  ')
            apert_apert_local=1.
            apert_dip_local=maxval(apert_dip(i_z,i_rho,:))
            in_aperture_global=(apert_dip_local /= 0.)
          case('both ')
            apert_apert_local=maxval(apert_apert_dip(i_z,i_rho,:,:))
            apert_dip_local=1.
            in_aperture_global=(apert_apert_local /= 0.)
          case('none ')
            in_aperture_global=.true.
          case default
            write(0,*)'aperture_type=',focusmes_par%aperture_type(1:5)
            call seperr ('Do not know what to do with this aperture_type')
          end select
!  if((i_mx == 1).and.(i_z == 35)) then
!  if((i_mx == 1)) then
!    write(0,*)'in_aperture_global =',in_aperture_global
!  end if
          in_aperture_global_1: if(in_aperture_global) then
          !{
            do_i_scan_dip_0: do i_scan_dip_0=1,n_scan_dip_0
            !{
              read_dips_0_6: if(focusmes_par%read_dips_0) then
                dip_0_curv= atan(dips_0(i_z,i_rho,1))
              else read_dips_0_6
                dip_0_curv= &
                  (scan_dip_0_0+(i_scan_dip_0-1)*d_scan_dip_0)*(pi/180.)
              end if read_dips_0_6
!  if((i_mx == n_mx/2).and.(i_z == 35)) then
!  if((i_mx == 35).and.(i_z == 35)) then
!  if((i_mx == 25).and.(i_z == 35).and.(i_rad_curv == n_rad_curv/2)) then
!    write(0,*)'i_z =',i_z
!    write(0,*)'rho =',rho
!!    write(0,*)'rad_curv (rel) =', (rad_curv_0+(i_rad_curv-1)*d_rad_curv)/pi
!!    write(0,*)'i_scan_dip_0 =',i_scan_dip_0
!    write(0,*)'dips_0 (raw) =',dips_0(i_z,i_rho,1)
!    write(0,*)'dip_0_curv (deg) =',dip_0_curv*180/pi
!  end if
!  
              num=0
              den=0
              do_i_z_avg: do i_z_avg=-n_z_avg,n_z_avg
              !{
                do_i_4_avg: do i_4_avg=1,n4_avg
              !{
!if((i_mx == 1)) then
!    write(0,*)'i_z =',i_z
!    write(0,*)'i_rho =',i_rho
!    write(0,*)'before num =',num
!    write(0,*)'before num_avg_neg =',num_avg_neg
!    write(0,*)'before num_avg_pos =',num_avg_pos
!end if
                num_avg_neg=0
                num_avg_pos=0
!if((i_mx == 1)) then
!    write(0,*)'num =',num
!    write(0,*)'num_avg_neg =',num_avg_neg
!    write(0,*)'num_avg_pos =',num_avg_pos
!end if
                do_i_4_sembl: do i_4_sembl=1,n4_sembl
                !{
                  i_4=(i_4_sembl-1)+(i_4_avg-1)+1
                  do_i_3: do i_3=1,n3_inp
                  !{

            aperture_type_1: select case (focusmes_par%aperture_type(1:5))
            case('apert') aperture_type_1
              in_aperture=(apert_apert(i_z+i_z_avg,i_rho,i_3) /=0 )
            case('dip  ') aperture_type_1
              in_aperture=(apert_dip(i_z+i_z_avg,i_rho,i_3) /=0 )

              curvature_image_1 : if(.not.focusmes_par%curvature_image) then
                dip_rad=dip_0_rad+(i_3-1)*d_dip_rad
                if (focusmes_par%global_straight_rays) then
                  delta_rho=((cos(dip_rad)*tan(dip_rad-dip_0_curv))**2)* &
                           0.5*rad_curv*inv_z_ref
                else
                  delta_rho=cos(dip_rad)*((tan(dip_rad-dip_0_curv))**2)* &
                           0.5*rad_curv*inv_z_ref
                end if
              else curvature_image_1
                delta_rho=0
              end if curvature_image_1

              if (i_z+i_z_avg <= i_z_dip_split_1) then
                split_pos=(i_3 >= i_dip_split)
              else
                split_pos=(i_3 >= i_dip_split_1)
              end if

            case('both ') aperture_type_1
              in_aperture=apert_apert_dip(i_z+i_z_avg,i_rho,i_3,i_4)/=0

              curvature_image_2 : if(.not.focusmes_par%curvature_image) then
                curvature_semblance_2 : if(focusmes_par%curvature_semblance) then
                  gamma_rad=gamma_0_rad+(i_3-1)*d_gamma_rad
                  dip_rad=dip_0_rad+(i_4-1)*d_dip_rad
                  if (focusmes_par%global_straight_rays) then
    
                    map_pseudo_depth_1 : if(focusmes_par%map_pseudo_depth) then
                      delta_rho=0.5*rad_curv*inv_z_ref
                    else map_pseudo_depth_1 
!                    delta_rho=(cos(dip_rad)**2-sin(gamma_rad)**2)* &
!                       (tan(dip_rad-dip_0_curv)**2)* &
!                             0.5*rad_curv*inv_z_ref
                      delta_rho=((cos(dip_rad)**2-sin(gamma_rad)**2)* &
                         tan(dip_rad-dip_0_curv)*sin(dip_rad-dip_0_curv)* &
                               0.5*rad_curv*inv_z_ref)/cos(dip_rad)
                    end if map_pseudo_depth_1 
    
!                delta_rho=((cos(dip_rad)**2-sin(gamma_rad)**2)/ &
!                delta_rho=((cos(dip_rad)**2-sin(gamma_rad)**2)/ &
!                           (sin(dip_rad)**2-sin(gamma_rad)**2)) * &
!                   (tan(dip_rad-dip_0_curv)**2)* &
!                         0.5*rad_curv*inv_z_ref
                  else
!                  delta_rho=cos(gamma_rad)**2*cos(dip_rad)* &
!                     (tan(dip_rad-dip_0_curv)**2)* &
!                           0.5*rad_curv*inv_z_ref
                    delta_rho=cos(gamma_rad)**2** &
                         tan(dip_rad-dip_0_curv)*sin(dip_rad-dip_0_curv)* &
                               0.5*rad_curv*inv_z_ref
                  end if
                else curvature_semblance_2
                    delta_rho=0.
                end if curvature_semblance_2
              else curvature_image_2
                delta_rho=0
              end if curvature_image_2
  
              if (i_z+i_z_avg <= i_z_dip_split_1) then
                split_pos=(i_4 >= i_dip_split)
              else
                split_pos=(i_4 >= i_dip_split_1)
              end if

            case('none ') aperture_type_1
              in_aperture= .true.
            case default aperture_type_1
              write(0,*)'aperture_type=',focusmes_par%aperture_type(1:5)
              call seperr ('Do not know what to do with this aperture_type')
            end select aperture_type_1


            aperture_type_2: select case (focusmes_par%aperture_type(1:5))
            case('apert') aperture_type_2
              sample_inp=cube_inp(i_z+i_z_avg,i_rho,i_3,1)
              split_pos=.true.
            case('dip  ','both ') aperture_type_2
              curvature_image_3 : if(.not.focusmes_par%curvature_image) then
                curvature_semblance_3 : if(focusmes_par%curvature_semblance) then
                  rho_corr=(rho+delta_rho-rho_0)/d_rho+1.
                  i_rho_corr=int(rho_corr)
                  w_rho_corr=(rho_corr-i_rho_corr)
                  if((i_rho_corr >=1) .and. (i_rho_corr <= n_rho-1)) then
                    sample_inp= &
                      cube_inp(i_z+i_z_avg,i_rho_corr  ,i_3,i_4)* &
                            (1.- w_rho_corr) + &
                      cube_inp(i_z+i_z_avg,i_rho_corr+1,i_3,i_4)* &
                            (    w_rho_corr)
                  else
                    if(n_rho == 1) then
                      sample_inp= cube_inp(i_z+i_z_avg,1,i_3,i_4)
                    else
                      in_aperture=.false.
                      sample_inp=0.
                    end if
                  end if
                else curvature_semblance_3
                  sample_inp= cube_inp(i_z+i_z_avg,i_rho,i_3,i_4)
                end if curvature_semblance_3
              else curvature_image_3
                sample_inp= cube_inp(i_z+i_z_avg,i_rho,i_3,i_4)
              end if curvature_image_3

            case('none ') aperture_type_2
              in_aperture= .true.
              sample_inp=cube_inp(i_z+i_z_avg,i_rho,i_3,1)
            case default aperture_type_2
              write(0,*)'aperture_type=',focusmes_par%aperture_type(1:5)
              call seperr ('Do not know what to do with this aperture_type')
            end select aperture_type_2

!  if((i_rho == i_rho_1)) then
!  if((i_mx == n_mx/2).and.(i_rho == i_rho_1)) then
!  if((i_mx == n_mx/2).and.(i_z == 35).and.(i_rho == i_rho_1)) then

!  if((i_mx == n_mx/2).and.(i_z == 35).and.(i_4 == n4_inp/2)) then
!  if((i_mx == 1).and.(i_z == 35).and.(i_4 == n4_inp/4).and. (i_z_avg == 0)) then
!    write(0,*)'sample_inp =',sample_inp
!    write(0,*)'in_aperture =',in_aperture
!    write(0,*)'dip =',dip_rad*180/pi
!    write(0,*)'gamma =',gamma_rad*180/pi
!    write(0,*)'rho =',rho,'  rho_corr (abs)=',rho+delta_rho
!    write(0,*)'i_rho_corr =',i_rho_corr,'  w_rho_corr=',w_rho_corr
!  end if


            in_aperture_1: if(in_aperture) then
              if(split_pos) then
                num_avg_pos=num_avg_pos+sample_inp
              else
                num_avg_neg=num_avg_neg+sample_inp
              end if
              den=den+sample_inp**2
            end if in_aperture_1

                  !}
                  end do do_i_3
                !}
                end do do_i_4_sembl
                num=num + num_avg_neg**2 + num_avg_pos**2
!                num=num + num_avg_neg + num_avg_pos
              !}
              end do do_i_4_avg
              !}
              end do do_i_z_avg
              if(focusmes_par%stack_power) then
!                cube_out(i_z,i_rho,1,1)=cube_out(i_z,i_rho,1,1)+num
                cube_out(i_z,i_rho,1,1)=cube_out(i_z,i_rho,1,1)+sqrt(num)
!                cube_out(i_z,i_rho,1,1)=cube_out(i_z,i_rho,1,1)+ &
!                         apert_apert_local*apert_dip_local*sqrt(num)
              else
                if(den > epsi_den) then
                  cube_out(i_z,i_rho,1,1)=cube_out(i_z,i_rho,1,1)+ &
                           apert_apert_local*apert_dip_local*num/den
                end if
              end if
            !}
            end do do_i_scan_dip_0
            cube_out(i_z,i_rho,1,1)=cube_out(i_z,i_rho,1,1)/n_scan_dip_0
          !}
          end if in_aperture_global_1
        !}
        end do do_i_z
      !}
      end do do_i_rho

      if(srite(trim(focusmes_par%tag_semblance),cube_out,n_bytes_cube_out)/= &
          n_bytes_cube_out) then
        write(0,*)'tag_cig=',trim(focusmes_par%tag_semblance)
        call seperr('Problems with reading tag_semblance file')
      end if

    !}
    end do do_i_rad_curv_semblance
    !}
    end do do_i_mx_semblance

    !}
    case('aperture ') oper_name_1
    !{

    do_i_mx_aperture: do i_mx=1,n_mx
    !{
      if(focusmes_par%verb) write(0,*)'i_mx=',i_mx
    do_i_rad_curv_aperture: do i_rad_curv=1,n_rad_curv
    !{
      if(focusmes_par%verb) write(0,*)'i_rad_curv=',i_rad_curv
      cube_out=0

      if(focusmes_par%curvature_image .or. (i_rad_curv ==1))  then
      !{

      if(focusmes_par%verb) write(0,*)'Read input cube'
        if(sreed(trim(focusmes_par%tag_cig),cube_inp,n_bytes_cube_inp)/= &
            n_bytes_cube_inp) then
          write(0,*)'tag_cig=',trim(focusmes_par%tag_cig)
          call seperr('Problems with reading tag_cig file')
        end if
  
        if(focusmes_par%dip_ang_image) then
          select case (focusmes_par%aperture_type(1:5))
          case('dip  ')
            read_dips_0_7: if(focusmes_par%read_dips_0) then
              if(sreed(trim(focusmes_par%tag_dips_0),dips_0, &
                  n_bytes_dips_0)/= n_bytes_dips_0) then
                write(0,*)'tag_dips_0=',trim(focusmes_par%tag_dips_0)
                  call seperr('Problems with reading tag_dips_0 file')
              end if
            end if read_dips_0_7
            if(i_dip_flip > 0) then
              cube_inp(:,:,1:i_dip_flip-1,:)=-cube_inp(:,:,1:i_dip_flip-1,:)
              cube_inp(:,:,i_dip_flip,:)=0.
            end if
            if((i_dip_split > 0) .and. focusmes_par%zero_split_dip) then
              if(i_z_dip_split_1 < n_z) then
                cube_inp(i_z_dip_split_1:n_z,:,i_dip_split_1,:)=0.
              else
                cube_inp(:,:,i_dip_split_1,:)=0.
              end if
            end if
          case('both ')
            read_dips_0_8: if(focusmes_par%read_dips_0) then
              if(sreed(trim(focusmes_par%tag_dips_0),dips_0, &
                  n_bytes_dips_0)/= n_bytes_dips_0) then
                write(0,*)'tag_dips_0=',trim(focusmes_par%tag_dips_0)
                  call seperr('Problems with reading tag_dips_0 file')
              end if
            end if read_dips_0_8
            if(i_dip_flip > 0) then
              cube_inp(:,:,:,1:i_dip_flip-1)=-cube_inp(:,:,:,1:i_dip_flip-1)
              cube_inp(:,:,:,i_dip_flip)=0.
            end if
            if((i_dip_split > 0) .and. focusmes_par%zero_split_dip) then
              if(i_z_dip_split_1 < n_z) then
  !              cube_inp(1:i_z_dip_split_1-1,:,:,i_dip_split)=0.
                cube_inp(i_z_dip_split_1:n_z,:,:,i_dip_split_1)=0.
              else
                cube_inp(:,:,:,i_dip_split)=0.
              end if
            end if
          case default
          end select
        end if
  
        do i_4=1,n4_inp
        !{
          do i_3=1,n3_inp
          !{
            gamma=gamma_0+(i_3-1)*d_gamma
  !      if(focusmes_par%verb) write(0,*)'gamma=',gamma
            do i_rho=1,n_rho
            !{
              do i_z=1,n_z
              !{
                read_dips_0_9: if(focusmes_par%read_dips_0) then
                  dip_0_curv= atan(dips_0(i_z,i_rho,1))*(180./pi)
                else read_dips_0_9
                  dip_0_curv= &
                    (scan_dip_0_0+(i_scan_dip_0-1)*d_scan_dip_0)
                end if read_dips_0_9
  
                if (focusmes_par%max_dip_prop > 0.) then
                  select case (focusmes_par%aperture_type(1:5))
                  case('apert')
                    if (abs(gamma) >= focusmes_par%max_dip_prop) then
                       cube_inp(i_z,i_rho,i_3,i_4)=0.
                    end if
                  case('dip  ')
                    dip=dip_0+(i_3-1)*d_dip
                    if ((abs(dip-dip_0_curv) >= focusmes_par%max_dip_range).or.&
                        (abs(dip) >= focusmes_par%max_dip_prop)) then
                       cube_inp(i_z,i_rho,i_3,i_4)=0.
                    end if
                  case('both ')
                    dip=dip_0+(i_4-1)*d_dip
                    if ((abs(dip-dip_0_curv) >= focusmes_par%max_dip_range).or.&
                        (abs(dip)+abs(gamma) >= focusmes_par%max_dip_prop)) then
                       cube_inp(i_z,i_rho,i_3,i_4)=0.
                    end if
                  case default
                  end select
                end if
  
               !}
               end do
             !}
             end do
           !}
           end do
         !}
         end do
  
        if(global_threshold /= 0.) then
          global_max_image_mx=maxval(abs(sum( sum(cube_inp,dim=4) ,dim=3)))
          global_max_image_mx=(global_max_image_mx**2)*(n_z_avg*2+1)
  !        write(0,*)'global_max_image_mx=',global_max_image_mx
        else
          global_max_image_mx=0.
        end if
  
      !}
      end if

!    do_i_rad_curv_aperture: do i_rad_curv=1,n_rad_curv
!    !
!      if(focusmes_par%verb) write(0,*)'i_rad_curv=',i_rad_curv

      do i_rho=1,n_rho
      !{
        do i_z=1,n_z
        !{
          local_max_image_mx=maxval(abs(cube_inp(i_z,i_rho,:,:)))

          read_dips_0_10: if(focusmes_par%read_dips_0) then
            dip_0_curv= atan(dips_0(i_z,i_rho,1))*(180./pi)
          else read_dips_0_10
            dip_0_curv= &
              (scan_dip_0_0+(i_scan_dip_0-1)*d_scan_dip_0)
          end if read_dips_0_10

           do i_4=1,n4_inp
           !{
             do i_3=1,n3_inp
             !{
               if (focusmes_par%max_dip_prop > 0.) then
                 select case (focusmes_par%aperture_type(1:5))
                 case('apert')
                   gamma=gamma_0+(i_3-1)*d_gamma
                   in_ang_range=(abs(gamma) < focusmes_par%max_dip_prop)
                 case('dip  ')
                   dip=dip_0+(i_3-1)*d_dip
                   in_ang_range= &
                     ((abs(dip-dip_0_curv) < focusmes_par%max_dip_range).and.&
                      (abs(dip) < focusmes_par%max_dip_prop))
                 case('both ')
                   gamma=gamma_0+(i_3-1)*d_gamma
                   dip=dip_0+(i_4-1)*d_dip
                   in_ang_range= &
                     ((abs(dip-dip_0_curv) < focusmes_par%max_dip_range).and.&
                      (abs(dip)+abs(gamma) < focusmes_par%max_dip_prop)) 
                 case default
                   in_ang_range=.true.
                 end select
               else
                 in_ang_range=.true.
               end if

               sample_inp=cube_inp(i_z,i_rho,i_3,i_4)
               if((abs(sample_inp) > local_max_image_mx*local_threshold).and. &
                   in_ang_range) then
                 apert(i_z,i_rho,i_3,i_4)=1.
               else
                 apert(i_z,i_rho,i_3,i_4)=0.
               end if
            !}
            end do
          !}
          end do
        !}
        end do
      !}
      end do
!        write(0,*)'after apert'

      do i_rho=1,n_rho
      !{
        do i_z=n_z_avg+1,n_z-n_z_avg
        !{
          num=0
          den=0
          do i_z_avg=-n_z_avg,n_z_avg
          !{
            num_avg_neg=0.
            num_avg_pos=0.
            do i_4=1,n4_inp
            !{
              do i_3=1,n3_inp
              !{
                select case (focusmes_par%aperture_type(1:5))
                case('apert')
                  split_pos=.true.
                case('dip  ')
                  if (i_z+i_z_avg <= i_z_dip_split_1) then
                    split_pos=(i_3 >= i_dip_split)
                  else
                    split_pos=(i_3 >= i_dip_split_1)
                  end if
                case('both ')
                  if (i_z+i_z_avg <= i_z_dip_split_1) then
                    split_pos=(i_4 >= i_dip_split)
                  else
                    split_pos=(i_4 >= i_dip_split_1)
                  end if
                case('none ')
                  in_aperture= .true.
                case default
                end select

                if(apert(i_z+i_z_avg,i_rho,i_3,i_4) == 1) then
                  sample_inp=cube_inp(i_z+i_z_avg,i_rho,i_3,i_4)
                  if(split_pos) then
                    num_avg_pos=num_avg_pos+sample_inp
                  else
                    num_avg_neg=num_avg_neg+sample_inp
                  end if
                  den=den + sample_inp**2
                end if
              !}
              end do
            !}
            end do
            num=num + num_avg_neg**2 + num_avg_pos**2
          !}
          end do

          if((den > tiny(den)*100000000000000000.) .and. &
             (den > global_max_image_mx*global_threshold)) then
!             (num > global_max_image_mx*global_threshold)) then
!            apert_width= sum(apert(i_z,i_mx,:))
            apert_width=0.
            do i_4=1,n4_inp
              do i_3=1,n3_inp
                do i_z_avg=-n_z_avg,n_z_avg
           apert_width=apert_width+apert(i_z+i_z_avg,i_rho,i_3,i_4)
                end do
              end do
            end do
            if(apert_width > 0.) then
              cube_out(i_z,i_rho,:,:)= apert(i_z,i_rho,:,:)/apert_width
            else
              cube_out(i_z,i_rho,:,:)=0.
            end if
          else
              cube_out(i_z,i_rho,:,:)=0.
          end if

        !}
        end do
      !}
      end do

      if(srite(trim(focusmes_par%tag_aperture),cube_out,n_bytes_apert)/= &
          n_bytes_apert) then
        write(0,*)'tag_cig=',trim(focusmes_par%tag_aperture)
        call seperr('Problems with reading tag_aperture file')
      end if

    !}
    end do do_i_rad_curv_aperture
    !}
    end do do_i_mx_aperture

    !}
    case default oper_name_1
      write(0,*)'oper_name=',focusmes_par%oper_name(1:9)
      call seperr ('Do not know what to do with this opern_name')
    !}
    end select oper_name_1


    status=0
    return	

  end function FocusMes_transp_oper

end module FocusMes_transp_mod
