subroutine param_focusmes(focusmes_par)
  use focusmes_types
  use sep

  implicit none
  type (focusmes_param_type) focusmes_par

  integer ::  omp_get_num_threads
  integer, external :: auxpar,getch,fetch,auxin,auxinout,sreed,sseek,auxout

  character(len=128) :: tag_name
  integer n_threads_env,n_threads_init
  integer n_z,n_mx,n_gamma,n_rho,n_rad_curv,n_dip,n_z_avg,n_mx_avg
  integer esize,zero_int,one_int,n4
  real z_0,mx_0,gamma_0,rho_0,rad_curv_0,dip_0
  real d_z,d_mx,d_gamma,d_rho,d_rad_curv,d_dip
  real zero_real,one_real

  n_threads_init=-1

  zero_int=0
  zero_real=0.
  one_int=1
  one_real=1.

  call from_param('verb',focusmes_par%verb,.false.)
  call from_param('debug1',focusmes_par%debug1,.false.)
  call from_param('debug2',focusmes_par%debug2,.false.)


  call from_param('apert_ang_image',focusmes_par%apert_ang_image,.true.)
  call from_param('dip_ang_image',focusmes_par%dip_ang_image,.true.)
  call from_param('curvature_image',focusmes_par%curvature_image,.false.)
  focusmes_par%sum_squares_input=.false.

  call from_param('semblance',focusmes_par%semblance,.false.)
  call from_param('kurtosis',focusmes_par%kurtosis,.false.)

  call from_param('n_threads',focusmes_par%n_threads,n_threads_init)
  call from_param('max_memory',focusmes_par%max_memory,100)

  ! checking parameters
!$omp parallel
  n_threads_env=omp_get_num_threads()
!$omp end parallel
  if(focusmes_par%n_threads == n_threads_init) then
     focusmes_par%n_threads=n_threads_env
  else
     if(n_threads_env > focusmes_par%n_threads) then
        call omp_set_num_threads(focusmes_par%n_threads)
        write(0,*)'WARNING: overwriting environment variable OMP_NUM_THREADS'
        write(0,*)'focusmes_par-n_threads=',focusmes_par%n_threads
     end if
     if(n_threads_env < focusmes_par%n_threads) then
        write(0,*)'focusmes_par-n_threads=',focusmes_par%n_threads
        write(0,*)'OMP_NUM_THREADS=',n_threads_env
        call seperr('ERROR: Asked for more threads than the environment variable allows')
     end if
  end if
  if(focusmes_par%debug1)write(0,*)'focusmes_par-n_threads=',focusmes_par%n_threads
  call to_history('n_threads',focusmes_par%n_threads)

   if(focusmes_par%n_threads /= 1) then
     call seperr("At the present n_threads must be == 1")
   end if

! Cig is an input

   call from_param('tag_cig',focusmes_par%tag_cig,'Cig')
   call from_param(trim(focusmes_par%tag_cig),tag_name,'Cig not found')
   focusmes_par%cig=auxin(trim(focusmes_par%tag_cig)) > 0
   if(.not. focusmes_par%cig) call seperr("Cig missing")

   call from_aux(trim(focusmes_par%tag_cig),'n4',n4)
   focusmes_par%prestack_image=(n4 /= 1) 
   call from_param('prestack_image',focusmes_par%prestack_image, &
        focusmes_par%prestack_image)
   call from_param('write_prestack',focusmes_par%write_prestack, &
        focusmes_par%prestack_image)
   if(focusmes_par%verb) write(0,*)'prestack_image=',focusmes_par%prestack_image
   if(focusmes_par%verb) write(0,*)'write_prestack=',focusmes_par%write_prestack

   call from_aux(trim(focusmes_par%tag_cig),'n1',focusmes_par%n_z)
   call from_aux(trim(focusmes_par%tag_cig),'d1',focusmes_par%d_z)
   call from_aux(trim(focusmes_par%tag_cig),'o1',focusmes_par%z_0)
   call from_aux(trim(focusmes_par%tag_cig),'label1',focusmes_par%label_z, &
              trim(focusmes_par%label_z))

   call to_history("From Cig: n1",focusmes_par%n_z)
   call to_history("From Cig: d1",focusmes_par%d_z)
   call to_history("From Cig: o1",focusmes_par%z_0)

   call from_aux(trim(focusmes_par%tag_cig),'n2',focusmes_par%n_mx)
   call from_aux(trim(focusmes_par%tag_cig),'d2',focusmes_par%d_mx)
   call from_aux(trim(focusmes_par%tag_cig),'o2',focusmes_par%mx_0)
   call from_aux(trim(focusmes_par%tag_cig),'label2',focusmes_par%label_mx, &
              trim(focusmes_par%label_mx))

   call to_history("From Cig: n2",focusmes_par%n_mx)
   call to_history("From Cig: d2",focusmes_par%d_mx)
   call to_history("From Cig: o2",focusmes_par%mx_0)

   call from_param('flip_sign_dip',focusmes_par%flip_sign_dip, &
        huge(focusmes_par%flip_sign_dip))
   call from_param('split_dip',focusmes_par%split_dip, &
        huge(focusmes_par%split_dip))
   call from_param('split_dip_1',focusmes_par%split_dip_1, &
        huge(focusmes_par%split_dip_1))
   call from_param('z_split_dip_1',focusmes_par%z_split_dip_1, &
        huge(focusmes_par%z_split_dip_1))
   call from_param('zero_split_dip',focusmes_par%zero_split_dip,.false.)

   if (.not. focusmes_par%curvature_image) then
     call from_param('n_rad_curv',focusmes_par%n_rad_curv,1)
     call from_param('rad_curv_0',focusmes_par%rad_curv_0,0.)
     call from_param('d_rad_curv',focusmes_par%d_rad_curv,1.)
     call from_param('rad_curv_scale',focusmes_par%rad_curv_scale,1.)
     call from_param('dip_0_curv',focusmes_par%dip_0_curv,0.)
   else
     focusmes_par%n_rad_curv=1
     focusmes_par%d_rad_curv=1.
     focusmes_par%rad_curv_0=0.
   end if

   prestack_image_1: if(focusmes_par%prestack_image) then
   !{
     dip_ang_image_1: if(focusmes_par%dip_ang_image) then
     !{
       apert_ang_image_1: if(focusmes_par%apert_ang_image) then
       !{
!          call seperr("dip_ang_image AND apert_ang_image not suported yet")
         call from_aux(trim(focusmes_par%tag_cig),'n3',focusmes_par%n_gamma)
         call from_aux(trim(focusmes_par%tag_cig),'d3',focusmes_par%d_gamma)
         call from_aux(trim(focusmes_par%tag_cig),'o3',focusmes_par%gamma_0)
         call from_aux(trim(focusmes_par%tag_cig),'label3', &
              focusmes_par%label_gamma,trim(focusmes_par%label_gamma))
    
         call to_history("From Cig: n3",focusmes_par%n_gamma)
         call to_history("From Cig: d3",focusmes_par%d_gamma)
         call to_history("From Cig: o3",focusmes_par%gamma_0)

         call from_aux(trim(focusmes_par%tag_cig),'n4',focusmes_par%n_dip)
         call from_aux(trim(focusmes_par%tag_cig),'d4',focusmes_par%d_dip)
         call from_aux(trim(focusmes_par%tag_cig),'o4',focusmes_par%dip_0)
         call from_aux(trim(focusmes_par%tag_cig),'label4', &
              focusmes_par%label_dip,trim(focusmes_par%label_dip))
    
         call to_history("From Cig: n4",focusmes_par%n_dip)
         call to_history("From Cig: d4",focusmes_par%d_dip)
         call to_history("From Cig: o4",focusmes_par%dip_0)
  
         call from_aux(trim(focusmes_par%tag_cig),'n5',focusmes_par%n_rho)
         call from_aux(trim(focusmes_par%tag_cig),'d5',focusmes_par%d_rho)
         call from_aux(trim(focusmes_par%tag_cig),'o5',focusmes_par%rho_0)
         call from_aux(trim(focusmes_par%tag_cig),'label5', &
              focusmes_par%label_rho,trim(focusmes_par%label_rho))
    
         call to_history("From Cig: n5",focusmes_par%n_rho)
         call to_history("From Cig: d5",focusmes_par%d_rho)
         call to_history("From Cig: o5",focusmes_par%rho_0)
       !}
       else apert_ang_image_1
       !{

         call from_aux(trim(focusmes_par%tag_cig),'n3',focusmes_par%n_dip)
         call from_aux(trim(focusmes_par%tag_cig),'d3',focusmes_par%d_dip)
         call from_aux(trim(focusmes_par%tag_cig),'o3',focusmes_par%dip_0)
         call from_aux(trim(focusmes_par%tag_cig),'label3', &
              focusmes_par%label_dip,trim(focusmes_par%label_dip))
    
         call to_history("From Cig: n3",focusmes_par%n_dip)
         call to_history("From Cig: d3",focusmes_par%d_dip)
         call to_history("From Cig: o3",focusmes_par%dip_0)

         curvature_image_1: if (focusmes_par%curvature_image) then
         !{
           call from_aux(trim(focusmes_par%tag_cig),'n4', &
                 focusmes_par%n_rad_curv)
           call from_aux(trim(focusmes_par%tag_cig),'d4', &
                 focusmes_par%d_rad_curv)
           call from_aux(trim(focusmes_par%tag_cig),'o4', &
                 focusmes_par%rad_curv_0)
           call from_aux(trim(focusmes_par%tag_cig),'label4', &
            focusmes_par%label_rad_curv,trim(focusmes_par%label_rad_curv))

           call from_aux(trim(focusmes_par%tag_cig),'n5',focusmes_par%n_rho)
           call from_aux(trim(focusmes_par%tag_cig),'d5',focusmes_par%d_rho)
           call from_aux(trim(focusmes_par%tag_cig),'o5',focusmes_par%rho_0)
           call from_aux(trim(focusmes_par%tag_cig),'label5', &
            focusmes_par%label_rho,trim(focusmes_par%label_rho))

           call to_history("From Cig: n4",focusmes_par%n_rad_curv)
           call to_history("From Cig: d4",focusmes_par%d_rad_curv)
           call to_history("From Cig: o4",focusmes_par%rad_curv_0)
      
           call to_history("From Cig: n5",focusmes_par%n_rho)
           call to_history("From Cig: d5",focusmes_par%d_rho)
           call to_history("From Cig: o5",focusmes_par%rho_0)
         !}
         else curvature_image_1
         !{
           call from_aux(trim(focusmes_par%tag_cig),'n4',focusmes_par%n_rho)
           call from_aux(trim(focusmes_par%tag_cig),'d4',focusmes_par%d_rho)
           call from_aux(trim(focusmes_par%tag_cig),'o4',focusmes_par%rho_0)
           call from_aux(trim(focusmes_par%tag_cig),'label4', &
            focusmes_par%label_rho,trim(focusmes_par%label_rho))
           call to_history("From Cig: n4",focusmes_par%n_rho)
           call to_history("From Cig: d4",focusmes_par%d_rho)
           call to_history("From Cig: o4",focusmes_par%rho_0)
         !}
         end if curvature_image_1
       !}
       end if apert_ang_image_1

     !}
     else dip_ang_image_1
     !{
       apert_ang_image_2: if(focusmes_par%apert_ang_image) then
         call from_aux(trim(focusmes_par%tag_cig),'n3',focusmes_par%n_gamma)
         call from_aux(trim(focusmes_par%tag_cig),'d3',focusmes_par%d_gamma)
         call from_aux(trim(focusmes_par%tag_cig),'o3',focusmes_par%gamma_0)
         call from_aux(trim(focusmes_par%tag_cig),'label3', &
              focusmes_par%label_gamma,trim(focusmes_par%label_gamma))
    
         call to_history("From Cig: n3",focusmes_par%n_gamma)
         call to_history("From Cig: d3",focusmes_par%d_gamma)
         call to_history("From Cig: o3",focusmes_par%gamma_0)
       else apert_ang_image_2
         call from_aux(trim(focusmes_par%tag_cig),'n3',focusmes_par%n_hx)
         call from_aux(trim(focusmes_par%tag_cig),'d3',focusmes_par%d_hx)
         call from_aux(trim(focusmes_par%tag_cig),'o3',focusmes_par%hx_0)
         call from_aux(trim(focusmes_par%tag_cig),'label3', &
              focusmes_par%label_hx,trim(focusmes_par%label_hx))
    
         call to_history("From Cig: n3",focusmes_par%n_hx)
         call to_history("From Cig: d3",focusmes_par%d_hx)
         call to_history("From Cig: o3",focusmes_par%hx_0)
       end if apert_ang_image_2
       call from_aux(trim(focusmes_par%tag_cig),'n4',focusmes_par%n_rho)
       call from_aux(trim(focusmes_par%tag_cig),'d4',focusmes_par%d_rho)
       call from_aux(trim(focusmes_par%tag_cig),'o4',focusmes_par%rho_0)
       call from_aux(trim(focusmes_par%tag_cig),'label4', &
          focusmes_par%label_rho,trim(focusmes_par%label_rho))
  
       call to_history("From Cig: n4",focusmes_par%n_rho)
       call to_history("From Cig: d4",focusmes_par%d_rho)
       call to_history("From Cig: o4",focusmes_par%rho_0)
     !}
     end if dip_ang_image_1
   !}
   else prestack_image_1
   !{
!     if(focusmes_par%dip_ang_image) then
!     !{
!       call from_aux(trim(focusmes_par%tag_cig),'n3',focusmes_par%n_dip)
!       call from_aux(trim(focusmes_par%tag_cig),'d3',focusmes_par%d_dip)
!       call from_aux(trim(focusmes_par%tag_cig),'o3',focusmes_par%dip_0)
!  
!       call to_history("From Cig: n3",focusmes_par%n_dip)
!       call to_history("From Cig: d3",focusmes_par%d_dip)
!       call to_history("From Cig: o3",focusmes_par%dip_0)
!
!       call from_aux(trim(focusmes_par%tag_cig),'n4',focusmes_par%n_rho)
!       call from_aux(trim(focusmes_par%tag_cig),'d4',focusmes_par%d_rho)
!       call from_aux(trim(focusmes_par%tag_cig),'o4',focusmes_par%rho_0)
!  
!       call to_history("From Cig: n4",focusmes_par%n_rho)
!       call to_history("From Cig: d4",focusmes_par%d_rho)
!       call to_history("From Cig: o4",focusmes_par%rho_0)
!     !}
!     else
!     !{
       call from_aux(trim(focusmes_par%tag_cig),'n3',focusmes_par%n_rho)
       call from_aux(trim(focusmes_par%tag_cig),'d3',focusmes_par%d_rho)
       call from_aux(trim(focusmes_par%tag_cig),'o3',focusmes_par%rho_0)
       call from_aux(trim(focusmes_par%tag_cig),'label3', &
          focusmes_par%label_rho,trim(focusmes_par%label_rho))
  
       call to_history("From Cig: n3",focusmes_par%n_rho)
       call to_history("From Cig: d3",focusmes_par%d_rho)
       call to_history("From Cig: o3",focusmes_par%rho_0)

       call from_aux(trim(focusmes_par%tag_cig),'n4',focusmes_par%n_rad_curv)
       call from_aux(trim(focusmes_par%tag_cig),'d4',focusmes_par%d_rad_curv)
       call from_aux(trim(focusmes_par%tag_cig),'o4',focusmes_par%rad_curv_0)
       call from_aux(trim(focusmes_par%tag_cig),'label4', &
          focusmes_par%label_rad_curv,trim(focusmes_par%label_rad_curv))

       call to_history("From Cig: n4",focusmes_par%n_rad_curv)
       call to_history("From Cig: d4",focusmes_par%d_rad_curv)
       call to_history("From Cig: o4",focusmes_par%rad_curv_0)
!     !}
!     end if
   !}
   end if prestack_image_1

   if(focusmes_par%semblance) then
     focusmes_par%oper_name(1:9)="semblance"
   else 
     if(focusmes_par%kurtosis) then
       focusmes_par%oper_name(1:9)="kurtosis "
     else
       focusmes_par%oper_name(1:9)="aperture "
     end if
   end if

  write(0,*)'oper_name=',focusmes_par%oper_name(1:9)
             
  
  oper_name_1: select case (focusmes_par%oper_name(1:9))
  !{
  case('semblance') oper_name_1
  !{

     call from_param('norm_semblance',focusmes_par%norm_semblance,.true.)

    if(focusmes_par%dip_ang_image) then
      focusmes_par%aperture_type="dip  "
    else
      if(focusmes_par%apert_ang_image) then
        focusmes_par%aperture_type="apert"
      else
        focusmes_par%aperture_type="none "
      end if
    end if
    if((focusmes_par%dip_ang_image).and.(focusmes_par%apert_ang_image)) then
      focusmes_par%aperture_type="both "
    end if
    if(focusmes_par%verb) write(0,*)'aperture_type=',focusmes_par%aperture_type

    aperture_type_1: select case (focusmes_par%aperture_type(1:5))
    !{
    case('dip  ') aperture_type_1
    !{
      call from_param('tag_apert_dip',focusmes_par%tag_apert_dip,'Apert_dip')
      call from_param(trim(focusmes_par%tag_apert_dip),tag_name,'Apert_dip not found')
      focusmes_par%cig=auxin(trim(focusmes_par%tag_apert_dip)) > 0
      if(.not. focusmes_par%cig) call seperr("Apert_dip missing")
  
      call from_aux(trim(focusmes_par%tag_apert_dip),'n1',n_z)
      call from_aux(trim(focusmes_par%tag_apert_dip),'d1',d_z)
      call from_aux(trim(focusmes_par%tag_apert_dip),'o1',z_0)
  
      call from_aux(trim(focusmes_par%tag_apert_dip),'n2',n_mx)
      call from_aux(trim(focusmes_par%tag_apert_dip),'d2',d_mx)
      call from_aux(trim(focusmes_par%tag_apert_dip),'o2',mx_0)

      call from_aux(trim(focusmes_par%tag_apert_dip),'n3',n_dip)
      call from_aux(trim(focusmes_par%tag_apert_dip),'d3',d_dip)
      call from_aux(trim(focusmes_par%tag_apert_dip),'o3',dip_0)

      curvature_image_2: if (focusmes_par%curvature_image) then
      !{

        call from_aux(trim(focusmes_par%tag_apert_dip),'n4',n_rad_curv)
        call from_aux(trim(focusmes_par%tag_apert_dip),'d4',d_rad_curv)
        call from_aux(trim(focusmes_par%tag_apert_dip),'o4',rad_curv_0)

        if(n_rad_curv /= focusmes_par%n_rad_curv) &
           call seperr("Apert_dip n4 /= Cig n4")
        if(d_rad_curv /= focusmes_par%d_rad_curv) &
           call seperr("Apert_dip d4 /= Cig d4")
        if(rad_curv_0 /= focusmes_par%rad_curv_0) &
           call seperr("Apert_dip o4 /= Cig o4")

        call from_aux(trim(focusmes_par%tag_apert_dip),'n5',n_rho)
        call from_aux(trim(focusmes_par%tag_apert_dip),'d5',d_rho)
        call from_aux(trim(focusmes_par%tag_apert_dip),'o5',rho_0)

        if(n_rho /= focusmes_par%n_rho) call seperr("Apert_dip n5 /= Cig n5")
        if(d_rho /= focusmes_par%d_rho) call seperr("Apert_dip d5 /= Cig d5")
        if(rho_0 /= focusmes_par%rho_0) call seperr("Apert_dip o5 /= Cig o5")
      !}
      else curvature_image_2
      !{
        call from_aux(trim(focusmes_par%tag_apert_dip),'n4',n_rho)
        call from_aux(trim(focusmes_par%tag_apert_dip),'d4',d_rho)
        call from_aux(trim(focusmes_par%tag_apert_dip),'o4',rho_0)

        if(n_rho /= focusmes_par%n_rho) call seperr("Apert_dip n4 /= Cig n4")
        if(d_rho /= focusmes_par%d_rho) call seperr("Apert_dip d4 /= Cig d4")
        if(rho_0 /= focusmes_par%rho_0) call seperr("Apert_dip o4 /= Cig o4")
      !}
      end if curvature_image_2
  
  
      if(n_z /= focusmes_par%n_z) call seperr("Apert_dip n1 /= Cig n1")
      if(d_z /= focusmes_par%d_z) call seperr("Apert_dip d1 /= Cig d1")
      if(z_0 /= focusmes_par%z_0) call seperr("Apert_dip o1 /= Cig o1")
  
      if(n_mx /= focusmes_par%n_mx) call seperr("Apert_dip n2 /= Cig n2")
      if(d_mx /= focusmes_par%d_mx) call seperr("Apert_dip d2 /= Cig d2")
      if(mx_0 /= focusmes_par%mx_0) call seperr("Apert_dip o2 /= Cig o2")

      if(n_dip /= focusmes_par%n_dip) call seperr("Apert_dip n3 /= Cig n3")
      if(d_dip /= focusmes_par%d_dip) call seperr("Apert_dip d3 /= Cig d3")
      if(dip_0 /= focusmes_par%dip_0) call seperr("Apert_dip o3 /= Cig o3")

  
      call from_aux(trim(focusmes_par%tag_apert_dip),'n_z_avg',n_z_avg)
      call from_aux(trim(focusmes_par%tag_apert_dip),'n_mx_avg',n_mx_avg)
  
      call to_history("From Apert_dip: n_z_avg",n_z_avg)
      call to_history("From Apert_dip: n_mx_avg",n_mx_avg)
  
      call from_param('n_z_avg',focusmes_par%n_z_avg,n_z_avg)
      call from_param('n_mx_avg',focusmes_par%n_mx_avg,n_mx_avg)
  
      if(n_z_avg /= focusmes_par%n_z_avg)  &
        write(0,*)'WARNING: Apert_dip n_z_avg /= parameter n_z_avg'
      if(n_mx_avg /= focusmes_par%n_mx_avg)  &
        write(0,*)'WARNING: Apert_dip n_mx_avg /= parameter n_mx_avg'


      call from_param('tag_sum_squares',focusmes_par%tag_sum_squares, &
           'Sum_squares')
      call from_param(trim(focusmes_par%tag_sum_squares),tag_name, &
           'Sum_squares not found')
      focusmes_par%sum_squares_input= &
          auxin(trim(focusmes_par%tag_sum_squares)) > 0
      if(focusmes_par%verb) write(0,*)'sum_squares_input=', &
          focusmes_par%sum_squares_input
      if(focusmes_par%sum_squares_input) then
      !{
        call from_aux(trim(focusmes_par%tag_sum_squares),'n1',n_z)
        call from_aux(trim(focusmes_par%tag_sum_squares),'d1',d_z)
        call from_aux(trim(focusmes_par%tag_sum_squares),'o1',z_0)
    
        call from_aux(trim(focusmes_par%tag_sum_squares),'n2',n_mx)
        call from_aux(trim(focusmes_par%tag_sum_squares),'d2',d_mx)
        call from_aux(trim(focusmes_par%tag_sum_squares),'o2',mx_0)

        call from_aux(trim(focusmes_par%tag_sum_squares),'n3',n_dip)
        call from_aux(trim(focusmes_par%tag_sum_squares),'d3',d_dip)
        call from_aux(trim(focusmes_par%tag_sum_squares),'o3',dip_0)
    
        call from_aux(trim(focusmes_par%tag_sum_squares),'n4',n_rho)
        call from_aux(trim(focusmes_par%tag_sum_squares),'d4',d_rho)
        call from_aux(trim(focusmes_par%tag_sum_squares),'o4',rho_0)

        if(n_z /= focusmes_par%n_z) call seperr("Sum_squares n1 /= Cig n1")
        if(d_z /= focusmes_par%d_z) call seperr("Sum_squares d1 /= Cig d1")
        if(z_0 /= focusmes_par%z_0) call seperr("Sum_squares o1 /= Cig o1")
    
        if(n_mx /= focusmes_par%n_mx) call seperr("Sum_squares n2 /= Cig n2")
        if(d_mx /= focusmes_par%d_mx) call seperr("Sum_squares d2 /= Cig d2")
        if(mx_0 /= focusmes_par%mx_0) call seperr("Sum_squares o2 /= Cig o2")

        if(n_dip /= focusmes_par%n_dip) call seperr("Sum_squares n3 /= Cig n3")
        if(d_dip /= focusmes_par%d_dip) call seperr("Sum_squares d3 /= Cig d3")
        if(dip_0 /= focusmes_par%dip_0) call seperr("Sum_squares o3 /= Cig o3")

        if(n_rho /= focusmes_par%n_rho) call seperr("Sum_squares n4 /= Cig n4")
        if(d_rho /= focusmes_par%d_rho) call seperr("Sum_squares d4 /= Cig d4")
        if(rho_0 /= focusmes_par%rho_0) call seperr("Sum_squares o4 /= Cig o4")

      !}
      end if
    !}
    case('apert') aperture_type_1
    !{

      call from_param('tag_apert_apert',focusmes_par%tag_apert_apert,'Apert_apert')
      call from_param(trim(focusmes_par%tag_apert_apert),tag_name,'Apert_apert not found')
      focusmes_par%cig=auxin(trim(focusmes_par%tag_apert_apert)) > 0
      if(.not. focusmes_par%cig) call seperr("Apert_apert missing")
  
      call from_aux(trim(focusmes_par%tag_apert_apert),'n1',n_z)
      call from_aux(trim(focusmes_par%tag_apert_apert),'d1',d_z)
      call from_aux(trim(focusmes_par%tag_apert_apert),'o1',z_0)
  
      call from_aux(trim(focusmes_par%tag_apert_apert),'n2',n_mx)
      call from_aux(trim(focusmes_par%tag_apert_apert),'d2',d_mx)
      call from_aux(trim(focusmes_par%tag_apert_apert),'o2',mx_0)

      call from_aux(trim(focusmes_par%tag_apert_apert),'n3',n_gamma)
      call from_aux(trim(focusmes_par%tag_apert_apert),'d3',d_gamma)
      call from_aux(trim(focusmes_par%tag_apert_apert),'o3',gamma_0)
  
      call from_aux(trim(focusmes_par%tag_apert_apert),'n4',n_rho)
      call from_aux(trim(focusmes_par%tag_apert_apert),'d4',d_rho)
      call from_aux(trim(focusmes_par%tag_apert_apert),'o4',rho_0)
  
      if(n_z /= focusmes_par%n_z) call seperr("Apert_apert n1 /= Cig n1")
      if(d_z /= focusmes_par%d_z) call seperr("Apert_apert d1 /= Cig d1")
      if(z_0 /= focusmes_par%z_0) call seperr("Apert_apert o1 /= Cig o1")
  
      if(n_mx /= focusmes_par%n_mx) call seperr("Apert_apert n2 /= Cig n2")
      if(d_mx /= focusmes_par%d_mx) call seperr("Apert_apert d2 /= Cig d2")
      if(mx_0 /= focusmes_par%mx_0) call seperr("Apert_apert o2 /= Cig o2")

      if(n_gamma /= focusmes_par%n_gamma) call seperr("Apert_apert n3 /= Cig n3")
      if(d_gamma /= focusmes_par%d_gamma) call seperr("Apert_apert d3 /= Cig d3")
      if(gamma_0 /= focusmes_par%gamma_0) call seperr("Apert_apert o3 /= Cig o3")

      if(n_rho /= focusmes_par%n_rho) call seperr("Apert_apert n4 /= Cig n4")
      if(d_rho /= focusmes_par%d_rho) call seperr("Apert_apert d4 /= Cig d4")
      if(rho_0 /= focusmes_par%rho_0) call seperr("Apert_apert o4 /= Cig o4")

      call from_aux(trim(focusmes_par%tag_apert_apert),'n_z_avg',n_z_avg)
      call from_aux(trim(focusmes_par%tag_apert_apert),'n_mx_avg',n_mx_avg)
  
      call to_history("From Apert_apert: n_z_avg",n_z_avg)
      call to_history("From Apert_apert: n_mx_avg",n_mx_avg)
  
      call from_param('n_z_avg',focusmes_par%n_z_avg,n_z_avg)
      call from_param('n_mx_avg',focusmes_par%n_mx_avg,n_mx_avg)
  
      if(n_z_avg /= focusmes_par%n_z_avg)  &
        write(0,*)'WARNING: Apert_apert n_z_avg /= parameter n_z_avg'
      if(n_mx_avg /= focusmes_par%n_mx_avg)  &
        write(0,*)'WARNING: Apert_apert n_mx_avg /= parameter n_mx_avg'

  
    !}
    case('both ') aperture_type_1
    !{

      call from_param('tag_apert_apert_dip',focusmes_par%tag_apert_apert_dip,&
                    'Apert_apert_dip')
      call from_param(trim(focusmes_par%tag_apert_apert_dip),tag_name,&
                    'Apert_apert_dip not found')
      focusmes_par%aperture=auxin(trim(focusmes_par%tag_apert_apert_dip)) > 0
      if(.not. focusmes_par%aperture) call seperr("Apert_apert_dip missing")
  
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'n1',n_z)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'d1',d_z)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'o1',z_0)
  
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'n2',n_mx)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'d2',d_mx)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'o2',mx_0)

      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'n3',n_gamma)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'d3',d_gamma)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'o3',gamma_0)
  
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'n4',n_dip)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'d4',d_dip)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'o4',dip_0)

      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'n5',n_rho)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'d5',d_rho)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'o5',rho_0)
  
      if(n_z /= focusmes_par%n_z) call seperr("Apert_apert_dip n1 /= Cig n1")
      if(d_z /= focusmes_par%d_z) call seperr("Apert_apert_dip d1 /= Cig d1")
      if(z_0 /= focusmes_par%z_0) call seperr("Apert_apert_dip o1 /= Cig o1")
  
      if(n_mx /= focusmes_par%n_mx) call seperr("Apert_apert_dip n2 /= Cig n2")
      if(d_mx /= focusmes_par%d_mx) call seperr("Apert_apert_dip d2 /= Cig d2")
      if(mx_0 /= focusmes_par%mx_0) call seperr("Apert_apert_dip o2 /= Cig o2")

      if(n_gamma/=focusmes_par%n_gamma)  &
              call seperr("Apert_apert_dip n3 /= Cig n3")
      if(d_gamma/=focusmes_par%d_gamma)  &
              call seperr("Apert_apert_dip d3 /= Cig d3")
      if(gamma_0/=focusmes_par%gamma_0)  &
              call seperr("Apert_apert_dip o3 /= Cig o3")

      if(n_dip /= focusmes_par%n_dip)  &
              call seperr("Apert_apert_dip n4 /= Cig n4")
      if(d_dip /= focusmes_par%d_dip)  &
              call seperr("Apert_apert_dip d4 /= Cig d4")
      if(dip_0 /= focusmes_par%dip_0)  &
              call seperr("Apert_apert_dip o4 /= Cig o4")

      if(n_rho /= focusmes_par%n_rho)  &
              call seperr("Apert_apert_dip n5 /= Cig n5")
      if(d_rho /= focusmes_par%d_rho)  &
              call seperr("Apert_apert_dip d5 /= Cig d5")
      if(rho_0 /= focusmes_par%rho_0)  &
              call seperr("Apert_apert_dip o5 /= Cig o5")

      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'n_z_avg',n_z_avg)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'n_mx_avg',n_mx_avg)
  
      call to_history("From Apert_apert_dip: n_z_avg",n_z_avg)
      call to_history("From Apert_apert_dip: n_mx_avg",n_mx_avg)
  
      call from_param('n_z_avg',focusmes_par%n_z_avg,n_z_avg)
      call from_param('n_mx_avg',focusmes_par%n_mx_avg,n_mx_avg)
  
      if(n_z_avg /= focusmes_par%n_z_avg)  &
        write(0,*)'WARNING: Apert_apert n_z_avg /= parameter n_z_avg'
      if(n_mx_avg /= focusmes_par%n_mx_avg)  &
        write(0,*)'WARNING: Apert_apert n_mx_avg /= parameter n_mx_avg'

    !}
    case default aperture_type_1
      write(0,*)'aperture_type=',focusmes_par%aperture_type(1:5)
      call seperr ('Do not know what to do with this aperture_type')
    !}
    end select aperture_type_1

    call from_param('tag_semblance',focusmes_par%tag_semblance,'Semblance')
    call from_param(trim(focusmes_par%tag_semblance),tag_name, &
          'Semblance not found')
    write(0,*)'focusmes_par%tag_semblance=',trim(focusmes_par%tag_semblance)
    
    if(auxinout(trim(focusmes_par%tag_semblance)) > 0) then
      if(focusmes_par%verb) write(0,*)'Semblance already exists'
    end if

!    call from_param('global_threshold',focusmes_par%global_threshold,zero_real)
!    call from_param('local_threshold',focusmes_par%local_threshold,zero_real)

    esize=4
    call to_history("esize",esize,trim(focusmes_par%tag_semblance))
    call to_history("n1",focusmes_par%n_z,trim(focusmes_par%tag_semblance))
    call to_history("d1",focusmes_par%d_z,trim(focusmes_par%tag_semblance))
    call to_history("o1",focusmes_par%z_0,trim(focusmes_par%tag_semblance))
    call to_history("label1",focusmes_par%label_z,trim(focusmes_par%tag_semblance))

    call to_history("n2",focusmes_par%n_mx,trim(focusmes_par%tag_semblance))
    call to_history("d2",focusmes_par%d_mx,trim(focusmes_par%tag_semblance))
    call to_history("o2",focusmes_par%mx_0,trim(focusmes_par%tag_semblance))
    call to_history("label2",focusmes_par%label_mx,trim(focusmes_par%tag_semblance))

    curvature_image_3: if (focusmes_par%curvature_image) then
    !{
      call to_history("n3",focusmes_par%n_rad_curv, &
                 trim(focusmes_par%tag_semblance))
      call to_history("d3",focusmes_par%d_rad_curv, &
                 trim(focusmes_par%tag_semblance))
      call to_history("o3",focusmes_par%rad_curv_0, &
                 trim(focusmes_par%tag_semblance))
      call to_history("label3",focusmes_par%label_rad_curv, &
                 trim(focusmes_par%tag_semblance))

      call to_history("n4",focusmes_par%n_rho,trim(focusmes_par%tag_semblance))
      call to_history("d4",focusmes_par%d_rho,trim(focusmes_par%tag_semblance))
      call to_history("o4",focusmes_par%rho_0,trim(focusmes_par%tag_semblance))
      call to_history("label4",focusmes_par%label_rho, &
                 trim(focusmes_par%tag_semblance))
    !}
    else curvature_image_3
    !{
      call to_history("n3",focusmes_par%n_rho,trim(focusmes_par%tag_semblance))
      call to_history("d3",focusmes_par%d_rho,trim(focusmes_par%tag_semblance))
      call to_history("o3",focusmes_par%rho_0,trim(focusmes_par%tag_semblance))
      call to_history("label3",focusmes_par%label_rho, &
                 trim(focusmes_par%tag_semblance))
    !}
    end if curvature_image_3


    if(.not. focusmes_par%prestack_image) call seperr("Computing Semblance needs a prestackimage")
    if(focusmes_par%write_prestack) call seperr("Cannot output prestack image when compting semblance")

  !}
  case('kurtosis') oper_name_1
  !{

    call from_param('tag_kurtosis',focusmes_par%tag_kurtosis,'Kurtosis')
    call from_param(trim(focusmes_par%tag_kurtosis),tag_name, &
          'Kurtosis not found')
    write(0,*)'focusmes_par%tag_kurtosis=',trim(focusmes_par%tag_kurtosis)
    
    if(auxinout(trim(focusmes_par%tag_kurtosis)) > 0) then
      if(focusmes_par%verb) write(0,*)'Kurtosis already exists'
    end if

    if((focusmes_par%write_prestack).and.(.not.focusmes_par%prestack_image))&
       call seperr("Cannot write prestack if input is not prestack image")
  
    if(focusmes_par%write_prestack) then
    !{
      esize=4
      call to_history("esize",esize,trim(focusmes_par%tag_kurtosis))
      if(focusmes_par%apert_ang_image) then
        call to_history("n1",focusmes_par%n_gamma, &
               trim(focusmes_par%tag_kurtosis))
        call to_history("d1",focusmes_par%d_gamma, &
               trim(focusmes_par%tag_kurtosis))
        call to_history("o1",focusmes_par%gamma_0, &
               trim(focusmes_par%tag_kurtosis))
        call to_history("label1",focusmes_par%label_gamma, &
                 trim(focusmes_par%tag_kurtosis))
      else
        call to_history("n1",focusmes_par%n_hx, &
               trim(focusmes_par%tag_kurtosis))
        call to_history("d1",focusmes_par%d_hx, &
               trim(focusmes_par%tag_kurtosis))
        call to_history("o1",focusmes_par%hx_0, &
               trim(focusmes_par%tag_kurtosis))
        call to_history("label1",focusmes_par%label_hx, &
                 trim(focusmes_par%tag_kurtosis))
      end if

      call to_history("n2",focusmes_par%n_rho,trim(focusmes_par%tag_kurtosis))
      call to_history("d2",focusmes_par%d_rho,trim(focusmes_par%tag_kurtosis))
      call to_history("o2",focusmes_par%rho_0,trim(focusmes_par%tag_kurtosis))
      call to_history("label2",focusmes_par%label_rho, &
                 trim(focusmes_par%tag_kurtosis))

    !}
    else
    !{
      call to_history("n1",focusmes_par%n_rho,trim(focusmes_par%tag_kurtosis))
      call to_history("d1",focusmes_par%d_rho,trim(focusmes_par%tag_kurtosis))
      call to_history("o1",focusmes_par%rho_0,trim(focusmes_par%tag_kurtosis))
      call to_history("label1",focusmes_par%label_rho, &
                 trim(focusmes_par%tag_kurtosis))

      call to_history("n2",focusmes_par%n_rad_curv,trim(focusmes_par%tag_kurtosis))
      call to_history("d2",focusmes_par%d_rad_curv,trim(focusmes_par%tag_kurtosis))
      call to_history("o2",focusmes_par%rad_curv_0,trim(focusmes_par%tag_kurtosis))
      call to_history("label2",focusmes_par%label_rad_curv, &
                 trim(focusmes_par%tag_kurtosis))
    !}
    end if

  !}
  case('aperture ') oper_name_1
  !{

    if(focusmes_par%dip_ang_image) then
      focusmes_par%aperture_type="dip  "
    else
      if(focusmes_par%apert_ang_image) then
        focusmes_par%aperture_type="apert"
      else
        focusmes_par%aperture_type="none "
      end if
    end if
    if((focusmes_par%dip_ang_image).and.(focusmes_par%apert_ang_image)) then
      focusmes_par%aperture_type="both "
    end if
    if(focusmes_par%verb) write(0,*)'aperture_type=',focusmes_par%aperture_type

    call from_param('tag_aperture',focusmes_par%tag_aperture,'Aperture')
    call from_param(trim(focusmes_par%tag_aperture),tag_name, &
          'Aperture not found')
    write(0,*)'focusmes_par%tag_aperture=',trim(focusmes_par%tag_aperture)
    
    if(auxinout(trim(focusmes_par%tag_aperture)) > 0) then
      if(focusmes_par%verb) write(0,*)'Aperture already exists'
    end if

    if(.not. focusmes_par%write_prestack)&
       call seperr("Must write prestack with aperture")

    call from_param('n_z_avg',focusmes_par%n_z_avg,zero_int)
    call from_param('n_mx_avg',focusmes_par%n_mx_avg,zero_int)
    call from_param('global_threshold',focusmes_par%global_threshold,zero_real)
    call from_param('local_threshold',focusmes_par%local_threshold,zero_real)

    call to_history("n_z_avg",focusmes_par%n_z_avg, &
                trim(focusmes_par%tag_aperture))
    call to_history("n_mx_avg",focusmes_par%n_mx_avg, &
                trim(focusmes_par%tag_aperture))

    esize=4
    call to_history("esize",esize,trim(focusmes_par%tag_aperture))
    call to_history("n1",focusmes_par%n_z,trim(focusmes_par%tag_aperture))
    call to_history("d1",focusmes_par%d_z,trim(focusmes_par%tag_aperture))
    call to_history("o1",focusmes_par%z_0,trim(focusmes_par%tag_aperture))
    call to_history("label1",focusmes_par%label_z,trim(focusmes_par%tag_aperture))

    call to_history("n2",focusmes_par%n_mx,trim(focusmes_par%tag_aperture))
    call to_history("d2",focusmes_par%d_mx,trim(focusmes_par%tag_aperture))
    call to_history("o2",focusmes_par%mx_0,trim(focusmes_par%tag_aperture))
    call to_history("label2",focusmes_par%label_mx,trim(focusmes_par%tag_aperture))


    aperture_type_2: select case (focusmes_par%aperture_type(1:5))
    case('apert') aperture_type_2
      call to_history("n3",focusmes_par%n_gamma,trim(focusmes_par%tag_aperture))
      call to_history("d3",focusmes_par%d_gamma,trim(focusmes_par%tag_aperture))
      call to_history("o3",focusmes_par%gamma_0,trim(focusmes_par%tag_aperture))
      call to_history("label3",focusmes_par%label_gamma, &
           trim(focusmes_par%tag_aperture))

      call to_history("n4",focusmes_par%n_rho,trim(focusmes_par%tag_aperture))
      call to_history("d4",focusmes_par%d_rho,trim(focusmes_par%tag_aperture))
      call to_history("o4",focusmes_par%rho_0,trim(focusmes_par%tag_aperture))
      call to_history("label4",focusmes_par%label_rho, &
           trim(focusmes_par%tag_aperture))
    case('dip  ') aperture_type_2
      call to_history("n3",focusmes_par%n_dip,trim(focusmes_par%tag_aperture))
      call to_history("d3",focusmes_par%d_dip,trim(focusmes_par%tag_aperture))
      call to_history("o3",focusmes_par%dip_0,trim(focusmes_par%tag_aperture))
      call to_history("label3",focusmes_par%label_dip, &
           trim(focusmes_par%tag_aperture))

    curvature_image_4: if (focusmes_par%curvature_image) then
    !{
      call to_history("n4",focusmes_par%n_rad_curv, &
                 trim(focusmes_par%tag_aperture))
      call to_history("d4",focusmes_par%d_rad_curv, &
                 trim(focusmes_par%tag_aperture))
      call to_history("o4",focusmes_par%rad_curv_0, &
                 trim(focusmes_par%tag_aperture))
      call to_history("label4",focusmes_par%label_rad_curv, &
           trim(focusmes_par%tag_aperture))

      call to_history("n5",focusmes_par%n_rho,trim(focusmes_par%tag_aperture))
      call to_history("d5",focusmes_par%d_rho,trim(focusmes_par%tag_aperture))
      call to_history("o5",focusmes_par%rho_0,trim(focusmes_par%tag_aperture))
      call to_history("label5",focusmes_par%label_rho, &
           trim(focusmes_par%tag_aperture))
    !}
    else curvature_image_4
    !{
      call to_history("n4",focusmes_par%n_rho,trim(focusmes_par%tag_aperture))
      call to_history("d4",focusmes_par%d_rho,trim(focusmes_par%tag_aperture))
      call to_history("o4",focusmes_par%rho_0,trim(focusmes_par%tag_aperture))
      call to_history("label4",focusmes_par%label_rho, &
           trim(focusmes_par%tag_aperture))
    !}
    end if curvature_image_4

    case('both ') aperture_type_2
      call to_history("n3",focusmes_par%n_gamma,trim(focusmes_par%tag_aperture))
      call to_history("d3",focusmes_par%d_gamma,trim(focusmes_par%tag_aperture))
      call to_history("o3",focusmes_par%gamma_0,trim(focusmes_par%tag_aperture))
      call to_history("label3",focusmes_par%label_gamma, &
           trim(focusmes_par%tag_aperture))
      call to_history("n4",focusmes_par%n_dip,trim(focusmes_par%tag_aperture))
      call to_history("d4",focusmes_par%d_dip,trim(focusmes_par%tag_aperture))
      call to_history("o4",focusmes_par%dip_0,trim(focusmes_par%tag_aperture))
      call to_history("label4",focusmes_par%label_dip, &
           trim(focusmes_par%tag_aperture))

      call to_history("n5",focusmes_par%n_rho,trim(focusmes_par%tag_aperture))
      call to_history("d5",focusmes_par%d_rho,trim(focusmes_par%tag_aperture))
      call to_history("o5",focusmes_par%rho_0,trim(focusmes_par%tag_aperture))
      call to_history("label5",focusmes_par%label_rho, &
           trim(focusmes_par%tag_aperture))
    case default aperture_type_2
      write(0,*)'aperture_type=',focusmes_par%aperture_type(1:5)
      call seperr ('Do not know what to do with this aperture_type')
    end select aperture_type_2

  
    if(.not. focusmes_par%prestack_image) call seperr("Computing Aperture needs a prestackimage")

  !}
  case default oper_name_1
    write(0,*)'oper_name=',focusmes_par%oper_name(1:9)
    call seperr ('Do not know what to do (#1 in param_focusmem)')
  !}
  end select oper_name_1

  return
end subroutine param_focusmes
