subroutine param_focusmes_transp(focusmes_par)
  use focusmes_types
  use sep

  implicit none
  type (focusmes_param_type) focusmes_par

  integer ::  omp_get_num_threads
  integer, external :: auxpar,getch,fetch,auxin,auxinout,sreed,sseek,auxout

  character(len=128) :: tag_name
  integer n_threads_env,n_threads_init
  integer n_z,n_mx,n_gamma,n_rho,n_rad_curv,n_dip,n_z_avg,n_mx_avg
  integer esize,zero_int,one_int,n4
  real z_0,mx_0,gamma_0,rho_0,rad_curv_0,dip_0
  real d_z,d_mx,d_gamma,d_rho,d_rad_curv,d_dip
  real zero_real,one_real,epsi_par

  n_threads_init=-1

  zero_int=0
  zero_real=0.
  one_int=1
  one_real=1.
  epsi_par=.001

  call from_param('verb',focusmes_par%verb,.false.)
  call from_param('debug1',focusmes_par%debug1,.false.)
  call from_param('debug2',focusmes_par%debug2,.false.)


  call from_param('apert_ang_image',focusmes_par%apert_ang_image,.true.)
  call from_param('dip_ang_image',focusmes_par%dip_ang_image,.true.)
  call from_param('curvature_image',focusmes_par%curvature_image,.false.)
  call from_param('curvature_semblance',focusmes_par%curvature_semblance,.false.)
  call from_param('scan_dips_0',focusmes_par%scan_dips_0,.false.)
  call from_param('read_dips_0',focusmes_par%read_dips_0,.false.)
  call from_param('global_straight_rays', &
             focusmes_par%global_straight_rays,.true.)
  focusmes_par%sum_squares_input=.false.

  call from_param('semblance',focusmes_par%semblance,.false.)
  call from_param('stack_power',focusmes_par%stack_power,.false.)
  call from_param('map_pseudo_depth',focusmes_par%map_pseudo_depth,.false.)

  call from_param('n_threads',focusmes_par%n_threads,n_threads_init)
  call from_param('max_memory',focusmes_par%max_memory,100)

  ! checking parameters
!$omp parallel
  n_threads_env=omp_get_num_threads()
!$omp end parallel
  if(focusmes_par%n_threads == n_threads_init) then
     focusmes_par%n_threads=n_threads_env
  else
     if(n_threads_env > focusmes_par%n_threads) then
        call omp_set_num_threads(focusmes_par%n_threads)
        write(0,*)'WARNING: overwriting environment variable OMP_NUM_THREADS'
        write(0,*)'focusmes_par-n_threads=',focusmes_par%n_threads
     end if
     if(n_threads_env < focusmes_par%n_threads) then
        write(0,*)'focusmes_par-n_threads=',focusmes_par%n_threads
        write(0,*)'OMP_NUM_THREADS=',n_threads_env
        call seperr('ERROR: Asked for more threads than the environment variable allows')
     end if
  end if
  if(focusmes_par%debug1)write(0,*)'focusmes_par-n_threads=',focusmes_par%n_threads
  call to_history('n_threads',focusmes_par%n_threads)

   if(focusmes_par%n_threads /= 1) then
     call seperr("At the present n_threads must be == 1")
   end if

  if(focusmes_par%scan_dips_0 .and. focusmes_par%read_dips_0) then
    call seperr("Cannot read dips and scan dips at the same time")
  end if

  if(focusmes_par%curvature_image .and. focusmes_par%read_dips_0) then
    call seperr("Cannot curvature_image and scan dips at the same time")
  end if

  if (focusmes_par%curvature_image .and. &               
     (.not.focusmes_par%curvature_semblance)) then
    call seperr("Cannot be curvature_image and not curvature_semblance")
  end if

  if (focusmes_par%curvature_image .and. &               
      focusmes_par%dip_ang_image .and. &
      focusmes_par%apert_ang_image) then
    call seperr("Cannot be curvature_image and both dip and aperture image")
  end if

! Cig is an input

   call from_param('tag_cig',focusmes_par%tag_cig,'Cig')
   call from_param(trim(focusmes_par%tag_cig),tag_name,'Cig not found')
   focusmes_par%cig=auxin(trim(focusmes_par%tag_cig)) > 0
   if(.not. focusmes_par%cig) call seperr("Cig missing")

   call from_aux(trim(focusmes_par%tag_cig),'n4',n4)
   focusmes_par%prestack_image=(n4 /= 1) 
   call from_param('write_prestack',focusmes_par%write_prestack, &
        focusmes_par%prestack_image)
   if(focusmes_par%verb) write(0,*)'prestack_image=',focusmes_par%prestack_image
   if(focusmes_par%verb) write(0,*)'write_prestack=',focusmes_par%write_prestack

   call from_aux(trim(focusmes_par%tag_cig),'n1',focusmes_par%n_z)
   call from_aux(trim(focusmes_par%tag_cig),'d1',focusmes_par%d_z)
   call from_aux(trim(focusmes_par%tag_cig),'o1',focusmes_par%z_0)
   call from_aux(trim(focusmes_par%tag_cig),'label1',focusmes_par%label_z, &
              trim(focusmes_par%label_z))

   call to_history("From Cig: n_z",focusmes_par%n_z)
   call to_history("From Cig: d_z",focusmes_par%d_z)
   call to_history("From Cig: z_0",focusmes_par%z_0)

   call from_aux(trim(focusmes_par%tag_cig),'n2',focusmes_par%n_rho)
   call from_aux(trim(focusmes_par%tag_cig),'d2',focusmes_par%d_rho)
   call from_aux(trim(focusmes_par%tag_cig),'o2',focusmes_par%rho_0)
   call from_aux(trim(focusmes_par%tag_cig),'label2',focusmes_par%label_rho, &
              trim(focusmes_par%label_rho))

   call to_history("From Cig: n_rho",focusmes_par%n_rho)
   call to_history("From Cig: d_rho",focusmes_par%d_rho)
   call to_history("From Cig: rho_0",focusmes_par%rho_0)

   call from_param('flip_sign_dip',focusmes_par%flip_sign_dip, &
        huge(focusmes_par%flip_sign_dip))
   call from_param('split_dip',focusmes_par%split_dip, &
        huge(focusmes_par%split_dip))
   call from_param('split_dip_1',focusmes_par%split_dip_1, &
        huge(focusmes_par%split_dip_1))
   call from_param('z_split_dip_1',focusmes_par%z_split_dip_1, &
        huge(focusmes_par%z_split_dip_1))
   call from_param('zero_split_dip',focusmes_par%zero_split_dip,.false.)

   if (focusmes_par%curvature_semblance) then
     if (focusmes_par%curvature_image) then

       focusmes_par%z_ref = 1.
       focusmes_par%rad_curv_scale=0.
       focusmes_par%dip_0_curv=0.
  
     else
       call from_param('linear_curvature_scale', &
                  focusmes_par%linear_curvature_scale,.false.)
       call from_param('dip_0_curv',focusmes_par%dip_0_curv,0.)
       call from_aux(trim(focusmes_par%tag_cig),'ResMig_z_ref',focusmes_par%z_ref)
       call to_history("From Cig: ResMig_z_ref",focusmes_par%z_ref)
       if(focusmes_par%z_ref == 0.) call seperr("ResMig_z_ref should be /= 0")
  
       if (focusmes_par%linear_curvature_scale) then
         call from_param('n_rad_curv',focusmes_par%n_rad_curv,1)
         call from_param('rad_curv_0',focusmes_par%rad_curv_0,0.)
         call from_param('d_rad_curv',focusmes_par%d_rad_curv,1.)
         focusmes_par%label_rad_curv="R (m)"
         call from_param('label_rad_curv',focusmes_par%label_rad_curv, &
           trim(focusmes_par%label_rad_curv))
         focusmes_par%rad_curv_scale=0.
       else
         call from_param('n_rad_curv',focusmes_par%n_rad_curv,1)
         call from_param('rad_curv_0',focusmes_par%rad_curv_0,0.)
         call from_param('d_rad_curv',focusmes_par%d_rad_curv,1.)
         call from_param('rad_curv_scale',focusmes_par%rad_curv_scale,1.)
         focusmes_par%label_rad_curv="arctan(R/scale)/pi"
         call from_param('label_rad_curv',focusmes_par%label_rad_curv, &
           trim(focusmes_par%label_rad_curv))
       end if
     end if
   else
     focusmes_par%z_ref = 1.
     focusmes_par%n_rad_curv=1
     focusmes_par%d_rad_curv=1.
     focusmes_par%rad_curv_0=0.
     focusmes_par%rad_curv_scale=0.
     focusmes_par%dip_0_curv=0.
   end if

   focusmes_par%n_scan_dip_0=1
   focusmes_par%scan_dip_0_0=focusmes_par%dip_0_curv
   focusmes_par%d_scan_dip_0=0.


   prestack_image_1: if(focusmes_par%prestack_image) then
   !{
     dip_ang_image_1: if(focusmes_par%dip_ang_image) then
     !{
       apert_ang_image_1: if(focusmes_par%apert_ang_image) then
       !{
!          call seperr("dip_ang_image AND apert_ang_image not suported yet")
         call from_param('only_apert_ang_sembl', &
                focusmes_par%only_apert_ang_sembl,.false.)
         call from_aux(trim(focusmes_par%tag_cig),'n3',focusmes_par%n_gamma)
         call from_aux(trim(focusmes_par%tag_cig),'d3',focusmes_par%d_gamma)
         call from_aux(trim(focusmes_par%tag_cig),'o3',focusmes_par%gamma_0)
         call from_aux(trim(focusmes_par%tag_cig),'label3', &
              focusmes_par%label_gamma,trim(focusmes_par%label_gamma))
    
         call to_history("From Cig: n_gamma",focusmes_par%n_gamma)
         call to_history("From Cig: d_gamma",focusmes_par%d_gamma)
         call to_history("From Cig: gamma_0",focusmes_par%gamma_0)

         call from_aux(trim(focusmes_par%tag_cig),'n4',focusmes_par%n_dip)
         call from_aux(trim(focusmes_par%tag_cig),'d4',focusmes_par%d_dip)
         call from_aux(trim(focusmes_par%tag_cig),'o4',focusmes_par%dip_0)
         call from_aux(trim(focusmes_par%tag_cig),'label4', &
              focusmes_par%label_dip,trim(focusmes_par%label_dip))
    
         call to_history("From Cig: n_dip",focusmes_par%n_dip)
         call to_history("From Cig: d_dip",focusmes_par%d_dip)
         call to_history("From Cig: dip_0",focusmes_par%dip_0)
  
         call from_aux(trim(focusmes_par%tag_cig),'n5',focusmes_par%n_mx)
         call from_aux(trim(focusmes_par%tag_cig),'d5',focusmes_par%d_mx)
         call from_aux(trim(focusmes_par%tag_cig),'o5',focusmes_par%mx_0)
         call from_aux(trim(focusmes_par%tag_cig),'label5', &
              focusmes_par%label_mx,trim(focusmes_par%label_mx))
    
         call to_history("From Cig: n_mx",focusmes_par%n_mx)
         call to_history("From Cig: d_mx",focusmes_par%d_mx)
         call to_history("From Cig: mx_0",focusmes_par%mx_0)
       !}
       else apert_ang_image_1
       !{
         if (focusmes_par%curvature_image) then
         !{
           call from_aux(trim(focusmes_par%tag_cig),'n3',focusmes_par%n_dip)
           call from_aux(trim(focusmes_par%tag_cig),'d3',focusmes_par%d_dip)
           call from_aux(trim(focusmes_par%tag_cig),'o3',focusmes_par%dip_0)
           call from_aux(trim(focusmes_par%tag_cig),'label3', &
                focusmes_par%label_dip,trim(focusmes_par%label_dip))
      
           call to_history("From Cig: n_dip",focusmes_par%n_dip)
           call to_history("From Cig: d_dip",focusmes_par%d_dip)
           call to_history("From Cig: dip_0",focusmes_par%dip_0)
  
           call from_aux(trim(focusmes_par%tag_cig),'n4',focusmes_par%n_rad_curv)
           call from_aux(trim(focusmes_par%tag_cig),'d4',focusmes_par%d_rad_curv)
           call from_aux(trim(focusmes_par%tag_cig),'o4',focusmes_par%rad_curv_0)
           call from_aux(trim(focusmes_par%tag_cig),'label4', &
                focusmes_par%label_rad_curv,trim(focusmes_par%label_rad_curv))
  
           call to_history("From Cig: n_rad_curv",focusmes_par%n_rad_curv)
           call to_history("From Cig: d_rad_curv",focusmes_par%d_rad_curv)
           call to_history("From Cig: rad_curv_0",focusmes_par%rad_curv_0)

           call from_aux(trim(focusmes_par%tag_cig),'n5',focusmes_par%n_mx)
           call from_aux(trim(focusmes_par%tag_cig),'d5',focusmes_par%d_mx)
           call from_aux(trim(focusmes_par%tag_cig),'o5',focusmes_par%mx_0)
           call from_aux(trim(focusmes_par%tag_cig),'label5', &
              focusmes_par%label_mx,trim(focusmes_par%label_mx))
  
           call to_history("From Cig: n_mx",focusmes_par%n_mx)
           call to_history("From Cig: d_mx",focusmes_par%d_mx)
           call to_history("From Cig: mx_0",focusmes_par%mx_0)
         !}
         else
         !{
           call from_aux(trim(focusmes_par%tag_cig),'n3',focusmes_par%n_dip)
           call from_aux(trim(focusmes_par%tag_cig),'d3',focusmes_par%d_dip)
           call from_aux(trim(focusmes_par%tag_cig),'o3',focusmes_par%dip_0)
           call from_aux(trim(focusmes_par%tag_cig),'label3', &
                focusmes_par%label_dip,trim(focusmes_par%label_dip))
      
           call to_history("From Cig: n_dip",focusmes_par%n_dip)
           call to_history("From Cig: d_dip",focusmes_par%d_dip)
           call to_history("From Cig: dip_0",focusmes_par%dip_0)
  
           call from_aux(trim(focusmes_par%tag_cig),'n4',focusmes_par%n_mx)
           call from_aux(trim(focusmes_par%tag_cig),'d4',focusmes_par%d_mx)
           call from_aux(trim(focusmes_par%tag_cig),'o4',focusmes_par%mx_0)
           call from_aux(trim(focusmes_par%tag_cig),'label4', &
              focusmes_par%label_mx,trim(focusmes_par%label_mx))
  
           call to_history("From Cig: n_mx",focusmes_par%n_mx)
           call to_history("From Cig: d_mx",focusmes_par%d_mx)
           call to_history("From Cig: mx_0",focusmes_par%mx_0)
         !}
         end if

       !}
       end if apert_ang_image_1

       if(focusmes_par%scan_dips_0) then
         call from_param('n_scan_dip_0',focusmes_par%n_scan_dip_0, &
                                        focusmes_par%n_dip)
         call from_param('scan_dip_0_0',focusmes_par%scan_dip_0_0, &
                                        focusmes_par%dip_0)
         call from_param('d_scan_dip_0',focusmes_par%d_scan_dip_0, &
                                        focusmes_par%d_dip)
       end if
       read_dips_0_1: if(focusmes_par%read_dips_0) then
       !{
        call from_param('tag_dips_0',focusmes_par%tag_dips_0,'Dips_0')
        call from_param(trim(focusmes_par%tag_dips_0),tag_name,'Dips_0 not found')
        focusmes_par%dips_0=auxin(trim(focusmes_par%tag_dips_0)) > 0
        if(.not. focusmes_par%dips_0) call seperr("Dips_0 missing")


        call from_aux(trim(focusmes_par%tag_dips_0),'n1',n_z)
        call from_aux(trim(focusmes_par%tag_dips_0),'d1',d_z)
        call from_aux(trim(focusmes_par%tag_dips_0),'o1',z_0)
  
        call from_aux(trim(focusmes_par%tag_dips_0),'n2',n_rho)
        call from_aux(trim(focusmes_par%tag_dips_0),'d2',d_rho)
        call from_aux(trim(focusmes_par%tag_dips_0),'o2',rho_0)

        call from_aux(trim(focusmes_par%tag_dips_0),'n3',n_mx)
        call from_aux(trim(focusmes_par%tag_dips_0),'d3',d_mx)
        call from_aux(trim(focusmes_par%tag_dips_0),'o3',mx_0)

        if(n_z /= focusmes_par%n_z) call seperr("Dips_0 n1 /= Cig n1")
        if(d_z /= focusmes_par%d_z) call seperr("Dips_0 d1 /= Cig d1")
        if(z_0 /= focusmes_par%z_0) call seperr("Dips_0 o1 /= Cig o1")
        if(abs((z_0-focusmes_par%z_0)/(z_0+epsi_par)) > epsi_par) &
               call seperr("Dips_0 o1 /= Cig o1")
  
        if(n_rho /= focusmes_par%n_rho) call seperr("Dips_0 n2 /= Cig n2")
        if(d_rho /= focusmes_par%d_rho) call seperr("Dips_0 d2 /= Cig d2")
        if(abs((rho_0-focusmes_par%rho_0)/(rho_0+epsi_par)) > epsi_par) &
               call seperr("Dips_0 o2 /= Cig o2")

        if(n_mx /= focusmes_par%n_mx) call seperr("Dips_0 n3 /= Cig n4")
        if(d_mx /= focusmes_par%d_mx) call seperr("Dips_0 d3 /= Cig d4")
        if(abs((mx_0-focusmes_par%mx_0)/(mx_0+epsi_par)) > epsi_par) &
               call seperr("Dips_0 o1 /= Cig o4")
!write(0,*)'mx_0 = ',mx_0,' from CIG mx_0=',focusmes_par%mx_0
!		call seperr("Dips_0 o3 /= Cig o4")
!        end if

  
       !}
       end if read_dips_0_1

     !}
     else dip_ang_image_1
     !{
       if (focusmes_par%curvature_image) then
       !{
         apert_ang_image_2: if(focusmes_par%apert_ang_image) then
           call from_aux(trim(focusmes_par%tag_cig),'n3',focusmes_par%n_gamma)
           call from_aux(trim(focusmes_par%tag_cig),'d3',focusmes_par%d_gamma)
           call from_aux(trim(focusmes_par%tag_cig),'o3',focusmes_par%gamma_0)
           call from_aux(trim(focusmes_par%tag_cig),'label3', &
                focusmes_par%label_gamma,trim(focusmes_par%label_gamma))
      
           call to_history("From Cig: n_gamma",focusmes_par%n_gamma)
           call to_history("From Cig: d_gamma",focusmes_par%d_gamma)
           call to_history("From Cig: gamma_0",focusmes_par%gamma_0)
         else apert_ang_image_2
           call from_aux(trim(focusmes_par%tag_cig),'n3',focusmes_par%n_hx)
           call from_aux(trim(focusmes_par%tag_cig),'d3',focusmes_par%d_hx)
           call from_aux(trim(focusmes_par%tag_cig),'o3',focusmes_par%hx_0)
           call from_aux(trim(focusmes_par%tag_cig),'label3', &
                focusmes_par%label_hx,trim(focusmes_par%label_hx))
      
           call to_history("From Cig: n_hx",focusmes_par%n_hx)
           call to_history("From Cig: d_hx",focusmes_par%d_hx)
           call to_history("From Cig: hx_0",focusmes_par%hx_0)
         end if apert_ang_image_2
         call from_aux(trim(focusmes_par%tag_cig),'n4',focusmes_par%n_rad_curv)
         call from_aux(trim(focusmes_par%tag_cig),'d4',focusmes_par%d_rad_curv)
         call from_aux(trim(focusmes_par%tag_cig),'o4',focusmes_par%rad_curv_0)
         call from_aux(trim(focusmes_par%tag_cig),'label4', &
                focusmes_par%label_rad_curv,trim(focusmes_par%label_rad_curv))
  
         call to_history("From Cig: n_rad_curv",focusmes_par%n_rad_curv)
         call to_history("From Cig: d_rad_curv",focusmes_par%d_rad_curv)
         call to_history("From Cig: rad_curv_0",focusmes_par%rad_curv_0)

         call from_aux(trim(focusmes_par%tag_cig),'n5',focusmes_par%n_mx)
         call from_aux(trim(focusmes_par%tag_cig),'d5',focusmes_par%d_mx)
         call from_aux(trim(focusmes_par%tag_cig),'o5',focusmes_par%mx_0)
         call from_aux(trim(focusmes_par%tag_cig),'label5', &
            focusmes_par%label_mx,trim(focusmes_par%label_mx))
    
         call to_history("From Cig: n_mx",focusmes_par%n_mx)
         call to_history("From Cig: d_mx",focusmes_par%d_mx)
         call to_history("From Cig: mx_0",focusmes_par%mx_0)

       !}
       else
       !{
         apert_ang_image_3: if(focusmes_par%apert_ang_image) then
           call from_aux(trim(focusmes_par%tag_cig),'n3',focusmes_par%n_gamma)
           call from_aux(trim(focusmes_par%tag_cig),'d3',focusmes_par%d_gamma)
           call from_aux(trim(focusmes_par%tag_cig),'o3',focusmes_par%gamma_0)
           call from_aux(trim(focusmes_par%tag_cig),'label3', &
                focusmes_par%label_gamma,trim(focusmes_par%label_gamma))
      
           call to_history("From Cig: n_gamma",focusmes_par%n_gamma)
           call to_history("From Cig: d_gamma",focusmes_par%d_gamma)
           call to_history("From Cig: gamma_0",focusmes_par%gamma_0)
         else apert_ang_image_3
           call from_aux(trim(focusmes_par%tag_cig),'n3',focusmes_par%n_hx)
           call from_aux(trim(focusmes_par%tag_cig),'d3',focusmes_par%d_hx)
           call from_aux(trim(focusmes_par%tag_cig),'o3',focusmes_par%hx_0)
           call from_aux(trim(focusmes_par%tag_cig),'label3', &
                focusmes_par%label_hx,trim(focusmes_par%label_hx))
      
           call to_history("From Cig: n_hx",focusmes_par%n_hx)
           call to_history("From Cig: d_hx",focusmes_par%d_hx)
           call to_history("From Cig: hx_0",focusmes_par%hx_0)
         end if apert_ang_image_3
         call from_aux(trim(focusmes_par%tag_cig),'n4',focusmes_par%n_mx)
         call from_aux(trim(focusmes_par%tag_cig),'d4',focusmes_par%d_mx)
         call from_aux(trim(focusmes_par%tag_cig),'o4',focusmes_par%mx_0)
         call from_aux(trim(focusmes_par%tag_cig),'label4', &
            focusmes_par%label_mx,trim(focusmes_par%label_mx))
    
         call to_history("From Cig: n_mx",focusmes_par%n_mx)
         call to_history("From Cig: d_mx",focusmes_par%d_mx)
         call to_history("From Cig: mx_0",focusmes_par%mx_0)
       !}
       end if
     !}
     end if dip_ang_image_1
   !}
   else prestack_image_1
   !{
!     if(focusmes_par%dip_ang_image) then
!     !{
!       call from_aux(trim(focusmes_par%tag_cig),'n3',focusmes_par%n_dip)
!       call from_aux(trim(focusmes_par%tag_cig),'d3',focusmes_par%d_dip)
!       call from_aux(trim(focusmes_par%tag_cig),'o3',focusmes_par%dip_0)
!  
!       call to_history("From Cig: n3",focusmes_par%n_dip)
!       call to_history("From Cig: d3",focusmes_par%d_dip)
!       call to_history("From Cig: o3",focusmes_par%dip_0)
!
!       call from_aux(trim(focusmes_par%tag_cig),'n4',focusmes_par%n_rho)
!       call from_aux(trim(focusmes_par%tag_cig),'d4',focusmes_par%d_rho)
!       call from_aux(trim(focusmes_par%tag_cig),'o4',focusmes_par%rho_0)
!  
!       call to_history("From Cig: n4",focusmes_par%n_rho)
!       call to_history("From Cig: d4",focusmes_par%d_rho)
!       call to_history("From Cig: o4",focusmes_par%rho_0)
!     !}
!     else
!     !{
       call from_aux(trim(focusmes_par%tag_cig),'n3',focusmes_par%n_mx)
       call from_aux(trim(focusmes_par%tag_cig),'d3',focusmes_par%d_mx)
       call from_aux(trim(focusmes_par%tag_cig),'o3',focusmes_par%mx_0)
       call from_aux(trim(focusmes_par%tag_cig),'label3', &
          focusmes_par%label_mx,trim(focusmes_par%label_mx))
  
       call to_history("From Cig: n_mx",focusmes_par%n_mx)
       call to_history("From Cig: d_mx",focusmes_par%d_mx)
       call to_history("From Cig: mx_0",focusmes_par%mx_0)
!     !}
!     end if
   !}
   end if prestack_image_1

   if(focusmes_par%semblance) then
     focusmes_par%oper_name(1:9)="semblance"
   else 
     focusmes_par%oper_name(1:9)="aperture "
   end if

  write(0,*)'oper_name=',focusmes_par%oper_name(1:9)


  oper_name_1: select case (focusmes_par%oper_name(1:9))
  !{
  case('semblance') oper_name_1
  !{

    if(focusmes_par%dip_ang_image) then
      focusmes_par%aperture_type="dip  "
    else
      if(focusmes_par%apert_ang_image) then
        focusmes_par%aperture_type="apert"
      else
        focusmes_par%aperture_type="none "
      end if
    end if
    if((focusmes_par%dip_ang_image).and.(focusmes_par%apert_ang_image)) then
      focusmes_par%aperture_type="both "
    end if
    if(focusmes_par%verb) write(0,*)'aperture_type=',focusmes_par%aperture_type

    aperture_type_1: select case (focusmes_par%aperture_type(1:5))
    !{
    case('dip  ') aperture_type_1
    !{
      call from_param('tag_apert_dip',focusmes_par%tag_apert_dip,'Apert_dip')
      call from_param(trim(focusmes_par%tag_apert_dip),tag_name,'Apert_dip not found')
      focusmes_par%aperture=auxin(trim(focusmes_par%tag_apert_dip)) > 0
      if(.not. focusmes_par%aperture) call seperr("Apert_dip missing")
  
      call from_aux(trim(focusmes_par%tag_apert_dip),'n1',n_z)
      call from_aux(trim(focusmes_par%tag_apert_dip),'d1',d_z)
      call from_aux(trim(focusmes_par%tag_apert_dip),'o1',z_0)
  
      call from_aux(trim(focusmes_par%tag_apert_dip),'n2',n_rho)
      call from_aux(trim(focusmes_par%tag_apert_dip),'d2',d_rho)
      call from_aux(trim(focusmes_par%tag_apert_dip),'o2',rho_0)

      call from_aux(trim(focusmes_par%tag_apert_dip),'n3',n_dip)
      call from_aux(trim(focusmes_par%tag_apert_dip),'d3',d_dip)
      call from_aux(trim(focusmes_par%tag_apert_dip),'o3',dip_0)

      call from_aux(trim(focusmes_par%tag_apert_dip),'n4',n_mx)
      call from_aux(trim(focusmes_par%tag_apert_dip),'d4',d_mx)
      call from_aux(trim(focusmes_par%tag_apert_dip),'o4',mx_0)



      if(n_z /= focusmes_par%n_z) call seperr("Apert_dip n1 /= Cig n1")
      if(d_z /= focusmes_par%d_z) call seperr("Apert_dip d1 /= Cig d1")
      if(z_0 /= focusmes_par%z_0) call seperr("Apert_dip o1 /= Cig o1")
      if(abs((z_0-focusmes_par%z_0)/(z_0+epsi_par)) > epsi_par) &
               call seperr("Apert_dip o1 /= Cig o1")

      if(n_rho /= focusmes_par%n_rho) call seperr("Apert_dip n2 /= Cig n2")
      if(d_rho /= focusmes_par%d_rho) call seperr("Apert_dip d2 /= Cig d2")
      if(abs((rho_0-focusmes_par%rho_0)/(rho_0+epsi_par)) > epsi_par) &
               call seperr("Apert_dip o2 /= Cig o2")

      if(n_dip /= focusmes_par%n_dip) call seperr("Apert_dip n3 /= Cig n3")
      if(d_dip /= focusmes_par%d_dip) call seperr("Apert_dip d3 /= Cig d3")
      if(abs((dip_0-focusmes_par%dip_0)/(dip_0+epsi_par)) > epsi_par) &
               call seperr("Apert_dip o3 /= Cig o3")


      if (focusmes_par%curvature_image) then
      !{
        call from_aux(trim(focusmes_par%tag_apert_dip),'n4',n_rad_curv)
        call from_aux(trim(focusmes_par%tag_apert_dip),'d4',d_rad_curv)
        call from_aux(trim(focusmes_par%tag_apert_dip),'o4',rad_curv_0)

        if(n_rad_curv /= focusmes_par%n_rad_curv) &
                    call seperr("Apert_dip n4 /= Cig n4")
        if(d_rad_curv /= focusmes_par%d_rad_curv) &
                    call seperr("Apert_dip d4 /= Cig d4")
        if(abs((rad_curv_0-focusmes_par%rad_curv_0)/(rad_curv_0+epsi_par)) &
                       > epsi_par) &
                 call seperr("Apert_dip o4 /= Cig o4")

        call from_aux(trim(focusmes_par%tag_apert_dip),'n5',n_mx)
        call from_aux(trim(focusmes_par%tag_apert_dip),'d5',d_mx)
        call from_aux(trim(focusmes_par%tag_apert_dip),'o5',mx_0)

        if(n_mx /= focusmes_par%n_mx) call seperr("Apert_dip n5 /= Cig n5")
        if(d_mx /= focusmes_par%d_mx) call seperr("Apert_dip d5 /= Cig d5")
        if(abs((mx_0-focusmes_par%mx_0)/(mx_0+epsi_par)) > epsi_par) &
                 call seperr("Apert_dip o5 /= Cig o5")
      !}
      else
      !{
        call from_aux(trim(focusmes_par%tag_apert_dip),'n4',n_mx)
        call from_aux(trim(focusmes_par%tag_apert_dip),'d4',d_mx)
        call from_aux(trim(focusmes_par%tag_apert_dip),'o4',mx_0)

        if(n_mx /= focusmes_par%n_mx) call seperr("Apert_dip n4 /= Cig n4")
        if(d_mx /= focusmes_par%d_mx) call seperr("Apert_dip d4 /= Cig d4")
        if(abs((mx_0-focusmes_par%mx_0)/(mx_0+epsi_par)) > epsi_par) &
                 call seperr("Apert_dip o4 /= Cig o4")
      !}
      end if

  
      call from_aux(trim(focusmes_par%tag_apert_dip),'n_z_avg',n_z_avg)
      call from_aux(trim(focusmes_par%tag_apert_dip),'n_mx_avg',n_mx_avg)
  
      call to_history("From Apert_dip: n_z_avg",n_z_avg)
      call to_history("From Apert_dip: n_mx_avg",n_mx_avg)
  
      call from_param('n_z_avg',focusmes_par%n_z_avg,n_z_avg)
      call from_param('n_mx_avg',focusmes_par%n_mx_avg,n_mx_avg)
  
      if(n_z_avg /= focusmes_par%n_z_avg)  &
        write(0,*)'WARNING: Apert_dip n_z_avg /= parameter n_z_avg'
      if(n_mx_avg /= focusmes_par%n_mx_avg)  &
        write(0,*)'WARNING: Apert_dip n_mx_avg /= parameter n_mx_avg'

      call from_param('tag_sum_squares',focusmes_par%tag_sum_squares, &
           'Sum_squares')
      call from_param(trim(focusmes_par%tag_sum_squares),tag_name, &
           'Sum_squares not found')
      focusmes_par%sum_squares_input= &
          auxin(trim(focusmes_par%tag_sum_squares)) > 0
      if(focusmes_par%verb) write(0,*)'sum_squares_input=', &
          focusmes_par%sum_squares_input
      if(focusmes_par%sum_squares_input) then
      !{
      !}
      end if
    !}
    case('apert') aperture_type_1
    !{

      call from_param('tag_apert_apert',focusmes_par%tag_apert_apert,'Apert_apert')
      call from_param(trim(focusmes_par%tag_apert_apert),tag_name,'Apert_apert not found')
      focusmes_par%aperture=auxin(trim(focusmes_par%tag_apert_apert)) > 0
      if(.not. focusmes_par%aperture) call seperr("Apert_apert missing")
  
      call from_aux(trim(focusmes_par%tag_apert_apert),'n1',n_z)
      call from_aux(trim(focusmes_par%tag_apert_apert),'d1',d_z)
      call from_aux(trim(focusmes_par%tag_apert_apert),'o1',z_0)
  
      call from_aux(trim(focusmes_par%tag_apert_apert),'n2',n_rho)
      call from_aux(trim(focusmes_par%tag_apert_apert),'d2',d_rho)
      call from_aux(trim(focusmes_par%tag_apert_apert),'o2',rho_0)

      call from_aux(trim(focusmes_par%tag_apert_apert),'n3',n_gamma)
      call from_aux(trim(focusmes_par%tag_apert_apert),'d3',d_gamma)
      call from_aux(trim(focusmes_par%tag_apert_apert),'o3',gamma_0)
  
      call from_aux(trim(focusmes_par%tag_apert_apert),'n4',n_mx)
      call from_aux(trim(focusmes_par%tag_apert_apert),'d4',d_mx)
      call from_aux(trim(focusmes_par%tag_apert_apert),'o4',mx_0)
  
      if(n_z /= focusmes_par%n_z) call seperr("Apert_apert n1 /= Cig n1")
      if(d_z /= focusmes_par%d_z) call seperr("Apert_apert d1 /= Cig d1")
      if(z_0 /= focusmes_par%z_0) call seperr("Apert_apert o1 /= Cig o1")
  
      if(n_rho /= focusmes_par%n_rho) call seperr("Apert_apert n2 /= Cig n2")
      if(d_rho /= focusmes_par%d_rho) call seperr("Apert_apert d2 /= Cig d2")
      if(rho_0 /= focusmes_par%rho_0) call seperr("Apert_apert o2 /= Cig o2")


      if(n_gamma/=focusmes_par%n_gamma) call seperr("Apert_apert n3 /= Cig n3")
      if(d_gamma/=focusmes_par%d_gamma) call seperr("Apert_apert d3 /= Cig d3")
      if(gamma_0/=focusmes_par%gamma_0) call seperr("Apert_apert o3 /= Cig o3")

      if(n_mx /= focusmes_par%n_mx) call seperr("Apert_apert n4 /= Cig n4")
      if(d_mx /= focusmes_par%d_mx) call seperr("Apert_apert d4 /= Cig d4")
      if(mx_0 /= focusmes_par%mx_0) call seperr("Apert_apert o4 /= Cig o4")

      call from_aux(trim(focusmes_par%tag_apert_apert),'n_z_avg',n_z_avg)
      call from_aux(trim(focusmes_par%tag_apert_apert),'n_mx_avg',n_mx_avg)
  
      call to_history("From Apert_apert: n_z_avg",n_z_avg)
      call to_history("From Apert_apert: n_mx_avg",n_mx_avg)
  
      call from_param('n_z_avg',focusmes_par%n_z_avg,n_z_avg)
      call from_param('n_mx_avg',focusmes_par%n_mx_avg,n_mx_avg)
  
      if(n_z_avg /= focusmes_par%n_z_avg)  &
        write(0,*)'WARNING: Apert_apert n_z_avg /= parameter n_z_avg'
      if(n_mx_avg /= focusmes_par%n_mx_avg)  &
        write(0,*)'WARNING: Apert_dip n_mx_avg /= parameter n_mx_avg'

  
    !}
    case('both ') aperture_type_1
    !{

      call from_param('tag_apert_apert_dip',focusmes_par%tag_apert_apert_dip,&
                    'Apert_apert_dip')
      call from_param(trim(focusmes_par%tag_apert_apert_dip),tag_name,&
                    'Apert_apert_dip not found')
      focusmes_par%aperture=auxin(trim(focusmes_par%tag_apert_apert_dip)) > 0
      if(.not. focusmes_par%aperture) call seperr("Apert_apert_dip missing")
  
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'n1',n_z)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'d1',d_z)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'o1',z_0)

      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'n2',n_rho)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'d2',d_rho)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'o2',rho_0)

      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'n3',n_gamma)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'d3',d_gamma)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'o3',gamma_0)
  
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'n4',n_dip)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'d4',d_dip)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'o4',dip_0)

      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'n5',n_mx)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'d5',d_mx)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'o5',mx_0)

  
      if(n_z /= focusmes_par%n_z) call seperr("Apert_apert_dip n1 /= Cig n1")
      if(d_z /= focusmes_par%d_z) call seperr("Apert_apert_dip d1 /= Cig d1")
      if(z_0 /= focusmes_par%z_0) call seperr("Apert_apert_dip o1 /= Cig o1")

      if(n_rho /= focusmes_par%n_rho)  &
              call seperr("Apert_apert_dip n2 /= Cig n2")
      if(d_rho /= focusmes_par%d_rho)  &
              call seperr("Apert_apert_dip d2 /= Cig d2")
      if(rho_0 /= focusmes_par%rho_0)  &
              call seperr("Apert_apert_dip o2 /= Cig o2")
  

      if(n_gamma/=focusmes_par%n_gamma)  &
              call seperr("Apert_apert_dip n3 /= Cig n3")
      if(d_gamma/=focusmes_par%d_gamma)  &
              call seperr("Apert_apert_dip d3 /= Cig d3")
      if(gamma_0/=focusmes_par%gamma_0)  &
              call seperr("Apert_apert_dip o3 /= Cig o3")

      if(n_dip /= focusmes_par%n_dip)  &
              call seperr("Apert_apert_dip n4 /= Cig n4")
      if(d_dip /= focusmes_par%d_dip)  &
              call seperr("Apert_apert_dip d4 /= Cig d4")
      if(dip_0 /= focusmes_par%dip_0)  &
              call seperr("Apert_apert_dip o4 /= Cig o4")

      if(n_mx /= focusmes_par%n_mx) call seperr("Apert_apert_dip n5 /= Cig n5")
      if(d_mx /= focusmes_par%d_mx) call seperr("Apert_apert_dip d5 /= Cig d5")
      if(mx_0 /= focusmes_par%mx_0) call seperr("Apert_apert_dip o5 /= Cig o5")

      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'n_z_avg',n_z_avg)
      call from_aux(trim(focusmes_par%tag_apert_apert_dip),'n_mx_avg',n_mx_avg)
  
      call to_history("From Apert_apert_dip: n_z_avg",n_z_avg)
      call to_history("From Apert_apert_dip: n_mx_avg",n_mx_avg)
  
      call from_param('n_z_avg',focusmes_par%n_z_avg,n_z_avg)
      call from_param('n_mx_avg',focusmes_par%n_mx_avg,n_mx_avg)
  
      if(n_z_avg /= focusmes_par%n_z_avg)  &
        write(0,*)'WARNING: Apert_apert n_z_avg /= parameter n_z_avg'
      if(n_mx_avg /= focusmes_par%n_mx_avg)  &
        write(0,*)'WARNING: Apert_apert n_mx_avg /= parameter n_mx_avg'

    !}
    case default aperture_type_1
      write(0,*)'aperture_type=',focusmes_par%aperture_type(1:5)
      call seperr ('Do not know what to do with this aperture_type')
    !}
    end select aperture_type_1

    call from_param('tag_semblance',focusmes_par%tag_semblance,'Semblance')
    call from_param(trim(focusmes_par%tag_semblance),tag_name, &
          'Semblance not found')
    write(0,*)'focusmes_par%tag_semblance=',trim(focusmes_par%tag_semblance)
    
    if(auxinout(trim(focusmes_par%tag_semblance)) > 0) then
      if(focusmes_par%verb) write(0,*)'Semblance already exists'
    end if

!    call from_param('global_threshold',focusmes_par%global_threshold,zero_real)
!    call from_param('local_threshold',focusmes_par%local_threshold,zero_real)

    esize=4
    call to_history("esize",esize,trim(focusmes_par%tag_semblance))
    call to_history("n1",focusmes_par%n_z,trim(focusmes_par%tag_semblance))
    call to_history("d1",focusmes_par%d_z,trim(focusmes_par%tag_semblance))
    call to_history("o1",focusmes_par%z_0,trim(focusmes_par%tag_semblance))
    call to_history("label1",focusmes_par%label_z,trim(focusmes_par%tag_semblance))

    call to_history("n2",focusmes_par%n_rho,trim(focusmes_par%tag_semblance))
    call to_history("d2",focusmes_par%d_rho,trim(focusmes_par%tag_semblance))
    call to_history("o2",focusmes_par%rho_0,trim(focusmes_par%tag_semblance))
    call to_history("label2",focusmes_par%label_rho, &
                 trim(focusmes_par%tag_semblance))

    curvature_semblance_1: if (focusmes_par%curvature_semblance) then
    !{
        call to_history("n3",focusmes_par%n_rad_curv, &
                   trim(focusmes_par%tag_semblance))
        call to_history("d3",focusmes_par%d_rad_curv, &
                   trim(focusmes_par%tag_semblance))
        call to_history("o3",focusmes_par%rad_curv_0, &
                   trim(focusmes_par%tag_semblance))
        call to_history("label3",focusmes_par%label_rad_curv, &
                   trim(focusmes_par%tag_semblance))

        call to_history("n4",focusmes_par%n_mx,trim(focusmes_par%tag_semblance))
        call to_history("d4",focusmes_par%d_mx,trim(focusmes_par%tag_semblance))
        call to_history("o4",focusmes_par%mx_0,trim(focusmes_par%tag_semblance))
        call to_history("label4",focusmes_par%label_mx, &
                 trim(focusmes_par%tag_semblance))

    !}
    else curvature_semblance_1
    !{
      call to_history("n3",focusmes_par%n_mx,trim(focusmes_par%tag_semblance))
      call to_history("d3",focusmes_par%d_mx,trim(focusmes_par%tag_semblance))
      call to_history("o3",focusmes_par%mx_0,trim(focusmes_par%tag_semblance))
      call to_history("label3",focusmes_par%label_mx, &
                 trim(focusmes_par%tag_semblance))

      call to_history("n4",one_int,trim(focusmes_par%tag_semblance))
      call to_history("d4",one_real,trim(focusmes_par%tag_semblance))
      call to_history("o4",zero_real,trim(focusmes_par%tag_semblance))
    !}
    end if curvature_semblance_1


    if(.not. focusmes_par%prestack_image)  &
           call seperr("Computing Semblance needs a prestack_image")
    if(focusmes_par%write_prestack)  &
           call seperr("Cannot output prestack image when computing semblance")

  !}
  case('aperture ') oper_name_1
  !{

    if(focusmes_par%dip_ang_image) then
      focusmes_par%aperture_type="dip  "
    else
      if(focusmes_par%apert_ang_image) then
        focusmes_par%aperture_type="apert"
      else
        focusmes_par%aperture_type="none "
      end if
    end if
    if((focusmes_par%dip_ang_image).and.(focusmes_par%apert_ang_image)) then
      focusmes_par%aperture_type="both "
    end if
    if(focusmes_par%verb) write(0,*)'aperture_type=',focusmes_par%aperture_type

    call from_param('max_dip_prop',focusmes_par%max_dip_prop,-1.)
    call from_param('max_dip_range',focusmes_par%max_dip_range,-1.)

    call from_param('tag_aperture',focusmes_par%tag_aperture,'Aperture')
    call from_param(trim(focusmes_par%tag_aperture),tag_name, &
          'Aperture not found')
    write(0,*)'focusmes_par%tag_aperture=',trim(focusmes_par%tag_aperture)
    
    if(auxinout(trim(focusmes_par%tag_aperture)) > 0) then
      if(focusmes_par%verb) write(0,*)'Aperture already exists'
    end if

    if(.not. focusmes_par%write_prestack)&
       call seperr("Must write prestack with aperture")

    call from_param('n_z_avg',focusmes_par%n_z_avg,zero_int)
    call from_param('n_mx_avg',focusmes_par%n_mx_avg,zero_int)

    if(focusmes_par%n_mx_avg /=0)  &
        call seperr('transp_image=y and n_mx_avg MUST be == 0')

    call from_param('global_threshold',focusmes_par%global_threshold,zero_real)
    call from_param('local_threshold',focusmes_par%local_threshold,zero_real)

    call to_history("n_z_avg",focusmes_par%n_z_avg, &
                trim(focusmes_par%tag_aperture))
    call to_history("n_mx_avg",focusmes_par%n_mx_avg, &
                trim(focusmes_par%tag_aperture))

    esize=4
    call to_history("esize",esize,trim(focusmes_par%tag_aperture))
    call to_history("n1",focusmes_par%n_z,trim(focusmes_par%tag_aperture))
    call to_history("d1",focusmes_par%d_z,trim(focusmes_par%tag_aperture))
    call to_history("o1",focusmes_par%z_0,trim(focusmes_par%tag_aperture))
    call to_history("label1",focusmes_par%label_z, &
           trim(focusmes_par%tag_aperture))

    call to_history("n2",focusmes_par%n_rho,trim(focusmes_par%tag_aperture))
    call to_history("d2",focusmes_par%d_rho,trim(focusmes_par%tag_aperture))
    call to_history("o2",focusmes_par%rho_0,trim(focusmes_par%tag_aperture))
    call to_history("label2",focusmes_par%label_rho, &
           trim(focusmes_par%tag_aperture))


    aperture_type_2: select case (focusmes_par%aperture_type(1:5))
    case('apert') aperture_type_2
      if (focusmes_par%curvature_image) then
      !{
        call to_history("n3",focusmes_par%n_gamma,trim(focusmes_par%tag_aperture))
        call to_history("d3",focusmes_par%d_gamma,trim(focusmes_par%tag_aperture))
        call to_history("o3",focusmes_par%gamma_0,trim(focusmes_par%tag_aperture))
        call to_history("label3",focusmes_par%label_gamma, &
             trim(focusmes_par%tag_aperture))
  
        call to_history("n4",focusmes_par%n_rad_curv, &
                   trim(focusmes_par%tag_aperture))
        call to_history("d4",focusmes_par%d_rad_curv, &
                   trim(focusmes_par%tag_aperture))
        call to_history("o4",focusmes_par%rad_curv_0, &
                   trim(focusmes_par%tag_aperture))
        call to_history("label4",focusmes_par%label_rad_curv, &
                   trim(focusmes_par%tag_aperture))

        call to_history("n5",focusmes_par%n_mx,trim(focusmes_par%tag_aperture))
        call to_history("d5",focusmes_par%d_mx,trim(focusmes_par%tag_aperture))
        call to_history("o5",focusmes_par%mx_0,trim(focusmes_par%tag_aperture))
        call to_history("label5",focusmes_par%label_mx, &
             trim(focusmes_par%tag_aperture))
      !}
      else
      !{
        call to_history("n3",focusmes_par%n_gamma,trim(focusmes_par%tag_aperture))
        call to_history("d3",focusmes_par%d_gamma,trim(focusmes_par%tag_aperture))
        call to_history("o3",focusmes_par%gamma_0,trim(focusmes_par%tag_aperture))
        call to_history("label3",focusmes_par%label_gamma, &
             trim(focusmes_par%tag_aperture))
  
        call to_history("n4",focusmes_par%n_mx,trim(focusmes_par%tag_aperture))
        call to_history("d4",focusmes_par%d_mx,trim(focusmes_par%tag_aperture))
        call to_history("o4",focusmes_par%mx_0,trim(focusmes_par%tag_aperture))
        call to_history("label4",focusmes_par%label_mx, &
             trim(focusmes_par%tag_aperture))
      !}
      end if

    case('dip  ') aperture_type_2

      call to_history("n3",focusmes_par%n_dip,trim(focusmes_par%tag_aperture))
      call to_history("d3",focusmes_par%d_dip,trim(focusmes_par%tag_aperture))
      call to_history("o3",focusmes_par%dip_0,trim(focusmes_par%tag_aperture))
      call to_history("label3",focusmes_par%label_dip, &
           trim(focusmes_par%tag_aperture))
  
      if (focusmes_par%curvature_image) then
      !{
        call to_history("n4",focusmes_par%n_rad_curv, &
                   trim(focusmes_par%tag_aperture))
        call to_history("d4",focusmes_par%d_rad_curv, &
                   trim(focusmes_par%tag_aperture))
        call to_history("o4",focusmes_par%rad_curv_0, &
                   trim(focusmes_par%tag_aperture))
        call to_history("label4",focusmes_par%label_rad_curv, &
                   trim(focusmes_par%tag_aperture))

        call to_history("n5",focusmes_par%n_mx,trim(focusmes_par%tag_aperture))
        call to_history("d5",focusmes_par%d_mx,trim(focusmes_par%tag_aperture))
        call to_history("o5",focusmes_par%mx_0,trim(focusmes_par%tag_aperture))
        call to_history("label5",focusmes_par%label_mx, &
             trim(focusmes_par%tag_aperture))

      !}
      else
      !{
        call to_history("n4",focusmes_par%n_mx,trim(focusmes_par%tag_aperture))
        call to_history("d4",focusmes_par%d_mx,trim(focusmes_par%tag_aperture))
        call to_history("o4",focusmes_par%mx_0,trim(focusmes_par%tag_aperture))
        call to_history("label4",focusmes_par%label_mx, &
             trim(focusmes_par%tag_aperture))
      !}
      end if

    case('both ') aperture_type_2
      if (focusmes_par%curvature_image) then
      !{
        call to_history("n3",focusmes_par%n_gamma,trim(focusmes_par%tag_aperture))
        call to_history("d3",focusmes_par%d_gamma,trim(focusmes_par%tag_aperture))
        call to_history("o3",focusmes_par%gamma_0,trim(focusmes_par%tag_aperture))
        call to_history("label3",focusmes_par%label_gamma, &
             trim(focusmes_par%tag_aperture))
        call to_history("n4",focusmes_par%n_dip,trim(focusmes_par%tag_aperture))
        call to_history("d4",focusmes_par%d_dip,trim(focusmes_par%tag_aperture))
        call to_history("o4",focusmes_par%dip_0,trim(focusmes_par%tag_aperture))
        call to_history("label4",focusmes_par%label_dip, &
             trim(focusmes_par%tag_aperture))

        call to_history("n5",focusmes_par%n_rad_curv, &
                   trim(focusmes_par%tag_semblance))
        call to_history("d5",focusmes_par%d_rad_curv, &
                   trim(focusmes_par%tag_semblance))
        call to_history("o5",focusmes_par%rad_curv_0, &
                   trim(focusmes_par%tag_semblance))
        call to_history("label5",focusmes_par%label_rad_curv, &
                   trim(focusmes_par%tag_semblance))

        call to_history("n6",focusmes_par%n_mx,trim(focusmes_par%tag_aperture))
        call to_history("d6",focusmes_par%d_mx,trim(focusmes_par%tag_aperture))
        call to_history("o6",focusmes_par%mx_0,trim(focusmes_par%tag_aperture))
        call to_history("label6",focusmes_par%label_mx, &
             trim(focusmes_par%tag_aperture))
      !}
      else
      !{
        call to_history("n3",focusmes_par%n_gamma,trim(focusmes_par%tag_aperture))
        call to_history("d3",focusmes_par%d_gamma,trim(focusmes_par%tag_aperture))
        call to_history("o3",focusmes_par%gamma_0,trim(focusmes_par%tag_aperture))
        call to_history("label3",focusmes_par%label_gamma, &
             trim(focusmes_par%tag_aperture))
        call to_history("n4",focusmes_par%n_dip,trim(focusmes_par%tag_aperture))
        call to_history("d4",focusmes_par%d_dip,trim(focusmes_par%tag_aperture))
        call to_history("o4",focusmes_par%dip_0,trim(focusmes_par%tag_aperture))
        call to_history("label4",focusmes_par%label_dip, &
             trim(focusmes_par%tag_aperture))
  
        call to_history("n5",focusmes_par%n_mx,trim(focusmes_par%tag_aperture))
        call to_history("d5",focusmes_par%d_mx,trim(focusmes_par%tag_aperture))
        call to_history("o5",focusmes_par%mx_0,trim(focusmes_par%tag_aperture))
        call to_history("label5",focusmes_par%label_mx, &
             trim(focusmes_par%tag_aperture))
      !}
      end if

    case default aperture_type_2
      write(0,*)'aperture_type=',focusmes_par%aperture_type(1:5)
      call seperr ('Do not know what to do with this aperture_type')
    end select aperture_type_2

  
    if(.not. focusmes_par%prestack_image) call seperr("Computing Aperture needs a prestack_image")

  !}
  case default oper_name_1
  !{
    write(0,*)'oper_name=',focusmes_par%oper_name(1:9)
    call seperr ('Do not know what to do (#1 in param_focusmes_transp)')
  !}
  end select oper_name_1

  return
end subroutine param_focusmes_transp
