module ResMig_mod

  use resmig_types
  use fft_types

  use fft_data_mod

  implicit none

  integer, external ::  sreed,srite,sseek_block

  logical,private :: full_3d_fft,cube_4d_fft
  integer, private :: n_z,n_mx,n_z_pos,n_mx_pos
  integer, private :: n_z_pad,n_mx_pad,n_z_pad_pos,n_mx_pad_pos
  integer, private :: n_hx,n_hx_pad,n_hx_pad_pos,n_gamma,n_rho,n_dip,n_rad_curv
  integer, private :: n_dip_0_curv,n3_inp,n4_out,n_cube
  integer, private :: n_bytes_cube_inp,n_bytes_cube_out,n_bytes_k_cube
  integer, private :: n_bytes_nz_cmpx,n_bytes_hx_cmpx,n_bytes_drho,n_bytes_dips_0
  integer, private :: i_hx_0,i_dip_flip
  integer, private :: n_threads
  real, private :: d_z,d_mx,d_hx,d_gamma,d_rho,d_dip,d_rad_curv,d_dip_0_curv
  real, private :: z_0,mx_0,hx_0,hz_0,gamma_0,rho_0,dip_0,rad_curv_0,dip_0_curv_0
  real, private :: dip_0_curv
  real, private :: dip_flip,rad_curv_scale
  real, private :: kz_0,kmx_0,khx_0,d_kz,d_kmx,d_khx
  real, private :: pi,max_dip,max_gamma,max_gamma_0,max_dip_0
  real, private :: max_prop_ang,max_prop_ang_0,z_ref

  real, allocatable, dimension (:), private :: k_mx,k_z,k_hx
  real, allocatable, dimension (:), private :: hx
  real, allocatable, dimension (:), private :: temp_gamma
  real, allocatable, dimension (:,:,:), private :: stack_cube
  real, allocatable, dimension (:,:,:), private :: dips_0
  real, allocatable, dimension (:,:,:,:,:), private :: cube
  complex, allocatable, dimension (:), private :: temp_hx
  complex, allocatable, dimension (:,:), private :: temp_z,temp_mx
  complex, allocatable, dimension (:,:,:), private :: k_cube_inp
  complex, allocatable, dimension (:,:,:,:), private :: k_cube
  complex, allocatable, dimension (:,:,:), private :: temp_mx_hx,temp_z_mx

  type (fft_type), dimension(:), pointer, private :: fft_nz_id,fft_nmx_id
  type (fft_type), dimension(:), pointer, private :: fft_nmx_nhx_id,fft_nz_nmx_id
!  type (fft_type), dimension(:), pointer, private :: fft_nz_id,fft_nmx_id,fft_nhx_id

contains

  function ResMig_init(resmig_par)  result(status)
    type (resmig_param_type) resmig_par
    integer status

    integer i_k_mx,i_k_z,i_k_hx,ierr,i_z,i_hx,i_thread
    real kz,kmx,khx

    status=0


    pi=acos(-1.)
    max_prop_ang=(resmig_par%max_prop_ang/180.)*pi
    max_dip=(resmig_par%max_dip/180.)*pi
    max_gamma=(resmig_par%max_gamma/180.)*pi

    max_prop_ang_0=min(.499*pi,1.2*max_prop_ang)
!    max_prop_ang_0=min(.499*pi,1.*max_prop_ang)
    max_dip_0=min(.499*pi,1.2*max_dip)
    max_gamma_0=min(.499*pi,1.2*max_gamma)

    z_ref=resmig_par%z_ref

    n_z=resmig_par%n_z
    z_0=resmig_par%z_0
    d_z=resmig_par%d_z

    n_mx=resmig_par%n_mx
    mx_0=resmig_par%mx_0
    d_mx=resmig_par%d_mx

    if (resmig_par%apert_ang_image) then
      n_gamma=resmig_par%n_gamma
      gamma_0=resmig_par%gamma_0*(pi/180.)
      d_gamma=resmig_par%d_gamma*(pi/180.)
      n3_inp=n_gamma
    else
      n_hx=resmig_par%n_hx
      hx_0=resmig_par%hx_0
      d_hx=resmig_par%d_hx

      n_hx_pad=resmig_par%n_hx_pad
      n_hx_pad_pos=(n_hx_pad/2)
      khx_0=0.
      d_khx=2.*pi/(d_hx*FLOAT(n_hx_pad))
      n3_inp=n_hx_pad
    end if

    n_rho=resmig_par%n_rho
    rho_0=resmig_par%rho_0
    d_rho=resmig_par%d_rho


    curvature_image_1 : if(resmig_par%curvature_image) then
    !{

      if(resmig_par%read_dips_0) then
        n_dip_0_curv=resmig_par%n_dip_0_curv
        dip_0_curv_0=resmig_par%dip_0_curv_0*(pi/180)
        d_dip_0_curv=resmig_par%d_dip_0_curv*(pi/180)

        n_bytes_dips_0=n_z*n_mx*n_rho*4
        allocate (dips_0(n_z,n_mx,n_rho),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of dips_0 ")

      else
        n_dip_0_curv=1
        dip_0_curv_0=resmig_par%dip_0_curv*(pi/180.)
        d_dip_0_curv=0.
      end if

      if (resmig_par%linear_curvature_scale) then
        n_rad_curv=resmig_par%n_rad_curv
        rad_curv_0=resmig_par%rad_curv_0
        d_rad_curv=resmig_par%d_rad_curv
        rad_curv_scale=resmig_par%rad_curv_scale
      else
        n_rad_curv=resmig_par%n_rad_curv
        rad_curv_0=pi*resmig_par%rad_curv_0
        d_rad_curv=pi*resmig_par%d_rad_curv
        rad_curv_scale=resmig_par%rad_curv_scale
        dip_0_curv=resmig_par%dip_0_curv*(pi/180.)
      end if
    !}
    else curvature_image_1 
    !{

      n_dip_0_curv=1
      dip_0_curv_0=0
      d_dip_0_curv=0.

      n_rad_curv=resmig_par%n_rad_curv
      rad_curv_0=pi*resmig_par%rad_curv_0
      d_rad_curv=pi*resmig_par%d_rad_curv
      rad_curv_scale=resmig_par%rad_curv_scale
    !}
    end if curvature_image_1 

    n_cube=resmig_par%n_cube

    dip_ang_image_1 : if (resmig_par%dip_ang_image) then
    !{
      n_dip=resmig_par%n_dip
      dip_0=resmig_par%dip_0
      d_dip=resmig_par%d_dip
      n4_out=n_dip
      cube_4d_fft=.true.


      flip_sign_dip_1:  &
      if(resmig_par%flip_sign_dip < huge(resmig_par%flip_sign_dip)) then
        dip_flip=resmig_par%flip_sign_dip 
        if(d_dip <= 0.) call seperr ("d_dip must be >=0 when flip_sign_dip is set")
        if(abs(mod(sign((dip_flip-dip_0),d_dip),d_dip)) > .01*d_dip) then
              write(0,*)'d_dip=', .01*d_dip
              write(0,*)'mod(sign(dip_flip-dip_0,d_dip),d_dip)=', &
                      mod(sign(dip_flip-dip_0,d_dip),d_dip)
              call seperr ("dip_flip-dip_0 must be multiple of d_dip")
        end if
        i_dip_flip=(dip_flip-dip_0)/d_dip+1
        if(resmig_par%debug1) write(0,*)'i_dip_flip=',i_dip_flip
        if((i_dip_flip <=1) .or. (i_dip_flip > n_dip)) then
          call seperr("Dip range must contain neg. dips when flip_sign_dip is set")
        end if
      else flip_sign_dip_1
        i_dip_flip=-1
      end if flip_sign_dip_1
      dip_0=dip_0*(pi/180.)
      d_dip=d_dip*(pi/180.)
    !}
    else dip_ang_image_1
    !{
      n4_out=1
      cube_4d_fft=.false.
    !}
    end if dip_ang_image_1

    n_z_pad=resmig_par%n_z_pad
    n_z_pos=resmig_par%n_z/2
    n_z_pad_pos=resmig_par%n_z_pad/2
    if(mod(n_z_pad,2) /=0) then
      write(0,*)'n_z_pad=',n_z_pad
      call seperr ("n_z_pad must be even")
    end if

    n_mx_pad=resmig_par%n_mx_pad
    n_mx_pos=resmig_par%n_mx/2
    n_mx_pad_pos=resmig_par%n_mx_pad/2
    if(mod(n_mx_pad,2) /=0) then
      write(0,*)'n_mx_pad=',n_mx_pad
      call seperr ("n_mx_pad must be even")
    end if

    kz_0=0.
    kmx_0=0.
    d_kz=2.*pi/(d_z*FLOAT(n_z_pad))
    d_kmx=2.*pi/(d_mx*FLOAT(n_mx_pad))

    n_bytes_cube_inp=n_z*n_mx*n3_inp*4

    write_prestack_1: if(resmig_par%write_prestack) then
      n_bytes_cube_out=n_z*n_mx*n3_inp*n4_out*4
    else write_prestack_1
      n_bytes_cube_out=n_z*n_mx*n4_out*4
!      if ((resmig_par%dip_ang_image).or.(resmig_par%apert_ang_image)) then
        allocate (stack_cube(n_z,n_mx,n4_out),stat=ierr)  
        if(ierr .ne. 0)  then
          write(0,*)'n_z=',n_z
          write(0,*)'n_mx=',n_mx
          write(0,*)'n4_out=',n4_out
          if(ierr .ne. 0) call seperr("Problems with allocation of stack_cube ")
        end if
        stack_cube=0.
!      else
!        allocate (stack_cube(n_z,n_mx,n4_out),stat=ierr)  
!          if(ierr .ne. 0)  then
!            write(0,*)'n_z=',n_z
!            write(0,*)'n_mx=',n_mx
!            write(0,*)'n4_out=',n4_out
!            call seperr("Problems with allocation of stack_cube ")
!          end if
!          stack_cube=0.
!        end if
!      end if
    end if write_prestack_1


    n_threads=resmig_par%n_threads

    allocate (cube(n_z,n_mx,n3_inp,n4_out,n_dip_0_curv),stat=ierr)  
    if(ierr .ne. 0) then
       write(0,*)'n_z=',n_z
       write(0,*)'n_mx=',n_mx
       write(0,*)'n3_inp=',n3_inp
       write(0,*)'n4_out=',n4_out
       write(0,*)'n_dip_0_curv=',n_dip_0_curv
       call seperr("Problems with allocation of cube ")
    end if
    cube=0.

  select case (resmig_par%oper_name(1:9))
  !{
    case('oper-rmov','oper-rmig')
    !{

      full_3d_fft=.not. resmig_par%apert_ang_image
    
      allocate (k_z(n_z_pad),stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of k_z ")
      allocate (k_mx(n_mx_pad),stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of k_mx ")

      allocate (k_cube(n_z_pad_pos+1,n_mx_pad,n3_inp,n4_out),stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of k_cube ")

      allocate (k_cube_inp(n_z_pad_pos+1,n_mx_pad,n3_inp),stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of k_cube_inp ")

      allocate (temp_z(n_z_pad,n_threads),stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of temp_z ")


      do i_k_z=1,n_z_pad
         kz  = i_k_z - 1.
         if(kz <= float(n_z_pad/2)) then
            k_z(i_k_z) =  kz_0+d_kz* kz
         else
            k_z(i_k_z) = -kz_0+d_kz*(kz-n_z_pad)
         end if
      end do

        call snap('K_z.H',n_z_pad,1,k_z)
  
      do i_k_mx=1,n_mx_pad
         kmx  = i_k_mx - 1.
         if(kmx <= float(n_mx_pad/2)) then
            k_mx(i_k_mx) =  kmx_0+d_kmx* kmx
         else
            k_mx(i_k_mx) = -kmx_0+d_kmx*(kmx-n_mx_pad)
         end if
      end do

        call snap('K_mx.H',n_mx_pad,1,k_mx)

      allocate (fft_nz_id(n_threads),stat=ierr)  
      if(ierr .ne. 0) call seperr("Problems with allocation of fft_nz_id")
      if(resmig_par%debug2) write(0,*)'Allocate fft_nz_id with shape=',shape(fft_nz_id)
  
      do i_thread=1,n_threads
        call init_fft_1d(fft_nz_id(i_thread),n_z_pad, &
                temp_z(:,i_thread),temp_z(:,i_thread),fast_init=.false.)  
        if(resmig_par%debug2) write(0,*)'scale (n_z_pad)=',fft_nz_id(i_thread)%scale
      end do

      if(resmig_par%apert_ang_image) then
      !{
        allocate (temp_mx(n_mx_pad,n_threads),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of temp_mx ")
        allocate (temp_gamma(n_gamma),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of temp_gamma ")

        allocate (fft_nmx_id(n_threads),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of fft_nmx_id")
        if(resmig_par%debug2) write(0,*)'Allocate fft_nmx_id with shape=',shape(fft_nmx_id)
        do i_thread=1,n_threads
          call init_fft_1d(fft_nmx_id(i_thread),n_mx_pad, &
                  temp_mx(:,i_thread),temp_mx(:,i_thread),fast_init=.false.)  
          if(resmig_par%debug2) write(0,*)'scale (n_mx_pad)=',fft_nmx_id(i_thread)%scale
        end do
  
      !}
      else
      !{
        allocate (temp_mx_hx(n_mx_pad,n_hx_pad,n_threads),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of temp_mx_hx ")
        allocate (temp_hx(n_hx_pad),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of temp_hx ")
        allocate (k_hx(n_hx_pad),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of k_hx ")
        do i_k_hx=1,n_hx_pad
           khx  = i_k_hx - 1.
           if(khx <= float(n_hx_pad/2)) then
              k_hx(i_k_hx) =  khx_0+d_khx* khx
           else
              k_hx(i_k_hx) = -khx_0+d_khx*(khx-n_hx_pad)
           end if
        end do
!        call snap('K_hx.H',n_hx_pad,1,k_hx)

        allocate (fft_nmx_nhx_id(n_threads),stat=ierr)  
        if(ierr .ne. 0) call seperr("Problems with allocation of fft_nmx_nhx_id")
        if(resmig_par%debug2) write(0,*)'Allocate fft_nmx_nhx_id with shape=',shape(fft_nmx_nhx_id)
  
        do i_thread=1,n_threads
          call init_fft_2d(fft_nmx_nhx_id(i_thread),n_mx_pad,n_hx_pad, &
                  temp_mx_hx(:,:,i_thread),temp_mx_hx(:,:,i_thread),fast_init=.false.)  
          if(resmig_par%debug2) write(0,*)'scale (n_mx_n_hx)=',fft_nmx_nhx_id(i_thread)%scale
        end do

      !}
      end if
    
    !}

      if(.not. resmig_par%apert_ang_image) then
      if(mod(hx_0,d_hx) /= 0.) &
            call seperr ("hx_0 must be multiple of d_hx")
      i_hx_0=1-hx_0/d_hx
      if((i_hx_0 <1) .or. (i_hx_0 > n_hx)) then
        call seperr ("CIG must contain zero offset")
      end if
    end if

    ! }
    case default
      call seperr ('Do not know what to do (#1)')
  !}
  end select

     if(resmig_par%debug1) then
write(0,*)'status=',status
write(0,*)'full_3d_fft=',full_3d_fft
write(0,*)'cube_4d_fft=',cube_4d_fft
write(0,*)'n3_inp=',n3_inp
write(0,*)'n4_out=',n4_out
write(0,*)'n_gamma=',n_gamma
write(0,*)'gamma_0=',gamma_0
write(0,*)'d_gamma=',d_gamma
write(0,*)'n_hx=',n_hx
write(0,*)'hx_0=',hx_0
write(0,*)'d_hx=',d_hx
write(0,*)'n_z=',n_z
write(0,*)'n_mx=',n_mx
write(0,*)'n_z_pad=',n_z_pad
write(0,*)'n_mx_pad=',n_mx_pad
write(0,*)'n_hx_pad=',n_hx_pad
write(0,*)'n_z_pos=',n_z_pos
write(0,*)'n_mx_pos=',n_mx_pos
write(0,*)'n_z_pad_pos=',n_z_pad_pos
write(0,*)'n_mx_pad_pos=',n_mx_pad_pos
write(0,*)'n_hx_pad_pos=',n_hx_pad_pos
write(0,*)'d_z=',d_z
write(0,*)'d_mx=',d_mx
write(0,*)'d_kz=',d_kz
write(0,*)'d_kmx=',d_kmx
write(0,*)'d_khx=',d_khx
write(0,*)'n_bytes_cube_inp=',n_bytes_cube_inp
write(0,*)'n_bytes_cube_out=',n_bytes_cube_out
write(0,*)'n_bytes_drho=',n_bytes_drho
write(0,*)'n_bytes_k_cube=',n_bytes_k_cube
write(0,*)'shape(cube)=',shape(cube)
write(0,*)'shape(k_cube)=',shape(k_cube)
write(0,*)'shape(stack_cube)=',shape(stack_cube)
write(0,*)'max_dip=',max_dip*(180./pi)
write(0,*)'max_gamma=',max_gamma*(180./pi)
write(0,*)'max_dip_0=',max_dip_0*(180./pi)
write(0,*)'max_gamma_0=',max_gamma_0*(180./pi)
write(0,*)'max_prop_ang=',max_prop_ang*(180./pi)
write(0,*)'max_prop_ang_0=',max_prop_ang_0*(180./pi)
write(0,*)'i_hx_0=',i_hx_0
write(0,*)'n_rad_curv=',n_rad_curv
write(0,*)'dip_0_curv=',dip_0_curv*(180./pi)
     end if


  end function ResMig_init

  function ResMig_oper(resmig_par) result(status)

    type (resmig_param_type) resmig_par
    integer status

!    logical hz_foc_inside,hx_foc_inside,hdip_inside,inside_k_hdip
    integer ret_sseek
    integer i_hx,i_mx,i_z,i_thread,i_k_hx,i_k_z,i_k_mx,i_3,i_gamma
    integer i_rho,i_rad_curv,i_dip_0_curv,i_cube
    integer i_dip,i_dip_p1,max_i_dip,i_interp_dip,i_interp_dip_p1
!    integer i_hfoc,i_hz_foc,i_hx_foc,i_hpar,i_hdip,i_mx,i_z,i_thread
!    integer i_k_hdip,i_gamma
!    real h_foc,hz_foc,hx_foc,hz_min,hz_max,hx_min,hx_max,kz,kmx,kz_out
!    real w_hz,w_hx,w_hz_foc,w_hx_foc,sin_rot_map,cos_rot_map
!    real hdip_min,hdip_max,h_par,h_dip
!    real tan_dip,dip,diff_dip,sin_diff,cos_diff,w_hdip,z_shift
!    real tan_gamma,k_hdip_inp,inv_d_khdip,inv_d_z
    real norm_shift,curv_z_shift,curv_norm_shift,z_shift,mx_shift
    real delta_rho,taper,rho_nl,cos_dip_nl
    real tan_dip,cos_dip,sin_dip,dip,w_dip,interp_dip
    real tan_dip_sh,cos_dip_sh,sin_dip_sh,dip_sh
!    real cos_dip_0_curv_sh,sin_dip_0_curv_sh
    real tan_gamma,cos_gamma,sin_gamma,gamma,floor,prop_ang
    real hx_max,sign_p1,sign_m1,sign_shift,lambda
    real ks_sq,kg_sq
    real rad_curv
    complex interp_temp

!    cos_dip_0_curv_sh=cos(dip_0_curv)
!    sin_dip_0_curv_sh=sin(dip_0_curv)

    do_i_cube_1: do i_cube=1,n_cube
    !{

    adj_1: if (resmig_par%adj)  then
    !{

    !}
    else adj_1
    !{

      if (resmig_par%inverse_mapping) then
        sign_shift=-1
      else
        sign_shift=1
      end if
!      sign_p1=1.
!      sign_m1=-1.

      if(resmig_par%verb) write(0,*)'sreed i_rho=',1
      if(sreed(trim(resmig_par%tag_cig),cube(:,:,:,1,1),n_bytes_cube_inp)/= &
          n_bytes_cube_inp) then
        write(0,*)'tag_cig=',trim(resmig_par%tag_cig)
        call seperr('Problems with reading tag_cig file')
      end if

      k_cube=cmplx(0.,0.)
      status= ResMig_transf_cube(.true.,full_3d_fft,.false.,resmig_par,i_5_cube=1)
      k_cube_inp(:,:,:)=k_cube(:,:,:,1)

      if(resmig_par%write_cig_transf) then
        if(srite(trim(resmig_par%tag_cig_transf),k_cube_inp,n_bytes_k_cube)/= &
            n_bytes_k_cube) then
          write(0,*)'tag_cig_transf=',trim(resmig_par%tag_cig_transf)
          call seperr('Problems with writing tag_resmig file')
        end if
        ret_sseek=sseek_block( &
              trim(resmig_par%tag_cig_transf),0,n_bytes_k_cube,0)
        if (ret_sseek .ne. 0) then
        !{
          write(0,*)'ret_sseek=',ret_sseek
          write(0,*)'tag_cig_transf=',trim(resmig_par%tag_cig_transf)
          call seperr("Error block seeking cig_transf in input")
        !}
        end if
      end if

      if(resmig_par%read_dips_0) then
          if(sreed(trim(resmig_par%tag_dips_0),dips_0, &
              n_bytes_dips_0)/= n_bytes_dips_0) then
            write(0,*)'tag_dips_0=',trim(resmig_par%tag_dips_0)
              call seperr('Problems with reading tag_dips_0 file')
        end if
!         write(0,*)'before minval (dips_0)=',minval (dips_0)
!         write(0,*)'before maxval (dips_0)=',maxval (dips_0)
        dips_0=atan(dips_0)
!         write(0,*)'after minval (dips_0)=',minval (dips_0)
!         write(0,*)'after maxval (dips_0)=',maxval (dips_0)
      end if

    !}
    end if adj_1

    do_i_rho_1: do i_rho=1,n_rho
    !{

    delta_rho=rho_0+(i_rho-1)*d_rho-1.
    if(resmig_par%verb) write(0,*)'i_rho=',i_rho
    if(resmig_par%verb) write(0,*)'delta_rho=',delta_rho

    if ((i_rho /= 1) .and. resmig_par%map_pseudo_depth_only) then
    if(resmig_par%verb) write(0,*)'sreed inside i_rho=',i_rho
      if(sreed(trim(resmig_par%tag_cig),cube(:,:,:,1,1),n_bytes_cube_inp)/= &
          n_bytes_cube_inp) then
        write(0,*)'tag_cig=',trim(resmig_par%tag_cig)
        call seperr('Problems with reading tag_cig file')
      end if

      k_cube=cmplx(0.,0.)
      status= ResMig_transf_cube(.true.,full_3d_fft,.false.,resmig_par,i_5_cube=1)
      k_cube_inp(:,:,:)=k_cube(:,:,:,1)
    end if


    do_i_rad_curv_1: do i_rad_curv=1,n_rad_curv
    !{
      if (resmig_par%linear_curvature_scale) then
        rad_curv=rad_curv_0+(i_rad_curv-1)*d_rad_curv
      else
        rad_curv=rad_curv_scale*tan(rad_curv_0+(i_rad_curv-1)*d_rad_curv)
      end if
      if(resmig_par%verb) write(0,*)'i_rad_curv=',i_rad_curv
      if(resmig_par%verb) write(0,*)'rad_curv=',rad_curv

    do_i_dip_0_curv_1: do i_dip_0_curv=1,n_dip_0_curv
    !{
      dip_0_curv=dip_0_curv_0+(i_dip_0_curv-1)*d_dip_0_curv
      if(resmig_par%debug1) write(0,*)'dip_0_curv=',dip_0_curv

      oper_name_1: select case (resmig_par%oper_name(1:9))
      !{
      case('oper-rmov','oper-rmig') oper_name_1
      !{
  
        floor=exp(-1.)
  
        k_cube=0.
        do_i_3: do i_3=1,n3_inp
        !{
          do_i_mx: do i_mx=1,n_mx_pad
          !{
            do_i_z: do i_z=1,n_z_pad_pos
            !{
              i_z_ne_1: if(i_z /= 1) then
              !{
                if(resmig_par%apert_ang_image) then
                  gamma=gamma_0+(i_3-1)*d_gamma
                else
                  tan_gamma=k_hx(i_3)/k_z(i_z)
                  gamma=atan(tan_gamma)
                end if
                cos_gamma=cos(gamma)
                sin_gamma=sin(gamma)
  
  !if((i_mx==n_mx_pad/4).and.(i_z==n_z_pad/4)) then
  ! write(0,*)'i_3=',i_3,'i_mx=',i_mx,'i_z=',i_z
  ! write(0,*)'gamma=',gamma,'cos_gamma=',cos_gamma,'sin_gamma=',sin_gamma
  !end if
  
                if (resmig_par%zero_dip) then
                  tan_dip=0
                  tan_dip_sh=0
                else
                  tan_dip=-k_mx(i_mx)/k_z(i_z)
                  tan_dip_sh=-k_mx(i_mx)/k_z(i_z)
                end if

                dip=atan(tan_dip)
                cos_dip=cos(dip)
                sin_dip=sin(dip)

                dip_sh=atan(tan_dip_sh)
                cos_dip_sh=cos(dip_sh)
                sin_dip_sh=sin(dip_sh)
                
  
  !if((i_mx==n_mx_pad/8).and.(i_z==n_z_pad/8)) then
  ! write(0,*)'i_3=',i_3,'i_mx=',i_mx,'i_z=',i_z
  ! write(0,*)'dip=',dip,'cos_dip=',cos_dip,'sin_dip=',sin_dip
  !end if
  
              !}
              else i_z_ne_1
              !{
                taper=0.
                if(resmig_par%apert_ang_image) then
                  gamma=gamma_0+(i_3-1)*d_gamma
                  tan_gamma=tan(gamma)
                  cos_gamma=cos(gamma)
                  sin_gamma=sin(gamma)
                else
                  gamma=pi/2.
                  tan_gamma=tan(.999*.5*pi)
                  cos_gamma=0.0001
                  sin_gamma=.9999
                end if
  
                dip=pi/2.
                dip_sh=pi/2.
                cos_dip=0.0001
                sin_dip=.9999
                cos_dip_sh=0.0001
                sin_dip_sh=.9999
  
              !}
              end if i_z_ne_1

              if (resmig_par%dip_ang_image) then
                interp_dip=(dip-dip_0)/d_dip+1
!                i_dip=max(min(nint((dip-dip_0)/d_dip+1),n_dip),1)
                i_dip=int(interp_dip)
                w_dip=(interp_dip-i_dip)
                i_dip_p1=i_dip+1
                max_i_dip=n_dip-1
              else
                i_dip=1
                i_dip_p1=1
                w_dip=1.
                max_i_dip=1
              end if

!    if((i_mx==n_mx_pad/4).and.(i_z==n_z_pad/4)) then
!     write(0,*)'i_3=',i_3,'i_mx=',i_mx,'i_z=',i_z
!     write(0,*)'gamma=',gamma*(180/pi),'dip=',dip*(180/pi),'prop_ang=',prop_ang*(180/pi)
!     write(0,*)'norm_shift=',norm_shift
!    end if
  
  
              i_dip_gt_1: if((i_dip >=1) .and. (i_dip <= max_i_dip)) then
              !{
                oper_name_2: select case (resmig_par%oper_name(1:9))
                !{
                case('oper-rmov') oper_name_2
                !{
                  prop_ang=max(abs(dip+gamma),abs(dip-gamma))
                  if (resmig_par%global_straight_rays) then
    !                if(abs(dip)+abs(gamma) <= max_prop_ang ) then
                    if((prop_ang <= max_prop_ang_0 )) then
                      norm_shift= &
                         ((sin_gamma**2)/(cos_dip*(cos_dip**2-sin_gamma**2)))
                    else
                      norm_shift=0.
                    end if
                  else
                      norm_shift=tan_gamma**2
                  end if
                  if (resmig_par%vertical_only) then
                    sin_dip_sh=0.
                  end if
                  curv_norm_shift=0.
                !}
                case('oper-rmig') oper_name_2
                !{
                  prop_ang=max(abs(dip+gamma),abs(dip-gamma))
                  if (resmig_par%global_straight_rays) then
!                    if((abs(dip+gamma) <= max_prop_ang ) .and. &
!                       (abs(dip-gamma) <= max_prop_ang )) then
                    if((prop_ang <= max_prop_ang_0 )) then
                      if(resmig_par%non_linear_resmig) then
                        if (delta_rho /= 0.) then
                          rho_nl=1+delta_rho
                          cos_dip_nl=cos(asin(rho_nl*sin_dip))
!                          cos_dip_nl=cos_dip
!                          cos_dip_nl=cos(asin(sin_dip))
!                          norm_shift=(rho_nl*cos_dip-cos_dip_nl)/(rho_nl*cos_dip*cos_dip_nl)
!                          norm_shift=(rho_nl*cos_dip-cos_dip_nl)/(cos_dip*cos_dip_nl)
                          norm_shift=(rho_nl*rho_nl*cos_dip-cos_dip_nl)/ &
                                      (rho_nl*rho_nl*cos_dip*cos_dip_nl)
!                          norm_shift=(cos_dip-cos_dip_nl)/(cos_dip*cos_dip_nl)
!                          norm_shift=(norm_shift/delta_rho)/rho_nl
                          norm_shift=(norm_shift/delta_rho)
                        else
                          norm_shift=0.
                        end if
                      else
                        norm_shift=((cos_dip)/(cos_dip**2-sin_gamma**2))
                      end if
                    else
                      norm_shift=0.
                    end if
                  else
                    norm_shift=1/cos_gamma**2
                  end if
                  curvature_image_2 : if(resmig_par%curvature_image) then
                    curv_z_shift=((tan(dip-dip_0_curv)**2)*(0.5*rad_curv))
                    curv_norm_shift=curv_z_shift*cos(dip-dip_0_curv)
                  else curvature_image_2
                    curv_norm_shift=0.
                  end if curvature_image_2
                !}
                case default oper_name_2
                  call seperr ('Do not know what to do (#2)')
                !}
                end select oper_name_2

  
                taper=1.
                if(  (abs(dip) < max_dip_0).and.(abs(dip) >= max_dip)) then
                  taper= taper* &
 (exp(-((abs(dip)-max_dip)/(max_dip_0-max_dip))**2)-floor)/(1.-floor)
                end if
                if  ((abs(gamma) < max_gamma_0).and. &
                     (abs(gamma) >= max_gamma)) then
                  taper= taper* &
 (exp(-((abs(gamma)-max_gamma)/(max_gamma_0-max_gamma))**2)-floor)/(1.-floor)
                end if
                if  ((abs(prop_ang) < max_prop_ang_0).and. &
                     (abs(prop_ang) >= max_prop_ang)) then
                  taper= taper* &
 (exp(-((abs(prop_ang)-max_prop_ang)/(max_prop_ang_0-max_prop_ang))**2)-floor)/&
                                                            (1.-floor)
                end if

                if  ((abs(dip) >= max_dip_0).or.&
                     (abs(gamma) >= max_gamma_0) .or. &
                     (prop_ang >= max_prop_ang_0) ) then
                  taper=0
                end if

!    if((i_mx==n_mx_pad/4).and.(i_z==n_z_pad/4)) then
!     write(0,*)'i_3=',i_3,'i_mx=',i_mx,'i_z=',i_z
!     write(0,*)'gamma=',gamma*(180/pi),'dip=',dip*(180/pi),'prop_ang=',prop_ang*(180/pi)
!     write(0,*)'norm_shift=',norm_shift,'taper=',taper
!    end if

                oper_name_3: select case (resmig_par%oper_name(1:9))
                !{
                case('oper-rmov','oper-rmig') oper_name_3
                !{
                  map_pseudo_depth_only_1: if(resmig_par%map_pseudo_depth_only) then

                     z_shift =-sign_shift*z_ref*delta_rho
                     mx_shift=0.
!                     z_shift=0.

!                     z_shift =sign_shift* &
!                            (z_ref*delta_rho* &
!                               (cos_dip_sh*(norm_shift + curv_norm_shift) -1.))
!
!                     mx_shift=-sign_shift*sin_dip_sh* &
!                            (z_ref*delta_rho*norm_shift + &
!                            curv_norm_shift)

                  else map_pseudo_depth_only_1 
                    map_pseudo_depth_1 : if(resmig_par%map_pseudo_depth) then
!                   z_shift =sign_shift*(&
!                          (z_ref*delta_rho*norm_shift + &
!                           curv_norm_shift)/cos_dip_sh-z_ref*delta_rho)
!                   mx_shift=0.

                     z_shift =sign_shift*cos_dip_sh* &
                            (z_ref*delta_rho*(norm_shift-cos_dip_sh) + &
                             curv_norm_shift)
  
                     mx_shift=-sign_shift*sin_dip_sh* &
                            (z_ref*delta_rho*(norm_shift-cos_dip_sh) + &
                            curv_norm_shift)
  
  
                   else map_pseudo_depth_1 
!                   z_shift =sign_shift*cos_dip_sh* &
!                          (z_ref*delta_rho*(norm_shift-1.) + &
!                           curv_norm_shift)

                     z_shift =sign_shift*cos_dip_sh* &
                            (z_ref*delta_rho*norm_shift + &
                             curv_norm_shift)
  
                     mx_shift=-sign_shift*sin_dip_sh* &
                            (z_ref*delta_rho*norm_shift + &
                            curv_norm_shift)
  
                    end if map_pseudo_depth_1 
                  end if map_pseudo_depth_only_1 

                   interp_temp=taper*k_cube_inp(i_z,i_mx,i_3)* &
                      cexp(cmplx(0,k_z(i_z)*z_shift+k_mx(i_mx)*mx_shift))

!                   k_cube(i_z,i_mx,i_3,i_dip   )=  &
!                     k_cube(i_z,i_mx,i_3,i_dip   ) + interp_temp

                   k_cube(i_z,i_mx,i_3,i_dip   )=  &
                     k_cube(i_z,i_mx,i_3,i_dip   ) + interp_temp*(1.-w_dip)
                   k_cube(i_z,i_mx,i_3,i_dip_p1)=  &
                     k_cube(i_z,i_mx,i_3,i_dip_p1) + interp_temp*(   w_dip)
                 !}
               case default oper_name_3
                 call seperr ('Do not know what to do (#3)')
                 !}
                 end select oper_name_3
              !}
              end if i_dip_gt_1
            !}
            end do do_i_z
!            k_cube(n_z_pad_pos+1,i_mx,i_3,:)=0.
          !}
          end do do_i_mx
!        call snap('Shift-ang.H',n_z_pad_pos,n_mx_pad,cube(1:n_z_pad_pos,:,i_3))
        !}
        end do do_i_3

        status=ResMig_transf_cube(.false.,full_3d_fft,cube_4d_fft,resmig_par, &
                                   i_5_cube=i_dip_0_curv)

      !}
      case default oper_name_1
        call seperr ('Do not know what to do #4')
      !}
      end select oper_name_1

    !}
    end do do_i_dip_0_curv_1

    adj_2: if (resmig_par%adj)  then
    !{
    !}
    else adj_2
    !{
      write_prestack_1:  if(resmig_par%write_prestack) then
      !{
        if((resmig_par%dip_ang_image).and.(i_dip_flip >0)) then
write(0,*)'flipping signs '
          cube(:,:,:,1:i_dip_flip-1,1)=-cube(:,:,:,1:i_dip_flip-1,1)
        end if
        if(srite(trim(resmig_par%tag_resmig),cube(:,:,:,:,1),n_bytes_cube_out)/= &
            n_bytes_cube_out) then
          write(0,*)'tag_resmig=',trim(resmig_par%tag_resmig)
          call seperr('Problems with writing tag_resmig file')
        end if
      !}
      else write_prestack_1
      !{
        if(resmig_par%apert_ang_image) then
        !{
          if(resmig_par%read_dips_0) then
          !{
            stack_cube=0.
            do_i_3_1: do i_3=1,n3_inp
            !{
              do_i_mx_1: do i_mx=1,n_mx
              !{
                do_i_z_1: do i_z=1,n_z
                !{
                  interp_dip= &
                   (dips_0(i_z,i_mx,i_rho)-dip_0_curv_0)/d_dip_0_curv+1
                  i_interp_dip=int(interp_dip)
                  i_interp_dip_p1=i_interp_dip+1
                  w_dip=(interp_dip-i_interp_dip)

                  if((i_interp_dip >=1).and.(i_interp_dip <= n_dip_0_curv)) then
                    stack_cube(i_z,i_mx,:)= stack_cube(i_z,i_mx,:)+ &
                      cube(i_z,i_mx,i_3,:,i_interp_dip   )*(1.-w_dip) + &
                      cube(i_z,i_mx,i_3,:,i_interp_dip_p1)*w_dip
                  else
                    write(0,*)'i_z=',i_z,'i_mx=',i_mx,'i_3=',i_3
                    write(0,*)'i_interp_dip=',i_interp_dip
                    write(0,*)'n_dip_0_curv=',n_dip_0_curv
                    call seperr("i_interp_dip out of range")
                  end if
             

                !}
                end do do_i_z_1
              !}
              end do do_i_mx_1
            !}
            end do do_i_3_1
          !}
          else
          !{
            stack_cube=sum(cube(:,:,:,:,1),dim=3)
          !}
          end if
        !}
        else
        !{
!          if (i_hx_0 /= 1) stack_cube=cube(:,:,i_hx_0,:,1)
          stack_cube=cube(:,:,i_hx_0,:,1)
        !}
        end if
        if((resmig_par%dip_ang_image).and.(i_dip_flip >0)) then
write(0,*)'flipping '
          stack_cube(:,:,1:i_dip_flip-1)=-stack_cube(:,:,1:i_dip_flip-1)
        end if
        if(srite(trim(resmig_par%tag_resmig),stack_cube,n_bytes_cube_out)/= &
            n_bytes_cube_out) then
          write(0,*)'tag_resmig=',trim(resmig_par%tag_resmig)
          call seperr('Problems with writing tag_resmig file')
        end if
        if(resmig_par%write_sum_squares)  then
          stack_cube=sum(cube(:,:,:,:,1)**2,dim=3)
          if(srite(trim(resmig_par%tag_sum_squares),stack_cube, &
              n_bytes_cube_out) /= n_bytes_cube_out) then
            write(0,*)'tag_sum_squares=',trim(resmig_par%tag_sum_squares)
            call seperr('Problems with writing tag_sum_squares file')
          end if
        end if

      !}
      end if write_prestack_1
    !}
    end if adj_2

    !}
    end do do_i_rad_curv_1
    !}
    end do do_i_rho_1
    !}
    end do do_i_cube_1

    status=0
    return	

  end function ResMig_oper

  function ResMig_transf_cube(fft_forward,full_3d_fft,cube_4d_fft,resmig_par,i_5_cube) &
              result(status)

    logical fft_forward,full_3d_fft,cube_4d_fft
    type (resmig_param_type) resmig_par
    integer i_5_cube
    integer status

    integer i_z,i_mx,i_3,i_4,n3_inp,n4,i_thread
    integer ret_sseek,i_sseek

    i_thread=1

    if (resmig_par%apert_ang_image) then
      n3_inp=n_gamma
    else
      n3_inp=n_hx
    end if
    if(cube_4d_fft) then
      n4=n_dip
    else
      n4=1
    end if

! zeroes temps
    temp_mx_hx=cmplx(0.,0.)
    temp_z=cmplx(0.,0.)
    temp_mx=cmplx(0.,0.)

   if(fft_forward) then
   !{

!      if(resmig_par%debug2) write(0,*)'Inside transforming forward'
!      if(resmig_par%debug2) write(0,*)'maxval(cube)=',maxval(cube)
  ! FFT along z
    do i_4=1,n4
    !{
      do i_3=1,n3_inp
      !{
        do i_mx=1,n_mx
        !{
  
          temp_z(1:n_z,i_thread)=cmplx(cube(1:n_z,i_mx,i_3,i_4,i_5_cube),0.)
          do i_z=n_z+1,n_z_pad
            temp_z(i_z,i_thread)=cmplx(0.,0.)
          end do
  
          call fft_1d_data(fft_nz_id(i_thread),FFT_BACK, &
                     n_z_pad,temp_z(:,i_thread))
  
          do i_z=1,n_z_pad_pos+1
            k_cube(i_z,i_mx,i_3,i_4)=temp_z(i_z,i_thread)
          end do
  
        !}
        end do 
      !}
      end do
    !}
    end do

  ! FFT along mx and h
    if(full_3d_fft) then
    !{
!      if(resmig_par%debug2) write(0,*)'Inside transforming forward (full_3d_fft)'
      do i_4=1,n4
      !{
        do i_z=1,n_z_pad_pos+1
        !{
          temp_mx_hx(:,:,i_thread)=k_cube(i_z,:,:,i_4)
          call fft_2d_data(fft_nmx_nhx_id(i_thread),FFT_BACK, &
              n_mx_pad,n3_inp,temp_mx_hx(:,:,i_thread))
          k_cube(i_z,:,:,i_4)=temp_mx_hx(:,:,i_thread)
        !}
        end do 
      !}
      end do 
    !}
    else
    !{
!      if(resmig_par%debug2) write(0,*)'Inside transforming forward ( not full_3d_fft)'
      do i_4=1,n4
      !{
        do i_3=1,n3_inp
        !{
          do i_z=1,n_z_pad_pos+1
          !{
            temp_mx(:,i_thread)=k_cube(i_z,:,i_3,i_4)
            call fft_1d_data(fft_nmx_id(i_thread),FFT_BACK, &
                     n_mx_pad,temp_mx(:,i_thread))
            k_cube(i_z,:,i_3,i_4)=temp_mx(:,i_thread)
          !}
          end do 
        !}
        end do 
      !}
      end do 
    !}
    end if

!      if(resmig_par%debug2) write(0,*)'Done transforming forward'
!      if(resmig_par%debug2) write(0,*)'maxval(cabs(k_cube)))=',maxval(cabs(k_cube))

   !}
   else
   !{

!      if(resmig_par%debug2) write(0,*)'Inside transforming backward'
!      if(resmig_par%debug2) write(0,*)'maxval(cabs(k_cube)))=',maxval(cabs(k_cube))
! INV FFT along mx and hx

!      if(resmig_par%debug2) write(0,*)'#0 maxval(k_cube)=',maxval(cabs(k_cube))

    if(full_3d_fft) then
    !{
!      if(resmig_par%debug2) write(0,*)'Inside transforming backward (full_3d_fft)'
      do i_4=1,n4
      !{
        do i_z=1,n_z_pad_pos+1
        !{
          temp_mx_hx(:,:,i_thread)=k_cube(i_z,:,:,i_4)
          call fft_2d_data(fft_nmx_nhx_id(i_thread),FFT_FORW, &
                   n_mx_pad,n3_inp,temp_mx_hx(:,:,i_thread))
  
           k_cube(i_z,:,:,i_4)=temp_mx_hx(:,:,i_thread)
        !}
        end do 
      !}
      end do 
    !}
    else
    !{
!      if(resmig_par%debug2) write(0,*)'Inside transforming backward (not full_3d_fft)'
      do i_4=1,n4
      !{
        do i_3=1,n3_inp
        !{
          do i_z=1,n_z_pad_pos+1
          !{
          temp_mx(:,i_thread)=k_cube(i_z,:,i_3,i_4)
          call fft_1d_data(fft_nmx_id(i_thread),FFT_FORW, &
                  n_mx_pad,temp_mx(:,i_thread))
    
           k_cube(i_z,:,i_3,i_4)=temp_mx(:,i_thread)
          !}
          end do 
        !}
        end do 
      !}
      end do 
    !}
    end if

!      if(resmig_par%debug2) write(0,*)'#1 maxval(k_cube)=',maxval(cabs(k_cube))

! NV FFT along z
    do i_4=1,n4
    !{
      do i_3=1,n3_inp
      !{
        do i_mx=1,n_mx
        !{
          do i_z=1,n_z_pad_pos+1
          !{
            temp_z(i_z,i_thread)=k_cube(i_z,i_mx,i_3,i_4)
          !}
          end do
          do i_z=n_z_pad_pos+2,n_z_pad
          !{
            temp_z(i_z,i_thread)= conjg(k_cube(n_z_pad-i_z+2,i_mx,i_3,i_4))
          !}
          end do
  !        temp_z(1,i_thread)=cmplx(0.,0.)
  
          call fft_1d_data(fft_nz_id(i_thread),FFT_FORW, &
                    n_z_pad,temp_z(:,i_thread))
  
          cube(1:n_z,i_mx,i_3,i_4,i_5_cube)=real(temp_z(1:n_z,i_thread))
  
        !}
        end do 
      !}
      end do 
    !}
    end do 
!      if(resmig_par%debug2) write(0,*)'maxval(cube)=',maxval(cube)
!      if(resmig_par%debug2) write(0,*)'Done transforming backward'

!      if(resmig_par%debug2) write(0,*)'#2 maxval(cube)=',maxval(cube)

   !}
   end if


   status=0
   return	

  end function ResMig_transf_cube

end module ResMig_mod
