subroutine param_resmig(resmig_par)
  use resmig_types
  use sep

  implicit none
  type (resmig_param_type) resmig_par

  integer ::  omp_get_num_threads
  integer, external :: auxpar,getch,fetch,auxin,auxinout,sreed,sseek,auxout

  character(len=128) :: tag_name
  integer n_threads_env,n_threads_init
  integer esize,one_int
  integer n_z_cig,n_mx_cig,n_hx_cig,n_gamma_cig,n_z_dips,n_mx_dips,n_rho_dips
  real d_z_cig,d_mx_cig,d_hx_cig,d_gamma_cig,d_z_dips,d_mx_dips,d_rho_dips
  real z_0_cig,mx_0_cig,hx_0_cig,gamma_0_cig,z_0_dips,mx_0_dips,rho_0_dips
  real kz_0,kmx_0,khx_0,d_kz,d_kmx,d_khx
  real pi,zero,one_real,epsi_par

  n_threads_init=-1

  pi=acos(-1.)
  one_int=1
  one_real=1.
  zero=0.
  epsi_par=.001

  resmig_par%label_z="Depth (m)"
  resmig_par%label_mx="Cmp X (m)"
  resmig_par%label_hx="Half Offset X (m)"
  resmig_par%label_gamma="Aperture Ang. (deg)"
  resmig_par%label_dip="Dip Ang. (deg)"
  resmig_par%label_rho="Rho"

  call from_param('verb',resmig_par%verb,.false.)
  call from_param('debug1',resmig_par%debug1,.false.)
  call from_param('debug2',resmig_par%debug2,.false.)

  call from_param('adj',resmig_par%adj,.false.)

  call from_param('apert_ang_image',resmig_par%apert_ang_image,.true.)
  call from_param('dip_ang_image',resmig_par%dip_ang_image,.true.)
  call from_param('write_prestack',resmig_par%write_prestack,.true.)
  call from_param('write_sum_squares',resmig_par%write_sum_squares,.false.)

  call from_param('rmo_ang_cig',resmig_par%rmo_ang_cig,.false.)
  call from_param('rmig_ang_cig',resmig_par%rmig_ang_cig,.false.)

  call from_param('curvature_image',resmig_par%curvature_image,.false.)
  call from_param('read_dips_0',resmig_par%read_dips_0,.false.)

  call from_param('n_threads',resmig_par%n_threads,n_threads_init)
  call from_param('max_memory',resmig_par%max_memory,100)

  if(resmig_par%rmo_ang_cig) then
    resmig_par%oper_name(1:9)='oper-rmov'
  else 
    if(resmig_par%rmig_ang_cig) then
      resmig_par%oper_name(1:9)="oper-rmig"
    end if
  end if

   write(0,*)'oper_name=',resmig_par%oper_name(1:9)
             
  ! checking parameters
!$omp parallel
  n_threads_env=omp_get_num_threads()
!$omp end parallel
  if(resmig_par%n_threads == n_threads_init) then
     resmig_par%n_threads=n_threads_env
  else
     if(n_threads_env > resmig_par%n_threads) then
        call omp_set_num_threads(resmig_par%n_threads)
        write(0,*)'WARNING: overwriting environment variable OMP_NUM_THREADS'
        write(0,*)'resmig_par-n_threads=',resmig_par%n_threads
     end if
     if(n_threads_env < resmig_par%n_threads) then
        write(0,*)'resmig_par-n_threads=',resmig_par%n_threads
        write(0,*)'OMP_NUM_THREADS=',n_threads_env
        call seperr('ERROR: Asked for more threads than the environment variable allows')
     end if
  end if
  if(resmig_par%debug1)write(0,*)'resmig_par-n_threads=',resmig_par%n_threads
  call to_history('n_threads',resmig_par%n_threads)

   if(resmig_par%n_threads /= 1) then
     call seperr("At the present n_threads must be == 1")
   end if

   if(resmig_par%read_dips_0 .and. (.not. resmig_par%curvature_image)) then
     call seperr("curvature_image must be true if read_dips_0 is true")
   end if
   if(resmig_par%read_dips_0 .and. (.not. resmig_par%apert_ang_image)) then
     call seperr("apert_ang_image must be true if read_dips_0 is true")
   end if

  resmig_par%n_cube=1

  resmig_par%n_rad_curv=1
  resmig_par%rad_curv_0=0.
  resmig_par%d_rad_curv=1.
  resmig_par%rad_curv_scale=1.

  select case (resmig_par%oper_name(1:9))
  !{
  case('oper-rmig','oper-rmov')
  !{
  
    call from_param('global_straight_rays', &
             resmig_par%global_straight_rays,.true.)
    call from_param('map_pseudo_depth',resmig_par%map_pseudo_depth,.false.)
    call from_param('map_pseudo_depth_only', &
                    resmig_par%map_pseudo_depth_only,.false.)
    call from_param('inverse_mapping',resmig_par%inverse_mapping,.false.)
    call from_param('non_linear_resmig',resmig_par%non_linear_resmig,.false.)

    call from_param('z_ref',resmig_par%z_ref,1000.)
    call from_param('max_dip',resmig_par%max_dip,80.)
    call from_param('max_gamma',resmig_par%max_gamma,80.)
    call from_param('max_prop_ang',resmig_par%max_prop_ang,89.)
  !}
    if (resmig_par%oper_name(1:9) == 'oper-rmov') then
    !{
      call from_param('zero_dip',resmig_par%zero_dip,.false.)
      call from_param('vertical_only',resmig_par%vertical_only,.false.)
    !}
    else
    !{
      resmig_par%zero_dip=.false.
      resmig_par%vertical_only=.false.
    !}
    end if
  case default
    call seperr ('Do not know what to do (#1 in param_resmig)')
  !}
  end select


  adj_1: if(resmig_par%adj) then
  !{

  !}
  else adj_1
  !{

! Cig is an input

    call from_param('n_rho',resmig_par%n_rho)
    call from_param('d_rho',resmig_par%d_rho)
    call from_param('rho_0',resmig_par%rho_0)
    call from_param('label_rho',resmig_par%label_rho,trim(resmig_par%label_rho))

    if (resmig_par%curvature_image) then
    !{
      call from_param('linear_curvature_scale', &
              resmig_par%linear_curvature_scale,.false.)
     if (resmig_par%linear_curvature_scale) then
     !{
        call from_param('n_rad_curv',resmig_par%n_rad_curv,1)
        call from_param('rad_curv_0',resmig_par%rad_curv_0,0.)
        call from_param('d_rad_curv',resmig_par%d_rad_curv,1.)
        call from_param('dip_0_curv',resmig_par%dip_0_curv,0.)
        resmig_par%label_rad_curv="R (m)"
        call from_param('label_rad_curv',resmig_par%label_rad_curv, &
           trim(resmig_par%label_rad_curv))
        resmig_par%rad_curv_scale=0
     !}
     else
     !{
        call from_param('n_rad_curv',resmig_par%n_rad_curv,1)
        call from_param('rad_curv_0',resmig_par%rad_curv_0,0.)
        call from_param('d_rad_curv',resmig_par%d_rad_curv,1.)
        call from_param('rad_curv_scale',resmig_par%rad_curv_scale,1.)
        call from_param('dip_0_curv',resmig_par%dip_0_curv,0.)
        resmig_par%label_rad_curv="arctan(R/scale)/pi"
        call from_param('label_rad_curv',resmig_par%label_rad_curv, &
           trim(resmig_par%label_rad_curv))
        if(resmig_par%rad_curv_0 <= -.5) &
                    call seperr("rad_curv_0 should be > -.5")
        if(resmig_par%rad_curv_0+ &
          (resmig_par%n_rad_curv-1)*resmig_par%d_rad_curv>=.5)&
                    call seperr("max of rad_curv_0 should be < .5")
      !}
      end if
    !}
    end if

    if(resmig_par%dip_ang_image) then
      if(.not. resmig_par%apert_ang_image)  &
         call seperr("dip_ang image must be with apert_ang_image")
      call from_param('n_dip',resmig_par%n_dip)
      call from_param('d_dip',resmig_par%d_dip)
      call from_param('dip_0',resmig_par%dip_0)
      call from_param('flip_sign_dip',resmig_par%flip_sign_dip, &
             huge(resmig_par%flip_sign_dip))
    else
      if(resmig_par%write_sum_squares)  &
         call seperr("write_sum_squares  must be with dip_ang_image")
    end if

    call from_param('tag_cig',resmig_par%tag_cig,'Cig')
    call from_param(trim(resmig_par%tag_cig),tag_name,'Cig not found')
    resmig_par%cig=auxin(trim(resmig_par%tag_cig)) > 0
    if(.not. resmig_par%cig) call seperr("Cig missing")

    call from_aux(trim(resmig_par%tag_cig),'n1',resmig_par%n_z)
    call from_aux(trim(resmig_par%tag_cig),'d1',resmig_par%d_z)
    call from_aux(trim(resmig_par%tag_cig),'o1',resmig_par%z_0)
    call from_aux(trim(resmig_par%tag_cig),'label1',resmig_par%label_z, &
              trim(resmig_par%label_z))


    call to_history("From Cig: n_z",resmig_par%n_z)
    call to_history("From Cig: d_z",resmig_par%d_z)
    call to_history("From Cig: z_0",resmig_par%z_0)

    call from_aux(trim(resmig_par%tag_cig),'n2',resmig_par%n_mx)
    call from_aux(trim(resmig_par%tag_cig),'d2',resmig_par%d_mx)
    call from_aux(trim(resmig_par%tag_cig),'o2',resmig_par%mx_0)
    call from_aux(trim(resmig_par%tag_cig),'label2',resmig_par%label_mx, &
              trim(resmig_par%label_mx))

    call to_history("From Cig: n_mx",resmig_par%n_mx)
    call to_history("From Cig: d_mx",resmig_par%d_mx)
    call to_history("From Cig: mx_0",resmig_par%mx_0)

    apert_ang_image_1: if(resmig_par%apert_ang_image) then
      call from_aux(trim(resmig_par%tag_cig),'n3',resmig_par%n_gamma)
      call from_aux(trim(resmig_par%tag_cig),'d3',resmig_par%d_gamma)
      call from_aux(trim(resmig_par%tag_cig),'o3',resmig_par%gamma_0)
      call from_aux(trim(resmig_par%tag_cig),'label3',resmig_par%label_gamma, &
              trim(resmig_par%label_gamma))

      call to_history("From Cig: n_gamma",resmig_par%n_gamma)
      call to_history("From Cig: d_gamma",resmig_par%d_gamma)
      call to_history("From Cig: gamma_0",resmig_par%gamma_0)
      if(resmig_par%dip_ang_image) then
        call from_aux(trim(resmig_par%tag_cig),'n4',resmig_par%n_cube)
        call from_aux(trim(resmig_par%tag_cig),'d4',resmig_par%d_cube)
        call from_aux(trim(resmig_par%tag_cig),'o4',resmig_par%cube_0)
        call from_aux(trim(resmig_par%tag_cig),'label4',resmig_par%label_cube, &
                trim(resmig_par%label_cube))
  
        call to_history("From Cig: n_cube",resmig_par%n_cube)
        call to_history("From Cig: d_cube",resmig_par%d_cube)
        call to_history("From Cig: cube_0",resmig_par%cube_0)
      end if
    else apert_ang_image_1
      call from_aux(trim(resmig_par%tag_cig),'n3',resmig_par%n_hx)
      call from_aux(trim(resmig_par%tag_cig),'d3',resmig_par%d_hx)
      call from_aux(trim(resmig_par%tag_cig),'o3',resmig_par%hx_0)
      call from_aux(trim(resmig_par%tag_cig),'label3',resmig_par%label_hx, &
              trim(resmig_par%label_hx))

      call to_history("From Cig: n_hx",resmig_par%n_hx)
      call to_history("From Cig: d_hx",resmig_par%d_hx)
      call to_history("From Cig: hx_0",resmig_par%hx_0)
    end if apert_ang_image_1 

    call from_param('n_z_pad',resmig_par%n_z_pad,resmig_par%n_z)
    call from_param('n_mx_pad',resmig_par%n_mx_pad,resmig_par%n_mx)
    kz_0=0.
    kmx_0=0.
    d_kz=2.*pi/(resmig_par%d_z*FLOAT(resmig_par%n_z_pad))
    d_kmx=2.*pi/(resmig_par%d_mx*FLOAT(resmig_par%n_mx_pad))

    if( .not. resmig_par%apert_ang_image) then
      call from_param('n_hx_pad',resmig_par%n_hx_pad,resmig_par%n_hx)
      khx_0=0.
      d_khx=2.*pi/(resmig_par%d_hx*FLOAT(resmig_par%n_hx_pad))
    end if

    if (resmig_par%curvature_image) then
    !{
      read_dips_0_1: if(resmig_par%read_dips_0) then
      !{
       call from_param('n_dip_0_curv',resmig_par%n_dip_0_curv)
       call from_param('dip_0_curv_0',resmig_par%dip_0_curv_0)
       call from_param('d_dip_0_curv',resmig_par%d_dip_0_curv)

       call from_param('tag_dips_0',resmig_par%tag_dips_0,'Dips_0')
       call from_param(trim(resmig_par%tag_dips_0),tag_name,'Dips_0 not found')
       resmig_par%dips_0=auxin(trim(resmig_par%tag_dips_0)) > 0
       if(.not. resmig_par%dips_0) call seperr("Dips_0 missing")

       call from_aux(trim(resmig_par%tag_dips_0),'n1',n_z_dips)
       call from_aux(trim(resmig_par%tag_dips_0),'d1',d_z_dips)
       call from_aux(trim(resmig_par%tag_dips_0),'o1',z_0_dips)

       call from_aux(trim(resmig_par%tag_dips_0),'n2',n_mx_dips)
       call from_aux(trim(resmig_par%tag_dips_0),'d2',d_mx_dips)
       call from_aux(trim(resmig_par%tag_dips_0),'o2',mx_0_dips)

       call from_aux(trim(resmig_par%tag_dips_0),'n3',n_rho_dips)
       call from_aux(trim(resmig_par%tag_dips_0),'d3',d_rho_dips)
       call from_aux(trim(resmig_par%tag_dips_0),'o3',rho_0_dips)

       if(n_z_dips /= resmig_par%n_z) call seperr("Dips_0 n1 /= Cig n1")
       if(d_z_dips /= resmig_par%d_z) call seperr("Dips_0 d1 /= Cig d1")
!       if(z_0_dips /= resmig_par%z_0) call seperr("Dips_0 o1 /= Cig o1")
       if(abs((z_0_dips-resmig_par%z_0)/(z_0_dips+epsi_par)) > epsi_par) &
              call seperr("Dips_0 o1 /= Cig o1")

       if(n_mx_dips /= resmig_par%n_mx) call seperr("Dips_0 n2 /= Cig n2")
       if(d_mx_dips /= resmig_par%d_mx) call seperr("Dips_0 d2 /= Cig d2")
!       if(mx_0_dips /= resmig_par%mx_0) call seperr("Dips_0 o2 /= Cig o2")
       if(abs((mx_0_dips-resmig_par%mx_0)/(mx_0_dips+epsi_par)) > epsi_par) &
              call seperr("Dips_0 o2 /= Cig o2")

       if(n_rho_dips /= resmig_par%n_rho) call seperr("Dips_0 n3 /= par n_rho")
       if(d_rho_dips /= resmig_par%d_rho) call seperr("Dips_0 d3 /= par d_rho")
!       if(rho_0_dips /= resmig_par%rho_0) call seperr("Dips_0 o3 /= Cig o3")
       if(abs((rho_0_dips-resmig_par%rho_0)/(rho_0_dips+epsi_par)) > epsi_par) &
              call seperr("Dips_0 o3 /= par rho_0" )
      !}
      end if read_dips_0_1
    !}
    end if 

! ResMig is an output

    call from_param('tag_resmig',resmig_par%tag_resmig,'ResMig')
    call from_param(trim(resmig_par%tag_resmig),tag_name, &
          'ResMig not found')
    write(0,*)'resmig_par%tag_resmig=',trim(resmig_par%tag_resmig)
    
    if(auxinout(trim(resmig_par%tag_resmig)) > 0) then
      if(resmig_par%verb) write(0,*)'ResMig already exists'
    end if
  
    esize=4
    call to_history("ResMig_z_ref",resmig_par%z_ref,trim(resmig_par%tag_resmig))
    call to_history("esize",esize,trim(resmig_par%tag_resmig))
    call to_history("n1",resmig_par%n_z,trim(resmig_par%tag_resmig))
    call to_history("d1",resmig_par%d_z,trim(resmig_par%tag_resmig))
    call to_history("o1",resmig_par%z_0,trim(resmig_par%tag_resmig))
    call to_history("label1",resmig_par%label_z,trim(resmig_par%tag_resmig))

    call to_history("n2",resmig_par%n_mx,trim(resmig_par%tag_resmig))
    call to_history("d2",resmig_par%d_mx,trim(resmig_par%tag_resmig))
    call to_history("o2",resmig_par%mx_0,trim(resmig_par%tag_resmig))
    call to_history("label2",resmig_par%label_mx,trim(resmig_par%tag_resmig))

    prestack_image_1: if(resmig_par%write_prestack) then
    !{
      apert_ang_image_2: if(resmig_par%apert_ang_image) then
        call to_history("n3",resmig_par%n_gamma, &
               trim(resmig_par%tag_resmig))
        call to_history("d3",resmig_par%d_gamma, &
               trim(resmig_par%tag_resmig))
        call to_history("o3",resmig_par%gamma_0, &
               trim(resmig_par%tag_resmig))
        call to_history("label3",resmig_par%label_gamma, &
               trim(resmig_par%tag_resmig))
      else apert_ang_image_2
        call to_history("n3",resmig_par%n_hx, &
               trim(resmig_par%tag_resmig))
        call to_history("d3",resmig_par%d_hx, &
               trim(resmig_par%tag_resmig))
        call to_history("o3",resmig_par%hx_0, &
               trim(resmig_par%tag_resmig))
        call to_history("label3",resmig_par%label_hx, &
               trim(resmig_par%tag_resmig))
      end if apert_ang_image_2
      dip_ang_image_1: if(resmig_par%dip_ang_image) then
      !{
        call to_history("n4",resmig_par%n_dip, &
              trim(resmig_par%tag_resmig))
        call to_history("o4",resmig_par%dip_0, &
              trim(resmig_par%tag_resmig))
        call to_history("d4",resmig_par%d_dip, &
            trim(resmig_par%tag_resmig))
        call to_history("label4",resmig_par%label_dip, &
               trim(resmig_par%tag_resmig))

        curvature_image_1 : if(resmig_par%curvature_image) then
        !{

          call to_history("n5",resmig_par%n_rad_curv, &
                trim(resmig_par%tag_resmig))
          call to_history("o5",resmig_par%rad_curv_0, &
                trim(resmig_par%tag_resmig))
          call to_history("d5",resmig_par%d_rad_curv, &
                trim(resmig_par%tag_resmig))
          call to_history("label5",resmig_par%label_rad_curv, &
                 trim(resmig_par%tag_resmig))

          call to_history("n6",resmig_par%n_rho, &
                trim(resmig_par%tag_resmig))
          call to_history("o6",resmig_par%rho_0, &
                trim(resmig_par%tag_resmig))
          call to_history("d6",resmig_par%d_rho, &
                trim(resmig_par%tag_resmig))
          call to_history("label6",resmig_par%label_rho, &
                 trim(resmig_par%tag_resmig))

        !}
        else curvature_image_1
        !{
          call to_history("n5",resmig_par%n_rho, &
                trim(resmig_par%tag_resmig))
          call to_history("o5",resmig_par%rho_0, &
                trim(resmig_par%tag_resmig))
          call to_history("d5",resmig_par%d_rho, &
                trim(resmig_par%tag_resmig))
          call to_history("label5",resmig_par%label_rho, &
                 trim(resmig_par%tag_resmig))
          if(resmig_par%n_cube /= 1) then
            call to_history("n6",resmig_par%n_cube, &
                  trim(resmig_par%tag_resmig))
            call to_history("o6",resmig_par%cube_0, &
                  trim(resmig_par%tag_resmig))
            call to_history("d6",resmig_par%d_cube, &
                  trim(resmig_par%tag_resmig))
            call to_history("label6",resmig_par%label_cube, &
                 trim(resmig_par%tag_resmig))
         end if
        !}
        end if curvature_image_1


      !}
      else dip_ang_image_1
      !{
        call to_history("n4",resmig_par%n_rho, &
              trim(resmig_par%tag_resmig))
        call to_history("o4",resmig_par%rho_0, &
              trim(resmig_par%tag_resmig))
        call to_history("d4",resmig_par%d_rho, &
              trim(resmig_par%tag_resmig))
        call to_history("label4",resmig_par%label_rho, &
               trim(resmig_par%tag_resmig))
      !}
      end if  dip_ang_image_1
    !}
    else prestack_image_1
    !{
      dip_ang_image_2: if(resmig_par%dip_ang_image) then
      !{
        call to_history("n3",resmig_par%n_dip, &
              trim(resmig_par%tag_resmig))
        call to_history("o3",resmig_par%dip_0, &
              trim(resmig_par%tag_resmig))
        call to_history("d3",resmig_par%d_dip, &
              trim(resmig_par%tag_resmig))
        call to_history("label3",resmig_par%label_dip, &
               trim(resmig_par%tag_resmig))

        curvature_image_2 : if(resmig_par%curvature_image) then
        !{
          call to_history("n4",resmig_par%n_rad_curv, &
                trim(resmig_par%tag_resmig))
          call to_history("o4",resmig_par%rad_curv_0, &
                trim(resmig_par%tag_resmig))
          call to_history("d4",resmig_par%d_rad_curv, &
                trim(resmig_par%tag_resmig))
          call to_history("label4",resmig_par%label_rad_curv, &
                 trim(resmig_par%tag_resmig))

          call to_history("n5",resmig_par%n_rho, &
                trim(resmig_par%tag_resmig))
          call to_history("o5",resmig_par%rho_0, &
                trim(resmig_par%tag_resmig))
          call to_history("d5",resmig_par%d_rho, &
                trim(resmig_par%tag_resmig))
          call to_history("label5",resmig_par%label_rho, &
               trim(resmig_par%tag_resmig))
        !}
        else curvature_image_2
        !{
          call to_history("n4",resmig_par%n_rho, &
                trim(resmig_par%tag_resmig))
          call to_history("o4",resmig_par%rho_0, &
                trim(resmig_par%tag_resmig))
          call to_history("d4",resmig_par%d_rho, &
                trim(resmig_par%tag_resmig))
          call to_history("label4",resmig_par%label_rho, &
                 trim(resmig_par%tag_resmig))
        !}
        end if curvature_image_2


        write_sum_squares_1: if(resmig_par%write_sum_squares) then
        !{
          call from_param('tag_sum_squares',resmig_par%tag_sum_squares, &
                'Sum_squares')
          call from_param(trim(resmig_par%tag_sum_squares),tag_name, &
                'Sum_squares not found')
          write(0,*)'resmig_par%tag_sum_squares=', &
                 trim(resmig_par%tag_sum_squares)
          
          if(auxinout(trim(resmig_par%tag_sum_squares)) > 0) then
            if(resmig_par%verb) write(0,*)'Sum_squares already exists'
          end if

          esize=4
          call to_history("esize",esize,trim(resmig_par%tag_sum_squares))
          call to_history("n1",resmig_par%n_z,trim(resmig_par%tag_sum_squares))
          call to_history("d1",resmig_par%d_z,trim(resmig_par%tag_sum_squares))
          call to_history("o1",resmig_par%z_0,trim(resmig_par%tag_sum_squares))
          call to_history("label1",resmig_par%label_z, &
                      trim(resmig_par%tag_resmig))
      
          call to_history("n2",resmig_par%n_mx,trim(resmig_par%tag_sum_squares))
          call to_history("d2",resmig_par%d_mx,trim(resmig_par%tag_sum_squares))
          call to_history("o2",resmig_par%mx_0,trim(resmig_par%tag_sum_squares))
          call to_history("label2",resmig_par%label_mx, &
                      trim(resmig_par%tag_resmig))

          call to_history("n3",resmig_par%n_dip,trim(resmig_par%tag_sum_squares))
          call to_history("o3",resmig_par%dip_0,trim(resmig_par%tag_sum_squares))
          call to_history("d3",resmig_par%d_dip,trim(resmig_par%tag_sum_squares))
          call to_history("label3",resmig_par%label_dip, &
                      trim(resmig_par%tag_resmig))
  
          call to_history("n4",resmig_par%n_rho,trim(resmig_par%tag_sum_squares))
          call to_history("o4",resmig_par%rho_0,trim(resmig_par%tag_sum_squares))
          call to_history("d4",resmig_par%d_rho,trim(resmig_par%tag_sum_squares))
          call to_history("label4",resmig_par%label_rho, &
                      trim(resmig_par%tag_resmig))

        !}
        end if write_sum_squares_1

      !}
      else dip_ang_image_2
      !{
        call to_history("n3",resmig_par%n_rho, &
              trim(resmig_par%tag_resmig))
        call to_history("o3",resmig_par%rho_0, &
              trim(resmig_par%tag_resmig))
        call to_history("d3",resmig_par%d_rho, &
              trim(resmig_par%tag_resmig))
          call to_history("label3",resmig_par%label_rho, &
              trim(resmig_par%tag_resmig))
      !}
      end if dip_ang_image_2
      write(0,*)'After'
    !}
    end if prestack_image_1


  !}
  end if adj_1

  call from_param('write_cig_transf',resmig_par%write_cig_transf,.false.)
  if((resmig_par%write_cig_transf).and. &
     (.not. resmig_par%apert_ang_image)) then
  !{
    call from_param('tag_cig_transf',resmig_par%tag_cig_transf,'Cig_transf')
    call from_param(trim(resmig_par%tag_cig_transf),tag_name,'Cig_transf not found')
    write(0,*)'resmig_par%tag_cig_transf=',trim(resmig_par%tag_cig_transf)

    if(auxinout(trim(resmig_par%tag_cig_transf)) > 0) then
      if(resmig_par%verb)write(0,*)'Cig_transf already exists'
    end if

    esize=8
    call to_history("esize",esize,trim(resmig_par%tag_cig_transf))
    call to_history("n1",resmig_par%n_z_pad/2+1,trim(resmig_par%tag_cig_transf))
    call to_history("n2",resmig_par%n_mx_pad,trim(resmig_par%tag_cig_transf))
    call to_history("n3",resmig_par%n_hx,trim(resmig_par%tag_cig_transf))
    call to_history("d1",d_kz,trim(resmig_par%tag_cig_transf))
    call to_history("d2",d_kmx,trim(resmig_par%tag_cig_transf))
    call to_history("d3",d_khx,trim(resmig_par%tag_cig_transf))
    call to_history("o1",kz_0,trim(resmig_par%tag_cig_transf))
    call to_history("o2",kmx_0,trim(resmig_par%tag_cig_transf))
    call to_history("o3",khx_0,trim(resmig_par%tag_cig_transf))
  !}
  end if
  
  return
end subroutine param_resmig
